/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft;

import java.util.*;

import jp.littlesoft.util.StringUtil;
import jp.littlesoft.util.res.MessageResource;

/**
 * LittleSoft J-Programming内で発生する例外の基底クラス<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class LSException extends Exception {

  private static final long serialVersionUID = 1L;
  private Locale locale = Locale.JAPAN;
  private MessageResource msgResource = null;

  /**
   * コンストラクタ<br/>
   */  
  public LSException() {
    super();
  }

  /**
   * コンストラクタ<br/>
   *
   * @param  message　メッセージ
   */
  public LSException(String message) {
    super(message);
  }

  /**
   * コンストラクタ<br/>
   *
   * @param  message　メッセージ
   * @param  cause  原因となったThrowableオブジェクト
   */
  public LSException(String message, Throwable cause) {
    super(message, cause);
  }

  /**
   * コンストラクタ<br/>
   *
   * @param  cause  原因となったThrowableオブジェクト
   */  
  public LSException(Throwable cause) {
    super(cause);
    if (cause instanceof LSException) {
      // 原因となったThrowableがLJExceptionなら、MessageResourceなどを引き継ぐ
      this.msgResource = ((LSException)cause).getMessageResource();
    }
  }

  /**
   * コンストラクタ<br/>
   *
   * @param cause  原因となったThrowableオブジェクト
   * @param msgResClassName メッセージリソースのクラス名
   * @param msgResFieldName メッセージリソースのフィールド名
   * @param msgArgs メッセージ中に表示する置換文字列の配列
   */
  public LSException(Throwable cause, String msgResClassName, String msgResFieldName, Object[] msgArgs) {
    super(cause);
    if (StringUtil.isNotNullString(msgResFieldName)) {
      this.msgResource = new MessageResource(msgResClassName, msgResFieldName, msgArgs);
    }
  }

  /**
   * コンストラクタ<br/>
   *
   * @param cause  原因となったThrowableオブジェクト
   * @param msgResource メッセージリソース
   */
  public LSException(Throwable cause, MessageResource msgResource) {
    super(cause);
    this.msgResource = msgResource;
  }

  /**
   * メッセージリソースを設定します。<br/>
   *
   * @param msgResource メッセージリソース
   */
  public void setMessageResource(MessageResource msgResource) {
    this.msgResource = msgResource;
  }
  
  /**
   * メッセージリソースを取得します。<br/>
   *
   * @return メッセージリソース
   */
  public MessageResource getMessageResource() {
    return this.msgResource;
  }
  
	/**
	 * LJExceptionがデータベースの更新エラーが原因で発生したか？<br/><br/>
	 * 
	 * ※注：INSERTのエラー時に更新エラー(INSERT)を返すのは、
	 * JDBCAdaptorのOptimisticLockTypeにJDBCAdaptor.OptimisticLockType.PRE_FETCH
	 * を指定した場合です。<br/>
	 * 指定しなかった場合は、更新エラー(INSERT)とはならず、
	 * キーが重複した、などのSQLエラーとなります。<br/>
	 * 
	 * 0:原因は更新エラー(DELETE)<br/>
	 * 1:原因は更新エラー(UPDATE)<br/>
	 * 2:原因は更新エラー(INSERT)<br/>
	 * 
	 * @return -1:原因は更新エラーでない<br/>
	 */
	public int getCauseUpdateError() {
    String[] updateErrorMsgs = new String[3];
    updateErrorMsgs[jp.littlesoft.data.event.ResolvingEvent.Type.DELETE.ordinal()] = MessageResource.getMessage(null, "ERROR_RESOLVING_DELETE", null, null);
    updateErrorMsgs[jp.littlesoft.data.event.ResolvingEvent.Type.UPDATE.ordinal()] = MessageResource.getMessage(null, "ERROR_RESOLVING_UPDATE", null, null);
    updateErrorMsgs[jp.littlesoft.data.event.ResolvingEvent.Type.INSERT.ordinal()] = MessageResource.getMessage(null, "ERROR_RESOLVING_INSERT", null, null);

    LSException we = this;
    while (we != null) {
      // getMessage()はLocaleを無視する。
      String errmsg = we.getMessage();
      if (errmsg != null) {
        for (int i=0; i<updateErrorMsgs.length; i++) {
          if (errmsg.indexOf(updateErrorMsgs[i]) >= 0) {
            return i;
          }
        }
      }

      Throwable e = we.getCause();
      if (!(e instanceof LSException))
        return -1;
      we = (LSException)e;
    }
    return -1;
  }

  /**
   * メッセージリソースのロケールを設定します。<br/><br/>
   * 
   * @param locale メッセージリソースのロケール
   */
  public void setLocale(Locale locale) {
    this.locale = locale;
    if (this.getCause() instanceof LSException) {
      // 原因となったThrowableがある場合は、そちらにもLocaleを設定する。
      ((LSException)this.getCause()).setLocale(locale);
    }
  }
  
  /**
   * メッセージリソースのロケールを取得します。<br/>
   * 
   * @return メッセージリソースのロケール
   */
  public Locale getLocale() {
    return this.locale;
  } 

  /**
   * エラーメッセージを取得します。<br/><br/>
   * 
   * メッセージリソースが設定されている場合、デフォルトロケールに応じたメッセージを取得します。<br/>
   * 
   * @return エラーメッセージ
   */
  public String getMessage() {
    if (this.msgResource == null) {
      return super.getMessage();
    }
    return getMessage(null);
  }
  
  /**
   * エラーメッセージを取得します。<br/><br/>
   * 
   * メッセージリソースが設定されている場合、ロケールに応じたメッセージを取得します。<br/>
   * 
   * @return エラーメッセージ
   */
  public String getLocalizedMessage() {
    return getMessage(getLocale());
  }
  
  /**
   * エラーメッセージを取得します。<br/><br/>
   * 
   * メッセージリソースが設定されている場合、ロケールに応じたメッセージを取得します。<br/>
   * 
   * @param locale ロケール
   * @return エラーメッセージ
   */
  private String getMessage(Locale locale) {
    if (this.msgResource == null) {
      return super.getMessage();
    } else {
      return this.msgResource.getMessage(locale);
    }
  }

}
