/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.util.*;
import java.util.regex.*;

import jp.littlesoft.data.ITableDef.ColumnDef;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.data.validator.DefaultValidator;
import jp.littlesoft.data.validator.Validator;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.sql.SQL;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;

/**
 * カラム定義クラス。<br/><br/>
 *
 * テーブルオブジェクトを構成する列の属性を保持するクラスです。<br/><br/>
 *
 * このクラス自体にはデータ値を持ちませんが、データ項目関する様々な属性を設定することができます。<br/>
 * 名前と型は必須のプロパティです。名前は任意のStringで設定し、型はColumn.Typeクラスの定数を使用して設定します。<br/><br/>
 * 
 * @see jp.littlesoft.data.Table
 * @see jp.littlesoft.data.Params
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class Column implements Serializable, Cloneable {

  private static final long serialVersionUID = 8911427062438774472L;
  private static final DefaultValidator defaultValidator = new DefaultValidator();
  //private AbstractTable parent = null;
  private String name;
  // true:0なら表示しない, false:0なら"0"と表示
  private boolean zeroSuppress = false;
  // チェック時に使用する正規表現
  private String regexp = null;
  // コンパイル済の正規表現のパターン 
  private Pattern regexpPattern = null;
  // パラメータタイプ. ディフォルトはIN パラメータ
  private ParamType paramType = ParamType.IN;
  // SQL文中にこのColumnがパラメタとして使用されている場合の位置を格納する内部変数
  private ArrayList<Integer> paramPos = null;
  
  /** 
   * TIMESTAMP型Columnのディフォルトフォーマットパターンを表す定数です。<br/><br/>
   *
   * final ではないので必要に応じて変更する事ができます。<br/>
   * デフォルト値: "yyyy/MM/dd HH:mm:ss.SSS"<br/>
   * 書き方については java.text.SimpleDateFormat クラスを参照してください。<br/>
   */
  public static String DEFAULT_FORMAT_PATTERN_TIMESTAMP = "yyyy/MM/dd HH:mm:ss.SSS";
  
  /** 
   * TIME型Columnのディフォルトフォーマットパターンを表す定数です。<br/><br/>
   *
   * final ではないので必要に応じて変更する事ができます。<br/>
   * デフォルト値: "HH:mm:ss"<br/>
   * 書き方については java.text.SimpleDateFormat クラスを参照してください。<br/>
   */
  public static String DEFAULT_FORMAT_PATTERN_TIME = "HH:mm:ss";
  
  /**
   * DATE型Columnのディフォルトフォーマットパターンを表す定数です。<br/><br/>
   *
   * final ではないので必要に応じて変更する事ができます。<br/>
   * デフォルト値: "yyyy/MM/dd"<br/>
   * 書き方については java.text.SimpleDateFormat クラスを参照してください。<br/>
   */
  public static String DEFAULT_FORMAT_PATTERN_DATE = "yyyy/MM/dd";
  
  public static String DEFAULT_REGEX_DATE = "\\d{4}/\\d{1,2}/\\d{1,2}";
  
  /**
   * Column名を保持する内部変数です。<br/>
   */
  //private String name = null;
  
  /**
   * データベースのカラム名を保持する内部変数です。<br/>
   */
  private String dbColumnName = null;
  
  /**
   * 見出し名を保持する内部変数です。<br/>
   */
  private String caption = null;
  
  /**
   * データタイプを保持する内部変数です。<br/>
   */
  private Type type = Type.UNKNOWN;
  
  /**
   * SQLデータタイプを保持する内部変数です。<br/>
   */
  private int sqlType = Const.NOT_ASSIGNED;
  
  /**
   * 表示幅（桁数）を保持する内部変数です。<br/>
   */
  private int displaySize = Const.NOT_ASSIGNED;
  
  /**
   * 数値項目の桁数を保持する内部変数です。<br/>
   */
  private int precision = Const.NOT_ASSIGNED;
  
  /**
   * 数値項目の小数点以下の桁数を保持する内部変数です。<br/>
   */
  private int scale = Const.NOT_ASSIGNED;
  
  /**
   * 文字列最大長を保持する内部変数です。<br/>
   */
  private int length = Const.NOT_ASSIGNED;
  
  /**
   * NULL値を許可するか否かのフラグを保持する内部変数です。<br/>
   */
  private Boolean nullable = null;
  
  /**
   * 読み取り専用か否かのフラグを保持する内部変数です。<br/>
   */
  private boolean readOnly = false;
  
  /**
   * 表示するか否かのフラグを保持する内部変数です。<br/>
   */
  private boolean visible = true;
  
  /**
   * JDBC ResultSetから自動生成されたか否かのフラグを保持する内部変数です。<br/>
   */
  private boolean fromDatabase = false;
  
  /**
   * 行を識別する項目か否かのフラグを保持する内部変数です。<br/>
   */
  private boolean resolveKey = false;
  
  /**
   * データが更新されたか否かのフラグを保持する内部変数です。<br/>
   */
  private boolean updated = false;
  
  /**
   * フォーマットパターンを保持する内部変数です。<br/>
   */
  private String formatPattern = null;
  
  /**
   * フォーマットパターンを保持する内部変数です。<br/>
   */
  private String parseRegex = null;
  
  /**
   * Columnの最大長より短い値の場合に空白文字で最大長分満たすか否かのフラグを保持する内部変数です。<br/>
   */
  private boolean fillChar = false;
  
  /**
   * Columnの値に対する制約条件。<br/><br/>
   * デフォルト値は、Column.Constraint.DEFAULT<br/>
   */
  private Constraint constraint = Constraint.DEFAULT;
  
  /**
   * 文字列値に対するトリム処理の指定。<br/><br/>
   *
   * デフォルト値は、Column.TrimType.NOTRIM<br/>
   */
  private TrimType trimType = TrimType.NOTRIM;
  
  /**
   * 値検査用オブジェクト。<br/>
   */
  private Validator validator = null;
  
  /**
   * デフォルト値。 <br/>
   */
  private String defaultValue;
  
  /**
   * DBサーバー側の自動連番か？<br/>
   */
  private boolean autoIncrement = false;
  
  /**
   * 数値に対するマイナス値を許可するか？<br/>
   */
  private boolean permitMinusValue = true;

  /**
   * エンティティオブジェクト項目の型とDBカラムの型が異なる場合、寛容な変換を行うか?.<br/>
   */
  private boolean lenientConvert = true;
  
  transient Field entityField = null;
  transient Method entityGetter = null;
  transient Method entitySetter = null;
  private boolean upperCaseDBColumnName = true;
  private String oldDBColumnName = null;
  private InputType defaultInputEditor = null;
  private String defaultInputStyle = null;
  private Class<? extends Code> optionType = null;
  
  /**
   * Column オブジェクトを生成します。<br/>
   */
  public Column() {}
  
  /**
   * 引数 name で指定されたカラム名で Column オブジェクトを生成します。<br/>
   *
   * @param  name  カラム名
   */
  public Column(String name) {
    try {
      setName(name);
    } catch (DataException ex) {
      ex.printStackTrace();
    }
  }
  
  /**
   * 引数 name、type で指定されたカラム名、データ型で Column オブジェクトを生成します。<br/><br/>
   * 
   * データ型はjp.littlesoft.data.Column.Typeで定義された定数を設定してください。<br/>
   *
   * @param  name  カラム名
   * @param  type  データ型
   * @see Column.Type
   */
  public Column(String name, Column.Type type) {
    this(name);
    setType(type);
  }
  
  /**
   * 引数 name、type、paramType で指定されたカラム名、データ型、パラメータ型で Column オブジェクトを生成します。<br/><br/>
   *
   * データ型はjp.littlesoft.data.Column.Typeで定義された定数を設定してください。<br/>
   * パラメータ型はjp.littlesoft.data.Column.ParamTypeで定義された定数を設定してください。<br/>
   *
   * @param  name      カラム名
   * @param  type      データ型
   * @param  paramType パラメータ型
   * @see Column.Type
   * @see Column.ParamType
   */
  public Column(String name, Column.Type type, ParamType paramType) {
    this(name);
    this.setType(type);
    this.setParamType(paramType);
  }
  
  /**
   * 引数 name、type、newFromResultSet で指定されたカラム名、データ型で Column オブジェクトを生成します。<br/><br/>
   *
   * newFromResultSet：true は問合わせ結果で自動的に生成されたカラム名である事を表します。<br/><br/>
   *
   * データ型はjp.littlesoft.data.Column.Typeで定義された定数を設定してください。<br/>
   * 
   * @param  name              Column名
   * @param  type              データタイプ
   * @param  newFromResultSet  true：JDBC ResultSetから自動生成された
   * @see Column.Type
   */
  public Column(String name, Column.Type type, boolean newFromResultSet) {
    this(name);
    setType(type);
    fromDatabase = newFromResultSet;
  }
  
  /**
   * Javaクラスのフィールドをもとにカラムオブジェクトを作成します。
   * @param field
   */
  public Column(Field entityField, boolean upperCaseDBColumnName) {
    this(entityField.getName(), classToColtypeMap(entityField.getType()));
    this.entityField = entityField;
    this.upperCaseDBColumnName = upperCaseDBColumnName;
    setReadOnly(Modifier.isFinal(entityField.getModifiers()));
    ColumnDef def = null;
    if (entityField.isAnnotationPresent(ColumnDef.class))
      def = (ColumnDef)entityField.getAnnotation(ColumnDef.class);
    defineColumnAttribute(def);
  }
  
  /**
   * JavaクラスのGetter,Setterをもとにカラムオブジェクトを作成します。
   * @param entityGetter
   * @param entitySetter
   */
  public Column(Method entityGetter, Method entitySetter, boolean upperCaseDBColumnName) {
    this();
    String getterName = entityGetter.getName();
    name = (getterName.startsWith("is")) ? 
      getterName.substring("is".length()) : getterName.substring("set".length());
    if (name.length()==0) throw new RuntimeException("[set] is a invalid method name.");
    if (Character.isUpperCase(name.charAt(0))) {
      for (int i=1; i<name.length(); i++) {
        if (Character.isLowerCase(name.charAt(i))) {
          name = name.substring(0,i).toLowerCase() + name.substring(i);
          break;
        }
      }
    }
    //this.type = classToColtypeMap.get(entityGetter.getReturnType());
    this.type = classToColtypeMap(entityGetter.getReturnType());
    this.entityGetter = entityGetter;
    this.entitySetter = entitySetter;
    setReadOnly(this.entitySetter==null);
    ColumnDef def = null;
    if (entityGetter.isAnnotationPresent(ColumnDef.class))
      def = (ColumnDef)entityGetter.getAnnotation(ColumnDef.class);
    
    if (def==null && entitySetter!=null &&
        entitySetter.isAnnotationPresent(ColumnDef.class))
      def = (ColumnDef)entitySetter.getAnnotation(ColumnDef.class);
    this.upperCaseDBColumnName = upperCaseDBColumnName;
    defineColumnAttribute(def);
    
  }
  
  private void defineColumnAttribute(ColumnDef def) {
    if (entityField==null && entityGetter==null) return;
    if (def!=null) {
      //アノテーションからカラム属性をプリセット
      if (!"".equals(def.oldName())) setOldDBColumnName(def.oldName());
      if (!"".equals(def.caption())) setCaption(def.caption());
      if (!"".equals(def.formatPattern())) setFormatPattern(def.formatPattern());
      if (!"".equals(def.defaultValue())) setDefaultValue(def.defaultValue());
      if (!def.nullable().equals(Nullable.DEFAULT)) 
        setNullable( def.nullable().equals(Nullable.TRUE) ? true : false );
      if (getType().equals(Type.BIGDECIMAL) && def.precision()>0) setPrecision(def.precision());  
      if (getType().equals(Type.BIGDECIMAL) && def.scale()>-1) setScale(def.scale());  
      if (getType().equals(Type.STRING) && def.length()>0) setLength(def.length());
      setLenientConvert(def.lenientConvert());
      if (!def.stringConstraint().equals(Constraint.DEFAULT)) setConstraint(def.stringConstraint());
      if (!def.defaultInputEditor().equals(InputType.NONE)) setDefaultInputEditor(def.defaultInputEditor());
      if (!"".equals(def.defaultInputStyle())) setDefaultInputStyle(def.defaultInputStyle());
      
      if (!def.optionType().equals(Code.class)) setOptionType(def.optionType());
      
      //DBカラム名を決定
      if (!"".equals(def.name()))       //アノテーションで指定されたカラムを設定する
        setDBColumnName(def.name()); 
      else                              //クラスフィールド名をもとにアンダースコア区切りに自動変換した名前を設定する
        setDBColumnName(SQL.camelToUnderscore(name, isUpperCaseDBColumnName()));
    } else {
      setDBColumnName(SQL.camelToUnderscore(name, isUpperCaseDBColumnName()));
    }
  }
  
  public Class<?> getObjectType() {
    if (entityField!=null) {
      return entityField.getType();
    } if (entityGetter!=null) {
      return entityGetter.getReturnType();
    } else {
      return null;
    }  
  }
  
  /**
   * このカラムのクローンを取得します。<br/><br/>
   *
   * クローンの親となる Table オブジェクトは null となります。<br/>
   *
   * @return Column オブジェクトのクローン
   */
  public Object clone() {
    try {
      Column df = (Column)super.clone();
      return df;
    } catch (Exception ex) {
      ex.printStackTrace();
      return null;
    }
  }
  
  /**
   * カラム名を設定します。<br/><br/>
   * 
   * <ul>
   *  <li> Table オブジェクト内のカラム名はユニークに設定する必要があります。</li>
   *  <li> "@OLD@_"で始まるカラム名を設定しないでください。</li>
   * </ul><br/>
   *
   * @param  name  カラム名
   * @throws DataException 例外発生時
   */
  public void setName(String name) throws DataException {
    //if (this.parent!=null) this.parent.checkDeactive();
    this.name = name;
  }
  
  /**
   * 対応するデータベース列名を設定します。<br/><br/>
   * 
   * 未設定の場合はカラム名がデータベース列名として適用されます。<br/><br/>
   * 
   * <em>※注意：</em>
   * 異なる複数のカラムに同一のデータベース列名を設定しないでください。<br/><br/>
   * 
   * @param dbColumnName データベース列名
   */
  public void setDBColumnName(final String dbColumnName) {
    this.dbColumnName = dbColumnName;
  }
  
  /**
   * 対応するデータベース列名を取得します。<br/>
   * 
   * @return データベース列名
   */
  public String getDBColumnName() {
    if (this.dbColumnName == null)
      return Const.NULLSTRING;
    return this.dbColumnName;
  }
  
  /**
   * 見出し名を取得します。<br/>
   *
   * @return  見出し名
   */
  public String getCaption() {
    if (StringUtil.isNullString(this.caption)) return this.getName();
    return this.caption;
  }
  
  /**
   * 見出し名を設定します。<br/><br/>
   *
   * @param  newCaption  見出し名
   *
   * @see #getCaption()
   */
  public void setCaption(String newCaption) {
    caption = newCaption;
  }
  
  /**
   * カラムのデータ型を設定します。<br/>
   * 
   * データ型はjp.littlesoft.data.Column.Typeで定義された定数を設定してください。<br/><br/>
   *
   * <em>※注意：</em>
   * 親である Table オブジェクトがアクティブの場合にこのメソッドを呼び出すと、
   * 親である Table Dataオブジェクトは非アクティブになるので注意してください。<br/>
   *
   * @param  newType  データ型
   * @see jp.littlesoft.data.Column.Type
   */
  public void setType(Column.Type newType) {
    // 親であるDataオブジェクトが active の場合、inactiveにします。
    /*
    {
      final AbstractTable _parentData = this.getParent();
      if (_parentData != null && _parentData.isActive()) {
        try {
          _parentData.setActive(false);
        } catch (DataException ex) {
          ex.printStackTrace();
        }
      }
    }*/

    type = newType;
    if (type==null) {
      type = Type.UNKNOWN;
    }
    if (displaySize == Const.NOT_ASSIGNED) {
      switch (type) {
        case TIMESTAMP :
          displaySize = DEFAULT_FORMAT_PATTERN_TIMESTAMP.length();
          break;
        case DATE :
          displaySize = DEFAULT_FORMAT_PATTERN_DATE.length();
          break;
        case TIME :
          displaySize = DEFAULT_FORMAT_PATTERN_TIME.length();
          break;
      }
    }

    if (formatPattern == null || Const.NULLSTRING.equals(formatPattern)) {
      switch (type) {
        case TIMESTAMP :
          formatPattern = DEFAULT_FORMAT_PATTERN_TIMESTAMP;
          break;
        case DATE :
          formatPattern = DEFAULT_FORMAT_PATTERN_DATE;
          parseRegex = DEFAULT_REGEX_DATE;
          break;
        case TIME :
          formatPattern = DEFAULT_FORMAT_PATTERN_TIME;
          break;
      }
    }
  }
  
  /**
   * カラムのデータ型を取得します。<br/>
   *
   * @return  データ型
   * @see jp.littlesoft.data.Column.Type
   */
  public Column.Type getType() {
    return type;
  }
  
  /**
   * カラムのSQLデータ型を設定します。<br/><br/>
   *
   * SQLデータ型はjava.sql.Types で定義された定数です。<br/>
   *
   * @param  newSqlType  SQLデータ型
   * @see java.sql.Types
   */
  public void setSqlType(int newSqlType) {
    sqlType = newSqlType;
  }
  
  /**
   * カラムのSQLデータ型を取得します。<br/>
   *
   * @return  SQLデータ型
   * @see jp.littlesoft.data.Column.Type#mapToSql
   * @see java.sql.Types
   */
  public int getSqlType() {
    if (sqlType == Const.NOT_ASSIGNED) {
      try {
        return Column.mapToSql(type);
      } catch (DataException x) {
        x.printStackTrace();
        return java.sql.Types.OTHER;
      }
    } else
      return sqlType;
  }
  
  /**
   * カラムの表示幅（桁数）を設定します。<br/><br/>
   *
   * データ表示コンポーネントの表示幅に影響します。<br/>
   * SQLDataProviderやWebSQLTableではJDBCメタデータから適切な値が自動設定されます。<br/>
   *
   * @param  newDisplaySize  表示幅（桁数）
   */
  public void setDisplaySize(int newDisplaySize) {
    displaySize = newDisplaySize;
  }
  
  /**
   * カラムの表示幅（桁数）を取得します。<br/><br/>
   *
   * @return  表示幅（桁数）
   */
  public int getDisplaySize() {
    return displaySize;
  }
  
  /**
   * カラムの数値項目の桁数を設定します。<br/><br/>
   *
   * SQLDataProviderやWebSQLTableではJDBCメタデータから適切な値が自動設定されます。<br/>
   *
   * @param  newPrecision  数値項目の桁数
   */
  public void setPrecision(int newPrecision) {
    precision = newPrecision;
  }
  
  /**
   * カラムの数値項目の桁数を取得します。<br/>
   *
   * @return  数値項目の桁数
   */
  public int getPrecision() {
    return precision;
  }
  
  /**
   * カラムの数値項目の小数点以下桁数を設定します。<br/><br/>
   *
   * SQLDataProviderやWebSQLTableではJDBCメタデータから適切な値が自動設定されます。<br/>
   *
   * <em>※注意：</em>
   * 0 以上を指定してください。-1 の場合は未定として扱われます。<br/>
   * 
   * @param  newScale  数値項目の小数点以下桁数
   * @see    Row#getBigDecimal
   */
  public void setScale(int newScale) {
    scale = newScale;
  }
  
  /**
   * カラムの数値項目の小数点以下桁数を取得します。<br/>
   *
   * @return  数値項目の小数点以下桁数
   */
  public int getScale() {
    return scale;
  }
  
  /**
   * カラムがNULL値を許可するかどうかを設定します。<br/><br/>
   *
   * SQLDataProviderやWebSQLTableではJDBCメタデータから適切な値が自動設定されます。<br/>
   *
   * @param  newNullable  True：許可する、False：許可しない
   */
  public void setNullable(boolean newNullable) {
    nullable = newNullable;
  }
  
  /**
   * カラムがNULL値を許可するかどうかを調べます。<br/><br/>
   * 
   * @return  True：許可する、False：許可しない
   */
  public boolean isNullable() {
    return (nullable == null) ? true : nullable;
  }
  
  /**
   * カラムがNULL値を許可するかどうかを調べます。<br/><br/>
   * ※isNullable()との違いは、明示的な設定が行われていない場合、nullを返す点です。
   * @return True：許可する、False：許可しない、null未設定
   */
  public Boolean getNullable() {
    return nullable;
  }
  
  /**
   * カラムが読取専用かどうかを設定します。<br/>
   *
   * @param  newReadOnly  True：読取専用、False：読取専用ではない
   */
  public void setReadOnly(boolean newReadOnly) {
    this.readOnly = newReadOnly;
  }
  
  /**
   * カラムが読取専用かどうかを調べます。<br/>
   *
   * @return  True：読取専用、False：読取専用ではない
   */
  public boolean isReadOnly() {
    return this.readOnly;
  }
  
  /**
   * カラムが可視項目かどうかを設定します。<br/><br/>
   *
   * 当Columnが表示コンポーネントにおいて、非表示として扱われるか？を設定します。<br/>
   *
   * @param  newVisible  True：可視、False：不可視
   */
  public void setVisible(boolean newVisible) {
    visible = newVisible;
  }
  
  /**
   * カラムが可視項目かどうかを調べます。<br/>
   *
   * @return  True：可視、False：不可視
   */
  public boolean isVisible() {
    return visible;
  }
  
  /**
   * カラムの文字列最大長を設定します。<br/><br/>
   *
   * String型カラムの最大バイト数を設定します。<br/>
   * SQLDataProviderやWebSQLTableではJDBCメタデータから適切な値が自動設定されます。<br/>
   * 設定されたバイト数以上の文字列が入力された場合、DataExceptionがスローされます。<br/>
   *
   * @param  newLength  文字列最大長
   */
  public void setLength(int newLength) {
    length = newLength;
  }
  
  /**
   * カラムの文字列最大長を取得します。<br/>
   *
   * @return  文字列最大長
   */
  public int getLength() {
    return length;
  }
  
  /**
   * カラムのフォーマットパターンを設定します。<br/><br/>
   *
   * yyyy/MM/dd、#,##0、など日付や数値のフォーマットを指定します。<br/><br/>
   * 
   * 詳細はjava.text パッケージの文書を参照してください。<br/>
   * 和暦のフォーマットについては、DataUtilクラスの文書を参照してください。<br/><br/>
   *
   * @param  newFormatPattern  フォーマットパターン
   */
  public void setFormatPattern(String newFormatPattern) {
    formatPattern = newFormatPattern;

    switch (type) {
      case DATE :
        if (DEFAULT_FORMAT_PATTERN_DATE.equals(formatPattern)) {
          parseRegex = DEFAULT_REGEX_DATE;
        }
        else {
          parseRegex = Const.NULLSTRING;
        }
        break;
    }
  }
  
  /**
   * カラムのフォーマットパターンを取得します。<br/>
   *
   * @return  フォーマットパターン
   */
  public String getFormatPattern() {
    if (formatPattern == null) return Const.NULLSTRING;
    return formatPattern;
  }
  
  /**
   * カラムのフォーマットパターンを取得します。<br/>
   *
   * @return  フォーマットパターン
   */
  public String getParseRegex() {
    if (parseRegex == null) return Const.NULLSTRING;
    return parseRegex;
  }
  
  /**
   * カラムが問合わせ結果から生成されたかどうかを設定します。<br/>
   *
   * @param  newFromDataBase  True：問合わせ結果(JDBC ResultSet)から生成された
   */
  public void setFromDatabase(boolean newFromDataBase) {
    fromDatabase = newFromDataBase;
  }
  
  /**
   * カラムが問合わせ結果から生成されたかどうかを調べます。<br/>
   *
   * @return  True：問合わせ結果(JDBC ResultSet)から生成された
   */
  public boolean isFromDataBase() {
    return fromDatabase;
  }
  
  /**
   * カラムが行を識別するキー項目であるかを設定します。<br/><br/>
   *
   * SQLDataResolverは、このプロパティをもとにSQL UPDATE文、DELETE文のWHERE句を自動生成します。<br/>
   * 一般的にはテーブルのPrimaryKey項目がこれにが該当します。<br/>
   * 
   * @param  resolveKey  True：行を識別する
   */
  public void setResolveKey(boolean resolveKey) {
    this.resolveKey = resolveKey;
  }
  
  /**
   * カラムが行を識別するキー項目であるかどうかを調べます。<br/>
   * 
   * @return  True：行を識別する
   */
  public boolean isResolveKey() {
    return this.resolveKey;
  }
  
  /**
   * カラムの値が変更されたかどうかを設定します。<br/><br/>
   *
   * カラムが変更された場合、このプロパティを ture に設定します。<br/>
   * SQLDataResolverでは、このプロパティをもとにSQL UPDATE文の更新カラムリストを自動生成します。<br/>
   *
   * @param  newUpdated  True：変更された
   */
  public void setUpdated(boolean newUpdated) {
    updated = newUpdated;
  }
  
  /**
   * カラムの値が変更されたかどうかを調べます。<br/>
   *
   * @return  True：変更された
   */
  public boolean isUpdated() {
    return updated;
  }
  
  /**
   * カラムにLengthのバイト数分、文字で満たすかどうかを設定します。<br/><br/>
   *
   * true に設定されたカラムはLengthのバイト数分、文字で満たすことが必須になります。 <br/>
   * 空白でない文字でLengthのバイト数分、値が設定されている事を検証します。<br/>
   *
   * @param  newFillChar  True：文字で満たす
   */
  public void setFillChar(boolean newFillChar) {
    fillChar = newFillChar;
  }
  
  /**
   * カラムをLengthのバイト数分、文字で満たすかどうかを調べます。<br/>
   *
   * @return  True：文字で満たす
   */
  public boolean isFillChar() {
    return fillChar;
  }
  
  /**
   * 文字列値の検証方法を設定します。<br/>
   *
   * @param constraint 制約の種別
   * @see Column.Constraint
   */
  public void setConstraint(Constraint constraint) {
    this.constraint = constraint;
  }
  
  /**
   * 文字列値の検証方法を取得します。<br/>
   *
   * @return 制約の種別
   * @see Column.Constraint
   */
  public Constraint getConstraint() {
    return this.constraint;
  }
  
  /**
   * 文字列値設定時の空白除去方法を設定します。<br/><br/>
   *
   * 設定したトリム処理の種別に応じて、table.setString()の際にトリム処理が行われます。<br/><br/>
   * 
   * @param trimType トリム処理の種別
   * @see Column.TrimType
   */
  public void setTrimType(TrimType trimType) {
    this.trimType = trimType;
  }
  
  /**
   * 文字列値設定時の空白除去方法を取得します。<br/>
   *
   * @return トリム処理の種別
   * @see Column.TrimType
   */
  public TrimType getTrimType() {
    return trimType;
  }
  
  /**
   * カラムに Validator オブジェクトを設定し、値の検査処理を拡張します。<br/><br/>
   * 
   * このメソッドで独自の値検査ロジックを設定した場合、デフォルトの振る舞いは
   * 無効化されます。<br/>
   *
   * @param validator Validator オブジェクト
   */
  public void setValidator(Validator validator) {
    this.validator = validator;
  }
  
  /**
   * カラムに設定されている Validator オブジェクトを取得します。<br/>
   *
   * @return Validator オブジェクト
   */
  public Validator getValidator() {
    if (this.validator==null) {
      return defaultValidator;
    } else {
      return this.validator;
    }
  }
  
  /*
  private void setParent(AbstractTable data) {
    this.parent = data;
  }
  
  private AbstractTable getParent() {
    return this.parent;
  }*/
  
  /**
   * ゼロサプレスするかどうかを設定します。<br/><br/>
   *
   * 数値型のカラムのときに値が 0 の場合に表示するかどうかを設定します。<br/><br/>
   * 
   * カラムが数値型でない場合に設定しても意味はありません。<br/>
   *
   * @param zeroSuppress True：ゼロサプレスします、False：ゼロサプレスしない
   */
  public void setZeroSuppress(boolean zeroSuppress) {
    this.zeroSuppress = zeroSuppress;
  }
  
  /**
   * ゼロサプレスするかどうかを調べます。<br/><br/>
   *
   * @return True：ゼロサプレスします、False：ゼロサプレスしない
   */
  public boolean isZeroSuppress() {
    return this.zeroSuppress;
  }
  
  /**
   * チェック用の正規表現において、先頭と末尾が'/'の場合に、その2文字を取り除くかどうかを表します。<br/><br/>
   * 
   * True：取り除く、False：取り除かない
   */
  public boolean isRemoveRegexpHeadTail = false;
  
  /**
   * カラムのチェック用の正規表現を設定します。<br/><br/>
   * 
   * 許可する文字列の正規表現を設定します。<br/>
   * マッチングにはJDKのjava.util.regexpパッケージを使用します。<br/>
   * 正規表現のフォーマットなどについてはJDKのマニュアルを参照してください。
   * 
   * <pre>
   * 例:
   *   column.setRegexp("^[a-z]+\\d+$");<br/>
   * </pre>
   * 
   * <em>※注意：</em>
   * 正規表現によるマッチングが行われるのはデータ型がSTRINGで、かつ長さ1以上の文字列が設定された場合のみです。<br/>
   *
   * @param regexp チェック用の正規表現
   */
  public void setRegexp(String regexp) {
    if (regexp == null) {
      this.regexp = null;
    } else {
      final int _len = regexp.length();
      if (_len < 3) {
        this.regexp = regexp;
      } else {
        // RemoveRegexpHeadTail=trueで、かつ先頭と末尾が'/'なら、その2文字を取り除く。
        if (isRemoveRegexpHeadTail && regexp.charAt(0) == '/' && regexp.charAt(_len-1) == '/') {
          this.regexp = regexp.substring(1, _len-1);
        } else {
          this.regexp = regexp;
        }
      }
    }
    if (StringUtil.isNullString(this.regexp)) {
      this.regexpPattern = null;
    } else {
      try {
        // あらかじめ、コンパイルしたパターンを設定しておく。
        this.regexpPattern = Pattern.compile(this.regexp);
      } catch (Throwable t) {
        this.regexpPattern = null;
        t.printStackTrace();
      }
    }
  }
  
  /**
   * カラムのチェック用の正規表現を取得します。<br/>
   *
   * @return チェック用の正規表現
   */
  public String getRegexp() {
    if (this.regexp == null)
      return Const.NULLSTRING;
    return this.regexp;
  }
  
  /**
   * カラムのチェック用のコンパイル済正規表現パターンを取得します。<br/>
   *
   * @return コンパイル済正規表現パターン
   */
  public Pattern getRegexpPattern() {
  	return this.regexpPattern;
  }
  
  /**
   * カラムのパラメータ型を設定します。<br/>
   *
   * @param  newParamType  パラメータ型
   * @see Column.ParamType
   */
  public void setParamType(ParamType newParamType) {
    paramType = newParamType;
  }
  
  /**
   * カラムのパラメータ型を取得します。<br/>
   *
   * @return  パラメータ型
   * @see Column.ParamType
   */
  public ParamType getParamType() {
    return paramType;
  }
  
  /**
   * SQL文中での当カラムのパラメータが位置する序数を設定します。<br/>
   *
   * @param  newParamPosArray  パラメータ位置情報リスト
   */
  public void setParamPos(int[] newParamPosArray) {
    if (newParamPosArray == null) {
      paramPos = null;
      return;
    }
    if (paramPos == null) {
      paramPos = new ArrayList<Integer>();
    } else {
      paramPos.clear();
    }
    for (int i = 0; i < newParamPosArray.length; i++) {
      addParamPos(newParamPosArray[i]);
    }
  }
  
  /**
   * SQL文中での当カラムのパラメータが位置する序数を設定（追加）します。<br/>
   *
   * @param  newParamPos  パラメータ位置
   */
  public void addParamPos(int newParamPos) {
    if (paramPos == null) {
      paramPos = new ArrayList<Integer>();
    }
    paramPos.add(new Integer(newParamPos));
  }
  
  /**
   * SQL文中での当カラムのパラメータが位置する序数を取得します。<br/>
   *
   * @return  パラメータ位置情報リスト
   */
  public int[] getParamPos() {
    if (paramPos == null || paramPos.size() == 0)
      return null;
    int[] pos = new int[paramPos.size()];
    for (int i = 0; i < paramPos.size(); i++) {
      pos[i] = ((Integer) paramPos.get(i)).intValue();
    }
    return pos;
  }
  
  /**
   * カラムのデフォルト値を設定します。<br/><br/>
   * 
   * Table#insertRow()を行った場合に、新規行にデフォルト値が設定されます。<br/>
   *
   * @param defaultValue デフォルト値
   */
  public void setDefaultValue(String defaultValue) {
    this.defaultValue = defaultValue;
  }
  
  /**
   * カラムのデフォルト値を取得します。<br/>
   * 
   * @return デフォルト値
   */
  public String getDefaultValue() {
    if (this.defaultValue == null)
      return Const.NULLSTRING;
    return this.defaultValue;
  }
  
  /**
   * カラムのカラム名を取得します。<br/>
   * 
   * @return カラム名
   */
  public String getName() {
    return name;
  }
  
  /**
   * カラムが自動採番かどうか調べます。<br/>
   * 
   * @return True：自動採番である
   */
  public boolean isAutoIncrement() {
    return autoIncrement;
  }
  
  /**
   * カラムが自動採番かどうか設定します。<br/>
   * 
   * @param serial True：自動採番である
   */
  public void setAutoIncrement(boolean serial) {
    this.autoIncrement = serial;
  }
  
  /**
   * マイナス値を許可するかどうか調べます。<br/>
   * 
   * ※「許可する」に設定されている場合、「textBind()」時にHTMLタグに出力されるmaxlength属性値に「+1」されます。<br/>
   * ※値の妥当性チェックは行いません。アプリケーション毎にチェック処理を実装してください。<br/>
   * 
   * @return True：許可する
   */
  public boolean isPermitMinusValue() {
    return permitMinusValue;
  }

  /**
   * マイナス値を許可するかどうか設定します。<br/><br/>
   * 
   * ※「許可する」に設定した場合、「textBind()」時にHTMLタグに出力されるmaxlength属性値に「+1」されます。<br/>
   * ※値の妥当性チェックは行いません。アプリケーション毎にチェック処理を実装してください。<br/>
   * 
   * @param permitMinusValue True：許可する
   */
  public void setPermitMinusValue(boolean permitMinusValue) {
    this.permitMinusValue = permitMinusValue;
  }
  
  @SuppressWarnings("serial")
  public static class NoAccessibleObjectException extends DataException {
    public NoAccessibleObjectException(String name) {
      super("No AccessibleObject[" + name + "]");
    }
  }
  
  @SuppressWarnings("serial")
  public static class TypeMismatchException extends DataException {
    public TypeMismatchException(String name) {
      super("No AccessibleObject[" + name + "]");
    }
  }

  public String getAccessibleObjectName() {
    if (entityField != null) {
      return entityField.getName();
    } else {
      return 
        ((entitySetter != null) ? ( entitySetter.getName() + "/") : "") +
        ((entityGetter != null) ? entityGetter.getName() : "");
    }
  }
  /**
   * エンティティオブジェクトに値を設定する。
   * @param entity
   * @param value
   */
  @SuppressWarnings("unchecked")
  public void setEntityValue(Serializable entity, Serializable value) throws NoAccessibleObjectException, TypeMismatchException, NullPointerException {
    Class targetType = null;
    try {
      if (entityField != null) {
        targetType = entityField.getType();
        entityField.set(entity, lenientConvert(value,targetType));
      } else if (entitySetter != null) {
        targetType = entitySetter.getParameterTypes()[0];
        try {
          entitySetter.invoke(entity, lenientConvert(value,targetType));
        } catch (InvocationTargetException e1) {
          if (e1.getCause() instanceof NullPointerException && value == null) {
            throw (NullPointerException)e1.getCause();
          }
          throw new RuntimeException(e1);
        }
      } else {
        throw new NoAccessibleObjectException(this.name);
      }
    } catch (IllegalArgumentException e2) {
      throw new TypeMismatchException(e2.getLocalizedMessage());
    } catch (IllegalAccessException e3) {
      throw new RuntimeException(e3.getLocalizedMessage() + "[" + getName() + "]");
    }
  }
  
  @SuppressWarnings("unchecked")
  private Serializable lenientConvert(Serializable value, Class destType) {
    if (value == null || !isLenientConvert() || value.getClass().equals(destType)) return value;
    return lenientConvertNumber(value, destType);
  }
  
  public static Serializable lenientConvertNumber(Serializable value, Class<?> destType) {
    if (value instanceof Number) {
      Number numValue = (Number)value;
      if (destType.equals(Byte.class)) {
        return new Byte(numValue.byteValue());
      } else if (destType.equals(Short.class)) {
        return new Short(numValue.shortValue());
      } else if (destType.equals(Integer.class)) {
        return new Integer(numValue.intValue());
      } else if (destType.equals(Long.class)) {
        return new Long(numValue.longValue());
      } else if (destType.equals(Float.class)) {
        return new Float(numValue.floatValue());
      } else if (destType.equals(Double.class)) {
        return new Double(numValue.doubleValue());
      } else if (destType.equals(BigDecimal.class)) {
        return new BigDecimal(numValue.toString());
      }
    }
    return value;    
  }
  
  
  
  public Serializable getEntityValue(Serializable entity) throws NoAccessibleObjectException {
    try {
      if (entityField != null) {
        return (Serializable) entityField.get(entity);
      } else if (entityGetter != null) {
        try {
          return (Serializable) entityGetter.invoke(entity);
        } catch (InvocationTargetException e1) {
          throw new RuntimeException(e1);
        }
      } else {
        throw new NoAccessibleObjectException(this.name);
      }
    } catch (IllegalArgumentException e2) {
      throw new RuntimeException(e2);
    } catch (IllegalAccessException e3) {
      throw new RuntimeException(e3);
    }
  }
  
  /**
   * SQLデータ型をデータ型値にマッピングします。<br/>
   *
   * @param  sqlType  SQLデータ型
   * @return マッピングされたデータ型値
   * @throws DataException 不正なSQLデータ型の場合
   * @see java.sql.Types
   */
  public static Column.Type mapToJava(int sqlType) throws DataException {
    switch (sqlType) {
      case java.sql.Types.CHAR :
      case java.sql.Types.VARCHAR :
      case java.sql.Types.LONGVARCHAR :
        return Type.STRING;
      case java.sql.Types.BIT :
      case java.sql.Types.BOOLEAN :
        return Type.BOOLEAN;
      case java.sql.Types.TINYINT :
        return Type.BYTE;
      case java.sql.Types.SMALLINT :
        return Type.SHORT;
      case java.sql.Types.INTEGER :
        return Type.INT;
      case java.sql.Types.BIGINT :
        return Type.LONG;
      case java.sql.Types.FLOAT :
        // JDBC の仕様に基づき、java.sql.Types.FLOAT→DOUBLEにマッピングします。
        return Type.DOUBLE;
      case java.sql.Types.REAL :
        return Type.FLOAT;
      case java.sql.Types.DOUBLE :
        return Type.DOUBLE;
      case java.sql.Types.NUMERIC :
      case java.sql.Types.DECIMAL :
        return Type.BIGDECIMAL;
      case java.sql.Types.DATE :
        return Type.DATE;
      case java.sql.Types.TIME :
        return Type.TIME;
      case java.sql.Types.TIMESTAMP :
      // for Oracle JDBC Driver 9.0.x
      case -100 :
      // for Oracle JDBC Driver 9.2.x
      case -101 :
        return Type.TIMESTAMP;
      case java.sql.Types.BINARY :
      case java.sql.Types.VARBINARY :
        return Type.BINARY;
      case java.sql.Types.LONGVARBINARY :
        return Type.LONGBINARY;
      case java.sql.Types.BLOB :
        return Type.BLOB;
      case java.sql.Types.CLOB :
        return Type.CLOB;
      case java.sql.Types.OTHER:
        return Type.JSONOBJECT;
      default :
        throw new DataException(null, null, "INVALID_DATAFIELD_TYPE",
            new String[]{Integer.toString(sqlType)});
    }
  }
  
  
  /**
   * データ型値をにSQLデータ型マッピングします。<br/>
   *
   * @param   type  データ型値
   * @return  SQLデータ型
   * @throws DataException 不正なデータ型値の場合
   * @see java.sql.Types
   */
  public static String mapToSQLString(Column.Type type) throws DataException {
    switch (type) {
      case JSONOBJECT :
      case STRING :
        return "VARCHAR";
      case BOOLEAN :
        return "BOOLEAN";
      case BYTE :
        return "TINYINT";
      case SHORT :
        return "SMALLINT";
      case INT :
        return "INTEGER";
      case LONG :
        return "BIGINT";
      case FLOAT :
        return "REAL";
      case DOUBLE :
        return "DOUBLE";
      case BIGDECIMAL :
        return "DECIMAL";
      case DATE :
        return "DATE";
      case TIME :
        return "TIME";
      case TIMESTAMP :
        return "TIMESTAMP";
      case BINARY :
        return "BINARY";
      case LONGBINARY :
        return "LONGVARBINARY";
      case BLOB :
        return "BLOB";
      case CLOB :
        return "CLOB";
      default :
        throw new DataException(null, null, "INVALID_DATAFIELD_TYPE",
            new String[]{type.toString()});
    }
  }
  /**
   * データ型値をにSQLデータ型マッピングします。<br/>
   *
   * @param   type  データ型値
   * @return  SQLデータ型
   * @throws DataException 不正なデータ型値の場合
   * @see java.sql.Types
   */
  public static int mapToSql(Column.Type type) throws DataException {
    switch (type) {
      case STRING :
        return java.sql.Types.VARCHAR;
      case BOOLEAN :
        return java.sql.Types.BOOLEAN;
      case BYTE :
        return java.sql.Types.TINYINT;
      case SHORT :
        return java.sql.Types.SMALLINT;
      case INT :
        return java.sql.Types.INTEGER;
      case LONG :
        return java.sql.Types.BIGINT;
      case FLOAT :
        return java.sql.Types.REAL;
      case DOUBLE :
        return java.sql.Types.DOUBLE;
      case BIGDECIMAL :
        return java.sql.Types.NUMERIC;
      case DATE :
        return java.sql.Types.DATE;
      case TIME :
        return java.sql.Types.TIME;
      case TIMESTAMP :
        return java.sql.Types.TIMESTAMP;
      case BINARY :
        return java.sql.Types.BINARY;
      case LONGBINARY :
        return java.sql.Types.LONGVARBINARY;
      case BLOB :
        return java.sql.Types.BLOB;
      case CLOB :
        return java.sql.Types.CLOB;

        /* 以下については未検討！
              java.sql.Types.NULL:
              java.sql.Types.ARRAY:        //JDBC 2.0 SQL 配列を表すタイプです。
              java.sql.Types.DISTINCT:     //JDBC 2.0 組み込み型に基づくタイプです。
              java.sql.Types.JAVA_OBJECT:  //JDBC 2.0 Java オブジェクトを表すタイプです。
              java.sql.Types.OTHER:        //OTHER は、SQL タイプがデータベース固有のものであることを示し、getObject メソッドと setObject メソッドを介してアクセスできる Java オブジェクトにマッピングされます。
              java.sql.Types.REF:          //JDBC 2.0 SQL REF<構造化型> を表すタイプです。
              java.sql.Types.STRUCT:       //JDBC 2.0 任意のタイプである属性で構成されるタイプです。
        */
      default :
        throw new DataException(null, null, "INVALID_DATAFIELD_TYPE",
            new String[]{type.toString()});
    }
  }
  
  /**
   * 指定したデータ型が数値型かどうか調べます。<br/>
   * 
   * @param type 判定するデータ型
   * @return True：数値型、False：数値型ではない
   */
  public static boolean isNumber(Column.Type type) {
    switch (type) {
      case BIGDECIMAL:
      case BYTE:
      case SHORT:
      case LONG:
      case INT:
      case FLOAT:
      case DOUBLE:
        return true;
      default:
        return false;
    }
  }
  
  /**
   * カラムのデータ型を定義する列挙型。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum Type { 
    /**
     * データ型が不明であることを表します。
     */
    UNKNOWN,
    /**
     * STRING型であることを表します。
     */
    STRING,
    /**
     * BINARY型であることを表します。
     */
    BINARY,
    /**
     * BYTE型であることを表します。
     */
    BYTE,
    /**
     * SHORT型であることを表します。
     */
    SHORT,
    /**
     * LONG型であることを表します。
     */
    LONG,
    /**
     * INT型であることを表します。
     */
    INT,
    /**
     * FLOAT型であることを表します。
     */
    FLOAT,
    /**
     * DOUBLE型であることを表します。
     */
    DOUBLE,
    /**
     * BOOLEAN型であることを表します。
     */
    BOOLEAN,
    /**
     * BIGDECIMAL型であることを表します。
     */
    BIGDECIMAL,
    /**
     * DATE型であることを表します。
     */
    DATE,
    /**
     * TIME型であることを表します。
     */
    TIME,
    /**
     * TIMESTAMP型であることを表します。
     */
    TIMESTAMP,
    /**
     * OBJECT型であることを表します。
     */
    JSONOBJECT,
    /**
     * LONGBINARY型であることを表します。
     */
    LONGBINARY,
    /**
     * BLOB型であることを表します。
     */
    BLOB,
    /**
     * CLOB型であることを表します。
     */
    CLOB
  }
  

  @SuppressWarnings("unchecked")
  static final private HashMap<Class, Column.Type> classToColtypeMap = new HashMap();
  @SuppressWarnings("unchecked")
  static public Column.Type classToColtypeMap(Class clazz) {
    Column.Type result = classToColtypeMap.get(clazz);
    if (result==null 
        && !clazz.equals(Class.class) 
        && Serializable.class.isAssignableFrom(clazz)) {
      return Column.Type.JSONOBJECT;
    }
    return result;
  }
  
  @SuppressWarnings("unchecked")
  static final private HashMap<Column.Type, Class> colTypeToClassMap = new HashMap();
  @SuppressWarnings("unchecked")
  static public Class colTypeToClassMap(Column.Type colType) {
    return colTypeToClassMap.get(colType);
  }
  
  static {
    classToColtypeMap.put(String.class, Column.Type.STRING);
    classToColtypeMap.put(Byte.class, Column.Type.BYTE);
    classToColtypeMap.put(byte.class, Column.Type.BYTE);
    classToColtypeMap.put(Short.class, Column.Type.SHORT);
    classToColtypeMap.put(short.class, Column.Type.SHORT);
    classToColtypeMap.put(Long.class, Column.Type.LONG);
    classToColtypeMap.put(long.class, Column.Type.LONG);
    classToColtypeMap.put(Integer.class, Column.Type.INT);
    classToColtypeMap.put(int.class, Column.Type.INT);
    classToColtypeMap.put(Float.class, Column.Type.FLOAT);
    classToColtypeMap.put(float.class, Column.Type.FLOAT);
    classToColtypeMap.put(Double.class, Column.Type.DOUBLE);
    classToColtypeMap.put(double.class, Column.Type.DOUBLE);
    classToColtypeMap.put(Boolean.class, Column.Type.BOOLEAN);
    classToColtypeMap.put(boolean.class, Column.Type.BOOLEAN);
    classToColtypeMap.put(BigDecimal.class, Column.Type.BIGDECIMAL);
    classToColtypeMap.put(java.sql.Date.class, Column.Type.DATE);
    classToColtypeMap.put(java.sql.Time.class, Column.Type.TIME);
    classToColtypeMap.put(java.sql.Timestamp.class, Column.Type.TIMESTAMP);
    classToColtypeMap.put(Serializable.class, Column.Type.JSONOBJECT);
    classToColtypeMap.put(byte[].class, Column.Type.BINARY);
    classToColtypeMap.put(char[].class, Column.Type.CLOB);
    classToColtypeMap.put(null, Column.Type.UNKNOWN);

    colTypeToClassMap.put(Column.Type.UNKNOWN, null);
    colTypeToClassMap.put(Column.Type.STRING, String.class);
    colTypeToClassMap.put(Column.Type.BINARY, byte[].class);
    colTypeToClassMap.put(Column.Type.BYTE, Byte.class);
    colTypeToClassMap.put(Column.Type.SHORT, Short.class);
    colTypeToClassMap.put(Column.Type.LONG, Long.class);
    colTypeToClassMap.put(Column.Type.INT, Integer.class);
    colTypeToClassMap.put(Column.Type.FLOAT, Float.class);
    colTypeToClassMap.put(Column.Type.DOUBLE, Double.class);
    colTypeToClassMap.put(Column.Type.BOOLEAN, Boolean.class);
    colTypeToClassMap.put(Column.Type.BIGDECIMAL, BigDecimal.class);
    colTypeToClassMap.put(Column.Type.DATE, java.sql.Date.class);
    colTypeToClassMap.put(Column.Type.TIME, java.sql.Time.class);
    colTypeToClassMap.put(Column.Type.TIMESTAMP, java.sql.Timestamp.class);
    colTypeToClassMap.put(Column.Type.JSONOBJECT, Serializable.class);
    colTypeToClassMap.put(Column.Type.LONGBINARY, byte[].class);
    colTypeToClassMap.put(Column.Type.BLOB, byte[].class);
    colTypeToClassMap.put(Column.Type.CLOB, String.class);
  }
  
  @SuppressWarnings("unchecked")
  public static Serializable createDefaultValue(Column column, int dummyValue) {
    switch(column.type) {
    case UNKNOWN : return String.valueOf(dummyValue);
    case STRING : return String.valueOf(dummyValue);
    case BINARY : return new byte[0];
    case BYTE : return new Byte((byte)dummyValue);
    case SHORT : return new Short((short)dummyValue);
    case LONG : return new Long((long)dummyValue);
    case INT : return new Integer(dummyValue);
    case FLOAT : return new Float(dummyValue);
    case DOUBLE : return new Double(dummyValue);
    case BOOLEAN : return new Boolean(false);
    case BIGDECIMAL : return new BigDecimal(dummyValue);
    case DATE : return new java.sql.Date((long)dummyValue);
    case TIME : return new java.sql.Time((long)dummyValue);
    case TIMESTAMP : return new java.sql.Timestamp((long)dummyValue);
    case JSONOBJECT : 
      if (column.getObjectType().isArray()) {
        return (Serializable)LSJson.createDummyArray(column.getObjectType());
      } else {
        try {
          return (Serializable) column.getObjectType().newInstance();
        } catch (Exception e) {
          throw new RuntimeException(e);
        }
      }
    case LONGBINARY : return new byte[0];
    case BLOB : return new byte[0];
    case CLOB : return String.valueOf(dummyValue);
    default : return null;
    }
  }

  public enum Nullable {
    DEFAULT, TRUE, FALSE
  }
  
  public enum InputType { 
    NONE, TEXT, PASSWORD, TEXTAREA, CHECKBOX, RADIO, COMBOBOX, LIST, HIDDEN, BUTTON, FILE
  }
  
  /**
   * パラメータの入出力タイプを定義する列挙型。<br/><br/>
   * 
   * <pre>
   *   // ストアドプロシージャを呼び出す例(ParamType.OUT)
   *   UpdateSQL updateSQL = new UpdateSQL();
   *   updateSql.setSQL("{call proc1(:PARAM1, :RESULT)}");
   *   updateSql.getColumn("PARAM1").setType(Column.Type.STRING);
   *   updateSql.getColumn("RESULT").setType(Operation.INT);
   *   updateSql.getColumn("RESULT").setParamType(Column.ParamType.OUT);
   *   updateSql.setSTRING("PARAM1", "xxxx");
   * </pre>
   */
  public enum ParamType {
    /**
     * パラメータが入力用であることを表します。
     */
    IN,
    /**
     * パラメータが出力用であることを表します。
     */
    OUT,
    /**
     * パラメータが入出力用であることを表します。
     */
    INOUT
  }
  
  /**
   * 文字列値に対しトリム処理を指定するための列挙型。<br/>
   */
  public enum TrimType { 
    /**
     * トリム処理を行わないことを表します。
     */
    NOTRIM,
    /**
     * 文字列の両端に対してトリム処理を行うことを表します。
     */
    BOTH,
    /**
     * 文字列の左端に対してトリム処理を行うことを表します。
     */
    LEFT,
    /**
     * 文字列の右端に対してトリム処理を行うことを表します。
     */
    RIGHT
  }
  
  /**
   * 文字列値の検証方法を指定するための列挙型。<br/>
   */
  public enum Constraint {
    /**
     * デフォルトの検証を行うことを表します。
     */
    DEFAULT,
    /**
     * 文字列値が英数字であることの検証を行うことを表します。
     */
    REQUIRED_ALPHA_NUMERICS,
    /**
     * 文字列値がシングルバイト文字であることの検証を行うことを表します。
     */
    REQUIRED_SINGLE_BYTES,
    /**
     * 文字列値が2バイト文字であることの検証を行うことを表します。
     */
    REQUIRED_DOUBLE_BYTES,
    /**
     * 文字列値がカナを除くシングルバイト文字であることの検証を行うことを表します。
     */
    REQUIRED_SINGLE_BYTES_WITHOUT_KANA
  }

  /**
   * データベースカラム名が大文字か？
   * @return
   */
  public boolean isUpperCaseDBColumnName() {
    return upperCaseDBColumnName;
  }

  /**
   * データベースカラム名が大文字か？
   * @param upperCaseDBColumnName
   */
  public void setUpperCaseDBColumnName(boolean upperCaseDBColumnName) {
    if (this.upperCaseDBColumnName == upperCaseDBColumnName || dbColumnName==null) return;
    this.upperCaseDBColumnName = upperCaseDBColumnName;
    dbColumnName = (this.upperCaseDBColumnName) ? dbColumnName.toUpperCase() : dbColumnName.toLowerCase();
  }

  /**
   * エンティティオブジェクト項目の型とDBカラムの型が異なる場合、寛容な変換を行うか?.<br/>
   * @return true:許可, false:許可しない。
   */
  public boolean isLenientConvert() {
    return lenientConvert;
  }

  /**
   * エンティティオブジェクト項目の型とDBカラムの型が異なる場合、寛容な変換を行うか?.<br/>
   * trueを設定すると、java.math.BigDecimalからIntegerへの値など、値や精度を失う可能性がある場合でも、
   * 可能なかぎり寛容な値の変換を行います。<br/>
   * falseを設定すると実行時に例外が発生します。
   * @param enientConvert true:許可, false:許可しない。
   */
  public void setLenientConvert(boolean lenientConvert) {
    this.lenientConvert = lenientConvert;
  }

  /**
   * リネーム前のDBカラム名
   * @return String
   */
  public String getOldDBColumnName() {
    return oldDBColumnName;
  }

  /**
   * リネーム前のDBカラム名
   * @param oldName
   */
  public void setOldDBColumnName(String oldDBColumnName) {
    this.oldDBColumnName = oldDBColumnName;
  }

  /**
   * ディフォルトの入力コンポーネントの型
   * @return defaultInputEditor
   */
  public InputType getDefaultInputEditor() {
    return defaultInputEditor;
  }

  /**
   * ディフォルトの入力コンポーネントの型
   * @param defaultInputEditor
   */
  public void setDefaultInputEditor(InputType defaultInputEditor) {
    this.defaultInputEditor = defaultInputEditor;
  }

  /**
   * このカラムに設定する値の候補リストの型
   * @return optionType
   */
  public Class<? extends Code> getOptionType() {
    return optionType;
  }

  /**
   * このカラムに設定する値の候補リストの型
   * @param optionType optionType
   */
  public void setOptionType(Class<? extends Code> optionType) {
    this.optionType = optionType;
  }
  
  /**
   * このカラムに設定する値の候補リスト
   * @return Code Object
   */
  public Code getOption() {
    try {
      return (optionType!=null) ? optionType.newInstance() : null;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  public String getDefaultInputStyle() {
    return defaultInputStyle;
  }

  public void setDefaultInputStyle(String defaultInputStyle) {
    this.defaultInputStyle = defaultInputStyle;
  }
}

