/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.Serializable;
import java.util.AbstractList;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;

/**
 * エンティティオブジェクトリスト<br/><br/>
 *
 * 型指定されたテーブルオブジェクトをもとに作成されるエンティティ格納リストです.<br/><br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class EntityList<E extends Serializable> extends AbstractList<E> implements Serializable {
  
  private Table<E> table;
  
  /**
   * 型指定されたテーブルオブジェクトのエンティティリストを作成する.<br/>
   * @param table
   */
  public EntityList(Table<E> table) {
    this.table = table;
  }

  /**
   * エンティティのクラス<br/>
   * @return Class<E>
   */
  public Class<E> getEntityType() {
    return table.getEntityType();
  }
  
  /**
   * このリストのもとになる型指定されたテーブルオブジェクトを取得します.<br/>
   * @return Table<E>
   */
  public Table<E> getTable() {
    return table;
  }

  /**
   * entityの編集を開始します.<br/>
   * このリストのもとになる型指定テーブルにエンティティの編集を開始することを知らせます.<br/>
   * getTable()で取得したテーブルオブジェクトを使用すると、このリストで編集されたエンティティをデータベースに永続化することができます.
   * @param entity
   * @throws DataException
   */
  public void startEditing(E entity) throws DataException {
    table.startEditing(entity);
  }
  
  /**
   * entityの編集内容を検証します.<br/>
   * このリストのもとになる型指定テーブルにエンティティのvalidateメソッドを実行し、編集内容のエラー有無をチェックします.<br/>
   * @param entity
   * @throws DataException
   */
  public void validate(E entity) throws DataException {
    table.validate(entity);    
  }
  
  private void checkActive() throws DataException {
    if (!table.isActive()) table.activate();
  }

  /**
   * indexの位置にあるエンティティオブジェクトを取得します。
   */
  @Override
  public E get(int index) {
   try {
     checkActive();
     table.moveTo(index);
     return (E)table.getRow().getEntity();
   } catch (Exception x) {
     throw new RuntimeException(x);
   }
  }

  /**
   * indexの位置にエンティティオブジェクトを設定します。
   */
  @Override
  public E set(int index, E entity) {
    try {
      checkActive();
      if (table.getRow(index).getEntity()==entity) {
        //同じエンティティオブジェクトなので何もしない。
      } else {
        //旧エンティティを削除、新エンティティを追加
        remove(index);
        add(index, entity);
      }
      return entity;
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }
  
  /**
   * リストに含まれるエンティティの数.<br/>
   */
  @Override
  public int size() {
    //if (!table.isActive()) return 0;
    try {
      checkActive();
      return table.getRowCount();
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }
  
  /**
   * エンティティオブジェクトを削除する。
   */
  @Override
  public void add(int index, E entity) {
    try {
      checkActive();
      table.insertRow(index, entity);
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }

  /**
   * エンティティオブジェクトを削除する。
   */
  @SuppressWarnings("unchecked")
  @Override
  public E remove(int index) {
    try {
      checkActive();
      //Object e = table.getCaptionResource()
      Object e = table.getRow(index).getEntity();
      table.deleteRow();
      return (E)e;
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }


}
