/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;

import jp.littlesoft.data.Column.Type;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.data.exception.InvalidValueException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.IndexedMap;

/**
 * パラメータクラス<br/><br/>
 *
 * 名前をキーにした項目の定義とその値の集合です。<br/><br/>
 * 
 * SQLDataProviderやWebSQLTableでは、このクラスをSQLパラメータを格納するために利用しています。<br/>
 * 名前、データタイプ等を定義したColumnオブジェクトと、それに対応する値を保持しています。<br/><br/>
 *
 * @see Column
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class Params<E extends Serializable> extends Row<E> {

  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  /**
   * SQLのパラメータ解析時の区切り文字です。<br/><br/>
   * 
   * 出現頻度の高いもの程、前に記述してください。<br/>
   */
  private static final char[] SQL_DELIM =
    {' ', '=', '<', '>', ',', '+', '-', '*', '/', '(', ')', '\'', '|', '\t', '\n', '\r', '\f'};
  
  /**
   * SQLパラメータ形式文字です。<br/><br/>
   *
   * SQLのパラメータにつける接頭子です。(値=':')<br/>
   */
  public static final char SQL_PARAM = ':';
  
  /**
   * JDBCパラメータ形式文字です。<br/>
   */
  private static final String JDBC_PARAM = "?";
  
  private final IndexedMap<String,Column> columns = new IndexedMap<String,Column>();
  
  private Class<E> entityType = null;
  
  /**
   * Params オブジェクトを生成します。<br/>
   */
  @SuppressWarnings("unchecked")
  public Params(E... e) {
    this.status = Row.Status.CREATING;
    try {
      entityType = (Class<E>) e.getClass().getComponentType();
      if (Serializable.class.equals(entityType)) {
        entityType = null;
      }
      initializeValues();
      initializeTypedRow();
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }

  private void initializeTypedRow() {
    if (isTypedRow()) {
      List<Column> entityCols = new ITableDef.Columns(this.entityType, true);
      for (Column col : entityCols) {
        try {
          col.setNullable(true);
          addColumn(col);
        } catch (DataException e) {
          throw new RuntimeException(e);
        }
      }
    }
  }
  public String[] names() {
    String[] result = new String[columns.size()];;
    columns.keySet().toArray(result);
    return result;
  }
  
  public boolean isTypedRow() {
    return (entityType!=null && 
           !entityType.isAnnotation() &&
           !entityType.isArray() &&
           !entityType.isPrimitive() &&
           !entityType.equals(Object.class) &&
           !entityType.equals(Serializable.class) &&
           !entityType.equals(Row.class));
  }
  
  protected void initializeValues() {
    if (isTypedRow()) {
      //Entity : javaクラスで型が定義された行の場合
      try {
        entity = entityType.newInstance();
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    } else {
      values = new Serializable[0];
    }
  }

  /**
   * sourceParamsの各項目値がnullでないとき、その値を当オブジェクトにコピーします.<br/>
   * @param sourceParams
   * @throws DataException
   */
  public void copyValues(Params<?> sourceParams) throws DataException {
    Column src;
    Column dest;
    for (int i=0; i< sourceParams.getColumnCount(); i++) {
      if (sourceParams.isNull(i)) continue;
      src = sourceParams.getColumn(i);
      if (indexOfColumn(src.getName())==-1) {
        addColumn(src);
      } 
      dest = getColumn(src.getName());
      if (dest.getType().equals(Column.Type.UNKNOWN)) {
        setValue(src.getName(), sourceParams.getValue(i));
      } else {
        Class<?> t = Column.colTypeToClassMap(dest.getType());
        setValue(src.getName(), Column.lenientConvertNumber(sourceParams.getValue(i), t));
      }
    }
  }
  
  @Override
  public Column getColumn(String name) throws DataException{
    return columns.get(name);
  }
  
  @Override
  public Column getColumn(int index) {
    return columns.getValue(index);    
  }
  
  @Override
  public int indexOfColumn(String name)  {
    return columns.indexOfKey(name);
  }
  
  @Override
  public int getColumnCount()  {
    return columns.size();    
  }

  /**
   * 名前を指定してColumnオブジェクトを追加します。<br/><br/>
   *
   * <em>※注意：</em>
   * データ型、パラメータ型はそれぞれディフォルト値のColumn.Type.UNKNOWN、
   * Column.ParamType.INが設定されます。<br/>
   *
   * @param   columnName  列名
   * @return  作成された Column オブジェクト
   * @throws  DataException  例外発生時
   * @see #addColumn(String, Column.Type, Column.ParamType)
   */
  public synchronized Column addColumn(String columnName) throws DataException {
    return addColumn(columnName, Column.Type.UNKNOWN , Column.ParamType.IN);
  }
  
  /**
   * 名前、データ型を指定してColumnオブジェクトを追加します。<br/><br/>
   *
   * <em>※注意：</em>
   * パラメータ型はColumn.ParamType.INが設定されます。<br/>
   *
   * @param   columnName  列名
   * @param   dataType   データ型
   * @return  作成された Column オブジェクト
   * @throws  DataException  例外発生時
   * @see Column.Type
   * @see Column.ParamType
   * @see #addColumn(String, Column.Type, Column.ParamType)
   */
  public synchronized Column addColumn(String columnName, Type dataType) throws DataException {
    return addColumn(columnName, dataType, Column.ParamType.IN);
  }
  
  /**
   * 名前、データ型、パラメータ型を指定してColumnオブジェクトを追加します。<br/><br/>
   *
   * <pre>
   *   例:
   *   dataMap.addColumn("EMPLOYEE_ID", Column.Type.STRING, Column.ParamType.IN);
   * </pre>
   *
   * @param   columnName  列名
   * @param   dataType   データ型
   * @param   paramType  パラメータ型
   * @return  作成された Column オブジェクト
   * @throws  DataException  例外発生時
   * @see Column.Type
   * @see Column.ParamType
   */
  public synchronized Column addColumn(String columnName, Type dataType, Column.ParamType paramType) throws DataException {
    if (dataType == null) dataType = Column.Type.UNKNOWN;
    Column col = new Column(columnName, dataType, paramType);
    //this.columns.put(columnName, col);
    addColumn(col);
    return col;
  }
  
  /**
   * Column オブジェクトを追加します。<br/>
   *
   * @param  column  Column オブジェクト
   * @throws  DataException  例外発生時
   */
  public synchronized void addColumn(Column column) throws DataException {
    columns.put(column.getName(), column);
    if (isTypedRow()) {
      
    } else if (values.length!=columns.size()) {
      Serializable[] v = new Serializable[columns.size()];
      System.arraycopy(values, 0, v, 0, values.length);
      this.values = v;
    }
  }
  
  /**
   * 全 Column オブジェクトを削除します。<br/>
   * 型付けされている場合は、エンティティクラスのフィールド、セッターよりディフォルトのカラムオブジェクトが生成されます.<br/>
   *
   * @throws  DataException  例外発生時
   */
  public synchronized void clearColumns() throws DataException {
    clear();
    columns.clear();
    initializeValues();
    initializeTypedRow();
  }
  

  /**
   * SQL文に含まれているパラメータを解析し、対応する Column オブジェクトを生成します。<br/><br/>
   *
   * SQL文には "SELECT * FROM EMP WHERE EMPNO = :EMPNO" の様な形式でパラメータが指定されている必要があります。<br/>
   * 戻り値は上記":EMPNO"が"?"にリプレースされたSQL文です。(上記の例では "SELECT * FROM EMP WHERE EMPNO = ?")<br/><br/>
   * 
   * Column オブジェクトは Column.Type.UNKNOWN で生成されます。<br/><br/>
   *
   * <em>※注意：</em>
   * SQL文中の "&#47;*$ XXX $*&#47;" もパラメータとして扱うことができます。<br/>
   * この例の場合は、"XXX"という名前の Column オブジェクトが Column.Type.STRING で生成されます。<br/>
   *
   * @param   sqlText  解析前のSQL文（:PARAM_NAMEを含むSQL文）
   * @return  解析後のSQL文（:PARAM_NAMEが?に変換されたSQL文）
   * @throws  DataException  例外発生時
   */
  public String parseSQL(String sqlText) throws DataException {
    
    HashMap<String,Serializable> saveValues = new HashMap<String,Serializable>();
    for (Column col:columns.values()) {
      if (col.isReadOnly()) continue;
      saveValues.put(col.getName(), this.getValue(col.getName()));
      col.setParamPos(null);
    }
    clear();
    
    if (sqlText==null) return Const.NULLSTRING;

    StringBuffer result = new StringBuffer();
    String colname;
    Column newParam;
    int postion = 0; // パラメタの番号
    int idx_s = 0; // 捜索開始の位置
    int idx_e = 0; // 捜索が終了した位置
    int maxLen = sqlText.length();

    // &#47;*$ XXX $*&#47; で囲まれた部分を置換パラメタとして扱うため
    // そのパターンを見つけたら Params に Column.Type.STRING として登録する。
    // ただし、PreparedStatementの'?'には置換しない。
    // "WHERE AAA IN ('A', 'B')" のような文字列に置換することもあるため。
    while (idx_e < maxLen) {
      int startIdx = -1; // 開始の "&#47;*" を見つけた位置('/'の位置)
      while (idx_e+1 < maxLen) {
        char c1 = sqlText.charAt(idx_e);
        if (c1 == '/' && sqlText.charAt(idx_e+1) == '*') {
          // "&#47;*"が見つかった。
          startIdx = idx_e;
          break;
        }

        idx_e++;
        if (c1 == '\'') {
          // '\''を見つけた場合は、終わりの'\''を探す。
          while (idx_e < maxLen) {
            if (sqlText.charAt(idx_e) == '\'') {
              idx_e++;
              break;
            }
            idx_e++;
          }
        }
      }

      // 開始の "&#47;*" が見つからなかった
      // (maxLen - startIdx) < 9 の場合も見つからなかったのと同じ
      if (startIdx < 0 || maxLen - startIdx < 9) {
        // 最後まで検索終了
        idx_e = maxLen + 1;
        break;
      }

      // 次の文字が "$ " と続かなかったら別のものなので次の文字から再検索
      if (!"$ ".equals(sqlText.substring(startIdx+2, startIdx+4))) {
        idx_e = startIdx+2;
        continue;
      }

      // 終わりの "*&#47;" を探す
      idx_e = startIdx+4;
      int endIdx = -1; // 終了の "*&#47;" を見つけた位置('/'の位置)
      while (idx_e+1 < maxLen) {
        if (sqlText.charAt(idx_e) == '*' && sqlText.charAt(idx_e+1) == '/') {
          // "*&#47;"が見つかった。
          endIdx = idx_e+1;
          break;
        }
        idx_e++;
      }

      // 終了の "*&#47;" が見つからなかった場合
      if (endIdx < 0) {
        // 最後まで検索終了
        idx_e = maxLen + 1;
        break;
      }

      // "*&#47;" の前の文字が " $" でなかったら別のものなので続きから再検索
      if (!" $".equals(sqlText.substring(endIdx-3, endIdx-1))) {
        idx_e = endIdx+1;
        continue;
      }

      // &#47;*$ ... $*&#47; で囲まれたキーワードを取得する。
      colname = sqlText.substring(startIdx+4, endIdx-3).trim();

      // colnameの長さが0なら、間が空白だったということ。
      if (colname.length()==0) {
        // 続きから再検索
        idx_e = endIdx+1;
        continue;
      }

      // パラメタをType.STRINGで追加する。
      // PreparedStatementには含めないのでpositionは指定しない。
      newParam = columns.get(colname);
      if (newParam==null) {
        newParam = new Column(colname, Column.Type.STRING, Column.ParamType.IN);
        this.addColumn(newParam);
      }

      // 次を検索する。
      idx_e = endIdx+1;
    }

    idx_e = 0; // この後、:付きのパラメタの走査を行うためポインタを戻す

    while (idx_e < maxLen) {
      int startIdx = 0; // ':'を見つけた位置+1
      while (idx_e < maxLen) {
        char c1 = sqlText.charAt(idx_e);
        if (c1 == SQL_PARAM) {
          // ':'が見つかった。
          startIdx = idx_e+1;
          break;
        }

        idx_e++;
        if (c1 == '\'') {
          // '\''を見つけた場合は、終わりの'\''を探す。
          while (idx_e < maxLen) {
            if (sqlText.charAt(idx_e) == '\'') {
              idx_e++;
              break;
            }
            idx_e++;
          }
        }
      }

      // ':'が見つからなかった
      if (startIdx <= 0) {
        // 最後まで検索終了
        result.append(sqlText.substring(idx_s));
        idx_e = maxLen + 1;
        break;
      }

      // ':'の前の文字までを SQL文に追加する。
      result.append(sqlText.substring(idx_s, startIdx-1));
      idx_e = startIdx;

      // ':'が見つかった場合は、終わりを見つける。
      int endIdx   = 0; // ':'から始まるパラメタ名の最後の位置+1
      while (idx_e < maxLen) {
        char c = sqlText.charAt(idx_e);
        for (int i=0; i<SQL_DELIM.length; i++) {
          if (c == SQL_DELIM[i]) {
            endIdx = idx_e;
            break;
          }
        }
        if (endIdx > 0) {
          break;
        }
        idx_e++;
      }

      if (endIdx <= 0) {
        // ':'で始まるパラメタ名の末尾が見つからなかった。→末尾までをパラメタ名とする。
        endIdx = maxLen;
      }

      if (endIdx > startIdx) {
        // ':'で始まるパラメタ名の末尾が見つかった。
        colname = sqlText.substring(startIdx, endIdx);
        postion++;
        newParam = this.columns.get(colname);
        if (newParam==null) {
          newParam = new Column(colname, Column.Type.UNKNOWN, Column.ParamType.IN);
          this.addColumn(newParam);
        }
        newParam.addParamPos(postion);
        result.append(JDBC_PARAM);  // パラメタ名の代わりに'?'をSQL文にセットする。
      } else {
        // ': 'の場合など、':'の後ろに名前がなかった。
        result.append(sqlText.substring(startIdx-1, endIdx));
      }
      idx_e = endIdx;
      idx_s = idx_e;
    }
    
    for (String name:saveValues.keySet()) {
      try {
        this.setValue(name, saveValues.get(name));
      } catch (Exception x) {}
    }
    
    return result.toString();
  }
  
  /**
   * Params オブジェクトの内容を文字列として取得します。<br/><br/>
   *
   * クラス名、各フィールド名とその値を返します。<br/>
   *
   * @return  Params オブジェクトの文字列表現
   */
  @Override
  public String toString() {
    StringBuffer temp = new StringBuffer();
    temp.append(this.getClass().getSimpleName());
    temp.append('{');
    try{
      for (int i=0; i<this.getColumnCount(); i++) {
        temp.append(this.getColumn(i).getName());
        temp.append(':');
        temp.append(this.getColumn(i).getType());
        temp.append("=[");
        temp.append(this.getValueAsString(i));
        temp.append("];");
      }
    } catch (DataException x) {
      x.printStackTrace();
    }
    temp.append('}');
    return temp.toString();
  }

  /**
   * パラメータオブジェクトに設定された全てのデータ値の検証を行います。<br/>
   * 
   * @see jp.littlesoft.data.Row#validate()
   */
  @Override
  public void validate()  throws DataException {
    String nulls = Const.NULLSTRING;
    for (int i = 0; i < getColumnCount(); i++) {
      // ヌル値チェックを実行
      if (!getColumn(i).isNullable() && isNull(i)) {
        if (!nulls.equals(Const.NULLSTRING)) nulls += ", ";
        nulls += getColumn(i).getCaption();
      }
    }
    if (!nulls.equals(Const.NULLSTRING)){
      throw new InvalidValueException("INVALID_NULL_VALUE",
          new String[]{nulls}, this, null, null);
    }
  }

  @Override
  public synchronized void setValue(String columnName, Serializable value) throws DataException, InvalidValueException {
    if (indexOfColumn(columnName)==-1) {
      this.addColumn(columnName, Column.classToColtypeMap(value.getClass()));
    }
    super.setValue(columnName, value);
  }


  
}
