/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import jp.littlesoft.soja.IJsonSerializable;

/**
 * SQLツリーモデルクラス。<br/><br/>
 *
 * 表構造のデータを格納する為のオブジェクトです。<br/>
 * 単純な２次元配列と異なる点は、データ項目の型や、その他の属性をColumnクラスによって定義できる点です。<br/>
 * また、データ項目値の設定、取得の他、行の追加、削除、並び替え、フィルタなど数多くのメソッドが用意されています。<br/><br/>
 * 
 * @author  LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class TreeModel implements IJsonSerializable {

  static long nodeSequence = 0L;

  static public class Node implements Serializable {

    private long id;
    private int originalRowIndex = -1;
    private Node parent = null;
    private ArrayList<Node> children = null;
    private boolean expanded = false;
    protected int level = -1;
    private String name = null;
    private String link = null;
    public HashMap<String, Serializable> attribute;
    public int tag;
    private Object object;

    public Object getObject() {
      return object;
    }

    public void setObject(Object object) {
      this.object = object;
    }


    public Node() {
      id = nodeSequence++;
    }

    public Node(String name) {
      this();
      this.name = name;
    }

    public String getPath(String nodeSeparator) {
      String temp = this.getName();
      Node parent = this.getParent();
      while(parent!=null) {
        temp = parent.getName() + nodeSeparator + temp;
        parent = parent.getParent();
      }
      return temp;
    }


    public ArrayList<Node> getChildren() {
      if (children == null) {
        children = new ArrayList<Node>();
      }
      return children;
    }

    protected Root getRoot() {
      Node node = this;
      while (true) {
        if (node instanceof Root) {
          return (Root)node;
        }
        node = node.parent;
        if (node == null) return null;
      }
    }

    protected void setLevel(int level) {
      if (this.level == level) return;
      this.level = level;
      for (Node child : getChildren()) {
        child.setLevel(this.level + 1);
      }
    }

    private void internalAdd(int index, Node newChild) {
      if (index > -1) {
        getChildren().add(index, newChild);
      } else {
        getChildren().add(newChild);
      }
      newChild.parent = this;
      newChild.setLevel(this.level+1);
    }

    public boolean existsChildName(String name) {
      for (Node child : getChildren()) {
        if (name.equals(child.getName())) {
          return true;
        }
      }
      return false;
    }

    public void addChild(Node newChild, boolean initialize) {
      internalAdd(-1, newChild);
    }

    public void addChild(Node newChild) {
      if (getChildren().contains(newChild)) return;
      Node position;
      if (getChildren().size()==0) {
        position = this;
      } else {
        position = getChildren().get(getChildren().size()-1);
      }
      getRoot().addNode(position, newChild);
      internalAdd(-1, newChild);
    }

    public void addChild(int index, Node newChild) {
      if (getChildren().contains(newChild)) return;
      Node position;
      if (getChildren().size()==0) {
        position = this;
      } else {
        position = getChildren().get(index);
      }
      getRoot().addNode(position, newChild);
      internalAdd(index, newChild);
    }

    public void removeChild(Node child) {
      if (getChildren().contains(child)) {
        child.clearChildren();
        getChildren().remove(child);
        getRoot().removeNode(child);
      }
    }

    public void clearChildren() {
      Node[] nodes = new Node[getChildren().size()];
      getChildren().toArray(nodes);
      for (Node node : nodes) {
        removeChild(node);
      }
      //getChildren().clear();
    }

    /**
     * ノードを移動します
     */
    public void moveNode(Node to) {
      moveNode(this, to);
    }

    /**
     *
     * @param node 移動対象のNode
     * @param to   移動先の親Node
     */
    public void moveNode(Node node, Node to) {
      //TODO：テスト中
      Node parent = node.getParent();
      //ROOTの場合
      if (parent == null && node instanceof Root) {
        return;
      } else {
        parent.removeChild(node);
        to.addChild(node);
      }
    }

    public boolean isExpanded() {
      return expanded;
    }

    public void expand() {
      this.expanded = true;
    }

    public void collapse() {
      this.expanded = false;
    }

    public String getName() {
      return name;
    }

    public void setName(String name) {
      this.name = name;
    }

    public Node getParent() {
      return parent;
    }

    public void setLink(String link) {
      this.link = link;
    }

    public boolean hasLink() {
      return (link!=null);
    }

    public String getLink() {
      return link;
    }

    public HashMap<String, Serializable> getAttribute() {
      return attribute;
    }

    public void setAttribute(HashMap<String, Serializable> attribute) {
      this.attribute = attribute;
    }

    public int getOriginalRowIndex() {
      return originalRowIndex;
    }

    public void setOriginalRowIndex(int originalRowIndex) {
      this.originalRowIndex = originalRowIndex;
    }

    public long getId() {
      return id;
    }
  }

  static public class Root extends Node{

    TreeModel tree;

    public Root(TreeModel tree) {
      setName("RootNode");
      this.tree = tree;

    }

    public void setLevel() {
      super.setLevel(0);
    }

    public void addNode(Node position, Node newNode) {
      tree.addNodeList(position, newNode);
    }

    public void removeNode(Node node) {
      tree.removeNodeList(node);
    }
  }

  static public class Leaf<T extends Serializable> extends Node {


    public Leaf() {}

    public Leaf(String name) {
      super(name);
    }



  }



  protected String key = this.getClass().getName();;
  private boolean showRoot = true;


  protected Root root = new Root(this);

  public TreeModel() {
    root.setLevel(0);
  }

  public void addNodeList(Node position, Node newNode) {}

  public void removeNodeList(Node node) {}

  @Override
  public void fromJSONString(String arg0) {
    // TODO Auto-generated method stub

  }

  @Override
  public String toJSONString() {
    // TODO Auto-generated method stub
    return null;
  }


  public String getKey() {
    return key;
  }


  public void setKey(String key) {
    this.key = key;
  }

  public String getRootName() {
    return getRoot().getName();
  }


  public void setRootName(String name) {
    getRoot().setName(name);
  }


  public Root getRoot() {
    return root;
  }



  public boolean isShowRoot() {
    return showRoot;
  }



  public void setShowRoot(boolean showRoot) {
    this.showRoot = showRoot;
  }


}
