/** Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data.validator;

import java.io.Serializable;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.exception.InvalidValueException;

/**
 * Validator デフォルト実装クラス。<br/><br/>
 * 
 * カラム値に対して暗黙的に適用されます.<br/>
 * カラムのデータ型に応じて、以下のようなチェックが行われます.<br/>
 * <br/>
 * <table border="1">
 * <tr><th class="TableSubHeadingColor">カラムのデータ型</th><th class="TableSubHeadingColor">使用されるValidatorクラス</th></tr>
 * <tr><td>Operation.STRING          </td><td>{@link jp.littlesoft.data.validator.StringValidator}</td></tr>
 * <tr><td>Operation.CLOB            </td><td>{@link jp.littlesoft.data.validator.StringValidator}</td></tr>
 * <tr><td>Operation.BIGDECIMAL      </td><td>{@link jp.littlesoft.data.validator.NumberValidator}</td></tr>
 * <tr><td>その他                    </td><td>なし</td></tr>
 * </table>
 * <br/>
 * 
 * デフォルトの実装に加え独自の値検証を実装したい場合は、以下のようにしてください.<br/>
 * <pre> 
 * // 独自Validatorクラスの実装 
 * public void MyValidator extends DefaultValidator {
 *
 *   public void validate(Column column, Object value) .. {
 *     // MyValidator独自の検証
 *     ......
 *     // デフォルトの検証ロジックを実行
 *     super.validate(column, value);
 *   }
 * }
 * -------
 * // ChangeListener実装クラスでtableのcolAにMyValidatorを設定する。
 * public void onColumnSetup(..) ... {
 *   table.getColumn("colA").setExtraValidator( new MyValidator() );
 *     :
 * }
 * </pre>
 * </pre>
 * 
 * @see jp.littlesoft.data.validator.StringValidator
 * @see jp.littlesoft.data.validator.NumberValidator
 * @see jp.littlesoft.data.Column#setValidator
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class DefaultValidator implements Validator, Serializable {
  /**
   * コンストラクタ。<br/>
   */
  public DefaultValidator() {  
  }

  /**
   * 値の検証を行います。<br/>
   * 
   * @param  column 検証対象のカラムオブジェクト
   * @param  value  検証対象値すべき値
   * @throws jp.littlesoft.data.exception.InvalidValueException エラーが検出された場合に送出される。
   */
  @SuppressWarnings("unchecked")
  public void execute(Row row, Column column, Object value) throws InvalidValueException {
  	// デフォルトの振る舞い
    switch (column.getType()) {
      case STRING:
      case CLOB:
				new StringValidator().execute(row, column, value);
        break;
      case BIGDECIMAL:
        new NumberValidator().execute(row, column, value);
        break;
      default:
    }
  }
}
