/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.file.server;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;
import javax.servlet.http.HttpSession;

import jp.littlesoft.file.common.SplitUploadParams;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.web.ActionController;
import jp.littlesoft.web.ActionPage;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;

/**
 * アップロードFilterクラス
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class UploadFilter implements Filter {
  
  /**
   * アップロード受け入れ可能か？を設定するセッション属性の名前
   */
  static final public String SA_ACCEPTABLE_UPLOAD = UploadFilter.class.getName() + ".acceptableUpload";
  
  /**
   * サーブレット側では処理せず、直ちに処理終了するリクエスト属性、分割アップロードで使用する.
   */
  static final public String RA_IMMEDIATELY_RETURN = UploadFilter.class.getName() + ".immediatelyReturn";
  
  static String FILE_SEP = System.getProperty("file.separator");
  
  @SuppressWarnings("serial")
  static public class Params implements Serializable {
    public String fileItemClassName = DefaultFileItem.class.getName();  //FileItem実装クラス名
    public String tempDirectory = null;               //アップロードファイル一時ディレクトリ
    public String saveDirectory = null;               //アップロードファイル保存ディレクトリ
    public String headerEncoding = "UTF-8";
    public Long sizeMax = new Long(1024*1024*16);     //The maximum allowed size, in bytes. The default value of -1 indicates, that there is no limit.
    public Long fileSizeMax = new Long(1024*1024*16); //the maximum allowed size of a single uploaded file, as opposed to getSizeMax()
    public String extension = null;                   //アップロードファイルの拡張子
  }

  private UploadFileItemFactory fileItemFactory;
  
  public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
    //マルチパートか確認
    if (ServletFileUpload.isMultipartContent((HttpServletRequest)request)) {
      if (fileItemFactory==null) {
        new ServletException("FileItemFactoryName not specified in the initialization parameter.");
      }
      request = new UploadRequestWrapper((HttpServletRequest) request, fileItemFactory);
    }  
    chain.doFilter(request, response);
  }

  public void init(FilterConfig config) throws ServletException {
    try {
      Params params;
      String paramJsonString = config.getInitParameter(Params.class.getName());
      if (paramJsonString!=null) {
          params = (Params)LSJson.deserialize(paramJsonString);
      } else {
        throw new ServletException(Params.class.getName() + " is not found in web.xml.");
      }
      if (params.tempDirectory==null) {
        params.tempDirectory =
          config.getServletContext().getRealPath(FILE_SEP + "WEB-INF" + FILE_SEP + "temp");
      } else if (params.tempDirectory.startsWith(FILE_SEP)) {
        params.tempDirectory =
          config.getServletContext().getRealPath(FILE_SEP + "WEB-INF" + params.tempDirectory);
      }
      if (params.saveDirectory==null) {
        params.saveDirectory =
          config.getServletContext().getRealPath(FILE_SEP + "WEB-INF" + FILE_SEP + "save");        
      } else if (params.saveDirectory.startsWith(FILE_SEP)) {
        params.saveDirectory = 
          config.getServletContext().getRealPath(FILE_SEP + "WEB-INF" + params.saveDirectory);
      }
      fileItemFactory = new UploadFileItemFactory(params);
    } catch (Exception e) {
      throw new ServletException(e);
    }
  }

  public void destroy() {
  }
  
  /**
   * 
   * @author littlesoft
   *
   */
  static public class UploadFileItemFactory implements FileItemFactory {

    private Params params;
    private Class<? extends FileItem> fileItemClass;
    
    @SuppressWarnings("unchecked")
    public UploadFileItemFactory(Params params) throws ClassNotFoundException {
      this.params = params;
      fileItemClass = (Class<? extends FileItem>) Class.forName(params.fileItemClassName);
    }

    /**
     * 渡されたパラメータと、このファクトリ内の設定から新規 FileItem を生成します。
     *
     * @param fieldName フォームフィールド名
     * @param contentType フォームフィールドのコンテントタイプ
     * @param isFormField 単純なフォームフィールドを示す場合はtrue、それ以外の場合はfalse
     * @param fileName ブラウザまたは他のクライアントから渡されるアップロードファイル名
     * @return 新規に作成されたファイルアイテム
     */
    public FileItem createItem(String fieldName, String contentType, boolean isFormField, String fileName) {
      try {
        Constructor<? extends FileItem> cons = fileItemClass.getConstructor(String.class, String.class, boolean.class, String.class);
        return (FileItem) cons.newInstance(fieldName, contentType, isFormField, fileName);
      } catch (Exception e) {
        throw new RuntimeException(e);
      }
    }
    
    String getHeaderEncoding() {
      return params.headerEncoding;
    }
    
    long getSizeMax() {
      return params.sizeMax;
    }

    long getFileSizeMax() {
      return params.fileSizeMax;
    }
  }

  /**
   * 
   * @author littlesoft
   *
   */
  static public class UploadRequestWrapper extends HttpServletRequestWrapper {


  
    private Map<String, String[]> paramMap = new HashMap<String, String[]>();

    @SuppressWarnings("unchecked")
    public UploadRequestWrapper(HttpServletRequest request, UploadFileItemFactory factory) throws ServletException {
      super(request);
      HttpSession session = request.getSession(true);
      if (session.getAttribute(SA_ACCEPTABLE_UPLOAD)==null 
          || !((Boolean)session.getAttribute(SA_ACCEPTABLE_UPLOAD)).booleanValue()) {
        throw new ServletException("Uploading is not acceptable in this Session.");
      }
      
      if (getRequest().getParameterMap().size() > 0) {
        for (Object key : getRequest().getParameterMap().keySet()) {
          paramMap.put((String)key, (String[])getRequest().getParameterMap().get(key));
        }
      };
      
      UploadFileManager fmanager = (UploadFileManager)session.getAttribute(UploadFileManager.class.getName());
      if (fmanager==null) {
        fmanager = new UploadFileManager();
      }
      ServletFileUpload fileUpload = new ServletFileUpload(factory);
      fileUpload.setSizeMax(factory.getSizeMax());
      fileUpload.setFileSizeMax(factory.getFileSizeMax());
      fileUpload.setHeaderEncoding(factory.getHeaderEncoding());
      try {
        // アップロードされたファイル情報をFileItemオブジェクトのリストとして取得
        String fieldName = null;
        String formFieldValue = null;
        List<AbstractFileItem> items = (List<AbstractFileItem>)fileUpload.parseRequest(request);
        for (AbstractFileItem fileItem : items) {
          fieldName = fileItem.getFieldName();
          if (fileItem.isFormField()) {   // 通常のフォーム項目
            formFieldValue = new String(fileItem.get(), factory.getHeaderEncoding());
            if (paramMap.containsKey(fieldName) && paramMap.get(fieldName)!=null) {
              String[] temp = new String[paramMap.get(fieldName).length+1];
              int i = 0;
              for (String s : paramMap.get(fieldName)) {
                temp[i++] = s;
              }
              temp[i] = formFieldValue;
              paramMap.put(fieldName, temp);
            } else {
              paramMap.put(fieldName, new String[] {formFieldValue} );
            }
          } else {  // アップロードファイルの場合
            fmanager.getItems().add(fileItem);
          }
        }
        // ファイルアップロードデータが存在する場合、Attributeに設定
        if (fmanager.getItems().size() > 0) {
          String jsonUploadParam = getParameter(ActionPage.PN_APPLICATION_PARAMS);
          SplitUploadParams uploadParam = null;
          if (jsonUploadParam!=null) {
            uploadParam = (SplitUploadParams) LSJson.deserialize(jsonUploadParam);
            fmanager.setParams(uploadParam);
            request.setAttribute(ActionController.RA_JAVA_APPLICATION_CALLBACK_PARAM, uploadParam.getAttributes());
          } 
          if (uploadParam!=null && uploadParam.isSplit) {
            if (uploadParam.isFinal) {
              //分割アップロードファイル最終アイテムの場合マージ処理を行う
              fmanager.mergeSplitFiles(fmanager.getItems().size()-1);
            } else {
              //分割アップロードファイル最終アイテムの場合マージ処理を行う
              request.setAttribute(UploadFilter.RA_IMMEDIATELY_RETURN, new Boolean(true));
            }
          }
          session.setAttribute(UploadFileManager.class.getName(), fmanager);
        }
      } catch (ServletException e) {
        throw e;
      } catch (Exception e) {
        throw new ServletException(e);
      }
    }


    /**
     * このリクエストから取得できるパラメータを java.util.Map で返します。<br/>
     * リクエストパラメータというのはリクエストに付けられて送られてくる付加情報です。<br/>
     * HTTP Servlet ではパラメータはクエリー文字列になっているか、 あるいは、フォームデータ形式でポストされます。<br/>
     *
     * @return パラメータ名が key 、パラメータ値がマップの値と なっている不変な java.util.Map。<br/>
     *         パラメータマップの key は String 型。 パラメータマップの値は String の配列です。
     */
    @Override
    public Map<String, String[]> getParameterMap() {
      return paramMap;
    }


    /**
     * リクエストパラメータの値を String 型のオブジェクトで返しますが、パラメータが存在しない場合は null を返します。<br/>
     * リクエストパラメータというのはリクエストに付けて送られてくる追加情報です。<br/>
     * HTTP Servlet ではパラメータはクエリー文字列か POST された FORM のデータに含まれています。<br/>
     * このメソッドはパラメータに確実に一つの値だけがある場合に限って使用するべきです。<br/>
     * パラメータに複数の値があるかもしれない場合には getParameterValues(java.lang.String) メソッドを使ってください。<br/><br/>
     *
     * もしも、このメソッドをマルチバリューのパラメータに対して使用すると、<br/>
     * getParameterValues() のメソッドの戻り値である配列の最初の値のみが返されます。<br/><br/>
     *
     * HTTP POST リクエストで送ったなど、 パラメータのデータがリクエストのメッセージボディで送られた場合、<br/>
     * getInputStream() や getReader() メソッドを使って直接メッセージボディを読み込む操作は、<br/>
     * このメソッドの実行に影響を与えることがあります。<br/>
     *
     * @param name パラメータの名前を指定する String
     * @return 単一のパラメータの値を表す String
     * @see javax.servlet.ServletRequest#getParameterValues(java.lang.String)
     */
    @Override
    public String getParameter(String name) {
      return paramMap.containsKey(name) ? paramMap.get(name)[0] : null;
    }


    /**
     * 指定された名前で取得できるリクエストパラメータ全ての値を String オブジェクトの配列で返しますが、<br/>
     * パラメータが存在しない場合は null を返します。<br/>
     * 受け取ったパラメータに値がひとつしかない場合、配列の大きさは 1 です。<br/>
     *
     * @param name 取得したいパラメータの名前を指定する String 
     * @return パラメータの値が入っている String オブジェクトの配列
     * @see javax.servlet.ServletRequest#getParameter(java.lang.String)
     */
    @Override
    public String[] getParameterValues(String name) {
      return paramMap.get(name);
    }


    /**
     * このリクエストに含まれているパラメータ名を表す String オブジェクトで構成される Enumeration を返します。<br/>
     * リクエストにパラメータが無い場合、空の Enumeration を返します。<br/>
     *
     * @return Stringオブジェクトで構成された Enumeration。<br/>
     *         Stringは各リクエストパラメータの名前。 パラメータが無い場合は空の Enumeration。
     */
    @Override
    public Enumeration<String> getParameterNames() {
      return new Vector<String>(paramMap.keySet()).elements();
    }

  }
}
