/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.StringReader;
import java.math.BigDecimal;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.ITableDef;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.ITableDef.Columns;
import jp.littlesoft.data.ITableDef.Index;
import jp.littlesoft.data.ITableDef.MetaData;
import jp.littlesoft.data.exception.DataException;

@SuppressWarnings("serial")
public class JDBCTableMetaData extends ITableDef.MetaData<Row<?>> {
  
  public static int SCHEMA_NAME = 1;
  
  public static int  TABLE_CATALOG = 1 ;
  public static int  TABLE_SCHEMA = 2 ;
  public static int  TABLE_NAME  = 3 ;
  public static int  TABLE_TYPE  = 4 ;
  public static int  TABLE_REMARKS = 5 ;
  public static int  TABLE_TYPE_CATALOG  = 6 ;
  public static int  TABLE_TYPE_SCHEMA  = 7 ;
  public static int  TABLE_TYPE_NAME = 8 ;
  public static int  TABLE_SELF_REFERENCING_COL_NAME = 9 ;
  public static int  TABLE_REF_GENERATION  = 10  ;

  public static int COL_COLUMN_NAME = 4;
  public static int COL_DATA_TYPE = 5;
  public static int COL_COLUMN_SIZE = 7;
  public static int COL_DECIMAL_DIGITS = 9;
  public static int COL_NULLABLE = 11;
  public static int COL_REMARKS = 12;
  public static int COL_DEFAULT_VALUE = 13;
  
  public static int PK_TABLE_NAME = 3;
  public static int PK_COLUMN_NAME = 4;
  public static int PK_KEY_SEQ = 5;
  public static int PK_PK_NAME = 6;
  
  public static int IX_TABLE_NAME = 3;
  public static int IX_INDEX_NAME = 6;
  public static int IX_ORDINAL_POS = 8;
  public static int IX_COLUMN_NAME = 9;
  public static int IX_NON_UNIQUE = 4;
  
  public static int FK_TABLE_NAME = 7;
  public static int FK_NAME = 12;
  public static int FK_KEY_SEQ = 9;
  public static int FK_COLUMN_NAME = 8;
  public static int FK_REF_TABLE_NAME = 3;
  public static int FK_REF_COLUMN_NAME = 4;
  public static int FK_UPDATE_RULE = 10;
  public static int FK_DELETE_RULE = 11;
  
  public JDBCTableMetaData(DatabaseMetaData dbMetaData, String catalog, String schema, String tableName) throws TableNotFound {
    try {
      ResultSet rsTable = dbMetaData.getTables(catalog, schema, tableName, null);
      ResultSet rsColumn = null; 
      if (!rsTable.next()) {
        throw new TableNotFound("table is not found [" + catalog + "." + schema + "."  + tableName + "]");
      } else {
        this.dbCatalog = catalog;
        this.dbSchema = schema;
        this.dbTableName = tableName;
        Table<Row<?>> dummyQuery = new Table<Row<?>>();
        SQLDataProviderImpl dummyQueryProvider = new SQLDataProviderImpl();
        dummyQueryProvider.setConnection(dbMetaData.getConnection());
        dummyQuery.setDataProvider(dummyQueryProvider);
        //空の問合せを実行し、resultSetMetaDataを取得
        dummyQuery.passivate();
        dummyQuery.clearColumns();
        dummyQueryProvider.setSQL("SELECT * FROM " + getDBTableName(true) + " WHERE NULL IS NOT NULL");
        dummyQuery.activate();
        this.columns = new Columns(dummyQuery);
        for (Column col : this.columns) {
          rsColumn = dbMetaData.getColumns(catalog, schema, tableName, col.getDBColumnName());
          if (rsColumn.next()) {
            String defaultValue = rsColumn.getString(COL_DEFAULT_VALUE);
            if (defaultValue == null || "".equals(defaultValue)) continue;
            try {
              switch (col.getType()) {
              case BOOLEAN:
                defaultValue = defaultValue.toLowerCase();
                if ("true".equals(defaultValue) || "false".equals(defaultValue)) {
                  col.setDefaultValue(defaultValue);
                }
                break;
              case BYTE:
              case SHORT:
              case INT:
              case LONG:
                Long l = Long.parseLong(defaultValue);
                col.setDefaultValue(l.toString());
                break;
              case DOUBLE:
              case FLOAT:
                Double f = Double.parseDouble(defaultValue);
                col.setDefaultValue(f.toString());
                break;
              case BIGDECIMAL:
                BigDecimal d = new BigDecimal(defaultValue);
                col.setDefaultValue(d.toString());
                break;
              case STRING:
                defaultValue = defaultValue.trim();
                if (defaultValue.startsWith("STRINGDECODE('")) {
                  // H2では２バイト文字が以下の式でディフォルトに設定されている
                  // STRINGDECODE('\u81ea\u7531\u5165\u529b')
                  defaultValue = defaultValue.substring(0, defaultValue.length() - 2).substring(14);
                  Properties p = new Properties();
                  p.load(new StringReader("value=" + defaultValue + "\n"));
                  col.setDefaultValue(p.getProperty("value"));
                } else if (defaultValue.startsWith("'") && defaultValue.endsWith("'")) {
                  // クォーテーションで囲まれた値だったらカラムにセット
                  col.setDefaultValue(defaultValue.substring(1, defaultValue.length() - 1));
                }
                break;
              case DATE:
              case TIME:
              case TIMESTAMP:
                if (defaultValue.startsWith("'") && defaultValue.endsWith("'")) {
                  // クォーテーションで囲まれた値だったらカラムにセット
                  col.setDefaultValue(defaultValue.substring(1, defaultValue.length() - 1));
                }
                break;
              }
            } catch (Exception x) { }
          }
          //System.out.println("Column.getDefaultValue:" + col.getDefaultValue());
        }
        boolean unique = false;
        String dbObjectName = null;
        ArrayList<Column> cols = new ArrayList<Column>();
        //PRIMARY KEY 情報の読み込み
        cols.clear();
        ResultSet pk = dbMetaData.getPrimaryKeys(catalog, schema, tableName);
        while(true) {
          if (pk.next()) {
            short pos = pk.getShort(PK_KEY_SEQ);
            while (cols.size()<pos) { cols.add(null); };
            cols.set((pos-1), this.columns.getByDBColumnName(pk.getString(PK_COLUMN_NAME)));
          } else if (cols.size()>0) {
            addPrimaryKey(cols);
            cols.clear();
          } else {
            break;
          }
        } 
        //FOREGIN KEY 情報の読み込み
        dbObjectName = null;
        cols.clear();
        String refTableName = null;
        ArrayList<String> refTableColumns = new ArrayList<String>();
        ResultSet fk = dbMetaData.getImportedKeys(catalog, schema, tableName);
        while(true) {
          if (fk.next()) {
            refTableName = fk.getString(FK_REF_TABLE_NAME);
            if (dbObjectName!=null && !dbObjectName.equals(fk.getString(FK_NAME))) {
              addForeginKey(cols)
                .references(refTableName, false)
                .referenceTableColumnNames(refTableColumns);
              cols.clear();
            }
            dbObjectName = fk.getString(FK_NAME);
            short pos = fk.getShort(FK_KEY_SEQ);
            while (cols.size()<pos) { 
              cols.add(null); 
              refTableColumns.add(null);
            };
            cols.set((pos-1), this.columns.getByDBColumnName(fk.getString(FK_COLUMN_NAME)));
            refTableColumns.set((pos-1), fk.getString(FK_REF_COLUMN_NAME));
          } else if (cols.size()>0) {
            addForeginKey(cols)
              .references(refTableName, false)
              .referenceTableColumnNames(refTableColumns);
            cols.clear();
          } else {
            break;
          }
        }
        //INDEX 情報の読み込み
        dbObjectName = null;
        cols.clear();
        ResultSet ix = dbMetaData.getIndexInfo(catalog, schema, tableName, false, true);
        while(true) {
          if (ix.next()) {
            if (dbObjectName!=null && !dbObjectName.equals(ix.getString(IX_INDEX_NAME))) {
              //System.out.println("### index name : " + dbObjectName);
              addIndex(cols).unique(unique);
              cols.clear();
            }
            dbObjectName = ix.getString(IX_INDEX_NAME);
            unique = !(ix.getBoolean(IX_NON_UNIQUE)); 
            short pos = ix.getShort(IX_ORDINAL_POS);
            while (cols.size()<pos) { cols.add(null); };
            cols.set((pos-1), this.columns.getByDBColumnName(ix.getString(IX_COLUMN_NAME)));
          } else if (cols.size()>0) {
            //System.out.println("### index name : " + dbObjectName);
            addIndex(cols).unique(unique);
            cols.clear();
          } else {
            break;
          }
        } 
      }
    } catch (SQLException e) {
      throw new RuntimeException(e);
    } catch (DataException e1) {
      throw new RuntimeException(e1);
    }
    
  }
  
  static public class TableNotFound extends Exception {
    public TableNotFound(String message) {
      super(message);
    }
  }

  @Override
  protected void initialize(Class<?> type) {
    //
  }
  
  public void checkDifference(MetaData<?> newMetaData, List<String> alterSQLs) throws DataException {
    String tableName = getDBTableName(true);
    if (!tableName.equals(newMetaData.getDBTableName(true))) {
      throw new DataException("Table names are different and therefore can not be compared.");
    }
    Column oldColumn;
    ArrayList<Column> modifiedColumns = new ArrayList<Column>();
    ArrayList<String> renamedColumnNames = new ArrayList<String>();
    //新しい構造をもとにthisの構造と比較する
    for (Column newColumn  : newMetaData.getColumns()) {
      oldColumn = getColumnByDBName(newColumn.getDBColumnName());
      //リネームの可能性あり
      if (oldColumn==null && newColumn.getOldDBColumnName()!=null) {
        oldColumn = getColumnByDBName(newColumn.getOldDBColumnName());
        if (oldColumn!=null) {
          StringBuffer buff = new StringBuffer();
          buff.append("ALTER TABLE ");
          buff.append(tableName);
          buff.append(" ALTER COLUMN ");
          buff.append(oldColumn.getDBColumnName());
          buff.append(" RENAME TO ");
          buff.append(newColumn.getDBColumnName());
          alterSQLs.add(buff.toString());
          modifiedColumns.add(newColumn);
          renamedColumnNames.add(oldColumn.getDBColumnName());
        }
      }
      //新しいカラムが追加？
      if (oldColumn==null) {
        StringBuffer buff = new StringBuffer();
        buff.append("ALTER TABLE ");
        buff.append(tableName);
        buff.append(" ADD ");
        newMetaData.getColumns().appendColumnDefString(newColumn, buff);
        alterSQLs.add(buff.toString());
        modifiedColumns.add(newColumn);
        continue;
      } 
      //データ型、桁数、精度が変更された ?
      if ( (newColumn.getType()!=oldColumn.getType()) 
        || (newColumn.getLength()!=oldColumn.getLength())
        || (newColumn.getPrecision()!=oldColumn.getPrecision()) 
        || (newColumn.getScale()!=oldColumn.getScale()) ) { 
//        || (newColumn.getLength()!=-1     && newColumn.getLength()!=oldColumn.getLength())
//        || (newColumn.getPrecision()!=-1  && newColumn.getPrecision()!=oldColumn.getPrecision()) 
//        || (newColumn.getScale()!=-1      && newColumn.getScale()!=oldColumn.getScale()) ) { 
        StringBuffer buff = new StringBuffer();
        buff.append("ALTER TABLE ");
        buff.append(tableName);
        buff.append(" ALTER COLUMN ");
        newMetaData.getColumns().appendColumnDefString(newColumn, buff);
        alterSQLs.add(buff.toString());
        modifiedColumns.add(newColumn);
        continue;
      } 
      //NULL制約が変更されているか?
      if (newColumn.isNullable()!=oldColumn.isNullable()) {
        StringBuffer buff = new StringBuffer();
        buff.append("ALTER TABLE ");
        buff.append(tableName);
        buff.append(" ALTER COLUMN ");
        buff.append(newColumn.getDBColumnName());
        buff.append(" SET ");
        buff.append( newColumn.isNullable() ? "NULL" : "NOT NULL");
        alterSQLs.add(buff.toString());
        modifiedColumns.add(newColumn);
      }
      //ディフォルト値が変更されているか?
      if (!newColumn.getDefaultValue().equals(oldColumn.getDefaultValue())) {
        //System.out.println("old[" + oldColumn.getDefaultValue() + "] new[" + newColumn.getDefaultValue() + "]");
        StringBuffer buff = new StringBuffer();
        buff.append("ALTER TABLE ");
        buff.append(tableName);
        buff.append(" ALTER COLUMN ");
        buff.append(newColumn.getDBColumnName());
        buff.append(" SET ");
        Columns.toDefaultString(newColumn, buff);
        alterSQLs.add(buff.toString());
        modifiedColumns.add(newColumn);
      }
    }
    //新Metaに存在しないカラムをドロップする.(リネームされたものを除く)
    for (Column old : getColumns()) {
      Column newCol = newMetaData.getColumnByDBName(old.getDBColumnName());
      if (newCol==null && !renamedColumnNames.contains(old.getDBColumnName())) {
        StringBuffer buff = new StringBuffer();
        buff.append("ALTER TABLE ");
        buff.append(tableName);
        buff.append(" DROP COLUMN ");
        buff.append(old.getDBColumnName());
        alterSQLs.add(buff.toString());
      }
    }
    //主キーを比較する
    String oldKey = (getPrimaryKey()==null) ? "" : getPrimaryKey().toString();
    String newKey = newMetaData.getPrimaryKey()==null ? "" : newMetaData.getPrimaryKey().toString();
    if (!oldKey.equals(newKey)) {
      if (!oldKey.equals("")) {
        alterSQLs.add(getPrimaryKey().toDropString());
      }
      alterSQLs.add(newKey);
    }
    //インデックスを比較する
    ArrayList<String> newIndexes = new ArrayList<String>();
    for (Index n : newMetaData.getIndexes()) {
      newIndexes.add(n.toString());
      //System.out.println("(new)" + n.toString());
    }
    //新メタに存在しないIndexはドロップする
    ArrayList<String> oldIndexes = new ArrayList<String>();
    for (Index o : getIndexes()) {
      String oldIndex = o.toString();
      oldIndexes.add(oldIndex);
      //System.out.println("(old)" + o.toString());
      if (!newIndexes.contains(oldIndex)) {
        alterSQLs.add(o.toDropString());
      } 
    }
    //旧メタに存在しないIndexはCteateする
    for (String n : newIndexes) {
      if (!oldIndexes.contains(n)) {
        //作成するIndex
        alterSQLs.add(n);
      }
    }
    /*
    //変更されたColumnを参照している主キー、インデックスを再作成するSQLを生成
    for (Column modCol : modifiedColumns) {
      if (newMetaData.getPrimaryKey()==null) {
        alterSQLs.add(newMetaData.getPrimaryKey().toDropString());
      } else if ( newMetaData.getPrimaryKey().contains(modCol)
               && !alterSQLs.contains(newMetaData.getPrimaryKey().toString())) {
        alterSQLs.add(newMetaData.getPrimaryKey().toDropString());
        alterSQLs.add(newMetaData.getPrimaryKey().toString());
      }
      for (Index index : newMetaData.getIndexes()) {
        if (index.contains(modCol) && !alterSQLs.contains(index.toString())) {
          alterSQLs.add(index.toDropString());
          alterSQLs.add(index.toString());
        }
      }
    }
    */
  }

}
