/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.DataProvider;
import jp.littlesoft.data.Params;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.exception.LSSQLException;
import jp.littlesoft.sql.logger.SQLLogger;
import jp.littlesoft.sql.logger.SQLTimeLogger;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;
import jp.littlesoft.util.IndexedMap;

/**
 * SQLデータベースからTableオブジェクトに読み込むクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 * @param <T>
 */
@SuppressWarnings({ "serial", "unchecked" })
public class SQLDataProviderImpl extends SQLAdaptor implements ISQLDataProvider {

  private String SQL             = null;         //設定したSQL文
  private String pageSQL         = null;         //設定したSQL文(ページ取得)
  private String jdbcSQL         = null;         //パラメータのプレースホルダをJDBC形式に変換したSQL文
  private String jdbcPageSQL     = null;         //パラメータのプレースホルダをJDBC形式に変換したSQL文(ページ取得)
  private String lastExecutedSql = null;         //前回実行したときのSQL文
  //private final Params params    = new Params(); //SQLパラメータオブジェクト
  private Params<?> params    = new Params(); //SQLパラメータオブジェクト
  
  private int maxRows = 0;    //JDBC ResultSetからデータを取得する場合の最大行数
  private int fetchSize = 0;  //JDBC ResultSetからデータを取得する場合に、何行ずつ取得するか
  
  private boolean executing = false;
  
  private static final transient Logger logger = LogManager.getLogger(SQLDataProviderImpl.class.getName());
  
  /**
   * SQL文をデータベースに送るためのjdbcStatementオブジェクトです。<br/>
   */
  protected transient Statement jdbcStatement      = null;
  
  /**
   * SQL文をデータベースに送るためのjdbcStatementオブジェクトです。<br/><br/>
   * 全件数取得時に使用します。<br/>
   */
  protected transient Statement jdbcCountStatement = null;
  
  /**
   * ResultSetオブジェクトです。<br/>
   */
  protected transient ResultSet jdbcResultSet = null;
  
  /**
   * ResultSetオブジェクトです。<br/><br/>
   * 全件数取得時に使用します。<br/>
   */
  protected transient ResultSet jdbcCountResultSet = null;
  
  /**
   * ResultSetオブジェクトのカラム名です。<br/>
   */
  protected String[] resultSetColumnNames;

  /** 
   * JDBC ResultSetMetaDataからカラム名が取得できなかった場合のカラム名を表す定数です。<br/>
   */
  public static final String UNKNOWN_COLUMN_NAME = "UNKNOWN_COLUMN";
  
  /** 
   * 全件数取得時のカラム名を表す定数です。<br/>
   */
  public static final String COUNT_COLUMN_NAME = "CNT";
  
  /**
   * JDBC ResultSetMetaDataから受け取ったカラム名を格納する内部変数です。<br/>
   */
  private static final Integer INTEGER_ZERO = new Integer(0);  // 使いまわすのでstaticにしておく。

  /** 
   * ページ取得時に内部で使用する置換用の定数です。<br/>
   */
  private static final String PARAM_NAME_LIMIT = "__LIMIT__";
  
  /** 
   * ページ取得時に内部で使用する置換用の定数です。<br/>
   */
  private static final String PARAM_NAME_OFFSET = "__OFFSET__";

  /**
   * このTableの基となるSQL問い合わせ文を取得します。<br/>
   *
   * @return  JDBC形式返還前のSQL文
   */
  public String getSQL() {
    if (this.SQL == null)
      return Const.NULLSTRING;
    return SQL;
  }

  /**
   * このTableの基となるSQL問い合わせ文を設定します。<br/><br/>
   *
   * テーブルオブジェクトに読み込むデータを取得するSELECT文を設定し、
   * パラメータの解析を行います。<br/>
   * 
   * @param  newSQL  SQL文
   * @exception  DataException  例外発生時
   */
  public void setSQL(String newSQL) throws DataException {
    setSQL(newSQL, true);
  }

  /**
   * このSQLTableの元となるSQL問い合わせ文を設定します。<br/><br/>
   *
   * テーブルオブジェクトに読み込むデータを取得するSELECT文を設定し、
   * 引数「parseParam」でTrueが指定された場合、パラメータの解析を行います。<br/>
   * 
   * @param  newSQL      SQL文
   * @param  parseParam  パラメータ解析を行うかどうか
   * @exception  DataException  例外発生時
   */
  public void setSQL(String newSQL, boolean parseParam) throws DataException {
    if (this.getSQL().equals(newSQL)) return;
    freeStatement();
    SQL = newSQL;
    if (StringUtil.isNullString(SQL))
      return;

    //boolean activeStatus = params.isActive();
    //if (params.isActive()) params.passivate();
    if (parseParam)  jdbcSQL = params.parseSQL(SQL);
    //if (activeStatus) params.isActive();
    return;
  }
  
  /**
   * このTableの基となるSQL問い合わせ文を設定します。<br/><br/>
   *
   * このTableの行列データを構築する基となるSELECT文を設定します。<br/>
   * SQLを設定するだけです。列データを再構築するなどの他の操作は全く行いません。<br/><br/>
   *
   * ※通常はこのメソッドは使用せずに、setSQL(Stirng newSQL)メソッドを使用してください。<br/>
   *
   * @see #setSQL(String newSQL)
   */
  public void setSQLText(String newSQL) {
    this.SQL = newSQL;
  }
  
  /**
   * ページ取得の際のSQL問い合わせ文を取得します。<br/>
   *
   * @return  JDBC形式返還前のページ取得SQL文
   */
  public String getPageSQL() {
    if (this.pageSQL == null)
      return Const.NULLSTRING;
    return pageSQL;
  }

  /**
   * ページ取得の際のSQL問い合わせ文を設定します。<br/><br/>
   * 
   * テーブルオブジェクトに読み込むデータを取得するページ取得SELECT文を設定し、
   * パラメータの解析を行います。<br/>
   * 
   * @param newPageSQL ページ取得SQL文
   * @param offset ページオフセット
   * @param limit ページリミット
   * @exception  DataException  例外発生時
   */
  public void setPageSQL(String newPageSQL, int offset, int limit) throws DataException {
    setPageSQL(newPageSQL, offset, limit, true);
  }

  /**
   * ページ取得の際のSQL問い合わせ文を設定します。<br/><br/>
   *
   * テーブルオブジェクトに読み込むデータを取得するページ取得SELECT文を設定し、
   * 引数「parseParam」でTrueが指定された場合、パラメータの解析を行います。<br/>
   * 
   * @param newPageSQL ページ取得SQL文
   * @param offset ページオフセット
   * @param limit ページリミット
   * @param parseParam パラメータ解析を行うかどうか
   * @throws DataException  例外発生時
   */
  public void setPageSQL(String newPageSQL, int offset, int limit, boolean parseParam) throws DataException {
    //ページ取得SQL文に変換し、オフセット及びリミットを引数で置換
    String parseSQL = convertPageSQL(newPageSQL);
    parseSQL = parseSQL.replaceAll(PARAM_NAME_OFFSET, String.valueOf(offset));
    parseSQL = parseSQL.replaceAll(PARAM_NAME_LIMIT, String.valueOf(limit));
    //ページ取得SQL文として設定
    pageSQL = parseSQL;
    if (StringUtil.isNullString(pageSQL)) return;
    //パラメータ解析を行う場合
    if (parseParam) {
      //解析後のページ取得SQL文を設定
      jdbcPageSQL = params.parseSQL(pageSQL);
    }
  }

  /**
   * 引数で指定されたSQL文をページ取得用のSQL文に変換して返却します。<br/>
   * 
   * @param sql 返還前SQL文
   * @return ページ取得SQL
   */
  private String convertPageSQL(String sql) {
    String pagingSQL = "";
    switch (getSQLServerType()) {
    case ORACLE:
      pagingSQL = "SELECT * FROM ( SELECT ROWNUM \"num\", \"t\".* FROM (" + sql + ") \"t\" ) WHERE \"num\" BETWEEN " + PARAM_NAME_OFFSET + " + 1 AND " + PARAM_NAME_OFFSET + " + " + PARAM_NAME_LIMIT;
      break;
    case POSTRGESQL:
      pagingSQL = "SELECT * FROM (" + sql + ") AS PAGE OFFSET " + PARAM_NAME_OFFSET + " LIMIT " + PARAM_NAME_LIMIT;
      break;
    case MYSQL:
      pagingSQL = "SELECT * FROM (" + sql + ") AS PAGE OFFSET " + PARAM_NAME_OFFSET + " LIMIT " + PARAM_NAME_LIMIT;
      break;
    case H2:
      pagingSQL = "SELECT * FROM (" + sql + ") AS PAGE LIMIT " + PARAM_NAME_LIMIT + " OFFSET " + PARAM_NAME_OFFSET;
      break;
    case OTHER:
      return Const.NULLSTRING;
    }
    return pagingSQL;
  }
  
  /**
   * 全件数取得の際のSQL問い合わせ文を取得します。<br/>
   *
   * @return  全件数取得SQL文
   */
  public String getCountSQL(String sql) {
    return "SELECT COUNT(1) AS " + COUNT_COLUMN_NAME + " FROM (" + sql + ") CNT_TBL";
  }
  
  /**
   * パラメータのプレースホルダがJDBC形式に変換された後のSQL文を取得します。<br/><br/>
   *
   * ※注意　このとき、SQL中の &#47;*$ xxxx $*&#47; というコメント部分もTable#getParams().getValueAsString("xxxx")の値に置換されます。<br/>
   * 例：table.setSQL("SELECT AAA FROM TBL_X WHERE BBB = :BBB &#47;*$ CCC_IN $*&#47;");<br/>
   *     table.getParams().getValueAsString("CCC_IN") の値が "AND CCC IN ('X', 'Y')"<br/>
   * の場合<br/>
   * "SELECT AAA FROM TBL_X WHERE BBB = ? AND CCC IN ('X', 'Y')"<br/>
   * が結果として返されます。<br/>
   *
   * @return  JDBC形式変換後のSQL文
   */
  public String getJDBCSQL() {
    return getJDBCSQL(jdbcSQL);
  }
  public String getJDBCSQL(String sql) {

    if (StringUtil.isNullString(sql))
      return Const.NULLSTRING;
    if (params == null || params.getColumnCount() == 0)
      return sql;

    //  "\/\*$ ～ $\*\/" が一つもなければ sql をそのまま返す。
    int firstIndex = sql.indexOf("/*$ ");
    if (firstIndex < 0)
      return sql;
    int lastIndex = sql.lastIndexOf(" $*/");
    if (lastIndex < firstIndex)
      return sql;

    if (logger.isDebugEnabled())
      logger.debug("getJdbcSQL():before replace = " + sql);

    // "\/\* xxxx \*\/" をparamSet.getValueAsString("xxxx")の文字列に置換する。
    // firstIndexは '\/\*' かもしれないので、頭から走査しなおす。
    StringBuffer strBuf = new StringBuffer();
    int beforeIndexE = 0;
    int i = -1;
    while (i < lastIndex - 1) {
      i++;
      char c = sql.charAt(i);
      // "'" ～ "'"は読み飛ばす。
      if (c == '\'') {
        i++;
        while (i < lastIndex) {
          if (sql.charAt(i) == '\'')
            break;
          i++;
        }
        continue;
      }

      // "\/\*$ "でなければ次を検索
      if (!"/*$ ".equals(sql.substring(i, i + 4))) {
        continue;
      }
      int indexS = i;

      // " $\*\/" があるか?
      // " $\*\/" がなければ終了（走査しても意味がない）
      int indexE = sql.indexOf(" $*/", indexS + 5);
      if (indexE < 0)
        break;

      // "\/\*$ ～ $\*\/"以前の部分を追加
      strBuf.append(sql.substring(beforeIndexE, indexS));
      beforeIndexE = indexE + 4;

      // "\/\* KEYWORD \*\/"のKEYWORD部分の取得
      String keyword = sql.substring(indexS + 4, indexE).trim();
      if (keyword.length() == 0) {
        // キーワードがない場合は何も追加しない。つまり""に置換される。
        continue;
      }

      // 同じ名前のParamがあるか？
      if (params.indexOfColumn(keyword) != -1) {
        // あれば置換する。
        try {
          String replacedWord = params.getValueAsString(keyword);
          strBuf.append(replacedWord);
        } catch (DataException ex) {}
      } else {
        // なければ \/\* ～ \*\/ はそのままとする。（ヒントの可能性もあるため）
        strBuf.append(sql.substring(indexS, indexE + 4));
      }

    }

    // "\/\* ～ \*\/"以降の部分を追加
    strBuf.append(sql.substring(beforeIndexE));

    if (logger.isDebugEnabled())
      logger.debug("getJdbcSQL():after  replace = " + strBuf.toString());

    return strBuf.toString();

  }

  /**
   * Tableに読み込む最大行数を設定します。<br/><br/>
   *
   * ０を設定すると無制限になります。（デフォルト）<br/><br/>
   *
   * このプロパティに対する実際の挙動は派生クラス側の実装に依存します。<br/><br/>
   *
   * DBActionTableおよびSQLTableの場合、MaxRowsプロパティを設定すると、
   * データベースから取得する行数を制限することができます。<br/><br/>
   *
   * ※注 java.sql.StatementクラスのsetMaxRows(int)メソッドを利用しています。<br/>
   * 　　　従って、java.sql.Statement#setMaxRows(int)メソッドが実装されていないJDBCドライバーを
   * 利用した場合には、全件が取得されます。<br/>
   * また、newMaxRowsに0より小さい値を設定することはできません。<br/>
   * 
   * @param  newMaxRows  最大行数(0以上の整数,0は無制限)
   */
  public void setMaxRows(int newMaxRows) {
    if (newMaxRows < 0)
      return;
    maxRows = newMaxRows;
  }

  /**
   * Tableに読み込む最大行数を取得します。<br/>
   *
   * @return  最大行数
   */
  public int getMaxRows() {
    return maxRows;
  }

  /**
   * Tableに一度にロードする行数を設定します。<br/><br/>
   *
   * ０を設定すると無制限になります。（デフォルト）<br/><br/>
   *
   * このプロパティに対する実際の挙動は派生クラス側の実装に依存します。<br/><br/>
   *
   * @param  fetchSize  ロード行数
   */
  public void setFetchSize(int fetchSize) {
    this.fetchSize = fetchSize;
  }

  /**
   * Tableに一度にロードする行数を取得します。<br/>
   *
   * @return  ロード行数
   */
  public int getFetchSize() {
    return this.fetchSize;
  }

  /**
   * 実行中のJDBCStatementを中断します。<br/><br/>
   * 
   * JDBCStatementの実行をただちに中断します。<br/>
   * 使用するJDBCドライバーがStatement.cancel()メソッドを実装していない場合、正常に動作しません。<br/>
   * 
   * @exception DataException
   */
  public void cancelQuery() throws DataException {
    try {
      if (executing) jdbcStatement.cancel();
    } catch (SQLException x) {
      throw new DataException(x);
    }
  }

  /**
   * カラムのセットアップを行います。<br/><br/>
   *
   * SQLを実行し、JDBC ResultSetMetaDataよりカラムのセットアップを行います。<br/><br/>
   * 
   * ResultSetMetaDataから取得したカラム名が重複していた場合は、
   * カラム名にシーケンスを付加することにより重複しないようにしています。<br/>
   * 例えば、"ABC"というカラムが２つあった場合は、"ABC"と"ABC1"になります。<br/>
   * 又、ResultSetMetaDataからカラム名が取得できなかった場合は、
   * "UNKNOWN_COLUMN"というカラム名が設定されます。<br/>
   * これは、このクラスの定数として宣言されています。<br/>
   * （public static final String UNKNOWN_COLUMN_NAME = "UNKNOWN_COLUMN";）<br/>
   *
   * @param  table この引数は使用していません。
   * @exception  jp.littlesoft..data.DataException  例外発生時
   * @see #UNKNOWN_COLUMN_NAME
   */
  public <T extends Serializable> void setupColumn(Table<?> table) throws DataException {
 
    for (Column col: table.getColumns().values() ) {
      col.setFromDatabase(false);
    }
    try {
      if (StringUtil.isNullString(this.getSQL())) {
        throw new DataException(null, null, "SQL_IS_NULL", new String[]{""});
      }
      if (!isConnecting()) connect();
      setAutoCommit(false);

      // 照会ログ出力
      if (SQLLogger.isQueryDebugEnabled()) {
        String execSql = getSQL();
        if (getPageSQL() != null && !StringUtil.isNullString(getPageSQL())) execSql = getPageSQL();
        SQLLogger.logForQuery(getClientId(true), execSql, this.getParams());
      }

      //照会時間のログ出力
      long start = System.currentTimeMillis();

      //パラメータが無い場合
      if (this.getParams().getColumnCount()==0) {
        if (jdbcStatement==null) jdbcStatement = createStatement();
        if (getMaxRows()==0) {
          jdbcStatement.setMaxRows(getMaxRows());
        }
        else {
          jdbcStatement.setMaxRows(getMaxRows()+1);
        }
        executing = true;
        //ページ取得SQL文が設定されている場合
        if (getPageSQL() != null && !StringUtil.isNullString(getPageSQL())) {
          jdbcResultSet = jdbcStatement.executeQuery(getPageSQL());
          //全件数が未取得の場合は取得
          if (table.getAllRowCount()==-1) {
            SQLLogger.logForQuery(getClientId(true), getCountSQL(getSQL()), this.getParams());
            if (jdbcCountStatement==null) jdbcCountStatement = createStatement();
            jdbcCountResultSet = jdbcCountStatement.executeQuery(getCountSQL(getSQL()));
            jdbcCountResultSet.next();
            table.setAllRowCount(jdbcCountResultSet.getInt(COUNT_COLUMN_NAME));
          }
        //ページ取得SQL文が設定されていない場合
        } else {
          jdbcResultSet = jdbcStatement.executeQuery(getSQL());
        }
        
      //パラメータがある場合
      } else {
        //----  SQL中の "\/\*$ xxxx $\*\/" を sqlParameters.getValueAsString("xxxx") の値に置換する機能を追加
        String jdbcSql = this.getJDBCSQL();
        //ページ取得SQL文が設定されている場合はページ取得SQL文を置換する
        if (jdbcPageSQL != null && !StringUtil.isNullString(jdbcPageSQL)) {
          jdbcSql = this.getJDBCSQL(jdbcPageSQL);
        }
        // 前回の実行SQLがnull(つまり今回が初めて)、または前回とは異なるSQLを実行する場合は、PreparedStatementを作り直す。
        if (jdbcStatement==null || this.lastExecutedSql==null || !this.lastExecutedSql.equals(jdbcSql)) {
          // \/\*$ xxxx $\*\/ 形式のSQLで、置換した結果が""だった場合など、
          // jdbcSqlがnull or ""だったらまずいと思うのだが、チェックはしなくていいのか?
          jdbcStatement = preparedStatement(jdbcSql);
          this.lastExecutedSql = jdbcSql;
          table.setAllRowCount(-1);
        }
        bindParamValues((PreparedStatement)jdbcStatement, this.getParams());
        if (getMaxRows()==0) {
          jdbcStatement.setMaxRows(getMaxRows());
        }
        else {
          jdbcStatement.setMaxRows(getMaxRows()+1);
        }
        executing = true;
        jdbcResultSet = ((PreparedStatement)jdbcStatement).executeQuery();
        
        //全件数が未取得の場合は取得
        if (table.getAllRowCount()==-1 && (jdbcPageSQL != null && !StringUtil.isNullString(jdbcPageSQL))) {
          SQLLogger.logForQuery(getClientId(true), getCountSQL(getSQL()), this.getParams());
          if (jdbcCountStatement==null) {
            jdbcCountStatement = createStatement();
            jdbcCountStatement = preparedStatement(getCountSQL(this.getJDBCSQL()));
          }
          bindParamValues((PreparedStatement)jdbcCountStatement, this.getParams());
          jdbcCountResultSet = ((PreparedStatement)jdbcCountStatement).executeQuery();
          jdbcCountResultSet.next();
          table.setAllRowCount(jdbcCountResultSet.getInt(COUNT_COLUMN_NAME));
        }
        
      }
      
      executing = false;
      //this.rowContainer.setSQL(getSQL());
      final ResultSetMetaData jdbcResultSetMetaData = jdbcResultSet.getMetaData();
      final int colCount = jdbcResultSetMetaData.getColumnCount();
      // ResultSetMetaDataからColumns, resultSetColumnNamesを構築する。
      resultSetColumnNames = doSetupColumnInternal(table, colCount, jdbcResultSetMetaData, table.getColumns());
      // ---- 照会時間のログ出力
      if (SQLTimeLogger.isQueryDebugEnabled()) {
        long provideTime = System.currentTimeMillis() - start;
        String qaName = "";
        SQLTimeLogger.logForQuery(SQLTimeLogger.PROVIDE, getClientId(true), qaName, "", provideTime);
      }
    } catch (SQLException x) {
      freeStatement();
      throw new LSSQLException(x);
    } finally {
      executing = false;
    }
  }

  /**
   * ResultSetMetaDataからColumns, resultSetColumnNamesを構築します。<br/>
   * 
   * @param colCount カラムの個数
   * @param jdbcResultSetMetaData ResultSetMetaData
   * @param columns 構築するColumns
   * @return カラム名の配列 (resultSetColumnNamesとなる)
   * @throws SQLException
   * @throws DataException
   */
  private String[] doSetupColumnInternal(Table table, int colCount, ResultSetMetaData jdbcResultSetMetaData, IndexedMap<String,Column> columns) throws SQLException,DataException {
    String[] resultSetColumnNames = new String[colCount];
    final java.util.Map<String, Integer> resultSetColumnNameTable = new java.util.HashMap<String, Integer>();
    for (int i=1; i<=colCount; i++) {
      //String colName = jdbcResultSetMetaData.getColumnName(i);
      String colName = jdbcResultSetMetaData.getColumnLabel(i); //2009.10.26 by Nagao
      // カラム名が取得できなかった場合に対応する。
      if (StringUtil.isNullString(colName)) {
        colName = UNKNOWN_COLUMN_NAME;
      }
      // リザルトセットに同じカラム名が存在している場合に対応する。
      Integer duplicateNameCount = (Integer) resultSetColumnNameTable.get(colName);
      if (duplicateNameCount == null) {
        duplicateNameCount = INTEGER_ZERO;
      } else {
        duplicateNameCount = new Integer(duplicateNameCount.intValue() + 1);
      }
      resultSetColumnNameTable.put(colName, duplicateNameCount);
      if (duplicateNameCount.intValue() > 0) {
        colName = colName + duplicateNameCount.toString();
      }
      int colIndex = table.indexOfDBColumnName(colName);
      // カラムの属性を設定する。
      // Operation, SqlType, Precision, Scale, Nullable, DisplaySize, Length
      Column col;
      if (colIndex < 0) {
        // 未生成のカラムを生成する。
        col = new Column(colName,Column.mapToJava(jdbcResultSetMetaData.getColumnType(i)), true);
        table.addColumn(col);
      } else {
        col = columns.getValue(colIndex);
        if (!table.isTypedRow()) {
          col.setType( Column.mapToJava(jdbcResultSetMetaData.getColumnType(i)) );
        }
      }
      col.setSqlType(jdbcResultSetMetaData.getColumnType(i));
      if (col.getSqlType()==java.sql.Types.NUMERIC || col.getSqlType()==java.sql.Types.DECIMAL) {
        //UNION等使用した複雑なSQLの結果だとPrecisionがゼロになる不具合に対応。
        //OracleでFLOATを使用するとScale = -127 になる？
        //厳密に言うと Precision=-2 Scale=3 のようなケースもあるが、これは取り扱わないものとする。
        //hsqldbではPrecision=646456993となるので、これを無視するよう変更
        final int _precision = jdbcResultSetMetaData.getPrecision(i);
        final int _scale     = jdbcResultSetMetaData.getScale(i);
        if (_precision>0 && _precision<=30000 && _scale>=0) {            
          col.setPrecision(_precision);
          col.setScale(_scale);
        }
      }

      // JDBCAdaptorで、文字列用検証クラスが設定されている場合、それを割り当てる。(Postgres 7.2対応)
      if (col.getType().equals(Column.Type.STRING) && getColumnValidator() != null) {
        col.setValidator(getColumnValidator());
      }
      
      if (col.getNullable()==null) {
        col.setNullable(!(jdbcResultSetMetaData.isNullable(i)==ResultSetMetaData.columnNoNulls));
      }

      if (col.getDisplaySize()==Const.NOT_ASSIGNED) {
        col.setDisplaySize(jdbcResultSetMetaData.getColumnDisplaySize(i));
      }
      if (col.getType().equals(Column.Type.STRING)) {
        if (col.getLength() == Const.NOT_ASSIGNED) {
          col.setLength(jdbcResultSetMetaData.getColumnDisplaySize(i));
        }
      }
      //自動インクリメントか？
      col.setAutoIncrement(jdbcResultSetMetaData.isAutoIncrement(i));
      if (col.isAutoIncrement()) {
        col.setNullable(true);
      }
      
      //DBから取り出された
      col.setFromDatabase(true);
      resultSetColumnNames[i-1] = colName;

      if (logger.isDebugEnabled()) {
        logger.debug("*** SQLDataProvider.setupColumn ***");
        logger.debug("name      : " + col.getName());
        logger.debug("type      : " + col.getType().toString());
        logger.debug("length    : " + col.getLength());
        logger.debug("precision : " + col.getPrecision());
        logger.debug("scale     : " + col.getScale());
        logger.debug("null able : " + col.isNullable());
      }
    }
    return resultSetColumnNames;
  }
  
  /**
   * データのロードを行います。<br/>
   *
   * @exception  jp.littlesoft..data.DataException  例外発生時
   */
  public void loadData(Table table) throws DataException {
    loadResultSet(table, false);
  }

  /**
   * 追加行が存在するかどうかを調べます。<br/>
   *
   * @param toLast True：最終行まで調べる、False：最終行まで調べない
   * @return  True：存在する、False：存在しない
   * @exception  jp.littlesoft..data.DataException  例外発生時
   */
  public boolean checkMoreRow(Table table, boolean toLast) throws DataException {
    if (jdbcResultSet==null) return false;
    return (loadResultSet(table, toLast)>0);
  }

  /**
   * ResultSetからデータをロードします。<br/>
   *
   * @param toLast True：最終行まで調べる、False：最終行まで調べない
   * @return  ロードした行数
   * @exception  jp.littlesoft..data.DataException  例外発生時
   */
  private <T extends Serializable> int loadResultSet(Table<T> table, boolean toLast) throws DataException {
    int loadCount = 0;
    try {
      boolean hasMoreRow = true;
      table.setHasMoreRow(false);
      while(true) {
        while ((getFetchSize()==0 || toLast || loadCount<getFetchSize()) && (hasMoreRow = jdbcResultSet.next())) {
          if (getMaxRows()!=0 && loadCount==getMaxRows()) {
            hasMoreRow = jdbcResultSet.next();
            table.setHasMoreRow(true);
            break;
          }
          Row<T> newRow = new Row<T>(table);
          newRow.setValueFromResultSet(jdbcResultSet, resultSetColumnNames);
          newRow.setStatus(Row.Status.LOADED);
          table.internalPut(newRow, true, false);
          loadCount++;
        }
        if (loadCount==getFetchSize())
          break;
        else if (!jdbcStatement.getMoreResults())
          break;
        else
          jdbcResultSet = jdbcStatement.getResultSet();
      }
      if (!hasMoreRow) freeStatement();

    } catch (SQLException x) {
      freeStatement();
      throw new LSSQLException(x);
    } catch (DataException x) {
      freeStatement();
      throw x;
    }
    return loadCount;
  }

  /**
   * 内部リソースを解放します。<br/>
   *
   * @exception  jp.littlesoft..data.DataException  例外発生時
   */
  public void free() throws DataException {
    freeStatement();
  }

  /**
   * ResultSetを解放します。<br/>
   */
  private void freeResultSet() {
    if (jdbcResultSet != null) {
      try { jdbcResultSet.close(); } catch (java.sql.SQLException x) {}
      jdbcResultSet = null;
    }
    if (jdbcCountResultSet != null) {
      try { jdbcCountResultSet.close(); } catch (java.sql.SQLException x) {}
      jdbcCountResultSet = null;
    }
  }

  /**
   * Statementを解放します。<br/>
   */
  protected void freeStatement() {
    freeResultSet();
    if (jdbcStatement != null) {
      try {  jdbcStatement.close(); } catch (java.sql.SQLException x) {}
      jdbcStatement = null;
    }
    if (jdbcCountStatement != null) {
      try {  jdbcCountStatement.close(); } catch (java.sql.SQLException x) {}
      jdbcCountStatement = null;
    }
  }

  /**
   * 設定されているパラメータを取得します。<br/>
   * 
   * @return 設定されているパラメータ
   */
  public Params<?> getParams() {
    return params;
  }
  
  public void setParams(Params<?> params) {
    this.params = params;
  }

}
