/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.*;
import java.sql.*;
import java.util.*;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.DataResolver;
import jp.littlesoft.data.Params;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.event.ResolvingEvent;
import jp.littlesoft.data.event.ResolvingListener;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.exception.LSSQLException;
import jp.littlesoft.sql.logger.SQLLogger;
import jp.littlesoft.sql.logger.SQLTimeLogger;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;
import jp.littlesoft.util.IndexedMap;
import jp.littlesoft.util.res.MessageResource;

/**
 * Tableオブジェクトの内容をSQLデータベースへ書き戻すクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class SQLDataResolverImpl extends SQLAdaptor implements ISQLDataResolver {

  private static final long serialVersionUID = 1L;
  
  private static final transient String CN_SELECT    =  " SELECT ";
  private static final transient String CN_DELETE    =  " DELETE FROM ";
  private static final transient String CN_INSERT    =  " INSERT INTO ";
  private static final transient String CN_VALUES    =  " VALUES ( ";
  private static final transient String CN_UPDATE    =  " UPDATE ";
  private static final transient String CN_WHERE     =  " WHERE ";
  private static final transient String CN_AND       =  " AND ";
  private static final transient String CN_SET       =  " SET ";
  private static final transient String CN_FROM      =  " FROM ";
  private static final transient String CN_FORUPDATE =  " FOR UPDATE ";
  private static final transient String CN_WITH_UPDLOCK = " WITH(UPDLOCK) ";
  private static final transient String CN_OLD       =  " :@OLD@_";
  private static final transient String CN_EQ        =  " = ";
  private static final transient String CN_COM       =  ", ";
  private static final transient String CN_P         =  " :";
  private static final transient String CN_S         =  " ( ";
  private static final transient String CN_E         =  " ) ";
  private static final transient String CN_D         =  ".";
  private static final transient int SQL_TEXT_WHERE_LIST =  1;
  private static final transient int SQL_TEXT_UPDATE_LIST =  2;
  private static final transient int SQL_TEXT_INSERT_INTO_LIST =  3;
  private static final transient int SQL_TEXT_INSERT_VALUES_LIST =  4;
  //private static final transient int SQL_TEXT_SELECT_LIST =  5;
  
  private boolean pendingCommit = true; 
  
  // カラム名を"で囲む必要があるか
  private boolean colNameRequireDblQuote = false;
  // 楽観的排他制御の種別
  private OptimisticLockType optimisticLockType = OptimisticLockType.DEFAULT;
  private String[] keyColumns = null;
  private String targetDBTable = null;
  private String targetDBSchema = null;
  private String targetDBCatalog = null;
  private String sqlInsert = null;
  private String sqlUpdate = null;
  private String sqlDelete = null;
  private Vector<ResolvingListener> sqlResolvingListeners;

  /**
   *  コンストラクタ。<br/>
   */
  public SQLDataResolverImpl() {
  }

  /**
   * データベース更新対象行を識別するカラム名配列を設定します。<br/><br/>
   *
   * executeメソッドで自動生成、適用されるUPDATE, DELETE文のWHERE句を構成するカラム名を指定します。<br/>
   * このプロパティを省略した場合、Tableオブジェクトと更新対象DBテーブルに両方含まれる全ての列によってWHERE句に含まれます。<br/>
   * (但し、BLOBなど一般的にWHERE句に含めることができない列は除かれます)<br/><br/>
   *
   * このプロパティは楽観的同期処理(楽観的な排他)を実現する上で重要です。<br/><br/>
   *
   * UPDATE, DELETEのWHERE句に与えられる値はTableがDBテーブルから読み取った値(OldValue)です。<br/>
   * アプリケーションプログラムがTableのデータ値を変更した場合でも、OldValueは保持されます。<br/>
   * SQLDataResolverは、このOldValueをもとに更新対象行を識別しているので、他のトランザクションから更新対象行の内容が先変更されている場合は、
   * 更新対象行を特定できない為、例外を発生させます。<br/>
   * アプリケーションプログラムはこの例外をキャッチして、自らのトランザクションをロールバックする必要があります。<br/><br/>
   *
   * このような楽観的同期処理のメカニズムを実現する為、このプロパティにはDBテーブルの主キー列に加えて、
   * トランザクション競合を検出するための列を追加するようにしてください。<br/>
   *
   * 注意：データ取得のSELECT文でカラムの別名を指定した場合でも、当プロパティには実際のカラム名を指定する必要があります。<br/>
   * 
   * @param  keyColumns カラム名配列
   * @exception  jp.littlesoft.data.DataException 例外発生時
   */
  public void setKeyColumns(String[] keyColumns) throws DataException {
    this.keyColumns = keyColumns;
  }

  /**
   * データベース更新対象行を識別するカラム名配列をを取得します。<br/>
   * 
   * @return  カラム名配列
   * @exception  jp.littlesoft.data.DataException 例外発生時
   */
  public String[] getKeyColumns() {
    return this.keyColumns;
  }

  /**
   * 更新対象とするデータベーステーブル名を設定します。<br/><br/>
   *
   * 注意：Oracleのリモートデータベース上にあるテーブルは指定できません。<br/>
   * その場合は、更新対象のリモートテーブルを取得するビューを作成し、ビュー経由でデータの更新を行ってください。<br/>
   * 
   * @param  targetDBTable 更新テーブル名
   * @throws DataException 例外発生時
   */
  public void setTargetDBTable(String targetDBTable) throws DataException {
    this.targetDBTable = targetDBTable;
  }

  /**
   * 更新対象とするデータベーステーブル名を取得します。<br/>
   * 
   * @return  更新テーブル名
   */
  public String getTargetDBTable() {
    return  (this.targetDBTable == null) ? Const.NULLSTRING : this.targetDBTable;
  }

  /**
   * 更新対象データベーステーブルが存在するスキーマ名を設定します。<br/>
   * 
   * @param targetDBSchema スキーマ名
   * @throws DataException 例外発生時
   */
  public void setTargetDBSchema(String targetDBSchema) throws DataException {
    this.targetDBSchema = targetDBSchema;
  }

  /**
   * 更新対象データベーステーブルが存在するスキーマ名を取得します。<br/>
   * 
   * @return  更新テーブルスキーマ名
   */
  public String getTargetDBSchema() {
    return  (this.targetDBSchema == null) ? Const.NULLSTRING : this.targetDBSchema;
  }

  /**
   * 更新対象データベーステーブルが存在するカタログ名を設定します。<br/><br/>
   *
   * 注意：MS-SQL Server 2000を使用する場合で、接続ユーザと異なるカタログのテーブルを更新する場合、更新テーブルのカタログ名を設定します。<br/>
   * MS-SQL Server 2000以外を使用する場合や、接続ユーザと同じカタログ内のテーブルを更新する場合には、設定する必要はありません。<br/>
   * 
   * @param  targetDBCatalog  更新テーブルカタログ名
   **/
  public void setTargetDBCatalog(String targetDBCatalog) throws DataException {
    this.targetDBCatalog = targetDBCatalog;
  }

  /**
   * 更新対象データベーステーブルが存在するカタログ名を取得します。<br/>
   * 
   * @return  更新テーブルカタログ名
   */
  public String getTargetDBCatalog() {
    return  (this.targetDBCatalog == null) ? Const.NULLSTRING : this.targetDBCatalog;
  }

  /**
   * 挿入処理で使用するINSERT文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとexecuteメソッド内で、自動作成されるINSERT文が使用されます。<br/>
   * ※注　1行だけ挿入されるINSERT文を設定してください。
   * 
   * <pre>
   * 例:
   * setSQLInsert(
   *    "INSERT INTO TBL_X VALUES(:COL_A, :COL_B, CURRENT_TIMESTAMP)");
   * </pre>
   * 
   * :COL_A, :COL_Bには、テーブルオブジェクト内の挿入対象行オブジェクトの"COL_A","COL_B"のカラム値が割り当てられます。<br/>
   * ※上記例では、CURRENT_TIMESTAMPは、データベースの現在時刻に置き換えられます。(Oracleの場合)<br/>
   * 
   * @param sqlInsert INSERT文
   */
  public void setSQLInsert(String sqlInsert) {
    this.sqlInsert = sqlInsert;
  }

  /**
   * 挿入処理で使用するINSERT文を取得します。<br/><br/>
   * 
   * setSQLInsertで設定した値を取得することができますが、executeメソッド内で、自動作成される内容を取得する事はできません。<br/>
   *
   * @return INSERT文
   */
  public String getSQLInsert() {
    return  (this.sqlInsert == null) ? Const.NULLSTRING : this.sqlInsert;
  }

  /**
   * 更新処理で使用するUPDATE文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとexecuteメソッド内で、自動作成されるUPDATE文が使用されます。<br/>
   * ※注　1行だけ更新されるUPDATE文を設定してください。<br/>
   * 
   * <pre>
   * 例:
   * setSQLUpdate(
   *    "UPDATE TBL_X"
   *  + " SET COL_A = FUNC(:COL_A), UPDT_TIMESTAMP = CURRENT_TIMESTAMP"
   *  + " WHERE UPDT_TIMESTAMP = :@OLD@_UPDT_TIMESTAMP"
   *  + " AND ROWIDTOCHAR(ROWID) = :@OLD@_ROWIDCHAR"
   *  );
   * </pre>
   * 
   * :COL_Aには、テーブルオブジェクト内の更新対象行オブジェクトの"COL_A"のフィールド値が割り当てられます。<br/>
   * WHERE句のパラメータ名が":@OLD@_"で開始されている点に注意してください。<br/>
   * ":@OLD@_"で開始するパラメータには更新対象行が、変更される前に保持していた値（データベースから取り出した際の値）が設定されます。<br/>
   * 
   * @param sqlUpdate UPDATE文
   */
  public void setSQLUpdate(String sqlUpdate) {
    this.sqlUpdate = sqlUpdate;
  }

  /**
   * 更新処理で使用するUPDATE文を取得します。<br/><br/>
   * 
   * setSQLUpdateで設定した値を取得することができますが、executeメソッド内で、自動作成される内容を取得する事はできません。
   *
   * @return UPDATE文
   */
  public String getSQLUpdate() {
    return  (this.sqlUpdate == null) ? Const.NULLSTRING : this.sqlUpdate;
  }

  /**
   * 削除処理で使用するDELETE文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとexecuteメソッド内で、自動作成されるDELETE文が使用されます。<br/>
   * ※注　1行だけ更新されるDELETE文を設定してください。<br/>
   * 
   * <pre>
   * 例:
   * setSQLDelete(
   *    "DELETE FROM TBL_X"
   *  + " WHERE UPDT_TIMESTAMP = :@OLD@_UPDT_TIMESTAMP"
   *  + " AND ROWIDTOCHAR(ROWID) = :@OLD@_ROWIDCHAR"
   *  );
   * </pre>
   * 
   * WHERE句のパラメータ名が":@OLD@_"で開始されている点に注意してください。<br/>
   * ":@OLD@_"で開始するパラメータには更新対象行が、変更される前に保持していた値（データベースから取り出した際の値）が設定されます。<br/>
   * 
   * @param sqlDelete DELETE文
   */
  public void setSQLDelete(String sqlDelete) {
    this.sqlDelete = sqlDelete;
  }

  /**
   * 削除処理で使用するDELETE文を取得します。<br/><br/>
   * 
   * setSQLDeleteで設定した値を取得することができますが、resolveChangeメソッド内で、自動作成される内容を取得するは事できません。<br/>
   * 
   * @return DELETE文
   */
  public String getSQLDelete() {
    return  (this.sqlDelete == null) ? Const.NULLSTRING : this.sqlDelete;
  }
  
   /**
   * テーブルオブジェクトの変更内容をもとにデータベーステーブルの更新・追加・削除を行います。<br/>
   * 
   * @param  table テーブルオブジェクト
   * @exception  jp.littlesoft.exception.data.DataException 例外発生時
   */
  public synchronized void execute(Table<?> table) throws DataException {
    execute(table, this.isPendingCommit(), ResolvingEvent.Type.ANY);
  }

  /**
   * テーブルオブジェクトの変更内容をもとにデータベーステーブルの更新・追加・削除を行います。<br/>
   * 
   * @param table テーブルオブジェクト
   * @param pendingCommit True：このメソッド内でコミットを行わない、False：行う
   * @throws jp.littlesoft.exception.data.DataException 例外発生時
   */
  public synchronized void execute(Table<?> table, boolean pendingCommit) throws DataException {
    execute(table, pendingCommit, ResolvingEvent.Type.ANY);
  }
  
  /**
   * テーブルオブジェクトの変更内容をもとにデータベーステーブルの更新・追加・削除を行います。<br/>
   * 
   * @param table   テーブルオブジェクト
   * @param pendingCommit True：このメソッド内でコミットを行わない、False：行う
   * @param resolveType テーブルオブジェクトデータ永続化の種類
   * @throws DataException 例外発生時
   * @see jp.littlesoft.data.event.ResolvingEvent.Type
   */
  @SuppressWarnings("unchecked")
  public synchronized <E extends Serializable> void execute(Table<E> table, boolean pendingCommit, ResolvingEvent.Type resolveType) throws DataException {
    if (StringUtil.isNullString(this.getTargetDBTable())) {
      throw new LSSQLException(null, null, "UPDATE_TABLENAME_NOT_SPECIFIED", new String[]{""});
    }

    if (this.keyColumns != null) {
      for (int i = 0; i < keyColumns.length; i++) {
        Column col;
        if ((col = table.getColumn(this.keyColumns[i],true)) != null) {
          col.setResolveKey(true);
        }
      }
    }
    final IndexedMap<String, Column> resolveColumns = new IndexedMap<String, Column>();
    {
      Column[] cols = table.getColumnArray();
      if (cols != null) {
        for (int i=0; i<cols.length; i++) {
          Column col = cols[i];
          if (col.getDBColumnName().length() == 0) {
            resolveColumns.put(col.getName(), col);
          } else {
            //DBColumnNameが定義されているものは、その名前をColumnの名前とする。
            Column newCol = (Column)col.clone();
            newCol.setName(col.getDBColumnName());
            resolveColumns.put(newCol.getName(), newCol);
          }
        }
      }
    }
    
    // 更新されたカラムが1つ以上あるか?
    boolean anyUpdate = false;
    for (int x=0; x<resolveColumns.size(); x++) {
      if (resolveColumns.getValue(x).isUpdated()) {
        anyUpdate = true;
        break;
      }
    }

    boolean execInsert = (resolveType == ResolvingEvent.Type.ANY || resolveType == ResolvingEvent.Type.INSERT);
    boolean execDelete = (resolveType == ResolvingEvent.Type.ANY || resolveType == ResolvingEvent.Type.DELETE);
    boolean execUpdate = (resolveType == ResolvingEvent.Type.ANY || resolveType == ResolvingEvent.Type.UPDATE);
    
    ArrayList[] resolveRows = {new ArrayList<Row>(), new ArrayList<Row>(), new ArrayList<Row>()};

    // 更新対象データを抽出
    for (Row resolveRow : table.allRows()) {
      switch (resolveRow.getStatus()) {
      case DELETED:
        if (execDelete) {
          resolveRows[ResolvingEvent.Type.DELETE.ordinal()].add(resolveRow);
        }
        break;
      case UPDATED:
        if (anyUpdate && execUpdate) {
          resolveRows[ResolvingEvent.Type.UPDATE.ordinal()].add(resolveRow);
        }
        break;
      case CREATED:
        if (execInsert) {
          resolveRows[ResolvingEvent.Type.INSERT.ordinal()].add(resolveRow);
        }
        break;
      }
    }
    
    // 更新対象データなしの場合、終了
    if (resolveRows[ResolvingEvent.Type.DELETE.ordinal()].size() +
        resolveRows[ResolvingEvent.Type.UPDATE.ordinal()].size() +
        resolveRows[ResolvingEvent.Type.INSERT.ordinal()].size() ==0) {
      return;
    }

    final PreparedStatement[] resolveStatements = new PreparedStatement[3];

    try {
      setAutoCommit(false);
      String schema = this.getTargetDBSchema();
      if (schema==null || Const.NULLSTRING.equals(schema)) {
        schema = getDBUserName();
      }
      // 更新対象テーブルのメタデータをもとに更新SQL文構築用カラムを生成
      IndexedMap<String,Column> updateColumns = getUpdateTableColumns(this.jdbcConnection.getMetaData(), this.getTargetDBCatalog(),
          schema, this.getTargetDBTable(), this.isSearchMetaDataWithinSchema(),
          this.isSearchMetaDataByUpperCase(), this.isSearchMetaDataByLowerCase());
      if (updateColumns.size()==0) {
        throw new LSSQLException(null, null, "INVALID_UPDATE_TABLENAME",
          new String[]{"", this.getTargetDBTable()});
      }

      // 主キー列を取得
      List pkeys = getPrimaryKeyColumns(this.getTargetDBCatalog(), schema, this.getTargetDBTable());

      final String sqls[] = new String[3];
      final Params[] resolveParams = {new Params(), new Params(), new Params()};

      // UPDATE/INSERT/DELETE全ての合計時間の測定を開始する。
      long start = System.currentTimeMillis();

      // 更新SQL文の生成、ステートメントの生成
      // 0:DELETE, 1:UPDATE, 2:INSERT
      for (int i=0; i<3; i++) {
        ResolvingEvent.Type sqlEventType = ResolvingEvent.Type.values()[i];
        if (resolveRows[i].size()>0) {
          String sql = createResolveSql(sqlEventType, table, pkeys, updateColumns, this.getSQLServerType(), this.isColNameRequireDblQuote());
          if (sql==null) {
            resolveRows[i].clear();
            continue;
          }
          String jsql = resolveParams[i].parseSQL(sql);
          resolveStatements[i] = preparedStatement(jsql);
          sqls[i] = sql;
        }
      }
      ResolvingEvent sqlResEvent = null;
      if (table!=null) sqlResEvent = new ResolvingEvent(table);

      // 更新処理
      // 0:DELETE, 1:UPDATE, 2:INSERT
      for (int i=0; i<3; i++) {
        ResolvingEvent.Type sqlEventType =  ResolvingEvent.Type.values()[i];
        if (table!=null) {
          sqlResEvent.setSQL(sqls[i]);
          sqlResEvent.setParams(resolveParams[i]);
        }
        for (int row=0; row<resolveRows[i].size(); row++) {
          Row resolveRow = (Row)resolveRows[i].get(row);
          Row targetRow = resolveRow.getOriginalRow();
          resolveParams[i].clear();
          // 更新用パラメータセットにデータ型を設定
          setResolveParamProperties(resolveParams[i], resolveColumns, resolveRow, targetRow);
          // TableのBeforeResolveRowイベントを送出
          if (table!=null) {
            sqlResEvent.setType(sqlEventType);
            sqlResEvent.setNewRow(resolveRow);
            sqlResEvent.setOldRow(targetRow);
            fireRowResolving(sqlResEvent);
          }

          // 更新用パラメータセットにデータ値を設定
          setResolveParamValues(resolveParams[i], resolveColumns, resolveRow, targetRow);
          // 更新用ステートメントにパラメータをバインド
          bindParamValues(resolveStatements[i], resolveParams[i]);

          try {
            // 楽観的排他制御
            // OptimisticLockTypeプロパティがPRE_FETCHの場合は、executeUpdate()の返り値ではなく
            // SELECT の件数で楽観的排他制御を行う。
            // (SQLServer 2000でトリガーを使用した場合、executeUpdate()でトリガーにより更新された行数が返されるため)
            int updateCount = -1; // SELECT文の件数 または executeUpdate()の返り値
            int expectUpdateCount = 1; // 期待値
            if (this.getOptimisticLockType() == OptimisticLockType.PRE_FETCH) {
              if (sqlEventType == ResolvingEvent.Type.INSERT) {
                //INSERTの場合は、0件が期待値。
                expectUpdateCount=0;
              }
              
              // 件数を数えるためのSELECT文の作成
              Params selectCountParam = new Params();
              // 行数を数えるカラム名は更新対象カラムの先頭カラム名とする。
              String _wkSelectCountSQL = createSelectCountSQL(this.getSQLServerType(), sqlEventType, table, pkeys, updateColumns, this.isColNameRequireDblQuote());
              String selectCountSQL = selectCountParam.parseSQL(_wkSelectCountSQL);
              PreparedStatement selectCountStatement = preparedStatement(selectCountSQL);
              // 件数を数えるための SELECT文 のパラメータセットにデータ型を設定
              setResolveParamProperties(selectCountParam, resolveColumns, resolveRow, targetRow);
              setResolveParamValues(selectCountParam, resolveColumns, resolveRow, targetRow);
              bindParamValues(selectCountStatement, selectCountParam);
              // 件数を数えるためのSELECT文をログ出力する。
              if (SQLLogger.isQueryDebugEnabled()) {
                SQLLogger.logForQuery(this.getClientId(true), _wkSelectCountSQL, selectCountParam);
              }
              ResultSet selectCountRs = selectCountStatement.executeQuery();
              int selectCount = 0;
              while (selectCountRs.next()) {
                selectCount++;
              }
              updateCount = selectCount;             
              selectCountRs.close();
              selectCountStatement.close();
            }

            if (this.getOptimisticLockType() != OptimisticLockType.PRE_FETCH || updateCount == expectUpdateCount) {
              // バインドした後ならパラメタの値が設定されているので、ここで更新ログを出力する。
              if (SQLLogger.isUpdateDebugEnabled()) SQLLogger.logForUpdate(this.getClientId(true), sqls[i], resolveParams[i]);

              // 更新用ステートメント実行
              int executeUpdateCount = resolveStatements[i].executeUpdate();
              if (this.getOptimisticLockType() != OptimisticLockType.PRE_FETCH) {
                // OptimisticLockTypeプロパティがPRE_FETCHでなければ、executeUpdateの返り値で更新件数をチェックする。
                updateCount = executeUpdateCount;
              }              
            }
            
            if (updateCount == expectUpdateCount) {
              //
            } else {
              // updateCountとexpectUpdateCountの値が異なる場合は外部から何らかの更新があったと判定し、エラーとする。
              String debugErrmsgResourceName = null;
              String errmsgResource = null;;
              switch (sqlEventType) {
                case DELETE:
                  debugErrmsgResourceName = "DELETE_COUNT_ERROR";
                  errmsgResource = "ERROR_RESOLVING_DELETE";
                  break;
                case UPDATE:
                  debugErrmsgResourceName = "UPDATE_COUNT_ERROR";
                  errmsgResource = "ERROR_RESOLVING_UPDATE";
                  break;
                case INSERT:
                  if (this.getOptimisticLockType() == OptimisticLockType.PRE_FETCH) {
                    debugErrmsgResourceName = "INSERT_COUNT_ERROR2";
                  } else {
                    debugErrmsgResourceName = "INSERT_COUNT_ERROR";                    
                  }
                  errmsgResource = "ERROR_RESOLVING_INSERT";
                  break;
              }              
              logger.debug(MessageResource.getMessage(null, debugErrmsgResourceName,
                  new String[]{Integer.toString(updateCount)}, null));
              // TableのOnResolvingErrorイベントを送出
              if (table!=null) {
                String errmsg = MessageResource.getMessage(null, errmsgResource, null, null);
                sqlResEvent.setAbort(true);
                sqlResEvent.setAbortMessage(errmsg);
                fireRowResolvingError(sqlResEvent, null);
                if (sqlResEvent.isAbort()) {
                  throw new LSSQLException(null, null, errmsgResource, null);
                }
              }
            }
            fireRowResolved(sqlResEvent);
          }
          catch (java.sql.SQLException x) {
            if (table!=null) {
              sqlResEvent.setAbort(true);
              sqlResEvent.setAbortMessage(null);
              fireRowResolvingError(sqlResEvent, x);
              if (sqlResEvent.isAbort()) {
                throw x;
              }
            }
          }
        } // end of for (row=0; row<resolveRows[i].size(); row++) {
      }  // end of for (i=0; i<3; i++) {

      // UPDATE/INSERT/DELETE全ての合計時間をログ出力する。
      if (SQLTimeLogger.isUpdateDebugEnabled()) {
        long executeTime = System.currentTimeMillis() - start;
        String qaName = "";
        SQLTimeLogger.logForUpdate(SQLTimeLogger.RESOLVE_CHANGES, this.getClientId(true), qaName, "", executeTime);
      } 

      if (!pendingCommit) {
        jdbcConnection.commit();
      }

      // 内部ステータスをリセット
      for (int i=0; i<3; i++) {
        for (int row=0; row<resolveRows[i].size(); row++) {
          Row resolveRow = (Row)resolveRows[i].get(row);
          if (i==ResolvingEvent.Type.DELETE.ordinal()) {
            table.internalRemove(resolveRow.getRowID());
          } else {
            resolveRow.setStatus(Row.Status.LOADED);
          }
        }
      }
    } catch(Throwable e) {
      try {
        if (!pendingCommit) jdbcConnection.rollback();
      } catch (SQLException s) {
      }
      if(e instanceof SQLException) throw new LSSQLException((SQLException)e);
      if(e instanceof DataException) throw (DataException)e;
      throw new DataException(e);
    } finally {
      // 更新SQL用のステートメントetc.を解放
      for (int i=0; i<3; i++) {
        if (resolveStatements[i] != null) {
          try {
            resolveStatements[i].close();
          } catch  (SQLException s) {
            logger.warn(s);
            s.printStackTrace();
          }
        }
      }
    }
  }

  /**
   * パラメータのTypeやScaleなどを設定します。<br/>
   * 
   * @param resolveParam
   * @param resolveColumns
   * @param resolveRow
   * @param oldRow
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  private static void setResolveParamProperties(Params resolveParam, IndexedMap<String,Column> resolveColumns,
      Row resolveRow, Row oldRow) 
      throws DataException
  {
    for (int pcol=0; pcol<resolveParam.getColumnCount(); pcol++) {
      Column param = resolveParam.getColumn(pcol);
      int dcol;
      if (param.getName().startsWith("@OLD@_")) {
        dcol = resolveColumns.indexOfKey(param.getName().substring(6,param.getName().length()));
        resolveParam.getColumn(pcol).setType( oldRow.getTable().getColumn(dcol).getType());
        resolveParam.getColumn(pcol).setSqlType( oldRow.getTable().getColumn(dcol).getSqlType());
        resolveParam.getColumn(pcol).setPrecision( oldRow.getTable().getColumn(dcol).getPrecision());
        resolveParam.getColumn(pcol).setScale( oldRow.getTable().getColumn(dcol).getScale());
      } else {
      	// 新しい値のパラメータ名は"@NEW@_"で始めた方がわかりやすいのでは？
        dcol = resolveColumns.indexOfKey(param.getName());
        resolveParam.getColumn(pcol).setType( resolveRow.getTable().getColumn(dcol).getType());
        resolveParam.getColumn(pcol).setSqlType( resolveRow.getTable().getColumn(dcol).getSqlType());
        resolveParam.getColumn(pcol).setPrecision( resolveRow.getTable().getColumn(dcol).getPrecision());
        resolveParam.getColumn(pcol).setScale( resolveRow.getTable().getColumn(dcol).getScale());
      }
      Column.TrimType dcolTrimType = resolveColumns.getValue(dcol).getTrimType();
      param.setTrimType(dcolTrimType);
    }
  }

  /**
   * パラメータに値を設定します。<br/>
   * 
   * @param resolveParam
   * @param resolveColumns
   * @param resolveRow
   * @param oldRow
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  private static void setResolveParamValues(Params resolveParam, IndexedMap<String,Column> resolveColumns,
      Row resolveRow, Row oldRow) 
  throws DataException {
    for (int pcol=0; pcol<resolveParam.getColumnCount(); pcol++){
      Column param = resolveParam.getColumn(pcol);
      int dcol;
      if (param.getName().startsWith("@OLD@_")) {
        dcol = resolveColumns.indexOfKey(param.getName().substring(6,param.getName().length()));
        resolveParam.setValue(pcol, oldRow.getValue(dcol));
      } else {
        dcol = resolveColumns.indexOfKey(param.getName());
        resolveParam.setValue(pcol, resolveRow.getValue(dcol));
      }
    }

  }
  
  /**
   * 更新対象の行数を数えるためのSELECT文を生成する。<br/>
   * 
   * @param sqlServerType データベースサーバの種類
   * @param statementTYpe 更新の種類(0:DELETE/1:UPDATE/2:INSERT)
   * @param valueCache    更新する内容であるValueCache
   * @param pkeys         更新対象テーブルの主キー名を格納したList
   * @param updateFields  更新対象テーブルのカラム集合
   * @return 更新対象の行数を取得するためのSELECT文
   * @throws DataException 例外発生時
   */
  @SuppressWarnings("unchecked")
  private String createSelectCountSQL(SQLServerType sqlServerType, ResolvingEvent.Type statementType, Table valueCache, List pkeys,
      IndexedMap<String,Column> updateFields, boolean isColNameRequireDblQuote) throws DataException {
    // 行数を数えるためのカラム名を決める。
    String countFieldName = null;
    Column countField = updateFields.getValue(0);
    countFieldName = countField.getDBColumnName();
    if (StringUtil.isNullString(countFieldName)) {
      countFieldName = countField.getName();
    }
    
    StringWriter sw = new StringWriter();
    sw.write(CN_SELECT);
    sw.write(getNameEnclosedDblQuoteInternal(isColNameRequireDblQuote, countFieldName)); // 行数を数えるためのカラム
    sw.write(CN_FROM);
    sw.write(getNameEnclosedDblQuoteInternal(isColNameRequireDblQuote, createResolveTableName(valueCache)));
    sw.write(getWithUpdlock(sqlServerType));
    sw.write(CN_WHERE);
    boolean forceNewRow;
    if (statementType == ResolvingEvent.Type.DELETE || statementType == ResolvingEvent.Type.UPDATE ) {
      // DELETE, UPDATEの場合は、変更前の値を見る。
      forceNewRow = false;
    } else {
      // INSERTの場合は、現在行の値を見る。
      forceNewRow = true;
    }
    createResolveSqlText(SQL_TEXT_WHERE_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, forceNewRow);
    sw.write(getForUpdate(sqlServerType));
    return sw.toString();
  }

  private static String getForUpdate(SQLServerType sqlServerType) {
    switch (sqlServerType) {
      case MS_SQL:
      case HSQLDB:
        // SQL Serverの場合、"WITH(UPDLOCK)"を使うので、" FOR UPDATE "句は使用しない。
        // HSQLDBでは"FOR UPDATE"が使えない
        return Const.NULLSTRING;
      default:
        return CN_FORUPDATE;
    }
  }

  private static String getWithUpdlock(SQLServerType sqlServerType) {
    if (sqlServerType == SQLServerType.MS_SQL) {
      // SQL Serverの場合、"WITH(UPDLOCK)"とする。
      return CN_WITH_UPDLOCK;
    } else {
      // SQL Server以外では、""を返す。
      return Const.NULLSTRING;
    }
  }

  /**
   * 更新テーブルのColumnsを作成します。<br/>
   *
   * @param metaData   メタデータ
   * @param catalog    カタログ名
   * @param schema     スキーマ名
   * @param tableName  テーブル名
   * @param isSearchMetaDataWithinSchema True：スキーマ名指定で検索する
   * @param isSearchMetaDataByUpperCase  True：テーブル名を大文字限定で検索する
   * @param isSearchMetaDataByLowerCase  True：テーブル名を小文字限定で検索する
   * @exception  java.sql.SQLException  SQL例外発生時
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  protected IndexedMap<String,Column> getUpdateTableColumns(DatabaseMetaData metaData, String catalog, String schema, String tableName,
      boolean isSearchMetaDataWithinSchema, boolean isSearchMetaDataByUpperCase, boolean isSearchMetaDataByLowerCase)
  throws SQLException, DataException {

    if (schema==null || Const.NULLSTRING.equals(schema)) {
      schema = metaData.getUserName();
    }
    String tables = tableName;
    String schemas = null;
    // searchMetaDataWithinSchemaプロパティがfalseの場合、スキーマパターンはnullとする
    if (isSearchMetaDataWithinSchema && StringUtil.isNotNullString(schema)) {
      schemas = schema;
    }

    //とりあえずORACLEは大文字でないとヒットしない。。。
    //他のDBMSは調査すべし。
    if (isSearchMetaDataByUpperCase) {
      // テーブル名を大文字限定で検索する。
      if (schemas != null) schemas = schemas.toUpperCase();
      tables = tableName.toUpperCase();
    } else if (isSearchMetaDataByLowerCase) {
      // テーブル名を小文字限定で検索する。
      if (schemas != null) schemas = schemas.toLowerCase();
      tables = tableName.toLowerCase();
    }

    // スキーマ名が null ならカタログ名も null とする。
    // カタログ名が "" の場合、検索できないので null とする。
    if (StringUtil.isNullString(schemas) || StringUtil.isNullString(catalog)) {
      catalog = null; 
    }
    ResultSet rs = metaData.getColumns(catalog, schemas, tables, null);

    IndexedMap<String,Column> updateTableFields = new IndexedMap<String,Column>();
    //updateTableColumnSet.removeAll();
    while (rs.next()) {
      String metaCatalogName = rs.getString(1);
      String metaSchemaName = rs.getString(2);
      String metaTableName  = rs.getString(3);
      String metaColumnName = rs.getString(4);

      if ( (StringUtil.isNullString(catalog) || catalog.equals(metaCatalogName))
        && (StringUtil.isNullString(schemas) || schemas.equals(metaSchemaName))
        && tables.equals(metaTableName) ) {
        Column col = new Column();
        col.setName(metaColumnName); //カラム名
        updateTableFields.put(col.getName(), col);
      }
    }
    rs.close();
    return updateTableFields;
  }

  /**
   * 更新テーブル名を作成します。<br/><br/>
   *
   * スキーマ名が設定されている場合、スキーマ名を付加します。<br/>
   *
   * @param  valueCache テーブルオブジェクト
   */
  @SuppressWarnings("unchecked")
  private String createResolveTableName(Table valueCache) {
    StringBuffer updateTableNameBuf = new StringBuffer();
    String catalog = this.getTargetDBCatalog();
    if (StringUtil.isNotNullString(catalog)) {
      updateTableNameBuf.append(catalog).append(CN_D);
    }
    
    String schema = this.getTargetDBSchema();
    if (StringUtil.isNullString(schema)) {
      // schemaがnullならカタログは無視する。
      return this.getTargetDBTable();
    } else {
      updateTableNameBuf.append(schema).append(CN_D).append(this.getTargetDBTable());
      return updateTableNameBuf.toString();
    }
  }

  /**
   * 更新SQLを作成します。<br/>
   *
   * @param   statementType  更新SQLのタイプ
   * @param   valueCache  RowContainer
   * @param   pkeys 主キー列名
   * @return  更新SQL
   * 
   * @see jp.littlesoft.data.event.ResolvingEvent
   */
  @SuppressWarnings("unchecked")
  private String createResolveSql(ResolvingEvent.Type statementType, Table valueCache, List pkeys, IndexedMap<String,Column> updateFields,
      SQLServerType sqlServerType, boolean isColNameRequireDblQuote) {
    java.io.StringWriter sw = new java.io.StringWriter();
    boolean isFirst;
    switch (statementType) {
      case DELETE:
        if (StringUtil.isNullString(this.getSQLDelete())) {
          sw.write(CN_DELETE);
          sw.write(createResolveTableName(valueCache));
          sw.write(CN_WHERE);
          createResolveSqlText(SQL_TEXT_WHERE_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, false);
        } else {
          sw.write(this.getSQLDelete());
        }
        break;
      case UPDATE:
        if (StringUtil.isNullString(this.getSQLUpdate())) {
          sw.write(CN_UPDATE);
          sw.write(createResolveTableName(valueCache));
          sw.write(CN_SET);
          isFirst = createResolveSqlText(SQL_TEXT_UPDATE_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, false);
          if (isFirst) {
            return null;
          }
          sw.write(CN_WHERE);
          createResolveSqlText(SQL_TEXT_WHERE_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, false);
        } else {
          sw.write(this.getSQLUpdate());
        }
        break;
      case INSERT:
        if (StringUtil.isNullString(this.getSQLInsert())) {
          sw.write(CN_INSERT);
          sw.write(createResolveTableName(valueCache));
          sw.write(CN_S);
          isFirst = createResolveSqlText(SQL_TEXT_INSERT_INTO_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, false);
          if (isFirst) {
            return null;
          }
          sw.write(CN_E);
          sw.write(CN_VALUES);
          createResolveSqlText(SQL_TEXT_INSERT_VALUES_LIST, sw, valueCache, pkeys, updateFields, sqlServerType, isColNameRequireDblQuote, false);
          sw.write(CN_E);
        } else {
          sw.write(this.getSQLInsert());
        }
        break;
    }
    sw.flush();
    String sql = sw.toString();
    return sql;
  }

  /**
   * 更新SQL文字列を作成します。<br/><br/>
   *
   * 作成した文字列は、sqlWriterに書き込まれます。<br/>
   *
   * @param  textType   テキストタイプ
   * @param  sw  StringWriter
   * @param  table  RowContainer
   * @param  pkeys 主キー列名
   * @param  forceNewRow 強制的に現在行の値を参照するか?
   * @return 最初のSQL句か否か
   */
  @SuppressWarnings("unchecked")
  private boolean createResolveSqlText(int textType, java.io.StringWriter sw, Table table, List pkeys,
      IndexedMap<String,Column> updateColList, SQLServerType sqlServerType, boolean isEnclosedDblQuote, boolean forceNewRow ) {
    int i;
    boolean isFirst = true;
    boolean isPKey = false;
    // ResolveKeyColumnsが一つも定義されていない
    boolean isNoResolveKeyColumns = true;

    Column col[] = table.getColumnArray();

    /*
      RowIDColumnNamesが設定されているか？を検査。
      何も存在しない場合はisNoRowIDColumnsがtrueとなり、
      すべてのカラムを使用してWhere文が構成される様になる。
      (BINARY、LOGNBINARY、BLOB、CLOB、OBJECT)
    */
    if (col == null) {
      return isFirst;
    }
    
   	for (i = 0; i<col.length; i++) {
     switch (col[i].getType()) {
       case BINARY:
       case BLOB:
       case LONGBINARY:
       //case CLOB:  CLOBは主キーにしてもよい？
       case JSONOBJECT:
       continue;
     }
     if (col[i].isResolveKey()) {
       isNoResolveKeyColumns = false;
       break;
     }
   	}

   	for (i=0; i<col.length; i++) {
     		String dbColumnName = col[i].getDBColumnName();
     		if (dbColumnName.length() == 0) {
       			dbColumnName = col[i].getName();
     		}
     		isPKey = (pkeys != null && pkeys.indexOf(dbColumnName)>-1);
     // NewRowを参照する場合は、":@OLD@_"を付けない。
     String oldColName = forceNewRow ? ":" + dbColumnName : CN_OLD + dbColumnName;
     // 必要に応じてカラム名をダブルクォーテーションで囲む
     String printColName = getNameEnclosedDblQuoteInternal(isEnclosedDblQuote, dbColumnName);
     switch (textType) {
       case SQL_TEXT_WHERE_LIST:
         if ((isNoResolveKeyColumns || col[i].isResolveKey() || isPKey) && (updateColList.indexOfKey(dbColumnName)!=-1)) {
            // BINARY、LOGNBINARY、BLOB、CLOB、OBJECTはWHERE句に含めない
            switch (col[i].getType()) {
              case BINARY:
              case BLOB:
              case LONGBINARY:
              //case CLOB:  CLOBは主キーにしてもよい？
              case JSONOBJECT:
                continue;
            }

            if (isFirst)
              isFirst = false;
            else
              sw.write(CN_AND);

            /*
              ・主キー構成カラムは暗黙でwhere句に含める
              ・主キー構成カラム以外がNULL可の場合であっても比較が成立する様、柔軟に処理する。
              ・以下は旧コード
              //sqlWriter.write(col[i].getName() + CN_EQ + CN_OLD + col[i].getName() );
            */
            if (isPKey || !col[i].isNullable()) {
              if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.BIGDECIMAL)) {
                //PostgreSQL BIGDECIMAL対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS NUMERIC) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIMESTAMP)) {
                //PostgreSQL TIMESTAMP対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIMESTAMP) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.DATE)) {
                //PostgreSQL DATE対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS DATE) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIME)) {
                //PostgreSQL TIME対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIME) ");
              } else {
                //その他のDBMS
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(oldColName);
              }
            } else {
              //以下のような条件分となります。
              // ( ( COLNAME = :@OLD@COLNAME ) OR ( :@OLD@COLNAME IS NULL AND COLNAME IS NULL ) )
              sw.write(" ( ( ");
              if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.BIGDECIMAL)) {
                //PostgreSQL BIGDECIMAL対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS NUMERIC) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIMESTAMP)) {
                //PostgreSQL TIMESTAMP対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIMESTAMP) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.DATE)) {
                //PostgreSQL DATE対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS DATE) ");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIME)) {
                //PostgreSQL TIME対応
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIME) ");
              } else {
                //その他のDBMS
                sw.write(printColName);
                sw.write(CN_EQ);
                sw.write(oldColName);
              }
              sw.write(" ) OR ( ");
              if (sqlServerType == SQLServerType.IBM_DB2) {
                // DB2では正しい型にCASTしないとエラーになる。
                sw.write(makeWhereClauseForDB2(col[i], printColName, oldColName));
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIMESTAMP)) {
                //PostgreSQL TIMESTAMP対応
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIMESTAMP) ");
                sw.write(" IS NULL AND (");
                sw.write(printColName);
                sw.write(" IS NULL) )");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.TIME)) {
                //PostgreSQL TIME対応
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS TIME) ");
                sw.write(" IS NULL AND (");
                sw.write(printColName);
                sw.write(" IS NULL) )");
              } else if (sqlServerType==SQLServerType.POSTRGESQL && col[i].getType().equals(Column.Type.DATE)) {
                //PostgreSQL DATE対応
                sw.write(" CAST(");
                sw.write(oldColName);
                sw.write(" AS DATE) ");
                sw.write(" IS NULL AND (");
                sw.write(printColName);
                sw.write(" IS NULL) )");
              } else {
                sw.write(oldColName);
                sw.write(" IS NULL AND (");
                sw.write(printColName);
                sw.write(" IS NULL");
                if (sqlServerType == SQLServerType.MYSQL && col[i].getType().equals(Column.Type.STRING)) {
                  // MySQLの場合は、NULLと''を区別するので、''とも比較する。
                  sw.write(" OR ");
                  sw.write(printColName);
                  sw.write(" = ''");
                }
                sw.write("))");
              }
              sw.write(")");
            }
          }
          break;

        case SQL_TEXT_UPDATE_LIST:
          if (col[i].isUpdated() && (updateColList.indexOfKey(dbColumnName)!=-1)) {
            if (isFirst)
              isFirst = false;
            else
              sw.write(CN_COM);
            sw.write(printColName);
            sw.write(CN_EQ);
            sw.write(CN_P);
            sw.write(dbColumnName);
          }
          break;

        case SQL_TEXT_INSERT_INTO_LIST:
            if (updateColList.indexOfKey(dbColumnName)!=-1 && !col[i].isAutoIncrement()) {
              if (isFirst)
                isFirst = false;
              else
                sw.write(CN_COM);
              sw.write(printColName);
            }
          break;

        case SQL_TEXT_INSERT_VALUES_LIST:
            if (updateColList.indexOfKey(dbColumnName)!=-1 && !col[i].isAutoIncrement()) {
              if (isFirst)
                isFirst = false;
              else
                sw.write(CN_COM);
              sw.write(CN_P);
              sw.write(dbColumnName);
            }
          break;
      }
    }
    return isFirst;
  }
  
  /**
   * DB2で、NULL許可のカラムをWHERE句に含める場合のWHERE句を生成する。<br/>
   */
  private static final String makeWhereClauseForDB2(Column col, String printColName, String oldColName) {
    String castType = null;
    switch (col.getType()) {
      case BIGDECIMAL:
        StringBuffer _decimalBuf = new StringBuffer();
        _decimalBuf.append("DECIMAL(").append(col.getPrecision()).append(Const.COMMA_CHAR).append(col.getScale()).append(')');
        castType = _decimalBuf.toString();
        break;
      case DATE:
        castType = "DATE";
        break;
      case DOUBLE:
      case FLOAT:
        castType = "DOUBLE";
        break;
      case INT:
        castType = "INTEGER";
        break;
      case LONG:       
        castType = "BIGINT";
        break;
      case SHORT:
        castType = "SMALLINT";
        break;
      case STRING:
        StringBuffer _varcharBuf = new StringBuffer();
        _varcharBuf.append("VARCHAR(").append(col.getLength()).append(')');
        castType = _varcharBuf.toString();
        break;
      case TIME:
        castType = "TIME";
        break;
      case TIMESTAMP:
        castType = "TIMESTAMP";
        break;
      //case BINARY:
      //case BLOB:
      //case BOOLEAN:
      //case BYTE:
      //case LONGBINARY:
      //case OBJECT:
      //case UNKNOWN:
        // これらの型が来ることはないはず。     
        //break;
    }

    StringBuffer _retBuf = new StringBuffer();
    _retBuf.append(printColName).append(" IS NULL AND CAST(").append(oldColName)
           .append(" AS ").append(castType).append(") IS NULL)");
    return _retBuf.toString();
  }
  
  /**
   * INSERT/DELETE/UPDATE時にテーブル名,カラム名をダブルクォーテーションで囲むかの設定を行ういます。<br/>
   * 
   * @param flag True：ダブルクォーテーションで囲む、False：囲まない
   */
  public void setColNameRequireDblQuote(boolean flag) {
    this.colNameRequireDblQuote = flag;
  }
  
  /**
   * INSERT/DELETE/UPDATE時にテーブル名,カラム名をダブルクォーテーションで囲むかどうかを調べます。<br/>
   * 
   * @return True：囲む、False：囲まない
   */  
  public boolean isColNameRequireDblQuote() {
    return this.colNameRequireDblQuote;
  }
  
  /**
   * 必要に応じてダブルクォーテーションで囲んだテーブル名、カラム名を返します。<br/><br/>
   * 
   * isColNameEnclosedDblQuoteがtrueを返す場合、ダブルクォーテーションで囲んだ名前を返します。<br/>
   * falseの場合はそのままの名前を返します。<br/>
   * 
   * @param isColNameEnclosedDblQuote True：ダブルクォーテーションで囲む、False：囲まない
   * @param name テーブル名またはカラム名
   * @return 必要に応じてダブルクォーテーションで囲んだ名前
   */
  private String getNameEnclosedDblQuoteInternal(boolean isColNameRequireDblQuote, String name) {
    if (isColNameRequireDblQuote) {
      StringBuffer _retBuf = new StringBuffer();
      _retBuf.append(Const.DOUBLE_QUOTE).append(name).append(Const.DOUBLE_QUOTE);
      return _retBuf.toString();
    } else {
      return name;
    }
  }

  /**
   * 楽観的排他制御の種別を設定します。<br/>
   * 
   * @param newOptimisticLockType 楽観的排他制御の種別
   * @see OptimisticLockType
   */
  public void setOptimisticLockType(OptimisticLockType newOptimisticLockType) {
    this.optimisticLockType = newOptimisticLockType;
  }

  /**
   * 楽観的排他制御の種別を取得します。<br/>
   * 
   * @return 楽観的排他制御の種別
   * @see OptimisticLockType
   */
  public OptimisticLockType getOptimisticLockType() {
    return this.optimisticLockType;
  }

  /**
   * ResolvingListenerを削除します。<br/>
   * 
   * @param  l  ResolvingListenerインターフェースを実装したオブジェクト
   * @see jp.littlesoft.data.event.ResolvingListener
   */
  @SuppressWarnings("unchecked")
  public synchronized void removeResolvingListener(ResolvingListener l) {
    if (sqlResolvingListeners != null && sqlResolvingListeners.contains(l)) {
      Vector v = (Vector) sqlResolvingListeners.clone();
      v.removeElement(l);
      sqlResolvingListeners = v;
    }
  }

  /**
   * ResolvingListenerを登録します。<br/>
   * 
   * @param  l  ResolvingListenerインターフェースを実装したオブジェクト
   * @see jp.littlesoft.data.event.ResolvingListener
   */
  @SuppressWarnings("unchecked")
  public synchronized void addResolvingListener(ResolvingListener l) {
    Vector v = sqlResolvingListeners == null ? new Vector(2) : (Vector) sqlResolvingListeners.clone();
    if (!v.contains(l)) {
      v.addElement(l);
      sqlResolvingListeners = v;
    }
  }

  @SuppressWarnings("unchecked")
  private void fireRowResolving(ResolvingEvent e) throws DataException {
    if (sqlResolvingListeners != null) {
      if (e.getSource() == null)
        e.setSource(this);
      Vector listeners = sqlResolvingListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        e.setAbort(false);
        e.setAbortMessage(Const.NULLSTRING);
        ((ResolvingListener) listeners.elementAt(i)).rowResolving(e);
        if (e.isAbort()) {
          throw new DataException(e.getAbortMessage());
        }
      }
    }
  }

  @SuppressWarnings("unchecked")
  private void fireRowResolvingError(ResolvingEvent e, java.sql.SQLException ex) throws DataException {
    if (sqlResolvingListeners != null) {
      if (e.getSource() == null)
        e.setSource(this);
      Vector listeners = sqlResolvingListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        e.setAbort(true);
        ((ResolvingListener) listeners.elementAt(i)).rowResolvingError(e, ex);
        if (e.isAbort()) {
          if (ex == null)
            throw new DataException(e.getAbortMessage());
          else
            throw new LSSQLException(ex);
        }
      }
    }
  }

  @SuppressWarnings("unchecked")
  private void fireRowResolved(ResolvingEvent e) throws DataException {
    if (sqlResolvingListeners != null) {
      if (e.getSource() == null)
        e.setSource(this);
      Vector listeners = sqlResolvingListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        e.setAbort(false);
        e.setAbortMessage(Const.NULLSTRING);
        ((ResolvingListener) listeners.elementAt(i)).rowResolved(e);
        if (e.isAbort()) {
          throw new DataException(e.getAbortMessage());
        }
      }
    }
  }

  /**
   * 楽観的ロックの種別を管理するクラス。<br/><br/>
   * 
   * DEFAULT : <br/>
   * java.sql.Statement.executeUpdate()の返り値により排他制御を行います。<br/>
   * 返り値が1でなければ、他から更新などの操作が行われたものと判定します <br/><br/>
   * 
   * PRE_FETCH :<br/>
   * java.sql.Statement.executeUpdate()を行う前に、SELECT文を実行し件数を確認します。<br/>
   * DELETE/UPDATE時には1件以外、INSERT時には0件以外の場合、他から更新などの操作が行われたものと判定します。<br/>
   * (java.sql.Statement.executeUpdate()の返り値は考慮しません。)<br/>
   * 
   * @see SQLDataResolverImpl#setOptimisticLockType
   */
  public enum OptimisticLockType {
    /**
     * executeUpdate()の返り値により排他制御を行うことを表す定数です。
     */
    DEFAULT,
    /**
     * executeUpdate()を行う前に、SELECT文を実行し件数を確認することを表す定数です。
     */
    PRE_FETCH;
  }

  /**
   * executeメソッド内でのコミットを保留するかどうか調べます。<br/>
   * 
   * @return True：コミットを保留する、False：保留しない
   */
  public boolean isPendingCommit() {
    return pendingCommit;
  }

  /**
   * executeメソッド内でのコミットを保留するかどうかを設定します。<br/>
   *  
   * @param pendingCommit True：コミットを保留する、False：保留しない
   */
  public void setPendingCommit(boolean pendingCommit) {
    this.pendingCommit = pendingCommit;
  }

}
