/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.Connection;

import jp.littlesoft.data.Params;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.exception.LSSQLException;

public class SQLTableSequentialReadOnly<E extends Serializable> extends Table<E> {

  private static final long serialVersionUID = 1L;
  
  // 現在の行を保持するRow
  private Row<E> cutenntRow = null;  
  
  /**
   * WebSQLTable オブジェクトを生成します。<br/>
   */
  public SQLTableSequentialReadOnly(E...e) {
    super(e);
    setDefault();
  }
    
  private void setDefault() {
    try { 
      if (this.getDataProvider()==null) {
        this.setDataProvider(new SQLDataSequentialProvider()); 
      }
    } catch (Exception x) {
      x.printStackTrace();
    };
  }
  
  /**
   * JDBCコネクションを設定します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void setConnection(Connection conn) throws DataException {
    setDefault();
    if (this.getDataProvider() instanceof SQLDataSequentialProvider) {
      ((SQLDataSequentialProvider)this.getDataProvider()).setConnection(conn);
    }
  }

  /**
   * JDBCコネクションを取得します。<br/>
   * 
   * @return JDBCコネクション 
   */
  public Connection getConnection() {
    if (this.getDataProvider() instanceof SQLDataSequentialProvider) {
      return ((SQLDataSequentialProvider)this.getDataProvider()).getConnection();
    }
    return null;
  }
  
  /**
   * JDBCコネクションをクローズします。<br/>
   */
  public void closeConnection() {
    try { 
      if (this.getDataProvider() instanceof SQLDataSequentialProvider) {
        ((SQLDataSequentialProvider)this.getDataProvider()).close();
      }
    } catch (LSSQLException e) {
      e.printStackTrace();
    }
  }
  
  /**
   * JDBCコネクションを設定すると同時にテーブルを活性化します。<br/><br/>
   * 
   * この際、SQLプロパティに設定したSELECT文が実行され、テーブルオブジェクトにデータ行が読み込まれた状態になります。<br/><br/>
   * 
   * ※但し、ページングを使用する(setShowRows(int)で「-1」以外の値が設定されている)場合、<br/>
   *   SQLプロパティに設定したSELECT文ではなくページ取得用SELECT文が生成され、実行されます。<br/><br/>
   * 
   * ※OFFSET = (ページ数 - 1) * 表示行数 + 表示行数
   * ※LIMIT  = 表示行数
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public synchronized void activate(Connection conn) throws DataException {
    activate(conn, true);
  }
  private synchronized void activate(Connection conn, boolean resetPage) throws DataException {
    if (isActive()) passivate();
    setConnection(conn);
    activate();
  }

  /**
   * SQLパラメータのセットを取得します。<br/><br/>
   * 
   * このメソッドで取得したParamsオブジェクトを利用して、SQLプロパティに設定したSQL文に含まれる
   * バインドパラメータへアクセスすることができます。<br/><br/>
   * 
   * 設定するSQLには
   * <pre>
   *   table.setSQL("select * from emp where empno = :EMPLOYEE_NO");
   * </pre>
   * の様に、バインドパラメータにする項目名の先頭にコロン(:)を指定します。<br/><br/>
   * 
   * EMPLOYEE_NOに値をセットする場合は
   * <pre>
   *   table.getParam.setValueAsString("1001");
   * </pre>
   * というコードになります。<br/>
   */
  @SuppressWarnings("unchecked")
  @Override
  public Params<E> getParams() {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      return (Params<E>) ((ISQLDataProvider)this.getDataProvider()).getParams();
    } else {
      return super.getParams();
    }
  }
  
  /**
   * WebSQLTable オブジェクトを活性化する際に実行されるSELECT文を設定します。<br/><br/>
   * 
   * activate() メソッドを実行すると、SQLプロパティに設定されたSELECT文が実行され、
   * その結果セットがテーブルオブジェクトの内容として設定されます。<br/><br/>
   * 
   * これらは実際のデータベーステーブルとは切り離された状態のオブジェクトなので、値の変更や、
   * 行の追加削除などを自由に行うことが可能です。<br/><br/>
   * 
   * WebSQLTable オブジェクトに対して行われた編集内容は resolveChange() メソッドによって
   * データベーステーブルへ書き戻すことができます。<br/>
   * 
   * @param selectSQL SELECT文
   * @throws DataException 例外発生時
   */
  public void setSQL(String selectSQL)  throws DataException {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      ((ISQLDataProvider)this.getDataProvider()).setSQL(selectSQL);
    }
  }
  
  /**
   * WebSQLTable オブジェクトを活性化する際に実行されるSELECT文を取得します。<br/>
   * 
   * @return String SELECT文
   */
  public String getSQL() {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      return ((ISQLDataProvider)this.getDataProvider()).getSQL();
    } else {
      return null;
    }
  }
  

  @Override
  public Row<E> getRow() throws DataException {
    return cutenntRow;
  }

  @Override
  public synchronized boolean next() throws DataException {
    return internalCurMove(false);
  }
  
  @Override
  //@Deprecated
  // jdbcResultSet.first で Operation requires a scrollable ResultSet, but this ResultSet is FORWARD_ONLY. のため
  public synchronized boolean first() throws DataException {
    return internalCurMove(true);
  }
  
 
  
  @SuppressWarnings("unchecked")
  private synchronized boolean internalCurMove(boolean isFirst) throws DataException{
    try {
      cutenntRow = ((SQLDataSequentialProvider)this.getDataProvider()).loadDataSeq(this,isFirst);
      return (cutenntRow != null);
    } catch (Exception x) {
      cutenntRow = null;
      return false;
    }
  }
  
}
