/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.exception.LSSQLException;
import jp.littlesoft.util.Const;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

/**
 * ストアドによりSQLデータベースからTableオブジェクトに読み込むクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class StoredProcDataProvider extends SQLDataProviderImpl {

  private static final long serialVersionUID = 1L;
  private static final Integer INTEGER_ZERO = new Integer(0); // よく使うので使いまわしする。
  private static final transient Logger logger = LogManager.getLogger(StoredProcDataProvider.class.getName());

  /**
   * コンストラクタ。<br/>
   */
  public StoredProcDataProvider() {
    super();
  }

  /**
   * カラムのセットアップを行います。<br/><br/>
   *
   * このメソッドを呼び出す前にsetJDBCResultSetメソッドにより、ResultSetオブジェクトを
   * 設定しておく必要があります。<br/><br/>
   * 
   * MetaDataから取得したカラム名が重複していた場合は、
   * カラム名にシーケンスを付加することにより重複しないようにしています。<br/>
   * 例えば、"ABC"というカラムが２つあった場合は、"ABC"と"ABC1"になります。<br/>
   * 又、MetaDataからカラム名が取得できなかった場合は、
   * "UNKNOWN_COLUMN"というカラム名が設定されます。<br/>
   *
   * @param table テーブルオブジェクト
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  @Override
  public <T extends Serializable> void setupColumn(Table<?> table) throws DataException {
    final java.util.Map<String,Integer> resultSetColumnNameTable = new java.util.HashMap<String,Integer>();

    try {
      //-- ログはStoredProcedure#executeSQL()側で出力しているので、ここでは出力しない。
      // ログはUpdateSQL#execute()側で出力しているので、ここでは出力しない。

      // SQLは取得不可。
      //this.rowContainer.setSQL(null);  
      ResultSetMetaData jdbcResultSetMetaData = this.jdbcResultSet.getMetaData();
      int colCount = jdbcResultSetMetaData.getColumnCount();
      resultSetColumnNames = new String[colCount];
      resultSetColumnNameTable.clear();
      for (int i=1; i<=colCount; i++) {
        String colName = jdbcResultSetMetaData.getColumnName(i);
        // カラム名が取得できなかった場合に対応する。
        if (colName == null || colName.equals(Const.NULLSTRING)) {
          colName = UNKNOWN_COLUMN_NAME;
        }
        // リザルトセットに同じカラム名が存在している場合に対応する。
        Integer duplicateNameCount = (Integer) resultSetColumnNameTable.get(colName);
        if (duplicateNameCount == null) {
          duplicateNameCount = INTEGER_ZERO;
        } else {
          duplicateNameCount = new Integer(duplicateNameCount.intValue() + 1);
        }
        resultSetColumnNameTable.put(colName, duplicateNameCount);
        if (duplicateNameCount.intValue() > 0) {
          colName = colName + duplicateNameCount.toString();
        }
        // 未生成のカラムを生成する。
        //Column col = columns.contains(colName);
        Column col;
        int colIndex = table.indexOfDBColumnName(colName);
        if (colIndex < 0) {
          col = new Column(colName,Column.mapToJava(jdbcResultSetMetaData.getColumnType(i)), true);
          table.addColumn(col);
        } else {
          col = table.getColumns().getValue(colIndex);
          col.setType( Column.mapToJava(jdbcResultSetMetaData.getColumnType(i)) );
        }
        col.setSqlType(jdbcResultSetMetaData.getColumnType(i));
        if (col.getSqlType()==java.sql.Types.NUMERIC || col.getSqlType()==java.sql.Types.DECIMAL) {
          if (jdbcResultSetMetaData.getPrecision(i)>0 && jdbcResultSetMetaData.getScale(i)>=0) {
            col.setPrecision(jdbcResultSetMetaData.getPrecision(i));
            col.setScale(jdbcResultSetMetaData.getScale(i));
          }
        }

        // JDBCAdaptorで、文字列用検証クラスが設定されている場合、それを割り当てる。(Postgres 7.2対応)
        if (col.getType().equals(Column.Type.STRING) && this.getColumnValidator() != null) {
          col.setValidator(this.getColumnValidator());
        }

        // NULL可 or 不可
        //  ResultSetMetaData.columnNoNulls         : NULL不可
        //  ResultSetMetaData.columnNullable        : NULL可
        //  ResultSetMetaData.columnNullableUnknown : NULL可
        //  とする。
        col.setNullable(!(jdbcResultSetMetaData.isNullable(i)==ResultSetMetaData.columnNoNulls));

        if (col.getDisplaySize()==Const.NOT_ASSIGNED) {
          col.setDisplaySize(jdbcResultSetMetaData.getColumnDisplaySize(i));
        }
        if (col.getType().equals(Column.Type.STRING)) {
          if (col.getLength() == Const.NOT_ASSIGNED) {
            col.setLength(jdbcResultSetMetaData.getColumnDisplaySize(i));            
          }
        }
        resultSetColumnNames[i-1] = colName;

        if (logger.isDebugEnabled()) {
          logger.debug("*** StoredProcDataProvider.setupColumn: ***");
          logger.debug("name      : " + col.getName());
          logger.debug("type      : " + col.getType().toString());
          logger.debug("length    : " + col.getLength());
          logger.debug("precision : " + col.getPrecision());
          logger.debug("scale     : " + col.getScale());
          logger.debug("null able : " + col.isNullable());
        }
      }
    } catch (SQLException x) {
      freeStatement();
      throw new LSSQLException(x);
    }
  }

  /**
   * データのロードを行います。<br/>
   *
   * @exception jp.littlesoft.data.DataException  例外発生時
   */
  @SuppressWarnings("unchecked")
  @Override
  public void loadData(Table table) throws DataException {
    this.loadResultSet(table, false);
  }

  /**
   * ResultSetからデータをロードします。<br/>
   *
   * @param toLast True：最終行までロードする、False：ロードしない
   * @return  ロードした行数
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  @SuppressWarnings("unchecked")
  private int loadResultSet(Table table, boolean toLast) throws DataException {
    int loadCount = 0;
    try {
      //boolean hasMoreRow = true;
      while ((getFetchSize()==0 || toLast || loadCount<getFetchSize()) && (this.jdbcResultSet.next())) {
        Row newRow = new Row(table);
        newRow.setValueFromResultSet(this.jdbcResultSet, this.resultSetColumnNames);
        newRow.setStatus(Row.Status.LOADED);
        table.internalPut(newRow, true, false);
        loadCount++;
      }
    } catch (SQLException x) {
      freeStatement();
      throw new LSSQLException(x);
    } catch (DataException x) {
      freeStatement();
      throw x;
    }
    return loadCount;
  }

}
