/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql.logger;

import java.io.Serializable;

import jp.littlesoft.util.Const;

import org.apache.log4j.*;

/**
 * 実行されたSQL文の実行時間のロギングを行うクラス。<br/><br/>
 * 
 * ログの出力自体は、Log4j管理下のLoggerクラスに委譲されます。<br/>
 * 本クラスは、実行されたSQL文の実行時間をログとして記録します。<br/>
 * 標準で、jp.littlesoft.sql.SQLDataResolver, jp.littlesoft.sql.UpdateSQL,
 * jp.littlesoft.sql.SQLTableで行われるSQL処理に組み込まれています。<br/>
 * ログの出力レベルを<B>DEBUG</B>に指定すると出力されます。<br/><br/>
 * 
 * log4j.xml設定ファイルにて、更新用ログのみ／照会用ログのみ／更新用ログ＋照会用ログ などの出力制御ができます。
 * 
 * <pre>
 * 設定例：
 * 
 *  &lt;!-- 更新用ログ＋照会用ログ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLTimeLogger&quot;&gt;
 *      &lt;level value =&quot;DEBUG&quot; /&gt;
 *  &lt;/logger>
 * 
 *  &lt;!-- 更新用ログのみ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLTimeLogger.update&quot;&gt;
 *      &lt;level value ="DEBUG" /&gt;
 *  &lt;/logger>
 *  
 *  &lt;!-- 照会用ログのみ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLTimeLogger.query&quot;&gt;
 *      &lt;level value ="DEBUG" /&gt;
 *  &lt;/logger&gt;
 *  
 * 出力例:
 * 
 *  2004-12-24 11:28:21,750,Provide,clientId,mypackege.EmployeeDBAction,tblEmployee,407
 *  2004-12-24 11:29:05,468,ResolveChanges,clientId,mypackege.EmployeeDBAction,tblEmployee,31
 *  
 *  (日時,時刻のミリ秒,Provide/ResolveChanges/ExecuteUpdate/ExecuteQueryの種別,クライアント識別子,
 *   DBActionのクラス名,SQLTable名,処理時間(ミリ秒))
 * </pre>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public final class SQLTimeLogger implements Serializable {
  
  /**
   * Provideを意味する文字列定数です。<br/>
   */
  public static final transient String PROVIDE = "Provide";
  
  /**
   * ResolveChangesを意味する文字列定数です。<br/>
   */
  public static final transient String RESOLVE_CHANGES = "ResolveChanges";
  
  /**
   * ExecuteUpdateを意味する文字列定数です。<br/>
   */
  public static final transient String EXECUTE_UPDATE = "ExecuteUpdate";
  
  /**
   * ExecuteUpdateを意味する文字列定数です。<br/>
   */
  public static final transient String EXECUTE_QUERY = "ExecuteQuery";

  private static final transient String CATEGORY        = SQLTimeLogger.class.getName();
  private static final transient String CATEGORY_UPDATE = CATEGORY+".update";
  private static final transient String CATEGORY_QUERY  = CATEGORY+".query";
  
  private static transient Logger logger = null;
  private static transient Logger updateLogger = null;
  private static transient Logger queryLogger  = null;
  
  private SQLTimeLogger() {
  }
	
  /**
   * 更新時間のログを出力します。<br/><br/>
   * 
   * 注意: ログレベルが、<b>DEBUG</b>以外の時は出力は無視されます。<br/>
   * 
   * @param kind     種別("ResolveChanges", "ExecuteUpdate", "ExecuteQuery"のいずれか) 
   * @param clientId 処理要求元を判別するための識別子.（null指定可）
   * @param dbActionName SQLを実行したDBAction名
   * @param tableName      SQLを実行したSQLTable名
   * @param executeTime 実行時間(単位:ミリ秒)
   */
  public static void logForUpdate(String kind, String clientId, String dbActionName, String tableName, long executeTime) {
    if (!isUpdateDebugEnabled()) return;
    log(updateLogger, kind, clientId, dbActionName, tableName, executeTime);
  }
	
  /**
   * 照会時間のログを出力します。<br/><br/>
   * 
   * 注意:　ログレベルが、<b>DEBUG</b>以外の時は出力は無視されます。<br/>
   * 
   * @param kind     種別(="Provide") 
   * @param clientId 処理要求元を判別するための識別子.（null指定可）
   * @param dbActionName SQLを実行したDBAction名
   * @param tableName      SQLを実行したSQLTable名
   * @param executeTime 実行時間(単位:ミリ秒)
   */
  public static void logForQuery(String kind, String clientId, String dbActionName, String tableName, long executeTime) {
    if (!isQueryDebugEnabled()) return;
    log(queryLogger, kind, clientId, dbActionName, tableName, executeTime);
  }
	
  private static void log(Logger logger, String kind, String clientId, String defaultActionName, String rowsetName, long executeTime) {
  	  //try {
      if (logger.isDebugEnabled()) {
        // 先頭の","はCSVとして読み込めるようにするため。
        StringBuffer _logBuf = new StringBuffer();
        _logBuf.append(Const.COMMA_CHAR).append(kind).append(Const.COMMA_CHAR).append(clientId).append(Const.COMMA_CHAR)
               .append(defaultActionName).append(Const.COMMA_CHAR).append(rowsetName).append(Const.COMMA_CHAR)
               .append(executeTime);
        logger.debug(_logBuf.toString());
      }
		    //} catch (Exception e){ e.printStackTrace(); }
  }

  /**
   * データ照会用ログが出力可能かどうか調べます。<br/>
   * 
   * @return True：出力可、False：出力不可
   */
  public static boolean isQueryDebugEnabled() {
    resetQueryLogger();
    if (queryLogger == null) return false;
    return queryLogger.isDebugEnabled();
  }

  /**
   * データ更新用ログが出力可能かどうか調べます。<br/>
   * 
   * @return True：出力可、False：出力不可
   */
  public static boolean isUpdateDebugEnabled() {
    resetUpdateLogger();
    if (updateLogger == null) return false;
    return updateLogger.isDebugEnabled();
  }

  private static void resetLogger() {
    if (logger == null)
      logger = LogManager.getLogger(CATEGORY);
  }

	  private static void resetQueryLogger() {
    if (queryLogger == null) {
      queryLogger = LogManager.getLogger(CATEGORY_QUERY);
      if (queryLogger == null)
        queryLogger = logger;    
    } 
	  }

	  private static void resetUpdateLogger() {
    if (updateLogger == null) {
      updateLogger = LogManager.getLogger(CATEGORY_UPDATE);
      if (updateLogger == null)
        updateLogger = logger;
    }
	  }

  static {
    resetLogger();
    resetQueryLogger();
    resetUpdateLogger();
  }

}

