/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.util;

import java.sql.Date;
import java.text.*;
import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.math.*;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.data.exception.InvalidColumnException;

import org.apache.log4j.*;

/**
 * データキャッシュコンポーネントから利用されるユーティリティクラス。<br/><br/>
 *
 * columnクラスをもとにした値のフォーマット、パース、チェックメソッドや
 * JDBCアクセス時の文字置換サポートメソッドを提供しています。<br/>
 * これらのメソッドはデータキャッシュコンポーネント内部から利用されています。<br/><br/>
 *
 * 【概要】<br/>
 * データベースからのデータ取得時およびデータベースへのデータ格納時、それぞれについて
 * 文字変換を設定することができます。<br/>
 * 一般的には '－' や '～' などの文字化けを防ぐために使用します。<br/><br/>
 *
 * 文字化けのパターンは、JDK, JDBC, OSの種類やバージョン、データベースの文字コード、
 * さらには実行時の環境(UNIXの環境変数など)によって異なります。<br/>
 * したがって、こうすれば絶対に文字化けしない、というパターンはありません。<br/>
 * 使用する環境に合わせて設定してください。<br/><br/>
 *
 * 【組み合わせて使用するクラス】<br/>
 * 通常は、ServletBase継承クラスのinitメソッド内で本クラスのsetReplaceCharsFromDBメソッドや
 * setReplaceCharsToDBメソッドを呼び出して文字置換のパターンを定義します。<br/><br/>
 *
 * <dl>
 * <dt>【主要なメソッド】
 * <dt>setReplaceCharsFromDB
 * <dd>データベースからのデータ取得時の文字置換を行うための設定をします。<br/><br/>
 *
 * <dt>setReplaceCharsToDB
 * <dd>データベースへのデータ格納時の文字置換を行うための設定をします。<br/><br/>
 * </dl>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public final class DataUtil implements Serializable {

  static private char[] fromDBSrc = new char[0];
  static private char[] fromDBDst = new char[0];
  static private char[] toDBSrc = new char[0];
  static private char[] toDBDst = new char[0];

  private static char[] doubleByteNum = {'０','１','２','３','４','５','６','７','８','９','／','，','－','：','．'};

  private static char[] singleByteNum = {'0','1','2','3','4','5','6','7','8','9','/',',','-',':','.'};
  
  private final static XWareki wareki = new XWareki();
  private final static SimpleDateFormat simpleDateFormat = new SimpleDateFormat();
  private final static DecimalFormat decimalFormat = new DecimalFormat();
  private final static ParsePosition parsePosition = new ParsePosition(0);
  private static final transient Logger logger = LogManager.getLogger(DataUtil.class.getName());

  private DataUtil() {
  }

  /**
   * データベースから読み込み時に行う文字置換リストを設定します。<br/><br/>
   *
   * このメソッドにより定義された変換パターンに基づく文字置換はDBから文字型のデータを
   * 読み込んだ後に行われます。<br/>
   * 　DBから文字型データを読み込み ===> 文字置換 ===> 画面表示など<br/><br/>
   *
   * src配列の各要素からdst配列の各要素に置換されますので、
   * 当然ながら、src配列の要素数とdst配列の要素数は同じにする必要があります。<br/>
   * 又、srcおよびdstにnullは設定できません。<br/>
   * 通常は、HttpServlet継承クラスのinit()メソッドから呼び出します。<br/>
   * 変換パターンはOS(および環境変数LANG),JDK,JDBCドライバー,データベースなどの環境に依存します.<br/>
   * <pre>
   * 例:
   *   // '－'と'～'の文字化け対応
   *   // &yen;u2212 → &yen;uff0d
   *   // &yen;u301c → &yen;uff5e
   *   DataUtil.setReplaceCharsFromDB(
   *     new char[] { '&yen;u2212', '&yen;u301c' }, // 変換前のコード
   *     new char[] { '&yen;uff0d', '&yen;uff5e' }  // 変換後のコード
   *   );
   * </pre>
   * 
   * @param  src  置換前文字の配列
   * @param  dst  置換後文字の配列
   * @return  設定に成功したかどうか
   */
  public static boolean setReplaceCharsFromDB(char[] src, char[] dst) {
    if (src != null && dst != null && src.length == dst.length) {
      fromDBSrc = src;
      fromDBDst = dst;
      return true;
    }
    return false;
  }
  /**
   * データベースへの書き込み時に行う文字列置換リストを設定します。<br/><br/>
   *
   * このメソッドにより定義された変換パターンに基づく文字置換はDBへ文字型のデータを
   * 保存する前に行われます。<br/>
   * 　保存する文字型データの決定 ===> 文字置換 ===> DBへの文字型データの保存<br/><br/>
   *
   * src配列の各要素からdst配列の各要素に置換されますので、
   * 当然ながら、src配列の要素数とdst配列の要素数は同じにする必要があります。<br/>
   * 又、srcおよびdstにnullは設定できません。<br/>
   * 通常は、HttpServlet継承クラスのinit()メソッドから呼び出します。<br/>
   * 変換パターンはOS(および環境変数LANG),JDK,JDBCドライバー,データベースなどの環境に依存します.<br/>
   * <pre>
   * 例:
   *   // '－'と'～'の文字化け対応
   *   // &yen;uff0d → &yen;u2212
   *   // &yen;uff5e → &yen;u301c
   *   DataUtil.setReplaceCharsToDB(
   *     new char[] { '&yen;uff0d', '&yen;uff5e' }, // 変換前のコード
   *     new char[] { '&yen;u2212', '&yen;u301c' }  // 変換後のコード
   *   );
   * </pre>
   * 
   * @param  src  置換前文字の配列
   * @param  dst  置換後文字の配列
   * @return  True：設定に成功した、False：設定に失敗した
   */
  public static boolean setReplaceCharsToDB(char[] src, char[] dst) {
    if (src != null && dst != null && src.length == dst.length) {
      toDBSrc = src;
      toDBDst = dst;
      return true;
    }
    return false;
  }
  /**
   * DB読み込み用の文字列置換配列を使用して文字列置換を行います。<br/>
   *
   * @param value 置換前文字列
   * @return 置換後文字列
   * @see #setReplaceCharsFromDB
   */
  public static String replaceFromDB(String value) {
    if (value == null) {
      return null;
    }
    for(int i=0; i<fromDBSrc.length; i++) {
      value = value.replace(fromDBSrc[i], fromDBDst[i]);
    }
    return value;
  }
  /**
   * DB書き込み用の文字列置換配列を使用して文字列置換を行います。<br/>
   *
   * @param value 置換前文字列
   * @return 置換後文字列
   * @see #setReplaceCharsToDB
   */
  public static String replaceToDB(String value) {
    if (value == null) {
      return null;
    }
    for(int i=0; i<toDBSrc.length; i++) {
      value = value.replace(toDBSrc[i], toDBDst[i]);
    }
    return value;
  }

  /**
   * valueをColumnに設定されたフォーマットパターンに基づきフォーマットします。<br/><br/>
   *
   * 内部でjava.text.DecimalFormatを使用し、フォーマットを行っています。<br/>
   * フォーマットパターンの書式についての詳細はjava.text.DecimalFormatのドキュメントを参照してください。<br/>
   *
   * Date, Time, Timestampの場合には、formatTime()メソッドを使用してください。<br/>
   *
   * 数値型のカラムがゼロサプレスする設定になっていた場合、フォーマットを試みて、その結果が"0"だったら
   * ""を返します。<br/>
   * ゼロサプレスしない設定であれば、フォーマットした結果をそのまま返します。<br/>
   * 
   * ※注 このメソッドは内部でstaticフィールドにsynchronizedを行っているため、同時に複数のスレッドから
   *      呼ばれると待ち状態となります。<br/>
   * <br/>
   * ※注 JDK1.4を利用している場合、かつ、valueがBigDecimal型の場合、
   *      doubleの精度を超えているものはdoubleの精度内に丸められてフォーマットされます。<br/>
   * 
   * @param   column    対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param   value  フォーマットするオブジェクト
   * @return  フォーマットした文字列
   * @exception  java.lang.Exception  フォーマットに失敗した場合
   * @see jp.littlesoft.data.Column#setFormatPattern
   */
  // -0.00000001などの場合、ゼロサプレスが効かない。
  public static String formatObject(final Column column, final Object value) throws InvalidColumnException {
    if (value==null) return Const.NULLSTRING;
    if (value instanceof String) return (String)value;

    if (value instanceof java.lang.Number) {
      try {
        //----  FormatPatternなしのBigDecimalに対応
        final String formatPattern = column.getFormatPattern();

        if (StringUtil.isNullString(formatPattern)) {
          String s = value.toString();
          // Doubleの場合は、"0.0"となる
          if (column.isZeroSuppress()) {
            if ("0".equals(s) || "0.0".equals(s)) {
              return Const.NULLSTRING;
            } else if (column.getType().equals(Column.Type.BIGDECIMAL) && column.getScale() > 1){
              BigDecimal bd = BigDecimal.valueOf(0L, column.getScale());
              if (bd.toString().equals(s)) {
                return Const.NULLSTRING;
              }
            }
          }
          return s;
        } else {
          synchronized (decimalFormat) {
            decimalFormat.applyPattern(formatPattern);
            String s = decimalFormat.format(value);
            if (column.isZeroSuppress()) {
              // -0.00のようになる可能性がある。
              String s2;
              if (s.length() > 1 && s.charAt(0) == '-') {
                s2 = s.substring(1);
              } else {
                s2 = s;
              }
              if (decimalFormat.format(0L).equals(s2)) {
                return Const.NULLSTRING;
              }
            }
            return s;
          }
        }
      } catch (Exception x) {
        String _caption = column.getCaption();
        throw new InvalidColumnException("FORMAT_ERROR_NUMBER",
            new String[]{_caption, value.toString()}, column);
      }
    } else if (value instanceof java.util.Date) {
    	return formatDate(column, (java.util.Date)value);
    } else {
      try {
        return value.toString();
      } catch (Exception ex) {
      }
      return Const.NULLSTRING;
    }
  }

  /**
   * valueをColumnに設定されたフォーマットパターンに基づきフォーマットします。<br/><br/>
   *
   * 内部でjava.text.SimpleDateFormatを使用し、フォーマットを行っています。<br/>
   * フォーマットパターンの書式についての詳細はjava.text.SimpleDateFormatのドキュメントを参照してください。<br/>
   *
   * このメソッドを呼ぶ前にcolumnがnullでないことを確認してから、このメソッドを呼び出してください。<br/>
   * nullの場合には事前に "" を返すように処理を行ってください。<br/><br/>
   *
   * ※ナノ秒部分はjava.text.SimpleDateFormatが対応していないためフォーマットすることはできません。<br/>
   *
   * ※注 dateにjava.sql.Timestampオブジェクトを指定した場合、JDK1.3.xではミリ秒が0として
   *      扱われます。(JDKの障害と思われます。)<br/>
   * ※注 このメソッドは内部でstaticフィールドにsynchronizedを行っているため、同時に複数のスレッドから
   *      呼ばれると待ち状態となります。<br/>
   * <br/>
   * なお、このメソッドは内部では使用していません。<br/>
   * 
   * @param   column    対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param   date  フォーマットするDateオブジェクト
   * @return  フォーマットした文字列
   * @exception  java.lang.Exception  フォーマットに失敗した場合
   * @see jp.littlesoft.data.Column#setFormatPattern
   */
  public static String formatDate(final Column column, final java.util.Date date) throws InvalidColumnException {
  	if (date == null) return Const.NULLSTRING;
    try {
      String formatPattern = column.getFormatPattern();
      if (XWareki.isWarekiFormatPattern(formatPattern)) {
        // wareki はメソッド内でsynchronizedしているので、ここでのsynchronizedは不要
        return wareki.format(date, formatPattern);
      } else {
        synchronized (simpleDateFormat) {
          if (StringUtil.isNullString(formatPattern)) {
            // フォーマットパターンが未指定の場合は、型に合わせたデフォルトのフォーマットパターンを適用する。
            if (date instanceof java.sql.Timestamp) {
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP);              
            } else if (date instanceof java.sql.Time) {
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIME);
            } else {
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_DATE);
            }
          } else {
            simpleDateFormat.applyPattern(formatPattern);
          }
          return simpleDateFormat.format(date);
        }
      }
    } catch (Exception x) {
      String _caption = column.getCaption();
      String _value = date.toString();
      throw new InvalidColumnException("FORMAT_ERROR_DATE",
        new String[]{_caption, _value}, column);
    }
  }

  /**
   * 文字列をColumnに設定されたフォーマットパターンに基づきパースします。<br/><br/>
   *
   * 内部でjava.textパッケージのDecimalFormat, DateFormat, SimpleDateFormatクラスを利用して
   * パースを行っています。<br/>
   * フォーマットパターンの書式については、java.textパッケージのドキュメントを参照してください。<br/>
   *
   * ※注 このメソッドは内部でstaticフィールドにsynchronizedを行っているため、同時に複数のスレッドから
   *      呼ばれると待ち状態となります。<br/>
   * 
   * @param   column    対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param   value  パースする文字列
   * @return  パースしたオブジェクト
   * @exception  java.lang.Exception  パースに失敗した場合
   * @see jp.littlesoft.data.Column#setFormatPattern
   */
  public static Object parseObject(Column column, String value) throws InvalidColumnException {
    if (value==null || Const.NULLSTRING.equals(value)) return null;
    switch (column.getType()) {
      case STRING:
        return (String)value;
      case BYTE:
      case SHORT:
      case LONG:
      case INT:
      case FLOAT:
      case DOUBLE:
        value = normalizeNumeric(value);
        checkDecimalLength(column, value);
        if  (column.getFormatPattern()==null || Const.NULLSTRING.equals(column.getFormatPattern())) {
          try {
            switch (column.getType()) {
              case BYTE:
                return Byte.valueOf(value);
              case SHORT:
                return Short.valueOf(value);
              case INT:
                return Integer.valueOf(value);
              case LONG:
                return Long.valueOf(value);
              case FLOAT:
                return Float.valueOf(value);
              case DOUBLE:
                return Double.valueOf(value);
              default:
                throw new Exception(); // ここに来ることはないはず。
            }
          } catch (Exception x) {
            String _caption = column.getCaption();
            throw new InvalidColumnException("PARSE_ERROR_INVALID_NUMBERFMT2", 
              new String[]{_caption, value}, column);
          }
        } else {
          try {
            Number num = parseNumber(column.getFormatPattern(), value);
            switch (column.getType()) {
              case BYTE:
                return new Byte(num.toString());
              case SHORT:
                return new Short(num.toString());
              case INT:
                return new Integer(num.toString());
              case LONG:
                return new Long(num.toString());
              case FLOAT:
                return new Float(num.doubleValue());
              case DOUBLE:
                return new Double(num.doubleValue());
              default:
                throw new Exception(); // ここに来ることはないはず。            
            }
          } catch (Exception ex) {
            String _caption = column.getCaption();
            throw new InvalidColumnException("PARSE_ERROR_INVALID_NUMBERFMT2", 
              new String[]{_caption, value}, column);            
          }
        }
      case BIGDECIMAL:
        value = normalizeNumeric(value);
        checkDecimalLength(column, value);
        java.math.BigDecimal dec = null;
        if  (column.getFormatPattern()==null || Const.NULLSTRING.equals(column.getFormatPattern())) {
          try {
            dec = new java.math.BigDecimal(value);
          } catch (Exception x) {
            String _caption = column.getCaption();
            throw new InvalidColumnException("PARSE_ERROR_INVALID_NUMBER",
                new String[]{_caption, value}, column);
          }
        } else {
          try {
          	// JDK5.0の場合は、BigDecimalを返すようにする。
          	// 未実装
            Number num = parseNumber(column.getFormatPattern(), value);
            dec = new java.math.BigDecimal(num.doubleValue());
            if (column.getScale()!=Const.NOT_ASSIGNED) {
              dec = dec.setScale(column.getScale(), java.math.BigDecimal.ROUND_HALF_EVEN);
            }
          } catch (Exception x) {
            String _caption = column.getCaption();
            throw new InvalidColumnException("PARSE_ERROR_INVALID_NUMBERFMT2",
                new String[]{_caption, value}, column);
          }
        }
        return dec;
      case DATE:
      case TIME:
      case TIMESTAMP:
        value = normalizeNumeric(value);
        String _formatPattern = column.getFormatPattern();
        java.util.Date dt = null;
        try {
          if (XWareki.isWarekiFormatPattern(_formatPattern)) {
            // wareki はメソッド内でsynchronizedしているので、ここでのsynchronizedは不要              
            dt = wareki.parse(value, column.getFormatPattern());
          } else {
            synchronized (simpleDateFormat) {
              if (StringUtil.isNullString(_formatPattern)) {
                // デフォルトのフォーマットパターンを適用する。
                switch (column.getType()) {
                  case TIMESTAMP:
                    simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP);
                    break;
                  case TIME:
                    simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIME);
                    break;
                  default: //DATE
                    simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_DATE);
                    break;
                }
              } else {
                simpleDateFormat.applyPattern(_formatPattern);
              }
              dt = simpleDateFormat.parse(value);
            }
          }
          switch (column.getType()) {
            case TIME:
              return new java.sql.Time(dt.getTime());
            case TIMESTAMP:
              return new java.sql.Timestamp(dt.getTime());
            default: //DATE
              return new java.sql.Date(dt.getTime());
          }
        } catch (Exception x) {
          String _caption = column.getCaption();
          throw new InvalidColumnException("PARSE_ERROR_INVALID_DATE",
              new String[]{_caption, value},
              column);
        }
      case BOOLEAN:
        return Boolean.valueOf(value);
      case UNKNOWN:
      case BINARY:
      case LONGBINARY:
      case BLOB:
      case JSONOBJECT:
        String _caption = column.getCaption();
        throw new InvalidColumnException("PARSE_ERROR_INVALID_TYPE",
            new String[]{_caption, column.getType().toString(), value},
            column);
      default:
        String _caption2 = column.getCaption();
        String[] errParams = { _caption2, column.getType().toString()};
        throw new InvalidColumnException("INVALID_DATAFIELD_TYPE2",
            errParams, column);
    }
  }

  /**
   * Columnに設定されたScale,Precisionに基づき検証します。<br/><br/>
   *
   * 不正な値の場合Exceptionがスローされる。<br/>
   *
   * @param  column    Columnオブジェクト
   * @param  value  チェックする値
   * @exception  java.lang.Exception  不正な値の場合
   */
  public static void checkDecimalLength(Column column, String value) throws InvalidColumnException{
    if (column.getPrecision() < 0 && column.getScale() < 0) {
      // Precision, Scaleともに未設定の場合はチェックしない。
      return;
    }
    
    final char[] ch = value.toCharArray();
    int integerDigits = 0;
    int fractionDigits = 0;
    boolean isInt = true;
    for (int i=0; i<ch.length; i++) {
      if (ch[i]=='.')
        isInt = false;
      else if (ch[i]=='-' || ch[i]<'0' || ch[i]>'9')
       continue;
      else if (isInt)
        integerDigits++;
      else
        fractionDigits++;
    }
    if (column.getScale()>=0) {
      if (column.getPrecision()>=0) {
        // 全桁＆小数点以下の両方チェック
        if (fractionDigits > column.getScale() ||
            integerDigits  > (column.getPrecision() - column.getScale()) ) {
          final String _caption = column.getCaption();
          throw new InvalidColumnException("PARSE_ERROR_INVALID_PRECISION",
            new String[]{_caption, value}, column);
        }
      } else { // Scale>=0 && Precision<0
        // 小数点以下のみチェック
        if (fractionDigits > column.getScale())  {
        final String _caption = column.getCaption();
          throw new InvalidColumnException("PARSE_ERROR_INVALID_PRECISION",
            new String[]{_caption, value}, column);
        }
      }
    } else { // Scale<0 && Precision>=0
    	// 全桁のみチェック
      if ((fractionDigits + integerDigits) > column.getPrecision()) {
      	final String _caption = column.getCaption();
        throw new InvalidColumnException("PARSE_ERROR_INVALID_PRECISION",
          new String[]{_caption, value}, column);
      }
    }
  }

  /**
   * 和暦（元号と年月日）よりタイムスタンプ値を取得します。<br/><br/>
   *
   * ※注 このメソッドは内部でstaticフィールドにsynchronizedを行っているため、同時に複数のスレッドから
   *      呼ばれると待ち状態となります。<br/>
   * 
   * @param   gengo  元号
   * @param   year   年
   * @param   month  月
   * @param   date   日
   * @return  タイムスタンプ値
   * @exception  java.lang.Exception  例外発生時
   */
  public static long parseWareki(String gengo, int year, int month, int date) throws Exception {
    // wareki はメソッド内でsynchronizedしているので、ここでのsynchronizedは不要
    return wareki.parse(gengo, year, month, date);
  }

  /**
   * 年月日が正しいかどうかを調べます。<br/><br/>
   *
   * 年月日が不正な場合、例外が発生します。<br/>
   *
   * @param  year   年
   * @param  month  月
   * @param  date   日
   * @exception  java.lang.Exception  年月日が不正な場合
   */
  public static void checkDate(int year, int month, int date) throws DataException{
    // months[0] はダミー
    int[] months = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    //  年月日をそれぞれ取得
    //  月に関する基本的なチェック
    if(month < 1 || 12 < month)
      throw new DataException(null, null, "PARSE_ERROR_MONTH",
          new String[]{Integer.toString(month)});

    //  うるう年であれば
    if(year % 4 == 0 && ( year % 100 != 0 || year % 400 == 0)) months[2] = 29;

    //  日に関するチェック
    if(date < 1 || months[month] < date)
      throw new DataException(null, null, "PARSE_ERROR_DATE",
          new String[]{Integer.toString(date)});
  }

  /**
   * 全角の数字を半角に変換します。<br/>
   *
   * @param   value  変換する文字列
   * @return  変換後の文字列
   */
  public static String normalizeNumeric(String value) {
    if (StringUtil.isNullString(value)) return value;
    char[] nums = value.toCharArray();
    for (int i=0; i<nums.length; i++) {
      for (int x=0; x<doubleByteNum.length; x++) {
        if (nums[i]==doubleByteNum[x]) {
          nums[i] = singleByteNum[x];
          continue;
        }
      }
    }
    return new String(nums);
  }

  private static Number parseNumber(String formatPattern, String value) throws DataException {
    Number answer;
    synchronized (decimalFormat) {
      decimalFormat.applyPattern(formatPattern);
      // 前回のindex値を元に戻す。
      parsePosition.setIndex(0);
      // JDK5.0の場合は、BigDecimalを返すことも可能。
      answer = decimalFormat.parse(value, parsePosition);
      if (parsePosition.getIndex() < value.length()) {
        throw new DataException(null, null, "PARSE_ERROR_INVALID_NUMBERFMT",
            new String[]{formatPattern, value});
      }
    }
    return answer;
  }

  static {
    //  日付チェックを厳しく行う。
    //DataUtil.dateFormat.setLenient(false);
    DataUtil.simpleDateFormat.setLenient(false);

    // リソースから日付を読み込むために使用するSimpleDateFormatインスタンス
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy/MM/dd");
    try {
      for (int i=0; i<XWareki.WAREKI_GENGO_START_DATE.length; i++) {
        java.util.Date dt = sdf.parse(XWareki.WAREKI_GENGO_START_DATE[i]);
        XWareki.startDates[i] = new java.util.GregorianCalendar();
        XWareki.startDates[i].setTime(dt);
      }      
    } catch (java.text.ParseException x) {
      logger.error("create error : " + DataUtil.class.getName(), x);
      x.printStackTrace();
    }
  }

  /**
   * 和暦変換クラス。<br/><br/>
   *
   * 日付型の和暦フォーマット、パースを行うクラスです。<br/><br/>
   *
   * 和暦のフォーマットパターンは、以下の９種類用意しています。<br/>
   * <pre>
   * public static final String FORMAT_DATETIME_LONG = "GGyy年MM月dd日 HH時mm分ss秒";
   * public static final String FORMAT_DATETIME_SHORT = "gyy/MM/dd HH:mm:ss";
   * public static final String FORMAT_DATE_LONG = "GGyy年MM月dd日";
   * public static final String FORMAT_DATE_SHORT = "gyy/MM/dd";
   * public static final String FORMAT_YEARMONTH_LONG = "GGyy年MM月";
   * public static final String FORMAT_YEARMONTH_SHORT = "gyy/MM";
   * public static final String FORMAT_YEAR_LONG = "GGyy年";
   * public static final String FORMAT_YEAR_LONG2 = "GGy年";
   * public static final String FORMAT_YEAR_SHORT = "gyy";
   * </pre>
   * これら以外でも、先頭が"gy"または"GGy"で始まるフォーマットであれば、自由に設定できます。<br/>
   * (例: "gy.M.d" "GGyy MM dd")<br/>
   * <br/> 
   * これらのフォーマットパターン文字列をColumnクラスのFormatPatternプロパティに設定することで、
   * 日付データを和暦で表示および入力することができます。<br/><br/>
   *
   * 又、和暦のチェック方法は、このクラスのstaticな変数"checkLevel"に設定します。<br/>
   * 次の４つの定数のいずれかをcheckLevelに設定することができます。<br/>
   * <pre>
   * ・WAREKI_CHECK_LEVEL_NO
   *    - 年月日をチェックしません。
   *    - 例: 昭和70年もOKです。
   * ・WAREKI_CHECK_LEVEL_YEAR
   *    - 年までチェックします。
   *    - 例: 昭和64年はOKですが、昭和65年はNGです。
   * ・WAREKI_CHECK_LEVEL_MONTH
   *    - 年月までチェックします
   *    - 例: 昭和64年1月はOKですが、昭和64年2月はNGです。
   * ・WAREKI_CHECK_LEVEL_DAY
   *    - 年月日までチェックします
   *    - 例: 昭和64年1月7日はOKですが、昭和64年1月8日はNGです。
   * </pre>
   * デフォルト値は、WAREKI_CHECK_LEVEL_YEARです。<br/><br/>
   *
   * 変換に必要なリソースがWareki.propertiesファイルに記述されています。<br/>
   * 和暦元号が変更になった際は上記リソースファイルの情報を変更します。<br/>
   * ※現在WAREKI_GENGO4が"平成"であり、GENGO5,GENGO6が新元号用に予約されています。 <br/>
   * WAREKI_GENGO5=****&nbsp;&nbsp;→&nbsp;&nbsp;"****"を新しい元号に置き換えます。UNICODEで直接指定します。(例:"平成"→&yen;u5e73&yen;u6210) <br/>
   * WAREKI_GENGO5_SHORT=X&nbsp;&nbsp;→&nbsp;&nbsp;Xを元号の短縮記号に置き換えます。(例:"平成"→"H")<br/>
   * WAREKI_GENGO5_START_DATE=2999/12/31&nbsp;&nbsp;→&nbsp;&nbsp;元号の開始する日付を設定します。(例:"平成"→"1989/01/08") <br/>
   *
   * <pre>
   * 【サンプル】
   * データベースに格納されている日付型のデータを和暦で表示する。
   * 　rsSeminar.getColumn(COL.KAISAIBI).setFormatPattern(XWareki.FORMAT_DATE_LONG);
   *
   * 現在の日時を和暦に変換する。
   * 　XWareki wareki = new XWareki();
   * 　String today = wareki.format(new java.util.Date(), XWareki.FORMAT_DATETIME_SHORT);
   *
   * 和暦の文字列をDateオブジェクトに変換する。
   * 　XWareki wareki = new XWareki();
   * 　java.util.Date date = wareki.parse("H16/09/08", XWareki.FORMAT_DATE_SHORT);
   * </pre>
   *
   * @author  LittleSoft Corporation
   */
  public static class XWareki implements Serializable {

    static ResourceBundle res = ResourceBundle.getBundle("jp.littlesoft.util.Wareki");

    /**
     * 年月日 時分秒のフォーマットです。<br/><br/>
     *
     * "GGyy年MM月dd日 HH時mm分ss秒"を表します。<br/>
     */
    public static final String FORMAT_DATETIME_LONG = "GGyy年MM月dd日 HH時mm分ss秒";

    /**
     * 年月日 時分秒のフォーマットです。<br/><br/>
     *
     * "gyy/MM/dd HH:mm:ss"を表します。<br/>
     */
    public static final String FORMAT_DATETIME_SHORT = "gyy/MM/dd HH:mm:ss";

    /**
     * 年月日のフォーマットです。<br/><br/>
     *
     * "GGyy年MM月dd日"を表します。<br/>
     */
    public static final String FORMAT_DATE_LONG = "GGyy年MM月dd日";

    /**
     * 年月日のフォーマットです。<br/><br/>
     *
     * "gyy/MM/dd"を表します。<br/>
     */
    public static final String FORMAT_DATE_SHORT = "gyy/MM/dd";

    /**
     * 年月のフォーマットです。<br/><br/>
     *
     * "GGyy年MM月"を表します。<br/>
     */
    public static final String FORMAT_YEARMONTH_LONG = "GGyy年MM月";

    /**
     * 年月のフォーマットです。<br/><br/>
     *
     * "gyy/MM"を表します。<br/>
     */
    public static final String FORMAT_YEARMONTH_SHORT = "gyy/MM";

    /**
     * 年のフォーマットです。<br/><br/>
     *
     * "GGyy年"を表します。<br/>
     */
    public static final String FORMAT_YEAR_LONG = "GGyy年";

    /**
     * 年のフォーマットです。<br/><br/>
     *
     * "GGy年"を表します。<br/>
     */
    public static final String FORMAT_YEAR_LONG2 = "GGy年";

    /**
     * 年のフォーマットです。<br/><br/>
     *
     * "gyy"を表します。<br/>
     */
    public static final String FORMAT_YEAR_SHORT = "gyy";

    /**
     * 年のフォーマットです。<br/><br/>
     *
     * "GGyy"を表します。<br/>
     */
    public static final String FORMAT_YEAR_SHORT2 = "GGyy";

    /**
     * 明治の元号です。<br/><br/>
     *
     * デフォルトでは"明治"が定義されていますが、変更可能です。<br/>
     */
    public  static final  String WAREKI_GENGO1 = res.getString("WAREKI_GENGO1");

    /**
     * 大正の元号です。<br/><br/>
     *
     * デフォルトでは"大正"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO2 = res.getString("WAREKI_GENGO2");

    /**
     * 昭和の元号です。<br/><br/>
     *
     * デフォルトでは"昭和"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO3 = res.getString("WAREKI_GENGO3");

    /**
     * 平成の元号です。<br/><br/>
     *
     * デフォルトでは"平成"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO4 = res.getString("WAREKI_GENGO4");

    /**
     * 平成の次の元号です。<br/><br/>
     *
     * デフォルトでは未定義です。<br/>
     */
    public static final String WAREKI_GENGO5 = res.getString("WAREKI_GENGO5");

    /**
     * 平成の次の次の元号です。<br/><br/>
     *
     * デフォルトでは未定義です。<br/>
     */
    public static final String WAREKI_GENGO6 = res.getString("WAREKI_GENGO6");

    /**
     * "明治","大正","昭和","平成"の元号の配列です。<br/><br/>
     *
     * 平成の次，平成の次の次までリザーブされています。
     */
    public static final String[] WAREKI_GENGO = {
                          WAREKI_GENGO1, WAREKI_GENGO2, WAREKI_GENGO3,
                          WAREKI_GENGO4, WAREKI_GENGO5, WAREKI_GENGO6 };

    /**
     * 明治の省略記号です。<br/><br/>
     *
     * デフォルトでは"M"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO1_SHORT = res.getString("WAREKI_GENGO1_SHORT");

    /**
     * 大正の省略記号です。<br/><br/>
     *
     * デフォルトでは"T"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO2_SHORT = res.getString("WAREKI_GENGO2_SHORT");

    /**
     * 昭和の省略記号です。<br/><br/>
     *
     * デフォルトでは"S"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO3_SHORT = res.getString("WAREKI_GENGO3_SHORT");

    /**
     * 平成の省略記号です。<br/><br/>
     *
     * デフォルトでは"H"が定義されていますが、変更可能です。<br/>
     */
    public static final String WAREKI_GENGO4_SHORT = res.getString("WAREKI_GENGO4_SHORT");

    /**
     * 平成の次の省略記号です。<br/><br/>
     *
     * デフォルトでは未定義です。<br/>
     */
    public static final String WAREKI_GENGO5_SHORT = res.getString("WAREKI_GENGO5_SHORT");

    /**
     * 平成の次の次の省略記号です。<br/><br/>
     *
     * デフォルトでは未定義です。<br/>
     */
    public static final String WAREKI_GENGO6_SHORT = res.getString("WAREKI_GENGO6_SHORT");

    /**
     * 明治,大正,昭和,平成の省略記号である"M","T","S","H"の配列です。<br/><br/>
     *
     * 平成の次，平成の次の次までリザーブされています。<br/>
     */
    public static final String[] WAREKI_GENGO_SHORT = {
                          WAREKI_GENGO1_SHORT, WAREKI_GENGO2_SHORT, WAREKI_GENGO3_SHORT,
                          WAREKI_GENGO4_SHORT, WAREKI_GENGO5_SHORT, WAREKI_GENGO6_SHORT };

    private static final String WAREKI_GENGO1_START_DATE = res.getString("WAREKI_GENGO1_START_DATE");
    private static final String WAREKI_GENGO2_START_DATE = res.getString("WAREKI_GENGO2_START_DATE");
    private static final String WAREKI_GENGO3_START_DATE = res.getString("WAREKI_GENGO3_START_DATE");
    private static final String WAREKI_GENGO4_START_DATE = res.getString("WAREKI_GENGO4_START_DATE");
    private static final String WAREKI_GENGO5_START_DATE = res.getString("WAREKI_GENGO5_START_DATE");
    private static final String WAREKI_GENGO6_START_DATE = res.getString("WAREKI_GENGO6_START_DATE");

    private static final String[] WAREKI_GENGO_START_DATE = {
                          WAREKI_GENGO1_START_DATE, WAREKI_GENGO2_START_DATE, WAREKI_GENGO3_START_DATE,
                          WAREKI_GENGO4_START_DATE, WAREKI_GENGO5_START_DATE, WAREKI_GENGO6_START_DATE };

    /**
     * チェックレベルを表す定数です。<br/><br/>
     * 
     * 年月日をチェックしません。<br/>
     */
    public static final int WAREKI_CHECK_LEVEL_NO    = 0; //0
    
    /**
     * チェックレベルを表す定数です。<br/><br/>
     * 
     * 年までチェックします。<br/>
     */
    public static final int WAREKI_CHECK_LEVEL_YEAR  = WAREKI_CHECK_LEVEL_NO + 1; //1
    
    /**
     * チェックレベルを表す定数です。<br/><br/>
     * 
     * 年月までチェックします。<br/>
     */
    public static final int WAREKI_CHECK_LEVEL_MONTH = WAREKI_CHECK_LEVEL_YEAR + 1; //2
    
    /**
     * チェックレベルを表す定数です。<br/><br/>
     * 
     * 年月日までチェックします。<br/>
     */
    public static final int WAREKI_CHECK_LEVEL_DAY   = WAREKI_CHECK_LEVEL_MONTH + 1; //3
    
    /**
     * チェックレベルです。<br/><br/>
     * 
     * 次の４つのいずれかを設定してください。<br/>
     * ・WAREKI_CHECK_LEVEL_NO<br/>
     * ・WAREKI_CHECK_LEVEL_YEAR (ディフォルト)<br/>
     * ・WAREKI_CHECK_LEVEL_MONTH<br/>
     * ・WAREKI_CHECK_LEVEL_DAY<br/>
     */
    public static int checkLevel = WAREKI_CHECK_LEVEL_YEAR;
    
    // Resourceから読み込むのだから static でよいはず
    // 各年号の開始日
    private static java.util.GregorianCalendar[] startDates = new java.util.GregorianCalendar[WAREKI_GENGO_START_DATE.length];
    //private final SimpleDateFormat wsimpleDateFormat = new SimpleDateFormat("yyyyMMddHHmmss");
    private static final Pattern gyyPattern = Pattern.compile("^(gy+)(.*)$");
    private static final Pattern GGyyPattern = Pattern.compile("^(GGy+)(.*)$");
    //private static transient Logger logger = LogManager.getLogger(XWareki.class.getName());

    /**
     * コンストラクタ。<br/>
     */
    public XWareki() {
    }

    /**
     * フォーマットパターンが和暦のものかどうかを調べます。<br/><br/>
     *
     * 当クラスの"FORMAT_"で始まる定数のフォーマットパターンと一致するかを調べます。<br/>
     *
     * @param   formatPattern  フォーマットパターン
     * @return  True：和暦、False：和暦ではない
     */
    public static boolean isWarekiFormatPattern(String formatPattern) {
      if (StringUtil.isNullString(formatPattern)) return false;
      if (formatPattern.startsWith("gy")) {
        return true;
      } else if (formatPattern.startsWith("GGy")) {
        return true;
      } else {
        return false;
      }	
    }

    /**
     * 値と和暦フォーマットパターンからjava.util.Dateオブジェクトを作成します。<br/>
     *
     * @param   value          値
     * @param   formatPattern  和暦フォーマットパターン
     * @return  java.util.Dateオブジェクト
     * @exception  java.lang.Exception  和暦フォーマットパターンでは無い場合
     */
    public java.util.Date parse(String value, String formatPattern) throws Exception {
      if (StringUtil.isNullString(value)) return null;
      if (!isWarekiFormatPattern(formatPattern))
        throw new DataException(null, null, "FORMAT_ERROR_WAREKI",
            new String[]{formatPattern});

      // parseは年だけ行えばよい。
      Matcher m = gyyPattern.matcher(formatPattern);
      int gengoLength;
      int yearLength;
      String postPattern;
      if (m.find()) {
        gengoLength = 1;
        yearLength = m.group(1).length() - 1;
        postPattern = m.group(2);
      } else {
        m = GGyyPattern.matcher(formatPattern);
        if (m.find()) {
          gengoLength = 2;
          yearLength = m.group(1).length() - 2;
          postPattern = m.group(2);      
        } else {
        // ここに来ることはあり得ない。
            throw new DataException(null, null, "FORMAT_ERROR_WAREKI",
                new String[]{formatPattern});      
        }
      }
      // 年を除いた部分
      String wFormatPattern = formatPattern.substring(gengoLength + yearLength);
      
      // yyの次がアルファベットなら、無条件に長さで取得する。
      // yyの次が.や/などの区切り文字なら、そこまでをyyとする。
      // (ex. "gy.M.d"で"H18.6.21"など)
      char delimiter = (char)0;
      if (StringUtil.isNullString(postPattern)) {
        // "gy"など年だけの場合
        yearLength = value.length() - gengoLength;
      } else {
        char c = postPattern.charAt(0);
        if ((c < 'A' || 'Z' < c)&&(c < 'a' || 'z' < c)) {
          delimiter = c;
        }
      }
      if (delimiter != (char)0) {
        // 区切り記号までが年
        for (int i=gengoLength; i<value.length(); i++) {
          char c = value.charAt(i);
          if (c == delimiter) {
            yearLength = i - gengoLength + 1;
            break;
          }
        }
      }
      String yearStr = value.substring(gengoLength, gengoLength + yearLength);    
      
      int yearIndex=0;
      for (; yearIndex<yearStr.length(); yearIndex++) {
        char c = yearStr.charAt(yearIndex);
        if (c < '0' || '9' < c) {
          break;
        }
      }
      int year = 0;
      if (yearIndex > 0) {
        	yearStr = yearStr.substring(0, yearIndex);
        	year = Integer.parseInt(yearStr);
        	yearLength = yearIndex;
      } else {
        	yearLength = 0;
      }
      
      // 年を除いた部分
      String wValue = value.substring(gengoLength + yearLength);
      java.util.Date wDate;
      if (StringUtil.isNullString(wValue)) {
      	  // "gy"など年だけを指定した場合
      	  SimpleDateFormat sdf = new SimpleDateFormat("MMdd");
      	  wDate = sdf.parse("0101");
      } else {
        	SimpleDateFormat sdf = new SimpleDateFormat(wFormatPattern);
        	try {
          		wDate = sdf.parse(wValue);
        	} catch (Exception ex) {
          		// parse Error
          		throw new DataException(null, null, "PARSE_ERROR_WAREKI", new String[]{formatPattern, value});
        	}
      }
      java.util.GregorianCalendar cal = new GregorianCalendar();
      cal.setTime(wDate);
      
      int[] num = new int[] {0,0,0,0,0,0,0};

      num[0] = year;
      num[1] = cal.get(Calendar.MONTH) + 1;
      num[2] = cal.get(Calendar.DAY_OF_MONTH);
      num[3] = cal.get(Calendar.HOUR_OF_DAY);
      num[4] = cal.get(Calendar.MINUTE);
      num[5] = cal.get(Calendar.SECOND);    
      num[6] = cal.get(Calendar.MILLISECOND);
            
      long _long = parse(value, num[0] , num[1] , num[2] , num[3] , num[4] , num[5], num[6]);
      return new Date(_long);
    }

    /**
     * 元号と年月日よりタイムスタンプ値を取得します。<br/>
     *
     * @param  gengo   元号
     * @param  year    年
     * @param  month   月
     * @param  date    日
     * @return  タイムスタンプ値
     * @exception  java.lang.Exception  不正な日付の場合
     */
    public long parse(String gengo, int year, int month, int date) throws Exception {
      return parse(gengo, year, month, date, 0, 0, 0);
    }

    /**
     * 元号と年月日と時分秒よりタイムスタンプ値を取得します。<br/>
     *
     * @param  gengo   元号
     * @param  year    年
     * @param  month   月
     * @param  date    日
     * @param  hour    時
     * @param  minute  分
     * @param  second  秒
     * @return  タイムスタンプ値
     * @exception  java.lang.Exception  不正な日付の場合
     */
    public long parse(String gengo, int year, int month, int date, int hour, int minute, int second) throws Exception {
    	  return parse(gengo, year, month, date, hour, minute, second, 0);
    }
    
    /**
     * 元号と年月日と時分秒よりタイムスタンプ値を取得します。<br/>
     *
     * @param  gengo   元号
     * @param  year    年
     * @param  month   月
     * @param  date    日
     * @param  hour    時
     * @param  minute  分
     * @param  second  秒
     * @param  millSecond ミリ秒
     * @return  タイムスタンプ値
     * @exception  java.lang.Exception  不正な日付の場合
     */
    public long parse(String gengo, int year, int month, int date, int hour, int minute, int second, int millSecond) throws Exception {
      int index = indexOf(gengo);
      if (index==-1)
        throw new DataException(null, null, "PARSE_ERROR_GENGO",
            new String[]{gengo});
      int startYear = startDates[index].get(java.util.Calendar.YEAR);
      checkWareki(gengo, year, month, date);
      DataUtil.checkDate(year + (startYear-1), month, date);
      java.util.GregorianCalendar cal = new java.util.GregorianCalendar(year + (startYear-1), (month-1), date, hour, minute, second);
      return cal.getTime().getTime() + millSecond;
    }

    /**
     * java.util.Dateオブジェクトを和暦フォーマットパターンに従って文字列に変換します。<br/>
     *
     * @param date java.util.Dateオブジェクト
     * @param formatPattern 和暦フォーマットパターン
     * @return 変換した文字列
     * @exception java.lang.Exception フォーマットパターンが和暦ではない場合など
     */
    public String format(java.util.Date date, String formatPattern) throws DataException {
      if (date==null) return Const.NULLSTRING;
      if (!isWarekiFormatPattern(formatPattern))
        throw new DataException(null, null, "FORMAT_ERROR_WAREKI",
            new String[]{formatPattern});
      int gengoIndex = indexOf(date.getTime());
      if (gengoIndex < 0) {
        throw new DataException(null, null, "FORMAT_ERROR_DATE_WAREKI",
            new String[]{formatPattern, date.toString()});
      }

      Matcher m = gyyPattern.matcher(formatPattern);
      String gengoPattern;
      int yearLength;
      String postPattern;
      if (m.find()) {
        gengoPattern = "g";
        yearLength = m.group(1).length() - 1;
        postPattern = m.group(2);
      } else {
        m = GGyyPattern.matcher(formatPattern);
        if (m.find()) {
          gengoPattern = "GG";
          yearLength = m.group(1).length() - 2;
          postPattern = m.group(2);      
        } else {
            throw new DataException(null, null, "FORMAT_ERROR_WAREKI",
                new String[]{formatPattern});      
        }
      }
      SimpleDateFormat sdf = new SimpleDateFormat();
      
      sdf.applyPattern("yyyy");
      int year = Integer.parseInt(sdf.format(date));
      int warekiYear = year - (startDates[gengoIndex].get(java.util.Calendar.YEAR) - 1);
      String warekiYearStr = Integer.toString(warekiYear);
      while (warekiYearStr.length() < yearLength) {
        warekiYearStr = "0" + warekiYearStr;
      }
      String gengo;
      if ("GG".equals(gengoPattern)) {
        gengo = WAREKI_GENGO[gengoIndex];
      } else {
        gengo = WAREKI_GENGO_SHORT[gengoIndex];
      }

      String postString = "";
      if (StringUtil.isNotNullString(postPattern)) {
        	sdf.applyPattern(postPattern);
      	  postString = sdf.format(date);
      }

      final StringBuffer strBuf = new StringBuffer();
      strBuf.append(gengo).append(warekiYearStr).append(postString);
      return strBuf.toString();
    }

    /**
     * 元号より元号インデックスを取得します。<br/><br/>
     *
     * 不正な場合は-1を返します。<br/>
     *
     * @param gengo 元号
     * @return 元号インデックス
     */
    public int indexOf(String gengo){
      if (gengo==null) return -1;
      int i = 0;
      for (i=0; i<WAREKI_GENGO.length; i++) {
        if (gengo.startsWith(WAREKI_GENGO[i])) return i;
      }
      for (i=0; i<WAREKI_GENGO_SHORT.length; i++) {
        if (gengo.startsWith(WAREKI_GENGO_SHORT[i])) return i;
      }
      return -1;
    }

    /**
     * タイムスタンプ値より元号インデックスを取得します。<br/><br/>
     *
     * 不正な場合は-1を返します。<br/>
     *
     * @param timestamp タイムスタンプ値(単位:ミリ秒)
     * @return 元号インデックス
     */
    public int indexOf(long timestamp){
      for (int i=0; i<(startDates.length-1); i++) {
        if ( startDates[i].getTime().getTime()<=timestamp &&
             timestamp < startDates[i+1].getTime().getTime())
          return i;
      }
      return -1;
    }

    /**
     * 和暦が正しい日付かどうかをチェックします。<br/>
     *
     * @param gengo 元号
     * @param year 年
     * @param month 月
     * @param date 日
     * @exception java.lang.Exception 不正な日付の場合
     */
    public void checkWareki(String gengo, int year, int month, int date)
            throws DataException {
      // 元号が不正な場合
      int index = indexOf(gengo);
      if (index == -1) {
        throw new DataException(null, null, "PARSE_ERROR_GENGO",
            new String[]{gengo});
      }

      // チェックレベルがノーチェックの場合はOK
      if(checkLevel == WAREKI_CHECK_LEVEL_NO) {
        return;
      }

      // 開始日との比較
      int startYear  = startDates[index].get(java.util.Calendar.YEAR);
      int startMonth = startDates[index].get(java.util.Calendar.MONTH)+1;
      int startDay   = startDates[index].get(java.util.Calendar.DAY_OF_MONTH);
      if (year < 1) {
        throw new DataException(null, null, "PARSE_ERROR_WAREKI_YEAR",
            new String[]{Integer.toString(year)});
      }
      if (year == 1) {
        if ((checkLevel >= WAREKI_CHECK_LEVEL_MONTH)
            && (month < startMonth)) {
          throw new DataException(null, null, "PARSE_ERROR_WAREKI_MONTH",
              new String[]{Integer.toString(month)});
        }
        if ((checkLevel >= WAREKI_CHECK_LEVEL_DAY)
            && (month == startMonth)
            && (date < startDay)) {
          throw new DataException(null, null, "PARSE_ERROR_WAREKI_DAY",
              new String[]{Integer.toString(date)});
        }
      }
      // 終了日との比較
      if (index<(WAREKI_GENGO.length-1)) {
        int nextStartYear  = startDates[index+1].get(java.util.Calendar.YEAR);
        int nextStartMonth = startDates[index+1].get(java.util.Calendar.MONTH)+1;
        int nextStartDay   = startDates[index+1].get(java.util.Calendar.DAY_OF_MONTH);
        if ((year + (startYear-1)) > nextStartYear) {
          throw new DataException(null, null, "PARSE_ERROR_WAREKI_YEAR",
              new String[]{Integer.toString(year)});
        }
        if ((year + (startYear-1)) == nextStartYear) {
          if ((checkLevel >= WAREKI_CHECK_LEVEL_MONTH)
              && (month > nextStartMonth)) {
            throw new DataException(null, null, "PARSE_ERROR_WAREKI_MONTH",
                new String[]{Integer.toString(month)});
          }
          if ((checkLevel >= WAREKI_CHECK_LEVEL_DAY)
              && (month == nextStartMonth)
              && (date >= nextStartDay)) {
            throw new DataException(null, null, "PARSE_ERROR_WAREKI_DAY",
                new String[]{Integer.toString(date)});
          }
        }
      }
    }
  }


  /** 
   * 指定された文字が半角文字(半角カナ含む)であるかどうか調べます。<br/>
   * 
   * @param c 判定対象
   * @return True：半角文字、False：全角文字
   */
  public static boolean isHalfSize( char c ) {
    return !isFullSize( c );
  }

  /** 
   * 指定された文字列が全て半角文字(半角カナ含む)で構成されているかどうか調べます。<br/>
   * 
   * @param value 判定対象の文字列
   * @return True：全て半角文字、False：全角を含む
   */
  public static boolean isHalfSize( String value ) {
    if ( value == null ) return true;
    for ( int i = 0; i < value.length(); i++ ) {
      if ( isFullSize( value.charAt( i ) ) ) return false;
    }
    return true;
  }

  /** 
   * 指定された文字列が全て半角文字(半角カナ含まず)で構成されているかどうか調べます。<br/>
   * 
   * @param value 判定対象の文字列
   * @return True：全て半角文字(半角カナ含まず)、False：全角または半角カナを含む
   */
  public static boolean isHalfSizeWithoutKana( String value ) {
    if ( value == null ) return true;
    for ( int i = 0; i < value.length(); i++ ) {
      char c = value.charAt(i);
      if (c > '\u007e') return false;
    }
    return true;
  }

  /** 
   * 指定された文字が全角文字であるかどうか調べます。<br/>
   * 
   * @param c 判定対象の文字
   * @return True：全角文字、False：半角文字
   */
  public static boolean isFullSize( char c ) {
    if ( c <= '\u007e' ) return false;  // 半角の英数記号
    if ( c >= '\uff61' && c <= '\uff9f' ) return false;  // 半角のカタカナ
    return true;
  }

  /** 
   * 指定された文字列が全て全角文字であるかどうか調べます。<br/>
   * 
   * @param value 判定対象の文字列
   * @return True：全て全角文字、False：半角文字を含む
   */
  public static boolean isFullSize( String value ) {
    if ( value == null ) return true;
    for ( int i = 0; i < value.length(); i++ ) {
      if ( isHalfSize( value.charAt( i ) ) ) return false;
    }
    return true;
  }

  /** 
   * 指定された文字が半角英字であるかどうか調べます。<br/>
   *
   * @param c 検査対象文字
   * @return True：半角英字、False：半角英字ではない
   */
  public static boolean isAlph( char c ) {
    if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ) return true;
    return false;
  }

  /** 
   * 指定された文字列が全て半角英字であるかどうか調べます。<br/>
   *
   * @param value 検査対象の文字列
   * @return True：全て半角英字、False：半角英字以外を含む
   */
  public static boolean isAlph( String value ) {
    if ( value == null ) return false;
    for ( int i = 0; i < value.length(); i++ ) {
      if ( !isAlph( value.charAt( i ) ) ) return false;
    }
    return true;
  }

  /** 
   * 指定された文字が半角英数字であるかどうか調べます。<br/>
   *
   * @param c 検査対象文字
   * @return True：半角英数字、False：半角英数字ではない
   */
  public static boolean isAlphOrNumber( char c ) {
    if ( isAlph(c) || isNumeric(String.valueOf(c)) ) return true;
    return false;
  }

  /**
   * 指定された文字列が全て半角英数字であるかどうか調べます。<br/>
   *
   * @param value 検査対象の文字列
   * @return True：全て半角英数字、False：半角英数字以外を含む
   */
  public static boolean isAlphOrNumber( String value ) {
    if ( value == null ) return false;
    for ( int i = 0; i < value.length(); i++ ) {
      if ( !isAlphOrNumber( value.charAt( i ) ) ) return false;
    }
    return true;
  }

  /**
   * 文字列が数値をあらわすものかどうか調べます。<br/>
   *
   * @param pStrValue 検査対象
   * @return True：数値、False：数値ではない
   */
  public static boolean isNumeric(String pStrValue) {
    // Doubleで解析できれば数値とする。
    return isDoubleValue(pStrValue);
  }

  /**
   * 文字列が数値(int型の範囲)をあらわすものかどうか調べます。<br/>
   *
   * @param pStrValue 検査対象
   * @return True：数値(int型の範囲)、False：数値(int型の範囲)ではない
   */
  public static boolean isIntValue(String pStrValue) {
    try {
      Integer.parseInt(pStrValue);
      return true;
    } catch (Exception e) {
      return false;
    }
  }

  /**
   * 文字列が数値(long型の範囲)をあらわすものかどうか調べます。<br/>
   *
   * @param pStrValue 検査対象
   * @return True：数値(long型の範囲)、False：数値(long型の範囲)ではない
   */
  public static boolean isLongValue(String pStrValue) {
    try {
      Long.parseLong(pStrValue);
      return true;
    } catch (Exception e) {
      return false;
    }
  }

  /**
   * 文字列が数値(double型の範囲)をあらわすものかどうか調べます。<br/>
   *
   * @param pStrValue 検査対象
   * @return True：数値(double型の範囲)、False：数値(double型の範囲)ではない
   */
  public static boolean isDoubleValue(String pStrValue) {
    try {
      Double.valueOf(pStrValue);
      return true;
    } catch (Exception e) {
      return false;
    }
  }
}
