/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.util;

import java.io.Serializable;
import java.util.*;
import java.util.Map.Entry;

import jp.littlesoft.data.exception.DataException;

/**
 * インデックス付きのハッシュマップクラス。<br/>
 *
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("unchecked")
public class IndexedMap<K,V> extends LinkedHashMap<K,V> implements Cloneable, Serializable  {

  private static final long serialVersionUID = 6028714385242672985L;
  private ArrayList<K> indexList = null;

  /**
   * コンストラクタ。<br/>
   */
  public IndexedMap() {
    super();
  }

  /**
   * キーを変更します。<br/>
   * 
   * @param oldKey 変更前のキー
   * @param newKey 変更後のキー
   */
  public void updateKey(K oldKey, K newKey) {
    V value = this.get(oldKey);
    if (value==null) return;
    int i = indexList.indexOf(oldKey);
    if (i==-1) return;
    this.remove(i);
    // newKeyが既に存在していたらまずいが、チェックしていない。
    super.put(newKey, value);
    indexList.add(i, newKey);
  }

  /**
   * 浅いクローンを返します。<br/><br/>
   * 
   * キーや値はクローンされません。<br/>
   * 
   * @return 浅いクローン
   */
  @Override
  public Object clone() {
    IndexedMap newObject = (IndexedMap)super.clone();
    if (this.indexList != null) {
      newObject.indexList = (ArrayList)this.indexList.clone();
    }
    return newObject;
  }

  /**
   * 全てのマッピングを削除します。<br/>
   */
  @Override
  public void clear() {
    super.clear();
    indexList = null;
  }

  /**
   * キーと値を登録します。<br/><br/>
   *
   * 新規のキーの場合はインデックスにも登録します。<br/>
   * キーが既に登録されている場合は値だけを変更し、インデックスには登録しません。<br/>
   *
   * <!--
   * ※注　value に null が指定されていた場合、順番が狂ってしまうため登録不可とする。<br/>
   * この制約はなくてもよいはず。
   * -->
   *
   * @param key 登録したいキー
   * @param value キーに関連付けたい値
   * @return 指定されたキーに関連していた値。(valueがnullの場合はnull)
   */
  @Override
  public V put(K key, V value) {
    if (!super.containsKey(key)) {
      // キーが存在しない、すなわち今回初めての登録の場合 Listの末尾にキーを登録する。
      if (indexList == null) indexList = new ArrayList();
      indexList.add(key);
    }
    return super.put(key, value);
  }

  /**
   * マッピングを削除します。<br/>
   *
   * @param key 削除したいキー
   * @return Object 削除したキーに対応する値（キーが見つからなければnull）
   */
  @Override
  public V remove(Object key) {
    if (super.containsKey(key)) {
      // ArrayListからも削除する。
       indexList.remove(key);
    }
    return super.remove(key);
  };
  
  @Override
  protected boolean removeEldestEntry(Entry<K, V> eldest) {
    // TODO Auto-generated method stub
    return super.removeEldestEntry(eldest);
  }

  /**
   * マッピングを削除します。<br/>
   *
   * @param index 削除したいインデックス
   * @return Object 削除したインデックスに対応する値（インデックスが不正ならばnull）
   */
  public V removeAt(int index) {
    K key = this.getKey(index);
    if (key == null) return null;
    return this.remove(key);
  }
  


  /**
   * インデックス指定で、値を取り出します。<br/>
   *
   * @param index インデックス
   * @return インデックスに対応した値
   * @throws DataException 
   */
  public V getValue(int index) {
    K key = this.getKey(index);
    if (key == null) return null;
    return super.get(key);
  }

  
  /**
   * インデックス指定で、キーを取り出します。<br/>
   *
   * @param index インデックス
   * @return インデックスに対応したキー
   */
  public K getKey(int index) {
    if (this.indexList == null) return null;
    return this.indexList.get(index);
  }

  /**
   * 指定した値が最初に出てくる位置を返します。<br/>
   * 
   * @param value 位置を調べたい値
   * @return valueの位置
   */
  public int indexOfValue(V value) {
    if (value == null) return -1;

    for (int i=0; i<this.size(); i++) {
      if (value.equals(this.getValue(i))) {
        return i;
      }
    }
    return -1;
  }

  /**
   * 指定したキーが最初に出てくる位置を返します。<br/>
   * 
   * @param key 位置を調べたいキー
   * @return keyの位置
   */
  public int indexOfKey(K key) {
    if (this.indexList == null) return -1;
    return indexList.indexOf(key);
  }

  

  
  /**
   * このIndexHashMapの値が全て格納されている配列を取得します。<br/>
   * 
   * @param a 格納先の配列
   * @return 格納先の配列(aをそのまま返します)
   */
  public V[] toArray(V[] a) {
    if (a == null) return null;
    if (a.length != this.size()) return null;
    for (int i=0; i<a.length; i++) {
      a[i] = this.getValue(i);
    }
    return a;
  }
  

}
