/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.util.res;

import java.io.*;
import java.util.*;

import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;

/**
 * リソースから文字列の取得を行うためのクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class ResourceString implements Serializable {
  private String resourceClassName = null;
  private String resourceFieldName = null;

  private static HashMap<String, String> customResourse = new HashMap<String, String>();
  
  /**
   * コンストラクタ。
   */
  public ResourceString() {
  }
  
  /**
   * コンストラクタ。
   * 
   * @param resourceClassField リソースクラス名とフィールド名(ex. "jp.littlesoft.util.Msg.INVALID_DATAFIELD_NAME")
   */
  public ResourceString(String resourceClassField) {
    this.setResourceClassField(resourceClassField);
  }

  /**
   * コンストラクタ。
   * 
   * @param resourceClassName リソースクラス名(ex. "jp.littlesoft.util.Msg")
   * @param resourceFieldName リソースフィールド名(ex."INVALID_DATAFIELD_NAME")
   */
  public ResourceString(String resourceClassName, String resourceFieldName) {
    this.setResourceClassName(resourceClassName);
    this.setResourceFieldName(resourceFieldName);
  }

  /**
   * リソースのクラス名とフィールド名を取得します。<br/>
   * 
   * @return リソースのクラス名とフィールド名(ex. "jp.littlesoft.util.Msg.INVALID_DATAFIELD_NAME")
   */
  public String getResourceClassField() {
    if (StringUtil.isNullString(this.resourceClassName)) {
      return Const.NULLSTRING;
    } else {
      StringBuffer _retBuf = new StringBuffer();
      _retBuf.append(this.resourceClassName).append('.').append(this.resourceFieldName);
      return _retBuf.toString();
    }
  }

  /**
   * リソースのクラス名とフィールド名を設定します。<br/>
   * 
   * @param resourceClassField リソースクラス名とフィールド名(ex. "jp.littlesoft.util.Msg.INVALID_DATAFIELD_NAME")
   */
  public void setResourceClassField(String resourceClassField) {
    if (StringUtil.isNullString(resourceClassField)) {
      this.resourceClassName = null;
      this.resourceFieldName = null;
    } else {
      // 最後の"."でクラス名とフィールド名に分割する。
      int idx = resourceClassField.lastIndexOf('.');
      if (idx < 0) {
        this.resourceClassName = null;
        this.resourceFieldName = null;
      } else {
        this.resourceClassName = resourceClassField.substring(0, idx);
        this.resourceFieldName = resourceClassField.substring(idx+1);
      }
    }
  }

  /**
   * リソースのクラス名を設定します。<br/>
   * 
   * @param resourceClassName リソースクラス名(ex. "jp.littlesoft.util.Msg")
   */
  public void setResourceClassName(String resourceClassName) {
    this.resourceClassName = resourceClassName;
  }
  
  /**
   * リソースのクラス名を取得します。<br/>
   * 
   * @return リソースクラス名(ex. "jp.littlesoft.util.Msg")
   */
  public String getResourceClassName() {
    return this.resourceClassName;
  }

  /**
   * リソースのフィールド名を設定します。<br/>
   * 
   * @param resourceFieldName リソースフィールド名(ex."INVALID_DATAFIELD_NAME")
   */
  public void setResourceFieldName(String resourceFieldName) {
    this.resourceFieldName = resourceFieldName;
  }
  
  /**
   * リソースのフィールド名を取得します。<br/>
   * 
   * @return リソースフィールド名(ex."INVALID_DATAFIELD_NAME")
   */
  public String getResourceFieldName() {
    return this.resourceFieldName;
  }

  /**
   * JVMのロケールを適用して、リソースの値を取得します。<br/><br/>
   * 
   * リソースファイルからの読み込みを試みます。<br/>
   * リソースファイルに未定義の場合は、Javaクラスから値を取得します。<br/>
   * ロケールにはJVMのロケールが適用されます。<br/>
   * 
   * @return リソースの値
   * @throws DataException
   */
  public String getValue() throws DataException {
    return getValue(null);
  }

  /**
   * リソースの値を取得します。<br/><br/>
   * 
   * リソースファイルからの読み込みを試みます。<br/>
   * リソースファイルに未定義の場合は、Javaクラスから値を取得します。<br/>
   * 
   * @param locale ロケール(nullの場合は、JVMのロケールが適用されます)
   * @return リソースの値
   * @throws DataException
   */
  public String getValue(Locale locale) throws DataException {
    return getValue(this.resourceClassName, this.resourceFieldName, locale, null);
  }

  /**
   * リソースクラス名とフィールド名とロケールを指定して、リソースの値を取得します。<br/>
   * 
   * @param resourceClassName リソースクラス名(ex. "jp.littlesoft.util.Msg")
   * @param resourceFieldName リソースフィールド名(ex."INVALID_DATAFIELD_NAME")
   * @param locale ロケール(nullの場合は、JVMのロケールが適用されます。)
   * @return リソースの値
   * @throws DataException
   */
  public static String getValue(String resourceClassName, String resourceFieldName, Locale locale) throws DataException {
    return getValue(resourceClassName, resourceFieldName, locale, null);
  }

  /**
   * リソースの値を取得します。<br/><br/>
   * 
   * リソースファイルからの読み込みを試みます。<br/>
   * リソースファイルに未定義の場合は、Javaクラスから値を取得します。<br/><br/>
   * 
   * リソースの値の取得には、java.util.ResouceBundle.getBundle()メソッドを利用しています。<br/>
   * 指定したロケールでリソースファイルが見つからなかった場合は、
   * デフォルトのロケール(Locale.getDefault())のリソースファイルを見つけようとします。<br/>
   * それでも見つからない場合は、ロケールなしのリソースファイルを見つけようとします。<br/>
   * それでも見つからない場合は、Javaクラスのstaticフィールドを返します。<br/>
   * staticフィールドが見つからない場合は、DataExceptionをスローします。<br/>
   * 
   * @param resourceClassName リソースの基底名 (例:"jp.littlesoft.util.Msg")
   * @param resourceFieldName リソースのフィールド名 (例:"INVALID_DATAFIELD_NAME")
   * @param locale ロケール(デフォルトのロケールを利用する場合はnull)
   * @param loader クラスローダー(デフォルトのクラスローダーを利用する場合はnull)
   * @return リソースの値
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public static String getValue(String resourceClassName, String resourceFieldName, Locale locale, ClassLoader loader) throws DataException {
    if (StringUtil.isNullString(resourceClassName) || StringUtil.isNullString(resourceFieldName)) {
      return Const.NULLSTRING;
    } else {
      
      // カスタムメッセージが設定されていたら、そこから値を取得する。
      if (customResourse.containsKey(resourceFieldName)) {
        return customResourse.get(resourceFieldName);
      }
      
      ResourceBundle rs;
      try {
        if (loader == null) {
          // デフォルトのクラスローダーを利用する場合
          if (locale == null) {
            // localeが未設定の場合
            rs = ResourceBundle.getBundle(resourceClassName);
          } else {
            rs = ResourceBundle.getBundle(resourceClassName, locale);
          }
        } else {
          // ClassLoaderが指定されている場合
          rs = ResourceBundle.getBundle(resourceClassName, locale, loader);
        }
      } catch (Throwable t) {
        rs = null;
      }
      
      if (rs != null) {
        // リソースが設定されていたら、そこから値を取得する。
        try {
          return (String)rs.getObject(resourceFieldName);
        } catch (Throwable t) {
        }
      }
      
      try {
        // リソースが見つからなかったら、Javaのクラスから値を取得する。
        Class c = Class.forName(resourceClassName);
        java.lang.reflect.Field resField = c.getField(resourceFieldName);
        return (String)resField.get(null);          
      } catch (Throwable t) {
        throw new DataException("Can not find resource string . : " + resourceClassName + "." + resourceFieldName + " (Locale=" + locale + ")");
      }
    }
  }

  /**
   * 各アプリケーション用のカスタムメッセージを設定します。
   * 
   * @param key キー
   * @param message メッセージ
   */
  public static void setCustomMessage(String key, String message) {
    ResourceString.customResourse.put(key, message);
  }

  /**
   * 各アプリケーション用のカスタムメッセージを削除します。
   * 
   * @param key キー
   */
  public static void removeCustomMessage(String key) {
    ResourceString.customResourse.remove(key);
  }
  
}

