/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.sql.Connection;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Properties;
import java.util.TimeZone;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import jp.littlesoft.file.server.UploadFilter;
import jp.littlesoft.sql.exception.LSSQLException;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.resource.loader.StringResourceLoader;
import org.apache.velocity.runtime.resource.util.StringResourceRepository;

/**
 * コントローラの基底クラス。<br/><br/>
 * 
 * 当クラスを継承してアプリケーション毎のコントローラを作成します。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public abstract class ActionController extends HttpServlet {
  

  
  /**
   * ファイルセパレータを表す定数です。<br/><br/>
   * System.getProperty("file.separator")で得られる値がセットされています。<br/>
   */
  public final static String FILE_SEP = System.getProperty("file.separator");
  
  /**
   * アクションコントローラオブジェクトがRequest属性にセットされる際の属性名を表す定数です。<br/>
   */
  public final static String RA_CURRENT_CONTROLLER = ActionController.class.getName() + ".CURRENT_CONTROLLER";
  
  /**
   * Javaアプリケーションからコールバックされた際、Request属性にセットされる際のアプリケーションパラメータオブジェクトを表す定数です。<br/>
   */
  public final static String RA_JAVA_APPLICATION_CALLBACK_PARAM = ActionController.class.getName() + ".JavaAppCallBack.Param";
  
  
  /**
   * サブセッションオブジェクトがRequest属性にセットされる際の属性名を表す定数です。<br/>
   */
  public final static String RA_CURRENT_SUBSESSION = ActionController.class.getName() + ".CURRENT_SUBSESSION";


  /**
   * アクションコントローラへ最初にアクセスした際に使用されるアクションページクラス名です。<br/>
   */
  protected String DefaultPageClassName;
  
  /**
   * 2009/08/10 tableのパラメータ名称を固定化するか。<br/>
   */
  private boolean useStaticParameterName;
  
  /**
   * 当サーブレットのVelocityEngineがRequest属性にセットされる際の属性名を表す定数です。<br/>
   */
  public final static String RA_VELOCITY_ENGINE = ActionController.class.getName() + ".velocityEngine";

  /**
   * 2010.02.07 VelocityEngine サーブレット毎のインスタンスに    
   */
  protected VelocityEngine velocityEngine = new VelocityEngine(); 
  public final static String PN_VelocityFileLoaderPath = "VelocityFileLoaderPath";
  public final static String PN_UseStaticParameterName = "UseStaticParameterName";
  public final static String PN_DefaultPageClassName   = "DefaultPageClassName";
  /**
   * HttpServletのinit(ServletConfig config)をオーバーライドしています。<br/><br/>
   * 
   * <em>※注意：</em>
   * このクラスではテンプレートエンジン(Velocity)の初期化が行われているので、
   * アプリケーションクラスでオーバーライドする場合は注意してください。<br/>
   * 
   * @see javax.servlet.GenericServlet#init(javax.servlet.ServletConfig)
   */
  @Override
  public void init(ServletConfig config) throws ServletException {
    String realPath = config.getServletContext().getRealPath("WEB-INF") + FILE_SEP;
    //  2009/08/10 tableのパラメータ名称を固定化するか。
    useStaticParameterName = (new Boolean(config.getInitParameter(PN_UseStaticParameterName))).booleanValue();
    DefaultPageClassName = config.getInitParameter(PN_DefaultPageClassName);
    if (DefaultPageClassName==null) {
      DefaultPageClassName = DefaultPage.class.getName();
    }    
    try {
      Properties p = new Properties();
      try {
        // テンプレートの読み込み(getTemplate())は、
        //「WEBコンテキストルートディレクトリ」+「web.xmlでのVelecityFileLoaderPath」
        // をルートとして行います。
        String webContextPath = config.getServletContext().getRealPath("");
        String settingPath = config.getInitParameter(PN_VelocityFileLoaderPath);
        String loaderPath = webContextPath;
        if (settingPath != null && !settingPath.equals("")) loaderPath += FILE_SEP + settingPath;
        p.load(new FileInputStream(realPath + "lsjVelocity.properties"));
        p.setProperty(Velocity.FILE_RESOURCE_LOADER_PATH, loaderPath);
        velocityEngine.init(p);
      } catch (IOException x) {}
    } catch (Exception e) {
      throw new ServletException(e);
    }
  }
  
  private boolean isTrue(HttpServletRequest request, String attrName) {
    return (request.getAttribute(attrName)!=null && ((Boolean)request.getAttribute(attrName)) );
  }
  /**
   * HttpServletのservice(request, response)をオーバーライドしています。<br/><br/>
   * 
   * <em>※注意：</em>
   * このクラスではLSJのセッション管理やアクション実行に必要な処理を実装してるので、
   * アプリケーションクラスでオーバーライドする場合は注意してください。<br/>
   * 
   * @see javax.servlet.http.HttpServlet#service(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
   * @see #onStartNewSession(HttpServletRequest, HttpServletResponse)
   * @see #createActionPage(String, HttpServletRequest, HttpServletResponse)
   * @see jp.littlesoft.web.ActionSubSession#execute(HttpServletRequest, HttpServletResponse)
   * @see jp.littlesoft.web.ActionSubSession#getID()
   * @see jp.littlesoft.web.ActionSubSession#getMaxInactiveInterval()
   * @see jp.littlesoft.web.ActionSubSession#getLastAccessedTime()
   */
  @SuppressWarnings("unchecked")
  @Override
  protected void service(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
    
    HttpSession session = request.getSession(true);
    synchronized (session) {
      String subSessionID = request.getParameter(ActionPage.PN_SUBSESSION_ID);
      ActionSubSession subSession = null;
      try {
        request.setAttribute(RA_CURRENT_CONTROLLER, this);
        request.setAttribute(RA_VELOCITY_ENGINE, this.velocityEngine);
        request.setAttribute("UseStaticParameterName", useStaticParameterName); // 2009/08/10 tableのパラメータ名称を固定化するか。
        
        if ( subSessionID!=null && session.getAttribute(subSessionID) instanceof ActionSubSession ) {
          // セッションの継続 - 識別子で対象のSubSessionを取り出します。
          subSession =  (ActionSubSession)session.getAttribute(subSessionID);
        } else {
          // 新しいセッションを開始します。
          onStartNewSession(request, response);
          subSession = new ActionSubSession(createActionPage(DefaultPageClassName,request,response));
        }
        
        if (isTrue(request, UploadFilter.RA_IMMEDIATELY_RETURN)) {
          response.getWriter().print("OK");
          response.getWriter().flush();
          response.getWriter().close();
        } else {
          subSession.execute(request, response);
        }
      } catch (Exception x) {
        x.printStackTrace();
        this.errorAction(request,response,x);
        return;
      } finally {
        session.setAttribute(subSession.getID(), subSession);
        ArrayList<String> names = new ArrayList<String>();
        try {
          // サブセッション毎のタイムアウトを検出し、不要なものはHttpSessionから削除します。
          for (Enumeration<String> e = session.getAttributeNames(); e.hasMoreElements() ;) {
            names.add(e.nextElement());
          }
          for (String name : names) {
            if (session.getAttribute(name)!=null && session.getAttribute(name) instanceof ActionSubSession) {
              ActionSubSession ss = (ActionSubSession)session.getAttribute(name);
              //ここでリソースを解放すると不整合が起こるのでコメントアウト
              ss.getActionPage().releaseJDBCConnection();
              
              if (ss.getMaxInactiveInterval()==0) {
                // サブセッションのMaxInactiveIntervalがゼロの場合は、セッションで指定されているMaxInactiveIntervalを設定します。
                ss.setMaxInactiveInterval(session.getMaxInactiveInterval());
              }
              if (ss.getMaxInactiveInterval()!=-1 && 
                  ( (ss.getLastAccessedTime() + ss.getMaxInactiveInterval() * 1000L)
                    < System.currentTimeMillis())) {
                session.removeAttribute(name);  // タイムアウトしているサブセッションをHttpSessionから削除します。
              }
            }
          }
          request.setAttribute(RA_CURRENT_CONTROLLER, null);
        }
        catch (Exception x) {
        }
      }
    }
  }
  
  /**
   * 例外発生時の処理を実装します。<br/><br/>
   * 
   * 以下の場合に呼び出されます。<br/>
   * <ul>
   *   <li>アクションページのメソッド内から、LSApplicationException 以外の例外がスローされた場合</li>
   *   <li>予期しない例外が発生した場合</li>
   * </ul><br/>
   * 
   * このメソッドの処理は空です。<br/>
   * 
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @param x 発生した例外
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   */
  protected void errorAction(HttpServletRequest request, HttpServletResponse response, Exception x) throws ServletException, IOException {} 

  /**
   * セッション新規作成時の処理を実装します。<br/><br/>
   * 
   * このメソッドの処理は空です。<br/>
   * 
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @throws Exception 例外発生時
   */
  protected void onStartNewSession(HttpServletRequest request, HttpServletResponse response) throws Exception{}    
  
  /**
   * 引数 className で指定された ActionPage オブジェクトを生成します。<br/>
   * 
   * @param className アクションページクラス名
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @return アクションページ
   * @throws Exception 例外発生時
   */
  @SuppressWarnings("unchecked")
  protected ActionPage createActionPage(String className, HttpServletRequest request, HttpServletResponse response) throws Exception{
    ActionPage actionPage = null;
    Constructor constructor = null;
    try{
      constructor = Class.forName(className).getConstructor(new Class[]{HttpServletRequest.class,HttpServletResponse.class});
      actionPage = (ActionPage)constructor.newInstance(new Object[]{request,response});
    } catch (NoSuchMethodException e){
      constructor = Class.forName(className).getConstructor(new Class[]{});
      actionPage = (ActionPage)constructor.newInstance(new Object[]{});
    }
    return actionPage;
  }
  
  /**
   * ブラウザの「戻る」や「更新」を使用した場合のエラー処理を実装します。<br/><br/>
   * 
   * アクションページがバックボタンによる不正なリクエストを受け付けない場合、サブセッションによって
   * リクエスト送信の順序チェックが行われ、エラーが発生した場合に呼び出されます。<br/><br/>
   * 
   * このメソッドの処理は空です。<br/>
   * 
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   */
  protected void onSequenceError(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {}

  /**
   * JDBCコネクションの取得処理を実装します。<br/><br/>
   * 
   * このメソッドの処理は空です。<br/>
   * 
   * @return null
   * @throws LSSQLException 例外発生時
   */
  @Deprecated
  public Connection getConnection() throws LSSQLException { return null; }
  
  /**
   * JDBCコネクションの取得処理を実装します。<br/><br/>
   * 
   * このメソッドの処理は空です。<br/>
   * 
   * @return null
   * @throws LSSQLException 例外発生時
   */
  public Connection getConnection(HttpServletRequest request) throws LSSQLException { return null; }
  
  
}
