/* Copyright (c) 2006-2010 LittleSoft Corporation
 *
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSessionBindingEvent;

import jp.littlesoft.LSApplicationException;
import jp.littlesoft.data.event.TableEvent;
import jp.littlesoft.data.event.TableListener;
import jp.littlesoft.file.common.Attributes;
import jp.littlesoft.file.common.IApplicationParams;
import jp.littlesoft.file.server.UploadFileManager;
import jp.littlesoft.soja.LSJson;

import org.apache.log4j.Logger;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.app.event.EventCartridge;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * アクションページの基底クラス。<br/>
 *
 * 当クラスを継承して機能毎のアクションページを作成します。<br/>
 *
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class ActionPage implements TableListener, Serializable {

  /**
   * リクエスト属性予約名 ： 子ページへ渡される親ページのインスタンス<br/>
   */
  public final static String RA_PARENT_PAGE = ActionPage.class.getName() + ".PARENT_PAGE";

  /**
   * リクエスト属性予約名 ： 親ページから子ページへの引数（HashMapオブジェクト）<br/>
   */
  public final static String RA_PARENT_ARGS = ActionPage.class.getName() + ".PARENT_ARGS";

  /**
   * JSESSIONIDを表す定数です。
   */
  public final static String PN_JSESSION_ID = "jsessionid";

  /**
   * サブセッションの識別子を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_SUBSESSION_ID = "_sbsid";

  /**
   * 自身のクラス名を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_CLASS = "_class";

  /**
   * 自身のURLを表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_SELF_URL = "_self";
  
  /**
   * 自身のURLを表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_SELF_GET_URL = "_get";

  /**
   * フォームタグ名を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_FORM_NAME = "_form";

  /*
   * フォームタグのメソッド名を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_METHOD = "_method";

  /**
   * アクションメソッドの引数を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_ARGS = "_args";

  /**
   * リクエストの送信順序判定用の識別子です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_SEQ = "_seq";

  /**
   * ロードするJavaScriptを表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_LOADING_SCRIPT = "_loadscript";

  /**
   * Ajaxの使用／未使用です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_USE_AJAX = "_useajax";

  /**
   * アクションページの識別子を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_ACTIONPAGE_ID = "_apid";

  /**
   * 親のアクションページの識別子を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_PARENT_ACTIONPAGE_ID = "_papid";

  /**
   * 親のアクションページの識別子を表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）<br/>
   */
  public final static String PN_APPLICATION_PARAMS = "ls_actionPage_APPLICATION_PARAMS";

  /**
   * アクションメソッドの引数のセパレータを表す定数です。
   * （Veloctyからアクセスできる予約済みの属性名）
   */
  public final static String DELIMITER_ARGS = "$$_%_$$";

  public final static String ARGS_ISNULL = "ls_ajax_arguments_is_NULLSTRING";

  private String contentType = "text/html;charset=UTF-8";

  /**
   * クライアントがアクセスする ActionSubSession オブジェクトです。<br/>
   */
  protected transient ActionSubSession subSession;

  /**
   * クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクトです。<br/>
   */
  protected transient HttpServletRequest request;

  /**
   * Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクトです。<br/>
   */
  protected transient HttpServletResponse response;

  private String templateName = "common/default.html";
  private VelocityContext velocityContext = null;

  private boolean strictSession = true;
  private String sequenceValue = null;
  private String alertMessage = null;
  private String formName = "form1";
  private boolean useAjax = false;

  transient protected Logger logger = null;

  /**
   * アクションページ毎の識別子です。<br/>
   */
  protected String actionPageId = null;

  /**
   * 親の ActionPage オブジェクトです。<br/><br/>
   *
   * 自身が子ページ（ダイアログ）の場合にのみ、オブジェクトがセットされます。<br/>
   */
  protected ActionPage parentPage = null;

  /**
   * 子の ActionPage オブジェクトのリストです。<br/><br/>
   *
   * 自身が親ページ（ダイアログを持つ）場合のみ、子ページがセットされます。<br/>
   */
  protected HashMap<String, ActionPage> subActionPage = new HashMap<String, ActionPage>();

  private transient Connection conn = null;

  private transient List<Connection> connList = new ArrayList<Connection>();

  /** 自クラス名 */
  public String selfClassName = this.getClass().getName();

  /**
   * executeメソッドでディスパッチするメソッド名
   */
  public String methodName;

  /**
   *
   */
  private String loadscript = "";



  /**
   * ActionPage オブジェクトを生成します。<br/>
   */
  public ActionPage() {
    this.actionPageId = Long.toHexString(System.currentTimeMillis() + this.hashCode());
  }

  /**
   * 引数 request、response で指定された ActionPage オブジェクトを生成します。<br/><br/>
   *
   * サブセッション取得のための初期処理を呼び出します。<br/>
   *
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @see #ActionPage()
   * @see #init(HttpServletRequest, HttpServletResponse)
   */
  public ActionPage(HttpServletRequest request, HttpServletResponse response) {
    this();
    this.init(request, response);
  }


  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    connList = new ArrayList<Connection>();
  }

  /**
   * サブセッション取得のための初期処理を行います。<br/><br/>
   *
   * リクエストにセットされている ActionSubSession オブジェクトを取得し、
   * メンバー変数セットのための初期処理を呼び出します。<br/>
   *
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @see #init(ActionSubSession, HttpServletRequest, HttpServletResponse)
   */
  public void init(HttpServletRequest request, HttpServletResponse response) {
    ActionSubSession subSession = (ActionSubSession)request.getAttribute(ActionController.RA_CURRENT_SUBSESSION);
    init(subSession, request, response);
  }

  /**
   * メンバー変数セットのための初期処理です。<br/><br/>
   *
   * メンバー変数に ActionSubSession オブジェクト、 HttpServletRequest オブジェクト、
   *  HttpServletResponse オブジェクトをセットします。<br/>
   *
   * @param subSession クライアントがアクセスする ActionSubSession オブジェクト
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   */
  public void init(ActionSubSession subSession, HttpServletRequest request, HttpServletResponse response) {
    logger = Logger.getLogger(this.getClass());
    this.subSession = subSession;
    this.request  = request;
    this.response = response;
    for (Field field : getClass().getFields()) {
      if (Modifier.isPublic(field.getModifiers())) {
        try {
          Object ctxobj = field.get(this);
          if (ctxobj instanceof ICheckableRequest) {
            ICheckableRequest checkableRequest = (ICheckableRequest)ctxobj;
            checkableRequest.setName(field.getName());
          }
        } catch (Exception x) {
          throw new RuntimeException(x);
        }
      }
    }
  }

  /**
   * リクエストからアクションコントローラを取得します。<br/>
   *
   * @return アクションコントローラ
   */
  public ActionController getController() {
    if (this.request!=null) {
      return (ActionController)request.getAttribute(ActionController.RA_CURRENT_CONTROLLER);
    }
    return null;
  }

  /**
   * セッションから、引数 name で指定された属性名のオブジェクトを取得します。<br/>
   *
   * @param name 属性名
   * @return オブジェクト
   */
  protected Object getSessionAttribute(String name) {
    return this.request.getSession(true).getAttribute(name);
  }

  /**
   * セッションへ、引数 name、value で指定されたオブジェクトをセットします。<br/>
   *
   * @param name 属性名
   * @param value オブジェクト
   */
  protected void setSessionAttribute(String name, Object value) {
    this.request.getSession(true).setAttribute(name, value);
  }

  /**
   * セッションから、引数 name で指定された属性名のオブジェクトを削除します。<br/>
   *
   * @param name 属性名
   */
  protected void removeSessionAttribute(String name) {
    this.request.getSession(true).removeAttribute(name);
  }

  /**
   * 解析・出力するVelocityテンプレートを設定します。<br/><br/>
   *
   * Velocityのファイル検索パスからの相対パスで記述します。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   */
  public void setTemplateName(String templateName) {
    this.templateName = templateName;
  }

  /**
   * JavaScriptのAlertつきで、ページを出力します。<br/>
   *
   * @param alertMessage 出力するAlertメッセージ
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #show()
   */
  public void showWithAlert(String alertMessage, boolean increment) throws ServletException, IOException  {
    try {
      this.alertMessage = alertMessage;
      show(this.templateName, this.response.getWriter(), increment);
    } finally {
      this.alertMessage = null;
    }
  }
  public void showWithAlert(String alertMessage) throws ServletException, IOException  {
    showWithAlert(alertMessage, true);
  }

  /**
   * 設定されているテンプレートを解析し、設定されている出力先にページを出力します。<br/>
   *
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #show(String, PrintWriter)
   */
  public void show() throws ServletException, IOException  {
    show(this.templateName, this.response.getWriter());
    if (this.subSession!=null && this.parentPage==null) this.subSession.setActionPage(this);
  }

  /**
   * 指定されたテンプレートを解析し、設定されている出力先にページを出力します。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #show(String, PrintWriter)
   */
  public void show(String templateName) throws ServletException, IOException  {
    show(templateName, this.response.getWriter());
    if (this.subSession!=null && this.parentPage==null) this.subSession.setActionPage(this);
  }

  /**
   * 設定されているテンプレートを解析し、指定された出力先（ファイル等）にページを出力します。<br/>
   *
   * @param out 出力先
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #show(String, PrintWriter)
   */
  public void show(PrintWriter out) throws ServletException, IOException  {
    show(this.templateName, out);
  }

  /**
   * 指定されたテンプレートを解析し、指定された出力先にページを出力します。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   * @param out 出力先
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #setTemplateName(String)
   * @see #merge(String, PrintWriter)
   */
  public void show(String templateName, PrintWriter out)  throws ServletException, IOException {
    show(templateName, out, true);
  }
  private void show(String templateName, PrintWriter out, boolean increment)  throws ServletException, IOException {
    setTemplateName(templateName);
    merge(this.templateName, out, increment);
  }

  /**
   * 指定されたテンプレートファイルを解析した内容（String）を返します。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   * @return マージしたString
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #merge(String, PrintWriter)
   */
  public String getMergedString(String templateName)  throws ServletException, IOException {
    java.io.StringWriter sw = new java.io.StringWriter();
    java.io.PrintWriter pw = new java.io.PrintWriter(sw);
    this.merge(templateName, pw);
    return sw.toString();
  }

  /**
   * 指定されたテンプレートを解析（Publicフィールドとテンプレートの合成）し、
   * 設定されている出力先にページを出力します。<br/><br/>
   *
   * showとの違いはthis.templateNameとSubSessionのActionPageが更新されない事です。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #merge(String, PrintWriter)
   */
  public void merge(String templateName)  throws ServletException, IOException {
    merge(templateName, this.response.getWriter());
  }

  /**
   * 指定されたテンプレートを解析（Publicフィールドとテンプレートの合成）し、
   * 指定された出力先にページを出力します。<br/><br/>
   *
   * showとの違いはthis.templateNameとSubSessionのActionPageが更新されない事です。<br/>
   *
   * @param templateName Velocityテンプレートファイル名
   * @param out 出力先
   * @throws ServletException 例外発生時
   * @throws IOException 例外発生時
   * @see #onStartShow()
   * @see #setContenxt(String, Object)
   */
  public void merge(String templateName, PrintWriter out)  throws ServletException, IOException {
    merge(templateName, out, true);
  }
  private void merge(String templateName, PrintWriter out, boolean increment)  throws ServletException, IOException {
    checkVelocityContext(increment);
    try {
      this.onStartShow(velocityContext);
      //this.onStartShow();
    } catch (Exception x) {
      throw new ServletException(x);
    }

    String tempLoadscript = getLoadscript();

    Field[] fields =  this.getClass().getFields();
    EventCartridge eventCartridge = new EventCartridge();
    for (int i=0; i<fields.length; i++) {
      if (Modifier.isPublic(fields[i].getModifiers())) {
        try {
          Object ctxobj = fields[i].get(this);
          if (ctxobj instanceof ICheckableRequest) {
            ICheckableRequest checkableRequest = (ICheckableRequest)ctxobj;
            checkableRequest.setName(fields[i].getName());
            if (checkableRequest.getLoadscript()!=null &&
                !(checkableRequest.getLoadscript().equals(""))) {
              tempLoadscript += checkableRequest.getLoadscript();
            }
          }

          if (ctxobj instanceof WebDataTable) {
            // 2009/08/10 WebDataTableのパラメータ名を固定にする対応
            if ((Boolean)(this.request.getAttribute("UseStaticParameterName"))){
              ((WebDataTable<?>)ctxobj).setId(fields[i].getName());
            }
            WebDataTable.PageContext pc  = ((WebDataTable<?>)ctxobj).toPageContext();
            ctxobj = pc;
            eventCartridge.addEventHandler(new WebDataTable.PageVelocityHandler(pc));
            this.setContenxt(fields[i].getName(), ctxobj, increment);
          }
          else {
            if (ctxobj instanceof String) {
              // @EscapeStringが付与されている場合
              if (fields[i].isAnnotationPresent(EscapeString.class)) {
                EscapeString escapeString = fields[i].getAnnotation(EscapeString.class);
                boolean isEscapeString = escapeString.isEscapeString();
                // isEscapeString=trueの場合、値をHTMLエスケープする
                if (isEscapeString) {
                  String strCtxobj = (String) ctxobj;
                  this.setContenxt(fields[i].getName(), HTML.escape(strCtxobj), increment);
                }
                // 値をHTMLエスケープしない
                else {
                  this.setContenxt(fields[i].getName(), ctxobj, increment);
                }
              }
              // @EscapeStringが付与されていない場合
              else {
                // 値をHTMLエスケープする
                String strCtxobj = (String) ctxobj;
                this.setContenxt(fields[i].getName(), HTML.escape(strCtxobj), increment);
              }

            }
            else {
              // 値をHTMLエスケープしない
              this.setContenxt(fields[i].getName(), ctxobj, increment);
            }
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }

    response.setContentType(getContentType());
    if (alertMessage!=null) {
      tempLoadscript += "  alert('" + alertMessage + "');\n";
    }
    velocityContext.put(PN_LOADING_SCRIPT,tempLoadscript);
    velocityContext.attachEventCartridge(eventCartridge);
    try {
      Template template;
      //template = Velocity.getTemplate(templateName);
      template = getVelocityEngine().getTemplate(templateName);
      template.merge(velocityContext,out);
    } catch (org.apache.velocity.exception.ResourceNotFoundException x) {
      out.println("<html><body>" + x.getLocalizedMessage() + "</body></html>");
    } catch (Exception e) {
      e.printStackTrace();
      throw new ServletException(e);
    } finally {
      velocityContext.attachEventCartridge(null);
      velocityContext = null;
      out.flush();
      out.close();
    }
  }

  private void checkVelocityContext(boolean increment) {
    if (this.velocityContext != null) return;
    String id = "";
    if (this.subSession != null) id = this.subSession.getID();
    this.sequenceValue = this.getSequenceValue(increment);
    this.velocityContext = new VelocityContext();
    this.velocityContext.put(PN_JSESSION_ID, this.getJsessionID());
    this.velocityContext.put(PN_SELF_URL,  this.request.getContextPath() + this.request.getServletPath());
    this.velocityContext.put(PN_SELF_GET_URL,  
        this.request.getContextPath() + this.request.getServletPath() +
        "?" + PN_SUBSESSION_ID + "=" + id + 
        "&" + PN_ACTIONPAGE_ID + "=" + actionPageId +
        "&" + PN_SEQ + "=" + sequenceValue +
        ((this.parentPage != null) ?  ("&" + PN_PARENT_ACTIONPAGE_ID + "=" + parentPage.actionPageId) : "" ));
    this.velocityContext.put(PN_FORM_NAME,  this.getFormName());
    this.velocityContext.put(PN_SUBSESSION_ID,  id);
    this.velocityContext.put(PN_CLASS,  this.getClass().getName());
    this.velocityContext.put(PN_SEQ,  this.sequenceValue);
    this.velocityContext.put(PN_USE_AJAX,  this.isUseAjax());
    this.velocityContext.put(PN_ACTIONPAGE_ID,  this.actionPageId);
    if (this.parentPage != null) {
      this.velocityContext.put(PN_PARENT_ACTIONPAGE_ID,  this.parentPage.actionPageId);
    }
    else {
      this.velocityContext.put(PN_PARENT_ACTIONPAGE_ID,  this.actionPageId);
    }
  }

  /**
   * コントロール項目の内容をHashMapで取得します.
   * @param increment true: sequenceValueが変化します.
   * @return HashMap
   */
  public HashMap<String, String> getControlParams(boolean increment) {
    HashMap<String, String> result = new HashMap<String, String>();
    sequenceValue = getSequenceValue(increment);
    result.put(PN_JSESSION_ID, getJsessionID());
    result.put(PN_SELF_URL,  request.getContextPath() + request.getServletPath());
    result.put(PN_FORM_NAME,  getFormName());
    result.put(PN_SUBSESSION_ID,  (subSession != null) ? subSession.getID() : "");
    result.put(PN_CLASS,  getClass().getName());
    result.put(PN_SEQ,  sequenceValue);
    result.put(PN_USE_AJAX,  new Boolean(isUseAjax()).toString());
    result.put(PN_ACTIONPAGE_ID,  actionPageId);
    if (this.parentPage != null) {
      result.put(PN_PARENT_ACTIONPAGE_ID,  parentPage.actionPageId);
    } else {
      result.put(PN_PARENT_ACTIONPAGE_ID,  actionPageId);
    }
    return result;
  }

  /**
   * Velocityに解析させるキーと値のペアを設定します。
   *
   * @param key キー
   * @param value 値
   */
  public void setContenxt(String key, Object value) {
    setContenxt(key, value, true);
  }
  private void setContenxt(String key, Object value, boolean increment) {
    checkVelocityContext(increment);
    this.velocityContext.put(key, value);
  }

  /**
   * サブセッションまたは親アクションページから処理要求がディスパッチされた際の処理を実装します。<br/><br/>
   *
   * メンバー変数セットのための初期処理、アクションメソッド前処理、アクションメソッド実行を行います。<br/>
   * アクションメソッドの指定が無い場合(初期表示時など)、指定されたテンプレートファイルでページを出力します。<br/><br/>
   *
   * 指定されたアクションメソッドがアクションページに存在しない場合、「doDefault」を
   * 対象のアクションメソッドとします。<br/><br/>
   *
   * アクションメソッドに「UseJDBC」アノテーションが付与されている場合、自動的にセッション管理を
   * 行います。<br/>
   *
   * @param subSession クライアントがアクセスする ActionSubSession オブジェクト
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @throws Exception 例外発生時
   * @see #init(ActionSubSession, HttpServletRequest, HttpServletResponse)
   * @see #show()
   */
  public void execute(ActionSubSession subSession, HttpServletRequest request, HttpServletResponse response) throws Exception {
    alertMessage = null;

    // リクエストを送信したアクションページのアクションページの識別子を取得
    String reqActionPageId = request.getParameter(PN_ACTIONPAGE_ID);
    //「送信元のアクションページ <> SubSessionのアクションページ」の場合は、
    // 子ページからの送信の為、子ページのexecuteへディスパッチする。
    if (reqActionPageId!=null && !this.actionPageId.equals(reqActionPageId)) {
      ActionPage subActionPage = this.subActionPage.get(reqActionPageId);
      if (subActionPage!=null) {
        subActionPage.execute(subSession, request, response);
        return;
      }
    }

    this.init(subSession, request, response);
    methodName = request.getParameter(PN_METHOD);
    if ("".equals(methodName)) {
      this.show();
      return;
    }
    String argStr = request.getParameter(PN_ARGS);
    Class<?>[] argsClass = null;
    Object[] argsObject = null;

    if (argStr != null && !argStr.equals("")){
      List<String> l = this.tokenToList(argStr,DELIMITER_ARGS);
      argsClass = new Class[l.size()];
      argsObject = new Object[l.size()];
      //2009.12.15 by tnagao argsにJSONオブジェクトをセット可能にする.
      for(int i=0 ; i<l.size(); i++){
        String arg = l.get(i);
        if (arg!=null && arg.startsWith("{") && arg.endsWith("}") && arg.indexOf("class")>-1) {
          argsObject[i] = LSJson.deserialize(arg);
          argsClass[i] = argsObject[i].getClass();
        } else {
          argsObject[i] = arg;
          argsClass[i] = String.class;
        }
      }
      if (argsObject.length > 0) {
        getRequest().setAttribute(ActionPage.class.getName() + "_args", argsObject);
      }
//      for(int i=0 ; i<l.size(); i++){
//        argsClass[i] = String.class;
//        argsObject[i] = l.get(i);
//      }
    }

    java.lang.reflect.Method method = null;
    if (methodName!=null) {
      try {
        method = this.getClass().getMethod(methodName, argsClass);
      } catch (java.lang.NoSuchMethodException e) {
        System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
        System.out.println("@@@ Error in ActionPage[" + this.getClass().getName());
        System.out.println("@@@ " + e.getClass().getName() );
        System.out.println("@@@ " + e.getLocalizedMessage() );
        System.out.println("@@@ " + argsToString(methodName, argsClass, argsObject));
        System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
        method = this.getClass().getMethod("doDefault", (Class[])null);
      }
    }
    if (method == null) {
      method = this.getClass().getMethod("doDefault",(Class[])null );
    }

    this.onStartAction(method.getName());

    //メソッドに「UseJDBC」アノテーションが付与されている場合は
    //コネクションを取得する。
    boolean useJdbcAndAuto = false;
    try {
      if (method.isAnnotationPresent(UseJDBC.class)) {
        //コネクションを取得
        conn = this.getController().getConnection();
        //アノテーションのメンバを取得
        UseJDBC useJdbcAnno = method.getAnnotation(UseJDBC.class);
        useJdbcAndAuto = useJdbcAnno.autoCommit(); //手動管理(commit、rollback)設定
      }
      if (checkRequest(method)) {
        try {
          method.invoke(this, argsObject);
        } catch (java.lang.IllegalArgumentException x) {
          System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
          System.out.println("@@@ Error in ActionPage[" + this.getClass().getName());
          System.out.println("@@@ " + x.getClass().getName() );
          System.out.println("@@@ " + x.getLocalizedMessage() );
          System.out.println("@@@ " + argsToString(method.getName(), argsClass, argsObject));
          System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
          throw new java.lang.IllegalArgumentException(argsToString(method.getName(), argsClass, argsObject));
        }
      }
      if (useJdbcAndAuto) conn.commit();
    } catch (InvocationTargetException x) {
      if (useJdbcAndAuto) conn.rollback();
      if (x.getCause() instanceof Exception ) {
        if (!(x.getCause() instanceof LSApplicationException)) {
          this.onExceptionAction((Exception)x.getCause());
        }
      } else {
        this.onExceptionAction(new Exception(x.getCause()));
      }
    } catch (Exception x) {
      if (useJdbcAndAuto) conn.rollback();
      this.onExceptionAction(x);
    } finally {
      //releaseJDBCConnection();
    }
    return;
  }
  
  static private String argsToString(String methodName, Class<?>[] argsClass, Object[] argsObject) {
    String args = "";
    if (argsClass==null) {
      args += "null";
    } else {
      int i=0;
      for (Class<?> clazz : argsClass) {
        args +=  ((i==0) ? " " : " , ");
        args += clazz.getName() + " : " 
                + ((argsObject!=null && argsObject[i]!=null) 
                   ?  ("'" + argsObject[i].toString() + "'")  : "'null'");        
        i++;
      }
    }
    return  methodName + "{" + args + "}";
  }

  /**
   * ls.exec, ls.ajax の引数を取得します.<br/>
   * @return 引数配列
   */
  public Object[] getActionMethodArguments(){
    return (Object[])getRequest().getAttribute(ActionPage.class.getName() + "_args");
  }

  @SuppressWarnings("unchecked")
  private boolean checkRequest(Method method) {
    boolean classShowAlert = true;
    boolean classDefaultCheckEnabled = true;
    boolean methodShowAlert = true;
    boolean methodDefaultCheckEnabled = true;

    logger.debug("");
    logger.debug("*** " + method.getName() + " ************");

    DefaultCheckReqest defaultCheck = null;
    if (getClass().isAnnotationPresent(DefaultCheckReqest.class)) {
      defaultCheck = getClass().getAnnotation(DefaultCheckReqest.class);
      classDefaultCheckEnabled = defaultCheck.isEnabled();
      classShowAlert = defaultCheck.isShowAlert();
    }

    if (method.isAnnotationPresent(DefaultCheckReqest.class) ) {
      defaultCheck = method.getAnnotation(DefaultCheckReqest.class);
      methodDefaultCheckEnabled = defaultCheck.isEnabled();
      methodShowAlert = defaultCheck.isShowAlert();
    }
    //ディフォルトのチェックリクエストは実行しません。
    if (!classDefaultCheckEnabled || !methodDefaultCheckEnabled) return true;

    ArrayList<ICheckableRequest> errorComponents = new ArrayList<ICheckableRequest>();
    logger.debug("*defaut checkRequest : true");

    ArrayList<ICheckableRequest> errorTableList = new ArrayList<ICheckableRequest>();
    for (Field f : getClass().getFields()) {
      try {
        if (Modifier.isPublic(f.getModifiers()) && f.get(this) instanceof ICheckableRequest) {
          ICheckableRequest webComponent = (ICheckableRequest) f.get(this);
          if (!webComponent.checkRequest(getRequest())) {
            errorComponents.add(webComponent);
          }
        }
      } catch (Exception e) {
        throw new RuntimeException(e);
      }
    }
    if (errorComponents.size()==0) {
      return true;
    } else {
      onDefaultCheckRequestError(method.getName(), errorComponents, (classShowAlert && methodShowAlert) );
      return false;
    }
  }

  /**
   * アクションページの自動チェックリクエストでエラーが発生した際、呼ばれるメソッド.<br/>
   * ブラウザにエラーメッセージをAlertで表示させるレスポンスを出力するのがディフォルトの動作です。<br/>
   * アプリケーションプログラムは当メソッドをオーバーライドして独自の振る舞いを実装してください。
   * @param methodName リクエストに含まれていた実行対象のメソッド名
   * @param errorComponents エラーが発生したWebDataTable
   */
  protected void onDefaultCheckRequestError(String methodName, ArrayList<ICheckableRequest> errorComponents, boolean showAlert) {
    StringBuilder errors = new StringBuilder();
    for (ICheckableRequest webComponent : errorComponents) {
      if (errors.length()>0) errors.append("\\n");
      errors.append(webComponent.getErrorMessages());
    }
    try {
      if (isAjaxRequest(getRequest())) {
        if (showAlert) {
          writeAjaxExceptionResponse(errors.toString());
        } else {
          JSONObject responseObject = new JSONObject();
          responseObject.put("errors", errors);
          writeAjaxResponse(responseObject);
        }
      } else if (showAlert){
        showWithAlert(errors.toString());
      } else {
        show();
      }
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * executeメソッド実行時の例外発生時処理
   *
   * @param x 例外オブジェクト
   * @throws Exception 例外発生時
   */
  private void onExceptionAction(Exception x) throws Exception {
    if (isAjaxRequest()) {
      this.onAjaxExceptionAction(x);
    } else {
      this.onExecExceptionAction(x);
    }
  }

  /**
   * executeメソッド実行時の例外発生時処理（ls.exec使用時）
   *
   * @param x 例外オブジェクト
   * @throws Exception 例外発生時
   */
  protected void onExecExceptionAction(Exception x) throws Exception {
    throw x;
  }

  /**
   * executeメソッド実行時の例外発生時処理（ls.ajax使用時）
   *
   * @param x 例外オブジェクト
   * @throws Exception 例外発生時
   */
  protected void onAjaxExceptionAction(Exception x) throws Exception {
    x.printStackTrace(System.err);
    String message = x.getClass().getName() + ":" + x.getLocalizedMessage() ;
    Throwable t = x;
    while (t.getCause()!=null) {
      message += "\n" + t.getClass().getName() + ":" + t.getLocalizedMessage() ;
      t = t.getCause();
    }
    this.writeAjaxExceptionResponse(message);
  }

  /**
   * Ajaxリクエストに対するレスポンスを出力します。<br/><br/>
   *
   * このメソッドでは返却されたレスポンスは、lsjcontrol.htmlにてonExceptionファンクションを
   * 呼び出します。
   *
   * @param errorMessage Exceptionの内容
   * @throws Exception 例外発生時
   */
  public void writeAjaxExceptionResponse(String errorMessage) throws Exception {
    org.json.JSONObject responseObject = new org.json.JSONObject();
    responseObject.put("__exception", errorMessage);
    responseObject.put("newSequence", this.getSequenceValue(true));
    this.response.getWriter().print(responseObject.toString());
    this.response.getWriter().flush();
    this.response.getWriter().close();
  }

  /**
   * Ajaxリクエストに対するレスポンスを出力します。<br/><br/>
   *
   * このメソッドでは返却されたレスポンスは、lsjcontrol.htmlにてonSequenceErrorファンクションを
   * 呼び出します。
   *
   * @param errorMessage シーケンスエラーの内容
   * @throws Exception 例外発生時
   */
  public void writeAjaxSequenceErrorResponse(String errorMessage) throws Exception {
    org.json.JSONObject responseObject = new org.json.JSONObject();
    responseObject.put("__sequenceerror", errorMessage);
    responseObject.put("newSequence", this.getSequenceValue(false));
    this.response.getWriter().print(responseObject.toString());
    this.response.getWriter().flush();
    this.response.getWriter().close();
  }

  /**
   * アクションページ内のJDBCコネクションを開放します。
   */
  public void releaseJDBCConnection() throws Exception {
    try {
      if (connList.size()>0) {
        for (int i=0; i < connList.size(); i++) {
          try {
            if (!connList.get(i).isClosed()) {
              connList.get(i).close();
            }
          }
          catch (Exception x) {
            onReleaseConnectionExcepion(x);
          }
        }
        connList.clear();
      }
      if (conn!=null && !conn.isClosed()) conn.close();
      conn = null;
      request = null;
      response = null;
    }
    catch (Exception x) {
      onReleaseConnectionExcepion(x);
    }
  }

  /**
   * JDBCコネクション解放エラー時の処理を行います。<br/>
   *
   * @param x 例外 オブジェクト
   * @throws Exception 例外発生時
   */
  protected void onReleaseConnectionExcepion(Exception x) throws Exception {}

  /**
   * 指定したアクションメソッドが、存在しなかった場合に実行されるデフォルトの
   * アクションメソッドです。<br/>
   *
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @throws Exception 例外発生時
   * @see #init(HttpServletRequest, HttpServletResponse)
   * @see #doDefault()
   */
  public void doDefault(HttpServletRequest request, HttpServletResponse response) throws Exception {
    this.init(request, response);
    doDefault();
  }

  /**
   * 指定したアクションメソッドが、存在しなかった場合に実行されるデフォルトの
   * アクションメソッドです。<br/>
   *
   * @throws Exception 例外発生時
   * @see #onStartAction(String)
   * @see #show()
   */
  public void doDefault() throws Exception {
      this.onStartAction("doDefault");
      if (isAjaxRequest()) {
        org.json.JSONObject responseObject = new org.json.JSONObject();
        responseObject.put("newSequence", this.getSequenceValue(false));
        for (Field f : this.getClass().getFields()) {
          if (Modifier.isPublic(f.getModifiers()) ) {
            Object field = f.get(this);
            if (field instanceof ICheckableRequest) {
              responseObject.put(f.getName() + "_defaultInformation", ((ICheckableRequest)field).getDefaultInformation());
            }
          }
        }
        this.response.getWriter().print(responseObject.toString());
        this.response.getWriter().flush();
        this.response.getWriter().close();
      } else {
        this.show();
      }
  }

  /**
   * 指定したアクションページを生成し、そのdoDefaultメソッドを実行します。<br/>
   *
   * @param className クラス名
   */
  @DefaultCheckReqest(isEnabled=false)
  public void forward(String className) throws Exception  {
    ActionPage newPage = this.getController().createActionPage(className, this.request, this.response);
    newPage.doDefault(this.request, this.response);
  }

  /**
   * 指定したアクションページを使って、新しいサブセッションを開始します.<br/>
   * @param className アクションクラス名
   * @throws Exception
   */
  @DefaultCheckReqest(isEnabled=false)
  public void startNewSubSession(HashMap args)  throws Exception {
    String className = (String) args.get("className");
    String areaId = (String) args.get("areaId");
    Constructor<?> constructor =
      Class.forName(className).getConstructor(new Class[]{HttpServletRequest.class,HttpServletResponse.class});
    ActionPage newPage = (ActionPage)constructor.newInstance(new Object[]{request, response});
    ActionSubSession newSubSession = new ActionSubSession(newPage);
    request.getSession().setAttribute(newSubSession.getID(), newSubSession);
    if (areaId!=null) {
      newPage.setLoadscript("ls.parentWidget = parent.document.getElementById('" + areaId + "')._lsUIWidget;  \n");
    }
    newPage.doDefault();
  }


  private  List<String> tokenToList(String value, String delimiter) {
    // 分割した文字を格納する変数
    List<String> list = new ArrayList<String>();
    int i = 0;
    int j = value.indexOf(delimiter);
    for (int h = 0; j >= 0; h++) {
      list.add(value.substring(i, j));
      i = j + DELIMITER_ARGS.length();
      j = value.indexOf(delimiter, i);
    }
    if (ARGS_ISNULL.equals(value.substring(i))) {
      list.add(null);
    } else {
      list.add(value.substring(i));
    }
    return list;
  }

  /**
   * アクションメソッドの実行前に呼び出される処理を実装します。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @param methodName アクションメソッド名
   * @throws Exception 例外発生時
   */
  protected void onStartAction(String methodName) throws Exception {}

  /**
   * ページ出力前に呼び出される処理を実装します。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @throws Exception 例外発生時
   */
  protected void onStartShow()  throws Exception {}

  /**
   * ページ出力前に呼び出される処理を実装します。<br/><br/>
   * VelocityContextに値をセットすることができます。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @throws Exception 例外発生時
   */
  protected void onStartShow(VelocityContext velocityContext)  throws Exception {
    onStartShow();
  }

  /**
   * アクションページがバックボタンによる不正なリクエストを受け付けるかどうかを取得します。<br/><br/>
   *
   * Trueの場合、自分が最後に出力したページのHTTPリクエストしか受け付ません。<br/>
   * ブラウザのバックボタンや再読込によるHTTPリクエストはエラーとなり、最終ページが強制的に再表示されます。<br/>
   *
   * @return True：不正リクエストを受け付ける、False：受け付けない
   */
  public boolean isStrictSession() {
    return strictSession;
  }

  /**
   * アクションページがバックボタンによる不正なリクエストを受け付けるかどうかを設定します。<br/><br/>
   *
   * Trueの場合、自分が最後に出力したページのHTTPリクエストしか受け付ません。<br/>
   * ブラウザのバックボタンや再読込によるHTTPリクエストはエラーとなり、最終ページが強制的に再表示されます。<br/>
   *
   * @param strictSession True：不正リクエストを受け付ける、False：受け付けない
   */
  public void setStrictSession(boolean strictSession) {
    this.strictSession = strictSession;
  }

  /**
   * リクエスト送信順序判定用の識別子を取得します。<br/>
   *
   * @return リクエスト送信順序判定用の識別子
   */
  public String getSequenceValue() {
    return sequenceValue;
  }

  /**
   * リクエスト送信順序判定用の識別子を取得します。<br/><br/>
   *
   * インクリメントを行った場合、新しい識別子が発行・設定されます。<br/>
   *
   * @param increment True：インクリメントを行う、False：行わない
   * @return リクエスト送信順序判定用の識別子
   */
  public String getSequenceValue(boolean increment) {
    if (increment) {
      this.sequenceValue = Long.toHexString(System.currentTimeMillis() + this.hashCode());
      return sequenceValue;
    } else {
      return sequenceValue;
    }
  }

  /**
   * 出力するページのContentTypeを取得します。<br/>
   *
   * @return ContentType
   */
  public String getContentType() {
    return contentType;
  }

  /**
   * 出力するページのContentTypeを設定します。<br/>
   *
   * @param contentType ContentType
   */
  public void setContentType(String contentType) {
    this.contentType = contentType;
  }

  /**
   * 出力するページのFormタグの名称を取得します。<br/>
   *
   * @return Formタグの名称
   */
  public String getFormName() {
    return formName;
  }

  /**
   * 出力するページのFormタグの名称を設定します。<br/>
   *
   * @param formName Formタグの名称
   */
  public void setFormName(String formName) {
    this.formName = formName;
  }

  /**
   * セッションへサブセッションが登録された際、またはサブセッションに当アクションページが
   * 設定された際の処理を実装します。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @param e HttpSessionBindingEvent オブジェクト
   */
  public void valueBound(HttpSessionBindingEvent e) {}

  /**
   * セッションからサブセッションが削除された際、またはサブセッションから当アクションページが
   * 設定された際の処理を実装します。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @param e HttpSessionBindingEvent オブジェクト
   */
  public void valueUnbound(HttpSessionBindingEvent e) {}

  /**
   * Ajaxリクエストに対するレスポンスを出力します。<br/><br/>
   * @throws Exception
   */
  public void writeAjaxResponse() throws Exception {
    writeAjaxResponse("");
  }

  /**
   * Ajaxリクエストに対するレスポンスを出力します。<br/><br/>
   * @param responseObject レスポンスしたいJSONObject
   * @throws Exception
   */
  public void writeAjaxResponse(org.json.JSONObject responseObject) throws Exception {
    writeAjaxResponse("", responseObject);
  }

  /**
   * Ajaxリクエストに対するレスポンスを出力します。<br/><br/>
   *
   * Ajaxを使用する場合は、
   * <pre>
   *   actionPage.setUseAjax(true);
   * </pre>
   * としくてください。また、HTMLテンプレートにはjavaScritタグでprototype.jsを指定してください。<br/><br/>
   *
   * Ajaxレスポンスにはshowメソッドを使用しません。<br/>
   * <pre>
   *   this.writeAjaxResponse(String);
   * </pre>
   * でHTMLの断片を書き出します。<br/><br/>
   *
   * HTMLの断片をテンプレートで編集したい場合は、
   * <pre>
   *   getMergedString("template/sample/xxxxxxxxxxx.html")
   * </pre>
   * で予めActionPage オブジェクトの内容とマージします。<br/><br/>
   *
   * 書き出した内容は、JSONオブジェクトとしてクライアントJavaScriptのonComplete(response)に返されます。<br/>
   * response.contentがthis.writeAjaxResponse()で書き出した文字列です。<br/><br/>
   *
   * responseには上記の他、ActionPageのパブリックフィールド、ActionPage内のWebDataTableのプロパティなどが
   * セットされています。<br/>
   * ※詳しくはwriteAjaxResponse()のソースを見てください。<br/><br/>
   *
   * もっと複雑な内容をresponseに入れたい場合は、
   * <pre>
   *   writeAjaxResponse(String content, org.json.JSONObject responseObject)
   * </pre>
   * を使用してくだい。予めJSONObjectを作成して、必要なオブジェクトをputしてからwriteAjaxResponseを呼び出すことで
   * どのような内容も、クライアントJavaScriptに返却できるようになります。<br/><br/>
   *
   * @param content 出力する文字列
   * @throws Exception 例外発生時
   * @see #writeAjaxResponse(String, JSONObject)
   */
  public void writeAjaxResponse(String content) throws Exception {
    org.json.JSONObject responseObject = new org.json.JSONObject();
    writeAjaxResponse(content, responseObject);
  }

  /**
   * Ajaxリクエストに対するレスポンスを出力します。
   *
   * @param content 出力する文字列
   * @param responseObject レスポンスオブジェクト
   * @throws Exception 例外発生時
   */
  public void writeAjaxResponse(String content, org.json.JSONObject responseObject) throws Exception {

    Field[] fields =  this.getClass().getFields();
    for (int i=0; i<fields.length; i++) {
      if (Modifier.isPublic(fields[i].getModifiers())) {
        try {
          Object ctxobj = fields[i].get(this);
          if (ctxobj instanceof WebDataTable) {
            WebDataTable<?> table = (WebDataTable<?>)ctxobj;
            org.json.JSONObject tableAttribute = new org.json.JSONObject();
            tableAttribute.put("active", table.isActive());
            tableAttribute.put("hasNextPage", table.hasNextPage());
            tableAttribute.put("hasPreviousPage", table.hasPreviousPage());
            tableAttribute.put("rowCount", table.getRowCount());
            tableAttribute.put("pageStartRowNum", table.getPageStartRowNum());
            tableAttribute.put("pageEndRowNum", table.getPageEndRowNum());
            tableAttribute.put("pager", table.getPagerTag());
            boolean hasError = !(table.getErrors().isEmpty());
            tableAttribute.put("hasError", hasError);
            tableAttribute.put("errors", table.getErrorMessages());
            responseObject.put(fields[i].getName(), tableAttribute);
          }
          else {
            if (ctxobj instanceof String) {
              // @EscapeStringが付与されている場合
              if (fields[i].isAnnotationPresent(EscapeString.class)) {
                EscapeString escapeString = fields[i].getAnnotation(EscapeString.class);
                boolean isEscapeString = escapeString.isEscapeString();
                // isEscapeString=trueの場合、値をHTMLエスケープする
                if (isEscapeString) {
                  String strCtxobj = (String) ctxobj;
                  responseObject.put(fields[i].getName(), HTML.escape(strCtxobj));
                }
                // 値をHTMLエスケープしない
                else {
                  responseObject.put(fields[i].getName(), ctxobj);
                }
              }
              // @EscapeStringが付与されていない場合
              else {
                // 値をHTMLエスケープする
                String strCtxobj = (String) ctxobj;
                responseObject.put(fields[i].getName(), HTML.escape(strCtxobj));
              }

            }
            else {
              // 値をHTMLエスケープしない
              responseObject.put(fields[i].getName(), ctxobj);
            }
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }
    responseObject.put("newSequence", this.getSequenceValue(true));
    responseObject.put(PN_SUBSESSION_ID, this.subSession.getID());
    responseObject.put(PN_ACTIONPAGE_ID, this.actionPageId);
    if (this.parentPage != null) {
      responseObject.put(PN_PARENT_ACTIONPAGE_ID, this.parentPage.actionPageId);
    }
    else {
      responseObject.put(PN_PARENT_ACTIONPAGE_ID, this.actionPageId);
    }
    responseObject.put("content", content);
    this.response.getWriter().print(responseObject.toString());
    this.response.getWriter().flush();
    this.response.getWriter().close();
  }


  /**
   * アクションページがAjaxを使用するかどうかを取得します。<br/>
   *
   * @return True：Ajaxを使用する、False：使用しない
   */
  public boolean isUseAjax() {
    return useAjax;
  }

  /**
   * アクションページがAjaxを使用するかどうかを設定します。<br/>
   *
   * @param useAjax True：Ajaxを使用する、False：使用しない
   */
  public void setUseAjax(boolean useAjax) {
    this.useAjax = useAjax;
  }

  /**
   * Table オブジェクトのデータが変更された際の処理を実装します。<br><br>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @see jp.littlesoft.data.event.TableListener#dataChanged(jp.littlesoft.data.event.TableEvent, TableEvent.Operation)
   */
  public <E extends Serializable> void dataChanged(TableEvent<?> e, TableEvent.Operation operation) {}

  /**
   * Table オブジェクトへの行追加、行変更の削除前の処理を実装します。<br><br>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @see jp.littlesoft.data.event.TableListener#rowDeleting(jp.littlesoft.data.event.TableEvent)
   */
  public <E extends Serializable> void rowDeleting(TableEvent<?> e) throws Exception {}

  /**
   * Table オブジェクトへのフィルタ適用時の処理を実装します。<br><br>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @see jp.littlesoft.data.event.TableListener#rowFiltering(jp.littlesoft.data.event.TableEvent)
   */
  public <E extends Serializable> boolean rowFiltering(TableEvent<?> e) { return true; }

  /**
   * Table オブジェクトへの行追加、行変更の確定前の処理を実装します。<br><br>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @see jp.littlesoft.data.event.TableListener#rowValidating(jp.littlesoft.data.event.TableEvent)
   */
  public <E extends Serializable> void rowValidating(TableEvent<?> e) throws Exception {}

  /**
   * アクションページの識別子を取得します。<br/>
   *
   * @return アクションページの識別子
   */
  public String getActionPageId() {
    return actionPageId;
  }

  /**
   * アクションページの識別子を設定します。<br/>
   *
   * @param actionPageId アクションページの識別子
   */
//  public void setActionPageId(String actionPageId) {
//    this.actionPageId = actionPageId;
//  }

  /**
   * 親のアクションページを取得します。<br/>
   *
   * @return 親のアクションページ
   */
  public ActionPage getParentPage() {
    return parentPage;
  }

  /**
   * 親のアクションページを設定します。<br/>
   *
   * @param parentPage 親のアクションページ
   */
  public void setParentPage(ActionPage parentPage) {
    this.parentPage = parentPage;
    this.parentPage.addSubPage(this);
  }

  /**
   * 親のアクションページが持つ、子アクションページのリストから自分自身の
   * アクションページを削除します。<br/>
   */
  @DefaultCheckReqest(isEnabled=false)
  public void closeSubPage() {
    parentPage.subActionPage.remove(actionPageId);
  }

  /**
   * 子アクションページリストに、引数 actionPage で指定されたアクションページを設定します。<br/>
   *
   * @param actionPage アクションページ
   */
  private void addSubPage(ActionPage actionPage) {
    subActionPage.put(actionPage.actionPageId, actionPage);
  }

  /**
   * JDBCコネクションを取得します。<br/>
   *
   * @return JDBCコネクション
   * @see #getController()
   * @see jp.littlesoft.web.ActionController#getConnection()
   */
  public Connection getConnection() throws Exception {
    if (conn==null) {
      Connection provideConn = getController().getConnection(request);
      connList.add(provideConn);
      return provideConn;
    }
    return conn;
  }

//  /**
//   * JDBCコネクションを取得します。<br/>
//   *
//   * @return JDBCコネクション
//   * @see #getController()
//   * @see jp.littlesoft.web.ActionController#getConnection()
//   */
//  public Connection getConnection(HttpServletRequest request) throws Exception {
//    if (conn==null) {
//      Connection provideConn = getController().getConnection(request);
//      connList.add(provideConn);
//      return provideConn;
//    }
//    return conn;
//  }
  /**
   * リクエストを取得します。<br/>
   *
   * @return リクエスト
   */
  public HttpServletRequest getRequest() {
    return request;
  }

  /**
   * レスポンスを取得します。<br/>
   *
   * @return レスポンス
   */
  public HttpServletResponse getResponse() {
    return response;
  }

  /**
   * サブセッションを取得します。<br/>
   *
   * @return サブセッション
   */
  public ActionSubSession getSubSession() {
    return subSession;
  }

  /**
   * セッションからサブセッションを削除します。<br/><br/>
   *
   * 画面の閉じるボタンが押されたら、サブセッションを終了します。<br/>
   *
   * @throws Exception 例外発生時
   * @see #removeSessionAttribute(String)
   */
  public void closeSession() throws Exception {
    removeSessionAttribute(getSubSession().getID());
  }

  /**
   * JDBCコネクションを使用するメソッドに付与するアノテーションです。<br/>
   *
   * @author LittleSoft Corporation
   */
  @Retention(RetentionPolicy.RUNTIME) //有効範囲
  public @interface UseJDBC {
    /**
     * Commit/Rollbackを自動化する場合は「true(デフォルト)」、
     * 自メソッド内で行う場合は「faise」を設定します。<br/>
     * ※JDBCのAutoCommit機能とは無関係です。
     */
    public boolean autoCommit() default true;
  }

  /**
   * Ajaxリクエストかどうかを判定します。
   *
   * @return Ajaxリクエスト：true、Ajaxリクエストではない：false
   */
  public boolean isAjaxRequest() {
    return isAjaxRequest(this.getRequest()) ;
  }

  /**
   * Ajaxリクエストかどうかを判定します。
   *
   * @param request リクエスト
   * @return Ajaxリクエスト：true、Ajaxリクエストではない：false
   */
  public boolean isAjaxRequest(HttpServletRequest request) {
    String xRequestWith = request.getHeader("x-requested-with");
    if (xRequestWith != null && xRequestWith.equals("XMLHttpRequest")) {
      return true;
    }
    return false;
  }

  /**
   * 引数で指定されたダイアログアクションページクラスを開きます。
   *
   * @param className ダイアログアクションページクラス名
   */
  public void showDialog(String className) throws Exception {
    HashMap<String,String> args = new HashMap<String,String>();
    args.put("className", className);
    showDialog(args);
  }

  /**
   * 引数で指定されたダイアログアクションページクラスを開きます。
   *
   * @param className ダイアログアクションページクラス名
   * @param args 引数、テンプレートファイルではJSONオブジェクトで指定してください 
   */
  @SuppressWarnings("unchecked")
  public void showDialog(HashMap args) throws Exception {
    Constructor constructor = 
      Class.forName((String)args.get("className"))
        .getConstructor(new Class[]{HttpServletRequest.class,HttpServletResponse.class});
    this.getRequest().setAttribute(ActionPage.RA_PARENT_PAGE, this);
    this.getRequest().setAttribute(ActionPage.RA_PARENT_ARGS, args);
    ActionPage dialogClass = (ActionPage)constructor.newInstance(new Object[]{this.getRequest(),this.getResponse()});
    dialogClass.setParentPage(this);
    dialogClass.show();
  }


  /**
   *
   * HTMLエスケープを行うかを付与するアノテーションです。<br/>
   *
   * @author LittleSoft Corporation
   */
  @Retention(RetentionPolicy.RUNTIME) //有効範囲
  public @interface EscapeString {
    /**
     * HTMLエスケープする場合は「true(デフォルト)」、
     * HTMLエスケープしない場合は「false」を設定します。
     */
    public boolean isEscapeString() default true;
  }

  /**
   * アクションメソッドを実行する前に自動のチェックリクエストを行うか？を示すアノテーションです.<br/>
   * @author littlesoft
   */
  @Retention(RetentionPolicy.RUNTIME) //有効範囲
  public @interface DefaultCheckReqest {
    /**
     * 自動チェックリクエストを行う場合は「true(デフォルト)」、行わない場合は「false」を設定します。<br/>
     * また当アノテーション自体を省略した場合はディフォルトで自動チェックリクエストが行われます。
     */
    public boolean isEnabled() default true;
    /**
     * 自動チェックリクエストによるエラーレスポンスで、alertによりエラーメッセージを表示するか？を設定します.<br/>
     * ※ この設定はAJAXリクエストには作用しません.
     */
    public boolean isShowAlert() default true;
  }

  /**
   * ページがロードされる際、実行されるJavaScriptコードを取得する.<br/>
   * @return ページがロードされる際、実行されるのJavaScriptコード
   */
  public String getLoadscript() {
    return loadscript;
  }

  /**
   * ページがロードされる際、実行されるJavaScriptコードを設定する.<br/>
   * @param loadscript
   */
  public void setLoadscript(String loadscript) {
    this.loadscript = loadscript;
  }

  /**
   * リクエストからJSESSIONIDの値を取得する.<br/>
   * @return JSESSION_ID
   */
  public String getJsessionID() {
    Cookie[] cookies = getRequest().getCookies();
    if (cookies==null) return null;
    for (Cookie c : cookies) {
      if (c.getName()!=null && "jsessionid".equals(c.getName().toLowerCase())) {
        return c.getValue();
      }
    }
    return null;
  }

  /**
   * アップロードファイルマネージャーを取得する.<br/>
   * @return
   */
  public UploadFileManager getUploadFileManager() {
    return (UploadFileManager)getSessionAttribute(UploadFileManager.class.getName());
  }

  /**
   * クライアント(WebBrowser)側でJavaアプリケーションを起動する.<br/>
   * JavaWebStartのjnlpファイルを生成し、レスポンスへ出力します.<br/>
   * ActionPageテンプレートに以下の様なスクリプトを記述し当メソッドを呼び出します.<br/><br/>
   * ls.startJavaApp({productId:'1',fileName:'test.zip'...})<br/><br/>
   * ※任意のパラメータをJSONオブジェクトで設定します.この内容はHashMapに変換されActionPageのメソッド引数になります.<br/>
   * @param args クライアントから送られたパラメータ
   */
  public void startJavaApp(HashMap<String, ?> args) {
    //
  }

  /**
   * クライアント(WebBrowser)側で起動されたJavaアプリケーションからのコールバック.<br/>
   */
  final public void onJavaAppCallBack() {
    Attributes attributes =
      (Attributes)request.getAttribute(ActionController.RA_JAVA_APPLICATION_CALLBACK_PARAM);
    
    if (attributes==null) {
      String jsonParams = getRequest().getParameter(ActionPage.PN_APPLICATION_PARAMS);
      IApplicationParams params = null;
      if (jsonParams!=null) {
        try {
          params = (IApplicationParams) LSJson.deserialize(jsonParams);
        } catch (JSONException e) {
          e.printStackTrace();
        }
        attributes = params.getAttributes();
        request.setAttribute(ActionController.RA_JAVA_APPLICATION_CALLBACK_PARAM, attributes);
      } 
    }
    onJavaAppCallBack(attributes);
  }

  /**
   * クライアント(WebBrowser)側で起動されたJavaアプリケーションからのコールバック.<br/>
   */
  public void onJavaAppCallBack(Attributes attributes) { }

  protected VelocityEngine getVelocityEngine() {
    return (VelocityEngine)getRequest().getAttribute(ActionController.RA_VELOCITY_ENGINE);
  }
  
}
