/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.Serializable;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionBindingListener;

/**
 * サブセッション管理クラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class ActionSubSession implements HttpSessionBindingListener, Serializable {
  
  //private Logger logger = Logger.getLogger(this.getClass().getName());
  private int maxInactiveInterval = 0; //ゼロにしておくとActionControllerがセッションに指定されているのと同値をセットします。
  private long lastAccessedTime;
  private String id = null;
  private ActionPage actionPage = null;
  
  private ActionSubSession() {
    this.id = Long.toHexString(System.currentTimeMillis() + this.hashCode());
    this.lastAccessedTime = System.currentTimeMillis();
  }
  
  /**
   * 引数 actionPageで指定された ActionSubSession オブジェクトを生成します。<br/><br/>
   * 
   * サブセッションがアクセスするアクションページを設定するメソッドを呼び出します。<br/>
   * 
   * @param actionPage アクションページ
   * @see #setActionPage(jp.littlesoft.web.ActionPage)
   */
  public ActionSubSession(ActionPage actionPage) {
    this();
    this.setActionPage(actionPage);
  }
  
  /**
   * アクションコントローラから処理要求がディスパッチされた際の処理を実装します。<br/><br/>
   * 
   * アクションページの初期処理、リクエスト送信順序のチェックを行い、<br/>
   * アクションページへ処理要求をディスパッチします。<br/><br/>
   * 
   * リクエスト送信順序が不正だった場合、アラート付きでページを再表示します。<br/>
   * 
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @param response Servletがクライアントに返すレスポンス内容を含む HttpServletResponse オブジェクト
   * @see jp.littlesoft.web.ActionPage#isStrictSession()
   * @see jp.littlesoft.web.ActionPage#getSequenceValue()
   * @see jp.littlesoft.web.ActionPage#showWithAlert(String)
   * @see jp.littlesoft.web.ActionController#onSequenceError(HttpServletRequest, HttpServletResponse)
   * @see jp.littlesoft.web.ActionPage#execute(ActionSubSession, HttpServletRequest, HttpServletResponse)
   */
  public synchronized void execute(HttpServletRequest request, HttpServletResponse response) throws Exception {
    this.lastAccessedTime = System.currentTimeMillis();
    request.setAttribute(ActionController.RA_CURRENT_SUBSESSION, this);
    try {
      actionPage.init(this, request, response);
      //送信の順序チェック
      if (actionPage.isStrictSession() && actionPage.getSequenceValue()!= null) {
        //アクションページID取得
        String actionPageId = request.getParameter(ActionPage.PN_ACTIONPAGE_ID);
        String conpareActionPageId = actionPage.actionPageId;
        //SEQ取得
        String sequenceValue = request.getParameter(ActionPage.PN_SEQ);
        String conpareSequenceValue = actionPage.getSequenceValue();
        ActionPage sub = null;
        //子ページからのリクエストの場合は、比較するSEQを子ページのSEQとする。
        if (actionPageId!=null && !conpareActionPageId.equals(actionPageId)) {
          sub = actionPage.subActionPage.get(actionPageId);
          if (sub!=null) {
            conpareActionPageId = sub.actionPageId;
            conpareSequenceValue = sub.getSequenceValue();
          }
        }
        //アクションページID比較
        if (!conpareActionPageId.equals(actionPageId)) {
          //System.out.println("[警告] 処理要求を割り当てるアクションページが存在しません。");
          //System.out.println("PN_ACTIONPAGE_ID: " + actionPageId);
          //System.out.println("PN_METHOD: " + request.getParameter(ActionPage.PN_METHOD));
          //return;
        }
        //SEQ比較
        if (!conpareSequenceValue.equals(sequenceValue)) {
          ActionController actionController = (ActionController)request.getAttribute(ActionController.RA_CURRENT_CONTROLLER);
          actionController.onSequenceError(request, response);
          //出力メッセージ
          String alertMessage = "ブラウザの「戻る」や「更新」を使用せず、現在表示されている画面から続行してください。";
          //レスポンスを返却するアクションページをセット
          ActionPage execPage = actionPage;
          //子ページからのリクエストの場合は、返却するアクションページは子ページとなる
          if (sub!=null) {
            execPage = sub;
            execPage.init(this, request, response);
          }
          //リクエストがAjaxの場合は
          if (execPage.isAjaxRequest(request)) {
            execPage.writeAjaxSequenceErrorResponse(alertMessage);
          }
          //それ以外の場合
          else {
            execPage.showWithAlert(alertMessage, false);
          }
          return;
        }
      }
      actionPage.execute(this, request, response);
    } finally {
      request.setAttribute(ActionController.RA_CURRENT_SUBSESSION, null);
    }
  }

  /**
   * サブセッションの識別子を取得します。
   * 
   * @return サブセッションの識別子
   */
  public String getID() {
    return id;
  }

  /**
   * サブセッションがアクセスするアクションページを取得します。<br/>
   * 
   * @return アクセスするアクションページ
   */
  public ActionPage getActionPage() {
    return actionPage;
  }

  /**
   * サブセッションがアクセスするアクションページを設定します。<br/><br/>
   * 
   * 指定されたアクションページがすでに設定されたいる場合、何も行いません。<br/>
   * 
   * @param actionPage アクセスするアクションページ
   * @see jp.littlesoft.web.ActionPage#valueBound(HttpSessionBindingEvent)
   * @see jp.littlesoft.web.ActionPage#valueUnbound(HttpSessionBindingEvent)
   */
  public void setActionPage(ActionPage actionPage) {
    if (this.actionPage == actionPage) return;
    if (this.actionPage!=null && this.actionPage.request!=null &&
        this.actionPage.request.getSession()!=null) {
      this.actionPage.valueUnbound(
          new HttpSessionBindingEvent(this.actionPage.request.getSession(), 
              this.actionPage.getClass().getName()));
    }
    this.actionPage = actionPage;
    if (this.actionPage!=null) {
      this.actionPage.subSession = this;
      if (this.actionPage.request!=null &&
          this.actionPage.request.getSession()!=null) {
        this.actionPage.valueBound(
            new HttpSessionBindingEvent(this.actionPage.request.getSession(), 
                this.actionPage.getClass().getName()));
      }
    }
  }

  /**
   * セッションにサブセッションが登録された際の処理を実装します。<br/><br/>
   * 
   * アクションページのセッション登録メソッドにディスパッチします。<br/>
   * 
   * @see javax.servlet.http.HttpSessionBindingListener#valueBound(javax.servlet.http.HttpSessionBindingEvent)
   * @see jp.littlesoft.web.ActionPage#valueBound(HttpSessionBindingEvent)
   */
  public void valueBound(HttpSessionBindingEvent e) {
    //System.out.println("*** valueBound:" + this.getClass().getName() + ":" + this.getID());
    if (this.actionPage!=null) {
      this.actionPage.valueBound(e);
    }
  }

  /**
   * セッションがらサブセッションが削除された際の処理を実装します。<br/><br/>
   * 
   * アクションページのセッション削除メソッドにディスパッチします。<br/>
   * 
   * @see javax.servlet.http.HttpSessionBindingListener#valueUnbound(javax.servlet.http.HttpSessionBindingEvent)
   * @see jp.littlesoft.web.ActionPage#valueUnbound(HttpSessionBindingEvent)
   */
  public void valueUnbound(HttpSessionBindingEvent e) {
    //System.out.println("*** valueUnbound:" + this.getClass().getName() + ":" + this.getID());
    if (this.actionPage!=null) {
      this.actionPage.valueUnbound(e);
    }
  }
  
  /**
   * クライアントがサブセッションに最後にアクセスしてからこのセッションを保ち続ける最大の秒数を取得します。<br/>
   * 
   * @return セッションを保ち続ける最大の秒数
   */  
  public int getMaxInactiveInterval() {
    return maxInactiveInterval;
  }

  /**
   * クライアントがサブセッションに最後にアクセスしてからこのセッションを保ち続ける最大の秒数を設定します。<br/>
   * 
   * @param maxInactiveInterval セッションを保ち続ける最大の秒数
   */
  public void setMaxInactiveInterval(int maxInactiveInterval) {
    this.maxInactiveInterval = maxInactiveInterval;
  }

  /**
   * クライアントがサブセッションに最後にアクセスした時間を取得します。<br/>
   * 
   * @return 最後にアクセスした時間
   */  
  public long getLastAccessedTime() {
    return lastAccessedTime;
  }
  
  /**
   * クライアントがサブセッションに最後にアクセスした時間を設定します。<br/>
   * 
   * @param lastAccessedTime 最後にアクセスした時間
   */
  public void setLastAccessedTime(long lastAccessedTime) {
    this.lastAccessedTime = lastAccessedTime;
  }

}
