/* Copyright (c) 2006-2010 LittleSoft Corporation
 *
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.Serializable;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;

import javax.servlet.http.HttpServletRequest;

import org.apache.velocity.app.event.InvalidReferenceEventHandler;
import org.apache.velocity.app.event.MethodExceptionEventHandler;
import org.apache.velocity.app.event.ReferenceInsertionEventHandler;
import org.apache.velocity.context.Context;
import org.apache.velocity.util.introspection.Info;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.util.StringUtil;
import jp.littlesoft.util.DataUtil.XWareki;

/**
 * Webデータテーブルクラス。<br/><br/>
 *
 * このクラスはjp.littlesoft.data.Tableを拡張して、HTMLテンプレートからのアクセスやHTTPリクエスト取り込みなど、
 * Webアプリケーション作成に必要となる機能を追加したものです。<br/>
 *
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class WebDataTable<E extends Serializable>  extends Table<E> implements ICheckableRequest {

  private String id;
  private int pageStartRow = -1;
  private int pageEndRow = -1;
  private int showRows = -1;
  private ArrayList<HashMap<String, Object>> errorMessage = new ArrayList<HashMap<String, Object>>();
  private HashMap<String, String> errorValues = new HashMap<String, String>();
  private boolean autoTrimRequest = true;
  static int COL_DELETION_RESERVATION = -2;
  static int ROW_PARAMS = -1;
  private final static DecimalFormat decimalFormat = new DecimalFormat();
  private final static XWareki wareki = new XWareki();
  private final static SimpleDateFormat simpleDateFormat = new SimpleDateFormat();

  private int currentPage = -1;
  private boolean usePageDivReserve = false;
  private boolean containRowNumInErrors = false;
  // 2009/08/10 WebDataTableのパラメータ名を固定にする対応
  private boolean useStaticParameterName = false;

  /**
   * WebDataTable オブジェクトを生成します。<br/>
   */
  public WebDataTable(E...e) {
    super(e);
    this.id = "t" + HTML.getRandomValue(this) + "z";
  }

  /**
   * 引数 colIndex で指定された列値をエスケープされた文字列として取得します。（現時行が対象となります）<br/>
   *
   * @param colIndex 列インデックス
   * @return エスケープされたデータ文字列
   * @throws DataException 例外発生時
   */
  public String getEscapeString(int colIndex) throws DataException {
    return HTML.escape(this.getRow().getValueAsString(colIndex));
  }

  /**
   * 引数 columnName で指定された列値をエスケープされた文字列として取得します。（現時行が対象となります）<br/>
   *
   * @param columnName 列名
   * @return エスケープされたデータ文字列
   * @throws DataException 例外発生時
   */
  public String getEscapeString(String columnName) throws DataException {
    return HTML.escape(this.getRow().getValueAsString(columnName));
  }

  /**
   * 引数 row、col で指定された列値をエスケープされた文字列として取得します。<br/>
   *
   * @param row 行インデックス
   * @param col 列インデックス
   * @return エスケープされたデータ文字列
   * @throws DataException 例外発生時
   */
  public String getEscapeString(int row, int col) throws DataException {
    return HTML.escape(this.getRow(row).getValueAsString(col));
  }

  @Override
  protected synchronized void setActive(boolean b) throws DataException {
    super.setActive(b);
    errorMessage.clear();
    errorValues.clear();
    if (b) {
      pageFirst();
    }
  }

  /**
   * 当 WebDataTable オブジェクトの内容を PageContext オブジェクトに変換します。<br/>
   *
   * @return PageContext オブジェクト
   */
  public PageContext toPageContext() {
    PageContext ctx = new PageContext(this);
    if (this.isActive()) {
    try {
        for (int i=pageStartRow; i<=pageEndRow; i++) {
          ctx.rows.add(toPageRow(ctx, i));
        }
      } catch (Exception x) {x.printStackTrace();}
      for (int i=0; i<this.errorMessage.size(); i++) {

        String rownum = this.errorMessage.get(i).get("rownum").toString();
        String message = this.errorMessage.get(i).get("message").toString();

        if (containRowNumInErrors && !rownum.equals("-1")) {
          ctx.errors.add(HTML.escape(rownum + "行目:" + message));
        }
        else {
          ctx.errors.add(HTML.escape(message));
        }
      }
    }
    for (int i=0; i<this.getParams().getColumnCount(); i++) {
      try {
        String name = this.getParams().getColumn(i).getName();
        ctx.params.put(name, this.getParams().getValueAsString(i));
        ctx.params.put("@" + name, toRequestParameterName(-1, i));
      } catch (Exception x) {x.printStackTrace();}
    }
    return ctx;
  }

  /**
   * 引数 row で指定された行オブジェクトを PageRow オブジェクトに変換します。<br/>
   *
   * @param row インデックス
   * @return PageRow オブジェクト
   * @throws DataException 例外発生時
   */
  public PageRow<String, String> toPageRow(PageContext ctx, int row) throws DataException {
    PageRow<String, String> result = new PageRow<String, String>(ctx, row, this.getRow(row).getStatus(), false);
    for (int x=0; x<this.getColumnCount(); x++) {
      String name = this.getColumn(x).getName();
      String value = this.errorValues.get(row + ":" + x);

      if (value==null) {
        value = this.getRow(row).getValueAsString(x);
      }
      result.put(name, value);
      result.put("@" + name, toRequestParameterName(row, x));
    }
    result.setIsNew((this.getRow(row).getStatus()==Row.Status.CREATING ||
                     this.getRow(row).getStatus()==Row.Status.CREATED));
    result.deletionDeserved = this.getRow(row).isDeletionReserved();
    return result;
  }

  /**
   * 引数 colName で指定された項目のリクエストパラメータ名を取得します。（現時行が対象となります）<br/>
   *
   * @param colName 列名
   * @return リクエストパラメータ名
   * @throws DataException 例外発生時
   */
  public String toRequestParameterName(String colName) throws DataException {
    if (!isActive()) {
      throw new DataException("Table is not active.");
    }
    return toRequestParameterName(this.getRowIndex(), this.indexOfColumn(colName));
  }

  /**
   * 引数 row、colName で指定された項目のリクエストパラメータ名を取得します。<br/>
   *
   * @param row 行インデックス
   * @param colName 列名
   * @return リクエストパラメータ名
   * @throws DataException 例外発生時
   */
  public String toRequestParameterName(int row, String colName) throws DataException  {
    if (row!=-1 && !isActive()) {
      throw new DataException("Table is not active.");
    }
    return toRequestParameterName(row, this.indexOfColumn(colName));
  }

  /**
   * 引数 row、col で指定された項目のリクエストパラメータ名を取得します。<br/>
   *
   * @param row 行インデックス
   * @param col 列インデックス
   * @return リクエストパラメータ名
   */
  public String toRequestParameterName(int row, int col) {
    // 2009/08/10 WebDataTableのパラメータ名を固定にする対応
    /*
    if (row==-1) {
      return this.id + "_p_" + col;
    } else {
      return this.id + "_" + row + "_" + col;
    }
    */
    if (row==-1) {
        if (!useStaticParameterName)
          return this.id + "_p_" + col;
        else
          return this.id + "_p_" + this.getParams().getColumn(col).getName();
    }
    else {
        if (!useStaticParameterName)
          return this.id + "_" + row + "_" + col;
        else {
          if (col == -2)
            // 削除予約 COL_DELETION_RESERVATION
            return this.id + "_" + row + "_" + col;
          else
            return this.id + "_" + row + "_" + this.getColumn(col).getName();
        }
    }
  }

  /**
   * HTMLテンプレートで取得できる行の範囲を先頭ページに位置づけます。<br/>
   *
   * @see PageContext#getrows()
   * @see WebDataTable#setShowRows(int)
   */
  public void pageFirst() {
    pageStartRow = 0;
    adjustEndRows();
    currentPage = 1;
  }

  /**
   * HTMLテンプレートで取得できる行の範囲を前のページに位置づけます。<br/>
   *
   * @see PageContext#getrows()
   * @see WebDataTable#setShowRows(int)
   */
  public void pagePrevious() {
    if (!this.isActive() || showRows==-1) return;
    if (pageStartRow==0) return;
    if ((pageStartRow - showRows) < 0)
      pageStartRow = 0;
    else
      pageStartRow = pageStartRow - showRows;
    adjustEndRows();
    currentPage = currentPage - 1;
  }

  /**
   * HTMLテンプレートで取得できる行の範囲を次のページに位置づけます。<br/>
   *
   * @see PageContext#getrows()
   * @see WebDataTable#setShowRows(int)
   */
  public void pageNext() {
    if (!this.isActive() || showRows==-1) return;
    if ( pageEndRow == (this.getRowCount()-1) )  return;
    pageStartRow = (pageEndRow + 1);
    adjustEndRows();
    currentPage = currentPage + 1;
  }

  /**
   * HTMLテンプレートで取得できる行の範囲を最終のページに位置づけます。<br/>
   *
   * @see PageContext#getrows()
   * @see WebDataTable#setShowRows(int)
   */
  public void pageLast() {
    if (!this.isActive() || showRows==-1) return;
    pageEndRow = this.getRowCount() - 1;

    //pageStartRow = pageEndRow - showRows + 1;
    pageStartRow = (getPageCount() - 1) * showRows;

    if (pageStartRow  < 0) pageStartRow = 0;
    currentPage = getPageCount();
  }

  private void adjustEndRows() {
    if (showRows==-1) {
      pageEndRow = this.getRowCount() - 1;
    } else {
      if  ( (pageStartRow + showRows) >= this.getRowCount() )
        pageEndRow = this.getRowCount() - 1;
      else
        pageEndRow = pageStartRow + showRows - 1;
    }
  }

  /**
   * HTMLテンプレートで取得できる行の範囲を指定ページに位置づけます。<br/>
   *
   * @see PageContext#getrows()
   * @see WebDataTable#setShowRows(int)
   */
  public void pageMoveTo(int page) {
    int num = page - this.getCurrentPage();
    int cnt = num;
    if (cnt < 0) cnt = cnt * -1;
    for (int i = 0; i < cnt; i++) {
      if (num > 0) {
        pageNext();
      } else {
        pagePrevious();
      }
    }
    currentPage = page;
  }

  /**
   * 次のページが存在するかどうか調べます。<br/>
   *
   * @return True：次のページが存在する、False：存在しない
   */
  public boolean hasNextPage() {
    if (!this.isActive() || showRows==-1) return false;
    return (pageEndRow < (this.getRowCount() - 1) );
  }

  /**
   * 前のページが存在するかどうか調べます。<br/>
   *
   * @return True：前のページが存在する、False：存在しない
   */
  public boolean hasPreviousPage() {
    if (!this.isActive() || showRows==-1 ) return false;
    return (pageStartRow > 0);
  }

  /**
   * HTMLテンプレートで取得できる先頭行のインデックスを取得します。<br/>
   *
   * @return 行インデックス
   */
  public int getPageStartRowIndex() {
    return pageStartRow;
  }

  /**
   * HTMLテンプレートで取得できる末尾行のインデックスを取得します。<br/>
   *
   * @return 行インデックス
   */
  public int getPageEndRowIndex() {
    return pageEndRow;
  }

  /**
   * HTMLテンプレートで取得できる先頭行のインデックスを取得します。<br/>
   *
   * @return 行インデックス
   */
  public int getPageStartRowNum() {
    if (this.getRowCount() == 0) return 0;
    int num = 0;
    if (!isPageDivReserve()) {
      num = pageStartRow + 1;
    } else {
      num = pageStartRow + 1 + showRows * (currentPage - 1);
    }
    return num;
  }

  /**
   * HTMLテンプレートで取得できる末尾行のインデックスを取得します。<br/>
   *
   * @return 行インデックス
   */
  public int getPageEndRowNum() {
    if (this.getRowCount() == 0) return 0;
    int num = 0;
    if (!isPageDivReserve()) {
      num = pageEndRow + 1;
    } else {
      num = pageEndRow + 1 + showRows * (currentPage - 1);
    }
    return num;
  }

  /**
   * HTMLテンプレートで取得できる行数を取得します。<br/><br/>
   *
   * ディフォルトでは -1 が設定されています。これは全ての行が取得できることを意味します。<br/>
   *
   * @return 行数
   */
  public int getShowRows() {
    return showRows;
  }

  /**
   * HTMLテンプレートで取得できる行数を設定します。<br/><br/>
   *
   * ディフォルトでは -1 が設定されています。これは全ての行が取得できることを意味します。<br/>
   *
   * @param showRows 行数
   */
  public void setShowRows(int showRows) {
    this.showRows = showRows;
    if (currentPage == -1) currentPage = 1;
    if (showRows == -1) currentPage = -1;
  }

  /**
   * checkRequest() メソッドで発生したエラーを解除します。<br/><br/>
   *
   * エラーメッセージ、エラーとなった値が消去され、変更保留の行オブジェクトが変更前の状態に戻されます。<br/>
   */
  public void cancelUpdates(boolean initial) throws DataException {
    this.errorMessage.clear();
    this.errorValues.clear();
    super.cancelUpdates(initial);
  }





  /**
   * リクエスト内容から、当 WebDataTable オブジェクトに対する変更内容を取り出し、
   * 各行・各列の値を自動的に変更します。<br/><br/>
   *
   * 値を変更する際の項目値検証や非NULL検証でエラーが検出された場合、その内容がエラーメッセージのリストに
   * 追加されて false が返却されます。<br/>
   *
   * @param request クライアントがServletへ要求したリクエスト内容を含む HttpServletRequest オブジェクト
   * @return True：エラー無し、False：エラーあり
   */
  @SuppressWarnings("unchecked")
  public boolean checkRequest(HttpServletRequest request) {

    errorMessage.clear();
    errorValues.clear();
    Enumeration<String> params = request.getParameterNames();
    String name;

    try {
      RequestParam reqParam = null;
      RequestParamTable<RequestParam> temp = new RequestParamTable<RequestParam>();
      while (params.hasMoreElements()) {
        name = params.nextElement();
        if (name.startsWith(this.id + "_")) {
          // 2009/08/10 tableのパラメータ名称を固定化するか。
          // 2009/11/10 refactor by tnagao
          reqParam = new RequestParam();
          int r = name.indexOf('_');
          int c = name.lastIndexOf('_');
          int rowIndex = (name.charAt(r+1)=='p') ? ROW_PARAMS : Integer.valueOf(name.substring(r+1,c));
          int colIndex;
          String colStr = name.substring(c+1);
          if (useStaticParameterName && !String.valueOf(COL_DELETION_RESERVATION).equals(colStr)) {
            colIndex = (rowIndex == ROW_PARAMS) ?
              getParams().indexOfColumn(colStr) : getColumns().indexOfKey(colStr);
          } else {
            colIndex = Integer.valueOf(colStr);
          }
          reqParam.rowIndex = rowIndex;
          reqParam.colIndex = colIndex;
          String[] values = request.getParameterValues(name);
          if (values!=null && values.length>0) {
            reqParam.value = values[0];
          }
          temp.validate(reqParam);
        }
      }

      if (temp.getRowCount()==0) return true;

      temp.sort(); // Row、Columnのインデックス順に並び替えを行います。
      int targetRow = -1;
      int targetCol = -1;
      String newValue = "";
      boolean init = true;
      boolean deletionReserved = false;
      for (int i=0; i<temp.getRowCount(); i++) {
        reqParam = temp.getRow(i).getEntity();
        newValue = reqParam.value;
        //前後の全角、半角スペースは削除
        if (autoTrimRequest) {
          newValue = newValue.replaceAll("^[\\s　]*", "").replaceAll("[\\s　]*$", "");
        }
        targetRow = reqParam.rowIndex;
        targetCol = reqParam.colIndex;
        // Pramasに対する送信内容です。
        if (targetRow==ROW_PARAMS) {
          if (!newValue.equals(getParams().getValueAsString(targetCol))) {
            try {
              getParams().setValueAsString(targetCol, newValue);
            } catch (DataException x) {
              addErrorMessage(x.getLocalizedMessage());
              addErrorValues(ROW_PARAMS, targetCol, newValue);
            }
          }
          continue;

        // 中身が無ければ終了です。
        } else if (!isActive() || getRowCount()<=0) {
          break;
        }
        // RowIndexが変化したときの処理です。
        if (getRowIndex()!= targetRow) {
          if (!init) {
            try {
              getRow().validate();
            } catch (DataException x) {
              addErrorMessage(getRowIndex(), x.getLocalizedMessage());
            }
          }
          this.moveTo(targetRow);
          deletionReserved = false;
        }
        init = false;
        if (deletionReserved) continue;
        // 削除予約の値を処理します。
        if (targetCol==COL_DELETION_RESERVATION) {
          getRow().setDeletionReserved(new Boolean(newValue));
          deletionReserved = getRow().isDeletionReserved();
          continue;
        }
        // 通常の値チェック＆受け入れです。
        try {
          if (!getRow().getValueAsString(targetCol).equals(newValue) && isTypedRow()) {
            getRow().startEditing();
          }
          getRow().setValueAsString(targetCol, newValue);
        } catch (DataException x) {
          addErrorMessage(targetRow, x.getLocalizedMessage());
          addErrorValues(targetRow, targetCol, newValue);
        }
      }

      //最終行のチェック
      if (this.isActive() && this.getRowCount()>0 && targetRow!=-1 && targetCol!=-1) {
        try {
          getRow().validate();
        } catch (DataException x) {
          addErrorMessage(targetRow, x.getLocalizedMessage());
        }
      }
      try {
        getParams().validate();
      } catch (DataException x) {
        addErrorMessage(-2, x.getLocalizedMessage());
      }
    } catch (DataException x) {
      x.printStackTrace();
    }
    return errorMessage.isEmpty();
  }

  private void addErrorValues(int row, int col, String value) {
    String key = row + ":" + col;
    if (errorValues.get(key)==null) errorValues.put(key, value);
  }

  /**
   * エラーメッセージを追加します。<br/>
   *
   * @param message エラーメッセージ
   */
  public void addErrorMessage(String message) {
    addErrorMessage(-2, message);
  }

  /**
   * エラーメッセージを追加します。<br/><br/>
   *
   * 行番号付きでメッセージを出力する場合、[行Index+1]が行番号になります。<br/>
   * 行Indexに[-2]を指定すると、行番号付きでメッセージを出力の場合でも行番号が付与されません。<br/>
   *
   * @param index 行Index
   * @param message エラーメッセージ
   */
  public void addErrorMessage(int index, String message) {

    if (index!=-2 && index < 0) return;

    HashMap<String, Object> errorContent = new HashMap<String, Object>();
    errorContent.put("rownum", index + 1);
    errorContent.put("message", message);

    if (this.errorMessage.contains(errorContent)) return;
    this.errorMessage.add(errorContent);
  }

  /**
   * エラーメッセージのリストを取得します。<br/>
   *
   * @return エラーメッセージリスト
   */
  public ArrayList<HashMap<String, Object>> getErrors() {
    return this.errorMessage;
  }

  /**
   * ひとつの文字列に編集したエラーメッセージを取得します.<br/>
   * この中には改行\nがふくまれる可能性があります。
   * @return errorMessages
   */
  public String getErrorMessages() {
    StringBuilder sb = new StringBuilder();
    for (HashMap<String, Object> err : getErrors()) {
      if (sb.length() > 0) sb.append("\n");
      if (isContainRowNumInErrors() && !"-1".equals(err.get("rownum"))){
        sb.append(err.get("rownum"));
        sb.append("行目:");
      }
      sb.append(HTML.escape(err.get("message").toString()));
    }
    return sb.toString();
  }

  /**
   * 出力するエラーメッセージに行番号を含めるか否かを取得します。<br/><br/>
   *
   * @return 出力するエラーメッセージに行番号を含めるか否か
   */
  public boolean isContainRowNumInErrors() {
    return containRowNumInErrors;
  }

  /**
   * 出力するエラーメッセージに行番号を含めるか否かを設定します。<br/><br/>
   *
   * 「デフォルト：false」<br/>
   * また、各アプリケーションからaddErrorMessageメソッドにより追加された
   * エラーメッセージは、行番号は出力されません。<br/>
   *
   * @param containRowNumInErrors 出力するエラーメッセージに行番号を含めるか否か
   */
  public void setContainRowNumInErrors(boolean containRowNumInErrors) {
    this.containRowNumInErrors = containRowNumInErrors;
  }

  @Override
  public synchronized Row<E> appendRow() throws DataException {
    Row<E> row = super.appendRow();
    adjustEndRows();
    return row;
  }

  @Override
  public synchronized Row<E> insertRow(int index) throws DataException {
    Row<E> row = super.insertRow(index);
    adjustEndRows();
    return row;
  }

  @Override
  public synchronized void deleteRow() throws DataException {
    super.deleteRow();
    if (showRows!=-1 && currentPage > getPageCount()) {
      if (getRowCount() == 0) {
        pageStartRow = 0;
        currentPage = 1;
      }
      else {
        pageStartRow = (getPageCount() - 1) * showRows;
        currentPage = getPageCount();
      }
    }
    adjustEndRows();
  }

  @Override
  public synchronized void validateAllRows() throws DataException {
    super.validateAllRows();
    adjustEndRows();
  }

  @Override
  public synchronized void applyFilter() throws DataException {
    super.applyFilter();
    adjustEndRows();
  }

  /**
   * WebDataTable オブジェクトの内容をHTMLテンプレートからアクセスできるように変換するクラス。<br/><br/>
   *
   * ActionPageクラスにpublicフィールドとして定義された WebDataTable オブジェクトは自動的に
   * HTMLテンプレートからアクセスできるようになっていますが、これは擬似的なものであり、
   * WebDataTable オブジェクトに対する直接アクセスではありません。<br/><br/>
   *
   * 実際には、ActionPage クラスの show() メソッドや getMergedString() メソッドの中で当PageContextクラスが生成され、
   * WebDataTable オブジェクトの内容がコピーされています。<br/><br/>
   *
   * <em>※注意：</em>
   * HTMLテンプレート内のスクリプトから呼び出される動作は、当クラスのpublicメソッドである点に注意してください。<br/>
   *
   * @author LittleSoft Corporation
   */
  public static class PageContext implements Serializable  {
    private static final long serialVersionUID = 1L;
    WebDataTable<?> dataTable;
    static class PageRowList<E> extends ArrayList<E> {
      private static final long serialVersionUID = 1L;
      private PageContext ctx;
      public PageRowList(PageContext ctx) {
        this.ctx = ctx;
      }
      @SuppressWarnings("unchecked")
      @Override
      public E get(int index) {
        Object result = super.get(index);
        ctx.cursor = (PageRow<String, String>)result;
        return (E)result;
      }
    }
    PageRowList<PageRow<String,String>> rows = new PageRowList<PageRow<String,String>>(this);
    ArrayList<String> errors = new ArrayList<String>();
    //PageRow params = new PageRow(dataTable, 0, Row.Status.UNKNOWN);
    PageRow<String,String> params;
    PageRow<String,String> cursor;
    /**
     * 引数 dataTable で指定されたWebデータテーブルを変換して、 PageContext オブジェクトを生成します。<br/>
     *
     * @param dataTable Webデータテーブル
     */
    public PageContext(WebDataTable<?> dataTable) {
      this.dataTable = dataTable;
      this.params = new PageRow<String, String>(this, 0, Row.Status.UNKNOWN, true);
    }

    /**
     * WebDataTable オブジェクトが活性化しているかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   #if($table.active) ...
     * </pre>
     *
     * @return True：活性化、False：非活性化
     */
    public boolean getactive() {
      return getActive();
    }

    /**
     * WebDataTable オブジェクトが活性化しているかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getactive() メソッドと同様です。<br/>
     *
     * @return True：活性化、False：非活性化
     */
    public boolean getActive() {
      return this.dataTable.isActive();
    }

    /**
     * スクロール可能な前ページが存在するかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   #if($table.hasPrevPage) ...
     * </pre>
     *
     * @return True：存在する、False：存在しない
     */
    public boolean gethasPrevPage() {
      return getHasPrevPage();
    }

    /**
     * スクロール可能な前ページが存在するかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き gethasPrevPage() メソッドと同様です。<br/>
     *
     * @return True：存在する、False：存在しない
     */
    public boolean getHasPrevPage() {
      return this.dataTable.hasPreviousPage();
    }

    /**
     * スクロール可能な次ページが存在するかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   #if($table.hasNextPage) ...
     * </pre>
     *
     * @return True：存在する、False：存在しない
     */
    public boolean gethasNextPage() {
      return getHasNextPage();
    }

    /**
     * スクロール可能な次ページが存在するかどうを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き gethasNextPage() メソッドと同様です。<br/>
     *
     * @return True：存在する、False：存在しない
     */
    public boolean getHasNextPage() {
      return this.dataTable.hasNextPage();
    }

    /**
     * WebDataTable オブジェクトの行数を取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.rowCount
     * </pre>
     *
     * @return 行数
     */
    public int getrowCount() {
      return getRowCount();
    }

    /**
     * WebDataTable オブジェクトの行数を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getrowCount() メソッドと同様です。<br/>
     *
     * @return 行数
     */
    public int getRowCount() {
      return this.dataTable.getRowCount();
    }

    /**
     * WebDataTable オブジェクトのページャーを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.pager
     * </pre>
     *
     * @return 行数
     */
    public String getpager() {
      return getPager();
    }

    /**
     * WebDataTable オブジェクトのページャーを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getpager() メソッドと同様です。<br/>
     *
     * @return 行数
     */
    public String getPager() {
      return this.dataTable.getPagerTag();
    }

    /**
     * 行オブジェクトの内容をリストで取得します。<br/><br/>
     *
     * HTMLテンプレートからは次のような記述でリストから行オブジェクトを取り出し、その内容にアクセスすることができます。<br/>
     * <pre>
     *   #foreach( $dept in $tableDept.rows )
     *     &lt;li&gt; $dept.get('dept_code') $dept.get('dept_name')&lt;/li&gt;
     *   #end
     * </pre>
     * <em>※注意：</em>
     * 当メソッドで取得できるリストの内容はWebデータテーブルのShowRowsプロパティによって変化します。<br/>
     * ディフォルトでは全ての行がリストに含まれますが、ShowRowsプロパティに値が設定されている場合、
     * 設定値を上限とした行のみを取得することができます。<br/><br/>
     *
     * これを利用して、リストのページング表示を行うことが可能となります。<br/><br/>
     *
     * リストに含まれる行の範囲を変更する場合は、Webデータテーブルの以下メソッドを使用します。<br/>
     * <ul>
     *   <li>pageFirst()</li>
     *   <li>pageNext()</li>
     *   <li>pageLast()</li>
     *   <li>pagePrevious()</li>
     * <ul><br/>
     *
     * @see WebDataTable#setShowRows(int)
     * @see WebDataTable#pageFirst()
     * @see WebDataTable#pageLast()
     * @see WebDataTable#pageNext()
     * @see WebDataTable#pagePrevious()
     * @return 行オブジェクトのArrayList
     */
    public java.util.ArrayList<PageRow<String, String>> getrows() {
      return getRows();
    }

    /**
     * 行オブジェクトの内容をリストで取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getrows() メソッドと同様です。<br/>
     *
     * @return 行オブジェクトのArrayList
     */
    public java.util.ArrayList<PageRow<String, String>> getRows() {
      return rows;
    }

    /**
     * 引数 index で指定された行オブジェクトを PageRow オブジェクトに変換して取得します。<br/><br/>
     *
     * HTMLテンプレートからは次のような記述でリストから行オブジェクトを取り出し、その内容にアクセスすることができます。
     * <pre>
     *   #set( $emp = $tableEmp.getRow('0') )
     * </pre>
     * <em>注意：</em>
     * 当メソッドで取得できる行オブジェクトはjp.littlesoft.data.Rowクラスではなく、PageRowクラスである点に注意してください。<br/>
     * WebDataTable内部の 行オブジェクトをHTMLテンプレートからアクセス可能にする為、PageRow オブジェクト へ変換しています。<br/>
     *
     * @param index 行インデックス
     * @return PageRow テンプレート用の行オブジェクト
     */
    public PageRow<String, String> getrow(String index) {
      return getRow(index);
    }

    /**
     * 引数 index で指定された行オブジェクトを PageRow オブジェクトに変換して取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getrow(String) メソッドと同様です。<br/>
     *
     * @param index 行インデックス
     * @return PageRow テンプレート用の行オブジェクト
     */
    public PageRow<String, String> getRow(String index) {
      PageRow<String, String> result = null;
      try {
        int i = Integer.parseInt(index);
        try { result = rows.get(i); } catch (IndexOutOfBoundsException a) {}
        if (result==null) {
          result = dataTable.toPageRow(this, i);
        }
        return result;
      } catch (Exception x) {
        //x.printStackTrace();
        return new PageRow<String,String>(this, 0, Row.Status.UNKNOWN, false);
      }
    }

    /**
     * getRows() メソッドで返される先頭行のインデックスを取得します。<br/><br/>
     *
     * このインデックスはWebDataTable内での Row オブジェクトの位置を示すものです。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.startRow
     * </pre>
     *
     * @return 開始行インデックス
     */
    public int getstartRow() {
      return getStartRow();
    }

    /**
     * getRows() メソッドで返される先頭行のインデックスを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getstartRow() メソッドと同様です。<br/>
     *
     * @return 開始行インデックス
     */
    public int getStartRow() {
      return this.dataTable.pageStartRow;
    }

    /**
     * getRows() メソッドで返される末尾行のインデックスを取得します。<br/><br/>
     *
     * このインデックスはWebDataTable内での Row オブジェクトの位置を示すものです。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.endRow
     * </pre>
     *
     * @return 末尾行インデックス
     */
    public int getendRow() {
      return getEndRow();
    }

    /**
     * getRows() メソッドで返される末尾行のインデックスを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getendRow() メソッドと同様です。<br/>
     *
     * @return 末尾行インデックス
     */
    public int getEndRow() {
      return this.dataTable.pageEndRow;
    }

    /**
     * getRows() メソッドで返される先頭行の行番号を取得します。<br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.startRowNum
     * </pre>
     *
     * @return 開始行インデックス
     */
    public int getstartRowNum() {
      return getStartRowNum();
    }

    /**
     * getRows() メソッドで返される先頭行の行番号を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getstartRowNum() メソッドと同様です。<br/>
     *
     * @return 開始行インデックス
     */
    public int getStartRowNum() {
      return this.dataTable.getPageStartRowNum();
    }

    /**
     * getRows() メソッドで返される末尾行の行番号を取得します。<br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.endRowNum
     * </pre>
     *
     * @return 末尾行インデックス
     */
    public int getendRowNum() {
      return getEndRowNum();
    }

    /**
     * getRows() メソッドで返される末尾行の行番号を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getendRowNum() メソッドと同様です。<br/>
     *
     * @return 末尾行インデックス
     */
    public int getEndRowNum() {
      return this.dataTable.getPageEndRowNum();
    }

    /**
     * getRows() メソッドで返される行数を取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.showRows
     * </pre>
     *
     * @return 行数
     */
    public int getshowRows() {
      return getShowRows();
    }

    /**
     * getRows() メソッドで返される行数を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getshowRows() メソッドと同様です。<br/>
     *
     * @return 行数
     */
    public int getShowRows() {
      return this.dataTable.showRows;
    }

    /**
     * エラーメッセージのリストを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   #foreach( $err in $table.errors )
     *      &lt;li&gt; $err &lt;/li&gt;
     *   #end
     * </pre>
     *
     * @return メッセージ文字列のリスト
     */
    public java.util.ArrayList<String> geterrors() {
      return getErrors();
    }

    /**
     * エラーメッセージのリストを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き geterrors() メソッドと同様です。<br/>
     *
     * @return メッセージ文字列のリスト
     */
    public java.util.ArrayList<String> getErrors() {
      return errors;
    }

    /**
     * エラーメッセージが存在するかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   #if($table.hasError) .....
     * </pre>
     *
     * @return True：エラーメッセージ有り、False：無し
     */
    public boolean gethasError() {
      return getHasError();
    }

    /**
     * エラーメッセージが存在するかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き gethasError() メソッドと同様です。<br/>
     *
     * @return True：エラーメッセージ有り、False：無し
     */
    public boolean getHasError() {
      return !this.errors.isEmpty();
    }

    /**
     * Webデータテーブルの Params オブジェクトの内容を取得します。<br/><br/>
     *
     * このメソッドで取得できるのは実際の Params オブジェクトではなく、その値がコピーされた PageRow オブジェクトです。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $table.params.[PageRowのメソッド]
     * </pre>
     *
     * @return Paramsの内容がコピーされた PageRow オブジェクト
     */
    public PageRow<String, String> getparams() {
      return getParams();
    }

    /**
     * Webデータテーブルの Params オブジェクトの内容を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getparams() メソッドと同様です。<br/>
     *
     * @return Paramsの内容がコピーされた PageRow オブジェクト
     */
    public PageRow<String, String> getParams() {
      return params;
    }

  }

  /**
   * 行オブジェクトの内容をHTMLテンプレートからアクセスできるように変換するクラス。<br/><br/>
   *
   * HTMLテンプレートでは以下のような記述で、行オブジェクトにアクセスします。<br/><br/>
   *
   * ループしながら各行オブジェクトにアクセスする。
   * <pre>
   *   #foreach( $emp in $tableEmp.rows )
   *     .......
   * </pre>
   *
   * インデックスを指定して、アクセスする。
   * <pre>
   *   #set( $emp = $tableEmp.getRow('0') )
   *     ......
   * </pre>
   *
   * @author LittleSoft Corporation
   */
  public final static class PageRow<K, V> extends HashMap<K, V> {

    private PageContext pageContext;
    private WebDataTable<?> dataTable;
    private boolean isNew = false;
    private int index;
    private int rowNumber;
    private boolean deletionDeserved;
    private Row.Status status;
    private boolean isParams = false;

    /**
     * 引数 dataTable、rowIndex で指定された行オブジェクトを変換して、 PageRow オブジェクトを生成します。<br/>
     *
     * @param dataTable Webデータテーブル
     * @param index 行インデックス
     * @param status 行の状態
     */
    public PageRow(PageContext pageContext, int index, Row.Status status, boolean isParams) {

      super();
      this.pageContext = pageContext;
      this.dataTable = pageContext.dataTable;
      this.index = index;
      this.status = status;
      if (dataTable==null || !dataTable.isPageDivReserve()) {
        this.rowNumber = index+1;
      } else {
        this.rowNumber = index+1+dataTable.showRows*(dataTable.currentPage-1);
      }
      this.isParams = isParams;
    }

    @SuppressWarnings("unchecked")
    private void setCursor() {
      pageContext.cursor = (PageRow<String, String>) this;
    }

    /**
     * 行の状態を取得します。<br/>
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.status
     * </pre>
     * @return 行の状態
     */
    public String getstatus() {
      return getStatus();
    }

    /**
     * 行の状態を取得します。<br/>
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getstatus() メソッドと同様です。<br/>
     * @return 行の状態
     */
    public String getStatus() {
      setCursor();
      return status.toString();
    }

    /**
     * 行オブジェクトの番号を取得します。<br/><br/>
     *
     * 先頭行が 1 から始まる値であり、一覧表示の行番号として利用することができます。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.rowNumber
     * </pre>
     *
     * @return 行番号
     */
    public int getrowNumber() {
      return getRowNumber();
    }

    /**
     * 行オブジェクトの番号を取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getrowNumber() メソッドと同様です。<br/>
     *
     * @return 行番号
     */
    public int getRowNumber() {
      //return index + 1;
      setCursor();
      return this.rowNumber;
    }

    /**
     * 行オブジェクトのインデックスを取得します。<br/><br/>
     *
     * この値はWebDataTable内での行オブジェクトの位置を示します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.index
     * </pre>
     *
     * @return 行インデックス
     */
    public int getindex() {
      return getIndex();
    }

    /**
     * 行オブジェクトのインデックスを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getindex() メソッドと同様です。<br/>
     *
     * @return 行インデックス
     */
    public int getIndex() {
      setCursor();
      return index;
    }

    /**
     * 行オブジェクトが新規作成されたものかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.isNew
     * </pre>
     *
     * @return True：新規作成である、False：新規ではない
     */
    public boolean getisNew() {
      return getIsNew();
    }

    /**
     * 行オブジェクトが新規作成されたものかどうかを取得します。<br/><br/>
     *
     * HTMLテンプレートでの記述が、先頭が大文字になる点を除き getisNew() メソッドと同様です。<br/>
     *
     * @return True：新規作成である、False：新規ではない
     */
    public boolean getIsNew() {
      setCursor();
      return isNew;
    }

    /**
     * 行オブジェクトが新規作成されたものかどうかを設定します。<br/><br/>
     *
     * @param isNew True：新規作成である、False：新規ではない
     */
    void setIsNew(boolean isNew) {
      setCursor();
      this.isNew = isNew;
    }

    /**
     * 引数 name で指定された列とその値を、HTMLのテキスト入力項目と連結します。<br/><br/>
     *
     * 行オブジェクトの列値をHTMLテンプレート内のテキスト入力項目のvalue属性に埋め込みます。<br/>
     * <em>※注意：</em>
     * この値はエスケープ文字列化されたデータ値です。</br>
     * 同時にテーブルオブジェクト、行、列を一意に特定する値を生成し、name属性に埋め込みます。<br/><br/>
     *
     * 行オブジェクトの桁数をHTMLテンプレート内のテキスト入力項目のmaxlength属性に埋め込みます。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   &lt;input type="text" $emp.textBind('salary')/&gt;
     * </pre>
     *
     * Webデータテーブルの checkRequest() メソッドを使用すると、textBindで生成されたHTMLフォームより送信された
     * HTTPリクエストに含まれるている変更値を取り出し、当該項目の内容を自動的に更新することが可能です。<br/>
     *
     * @param name 列名
     * @return HTMLInputタグのvalue属性とname属性
     * @see WebDataTable#checkRequest(HttpServletRequest)
     */
    public String textBind(String name) {
      return textBind(name, true);
    }

    /**
     * 引数 name で指定された列とその値を、HTMLのテキスト入力項目と連結します。<br/><br/>
     *
     * 行オブジェクトの列値をHTMLテンプレート内のテキスト入力項目のvalue属性に埋め込みます。<br/>
     * <em>※注意：</em>
     * この値はエスケープ文字列化されたデータ値です。</br>
     * 同時にテーブルオブジェクト、行、列を一意に特定する値を生成し、name属性に埋め込みます。<br/><br/>
     *
     * 引数 maxLength が true の場合、行オブジェクトの桁数をHTMLテンプレート内のテキスト入力項目のmaxlength属性に埋め込みます。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   &lt;input type="text" $emp.textBind('salary', true)/&gt;
     * </pre>
     *
     * Webデータテーブルの checkRequest() メソッドを使用すると、textBindで生成されたHTMLフォームより送信された
     * HTTPリクエストに含まれるている変更値を取り出し、当該項目の内容を自動的に更新することが可能です。<br/>
     *
     * @param name 列名
     * @return HTMLInputタグのvalue属性とname属性
     * @see WebDataTable#checkRequest(HttpServletRequest)
     */
    public String textBind(String name, boolean maxLength) {
      String v = (String)this.get(name);
      // MaxLength対応
      String m = "";
      if (maxLength && this.dataTable != null) {
        try {
          Column column;
          // param の場合
          if (this.isParams) {
            column = this.dataTable.getParams().getColumn(name);
          }
          // column の場合
          else {
            column = this.dataTable.getColumn(name);
          }
          if (column != null) {
            m = HTML.toMaxLength(column);
          }
        } catch (DataException x) {
          //x.printStackTrace();
        }
      }
      return this.nameBind(name) + " " + m + " value=\"" + v + "\"";
    }

    /**
     * 引数 name、checkValue で指定された列とその値を、HTMLのチェック項目と連結します。<br/><br/>
     *
     * 行オブジェクトのカラム値をもとにHTMLテンプレート内のチェック項目状態を決定します。<br/>
     * 同時にテーブルオブジェクト、行、列を一意に特定する値を生成し、name属性に埋め込みます。<br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   &lt;input type="checkbox" $emp.checkBind('manager'，'01')/&gt;
     *   &lt;input type="hidden" $emp.checkBind('manager','00')/&gt;
     * </pre>
     *
     * <em>※注意：</em>
     * この例では、カラム[manager]の値が '01' の場合にチェックされた状態がWebブラウザに表示されます。<br/>
     * 2行目のhidden項目はチェックが解除された際、カラム[manager]に設定される値を示しています。
     * この例では '00' が設定されます。<br/><br/>
     *
     * Webデータテーブルの checkRequest() メソッドを使用すると、checkBindで生成されたHTMLフォームより送信された
     * HTTPリクエストに含まれるている変更値を取り出し、当該項目の内容を自動的に更新することが可能です。<br/>
     *
     * @param name 列名
     * @param checkValue チェック状態とする値、あるいは未チェック状態とする値
     * @return HTMLInputタグのvalue属性とname属性および"checked"
     * @see WebDataTable#checkRequest(HttpServletRequest)
     */
    public String checkBind(String name, String checkValue) {
      String v = (String)this.get(name);
      //String n = this.nameBind(name);
      return this.nameBind(name) + " value=\"" + checkValue + "\"" +
             (v.equals(checkValue) ? " checked" : "" );
    }

    /**
     * 行オブジェクトの削除予約プロパティをHTMLのチェック項目と連結します。<br/><br/>
     *
     * 行オブジェクトの削除予約プロパティ(DeletionDeserved)をもとにのチェック項目状態を決定します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   &lt;input type="checkbox" $emp.deleteCheckBind('true')  /&gt;
     *   &lt;input type="hidden"   $emp.deleteCheckBind('false') /&gt;
     * </pre>
     *
     * WebDataTableのcheckRequestメソッドを使用すると、deleteCheckBindで生成されたHTMLフォームより送信された
     * HTTPリクエストに含まれるている変更値を取り出し、当該行オブジェクトの削除予約プロパティを自動的に更新することが可能です。<br/>
     *
     * @param checkValue 'true'あるいは'false'
     * @return HTMLInputタグのvalue属性とname属性および"checked"
     * @see WebDataTable#checkRequest(HttpServletRequest)
     */
    public String deleteCheckBind(String checkValue) {
      setCursor();
      String name = this.dataTable.toRequestParameterName(this.index, COL_DELETION_RESERVATION);
      checkValue = checkValue.toLowerCase();
      if ("true".equals(checkValue)) {
        return "name=\"" + name + "\" value=\"true\"" + (this.deletionDeserved ? " checked" : "");
      } else {
        return "name=\"" + name + "\" value=\"false\" ";
      }
    }

    /**
     * 引数 name で指定された列を、HTML項目と連結します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   &lt;select $emp.nameBind('dept_code')&gt;
     *     &lt;option .......
     *   &lt;/select&gt;
     * </pre>
     *
     * @param name 列名
     * @return HTMLInputタグのname属性
     */
    public String nameBind(String name) {
      setCursor();
      return "name=\"" + (String)this.get("@" + name) + "\"";
    }

    /**
     * 引数 name で指定された列値を、行オブジェクトから取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.value('memo')
     * </pre>
     *
     * @param name 列名
     * @return エスケープ文字列化された列値
     */
    public String value(String name) {
      setCursor();
      return (String)this.get(name);
    }

    /**
     * 引数 name で指定された列値を、行オブジェクトから取得します。<br/><br/>
     *
     * HTMLテンプレートでは以下のような記述となります。
     * <pre>
     *   $emp.get('memo')
     * </pre>
     *
     * @param name 列名
     * @return エスケープ文字列化された列値
     */
    @SuppressWarnings("unchecked")
    public V get(Object name) {
      setCursor();
      String value = (String)super.get(name);
      return (V)HTML.escape(value);
    }

    /**
     * 引数 name、rawValueで指定された列{値を、行オブジェクトから取得します。<br/><br/>
     *
     * パラメータrawValueをtrueに設定するとエスケープ前のデータを取得できます。
     * <pre>
     *   $emp.get('memo',true)
     * </pre>
     *
     * @param name  列名
     * @param rawValue True：エスケープしない、False：エスケープする
     * @return 文字列化された列値
     */
    public String get(String name, boolean rawValue) {
      setCursor();
      if (rawValue) {
        return (String)super.get(name);
      } else {
        return (String)get(name);
      }
    }
  }

  /**
   * Velecityの動作をカスタマイズするﾊﾝﾄﾞﾗ.
   * LSJのHTMLテンプレートに必要なメソッドを拡張します。
   * @author littlesoft
   */
  public static class PageVelocityHandler implements ReferenceInsertionEventHandler, InvalidReferenceEventHandler, MethodExceptionEventHandler, Serializable  {

    private PageContext pageContext;
    private Context context;

    public PageVelocityHandler(PageContext pageContext) {
      this.pageContext = pageContext;
    }

    private String extractFieldName(String reference) {
      try {
        int rightDot = reference.lastIndexOf('.');
        int leftDot = reference.lastIndexOf('.', (rightDot-1));
        return reference.substring(leftDot + 1, rightDot);
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    }

    private String extractTableRef(String reference) {
      int leftDot = reference.indexOf('.');
      if (leftDot==-1) return null;
      String name = reference.substring(0, leftDot);
//      if (reference.substring(leftDot).toLowerCase().startsWith(".params.")) {
//        name += ".params";
//      }
      if (name.charAt(0)=='$') name = name.substring(1);
      return name;
    }

    @Override
    public Object referenceInsert(String reference, Object value) {
      if (value!=null || !checkEventSource(this.context, reference)) return value;
      String ref1 = reference.toLowerCase().replaceAll("\"", "");
      ref1 = ref1.toLowerCase().replaceAll("'", "");
      int p1 = ref1.lastIndexOf("(");
      int p2 = ref1.lastIndexOf(")");
      if (p1 < p2) {
        if (ref1.substring(0, p1).endsWith(".checkbind")) {
          String args = ref1.substring(p1+1, p2);
          String name = extractFieldName(reference.substring(0, p1));
          return pageContext.cursor.checkBind(name, args);
        }
      } else if (p1 == p2) {
      } else {
        throw new RuntimeException("invalid reference[" + reference + "]");
      }
      return value;
    }

    private boolean checkEventSource(Context context, String reference) {
      if (context==null) return false;
      String tableRef = extractTableRef(reference);
      Object tartget = context.get(tableRef);
      if (tartget!=pageContext && tartget!=pageContext.cursor) {
        return false;
      } else {
        this.context = context;
        return true;
      }
    }

    @Override
    public Object invalidGetMethod(Context context, String reference, Object object, String property, Info info) {
      if (property==null || !checkEventSource(context, reference)) return null;
      String ref = reference.toLowerCase();
      String result = null;
      //todo Captionとれるようにする。
      if (ref.endsWith(".namebind")) {
        result = pageContext.cursor.nameBind(extractFieldName(reference));
      } else if (ref.endsWith(".textbind")) {
        result = pageContext.cursor.textBind(extractFieldName(reference));
      } else if (ref.endsWith(".textbindnolength")) {
        result = pageContext.cursor.textBind(extractFieldName(reference),false);
      } else if (ref.endsWith(".rawvalue")) {
        result = pageContext.cursor.get(extractFieldName(reference),true);
      } else if (ref.endsWith(".inputname") || ref.endsWith(".name")) {
        result = (String) pageContext.cursor.get("@" + extractFieldName(reference));
      }
      return result;
    }


    @Override
    public Object invalidMethod(Context context, String reference, Object object, String method, Info info) {
      //System.out.println("*** invalidMethod:" + reference + ", object:" + object + ", method:" + method + ", info:" + info);
      return null;
    }

    @Override
    public boolean invalidSetMethod(Context context, String leftreference, String rightreference, Info info) {
      //System.out.println("*** invalidSetMethod:" + leftreference + ", right:" + rightreference + ", Info:" + info );
      return false;
    }

    @Override
    public Object methodException(Class clazz, String method, Exception exception) throws Exception {
      System.out.println("*** methodException:" + clazz.getSimpleName() + ", method:" + method + ", Exception:" + exception.toString() );
      return null;
    }

  }

  static public class RequestParam implements Serializable {
    public Integer rowIndex;
    public Integer colIndex;
    public String value;
  }
  /**
   * RequestData
   *
   * @author LittleSoft Corporation
   */
  static class RequestParamTable<E extends Serializable> extends Table<E> {
    static SortKey[] skey = new SortKey[]{new Table.SortKey("rowIndex"), new Table.SortKey("colIndex")};
    RequestParamTable(E...e) {
      super(e);
    }
    void sort() {
      try { this.sortRows(skey ); } catch (Exception x) {}
    }
  }

  /**
   * ページの最終行のIndexを設定します。<br/>
   *
   * @param pageEndRow  最終行のIndex
   */
  protected void setPageEndRowIndex(int pageEndRow) {
    this.pageEndRow = pageEndRow;
  }

  /**
   * 現在のページ番号を取得します。<br/>
   *
   * @return  ページ番号
   */
  public int getCurrentPage() {
    return currentPage;
  }

  /**
   * 現在のページ番号を設定します。<br/>
   *
   * @param currentPage  ページ番号
   */
  protected void setCurrentPage(int currentPage) {
    this.currentPage = currentPage;
  }

  /**
   * ページの分割取得を使用するか否かを設定します。<br/>
   *
   * @param usePageDivReserve  ページの分割取得を使用するか否か
   * @throws DataException  例外発生時
   */
  public void setPageDivReserve(boolean usePageDivReserve) throws DataException {
    checkDeactive();
    this.usePageDivReserve = usePageDivReserve;
  }

  /**
   * ページの分割取得を使用しているか否かを取得します。<br/>
   *
   * @return  ページの分割取得を使用しているか否か
   */
  public boolean isPageDivReserve() {
    return !(currentPage==-1 || !usePageDivReserve);
  }

  /**
   * ページングを実行するメソッドをコールするタイプの列挙型
   */
  public enum PagingCallMethodType { EXEC, AJAX }

  static String strTagSpc = "&nbsp;";
  static String strPrev   = "前へ";
  static String strNext   = "次へ";
  private String pagingJavaMethodName     = "paging";
  private String pagingAjaxCompMethodName = "ajaxSearchUI";
  private PagingCallMethodType pagingCallMethodType = PagingCallMethodType.AJAX;
  private int pagingMaxPageCount = 10;

  /**
   * ページングを行うアクションメソッドを設定します。<br/>
   *
   * @param name メソッド名
   */
  public void setPagingJavaMethodName(String name) {
    this.pagingJavaMethodName = name;
  }

  /**
   * Ajax呼び出し時、レスポンス返却時に呼び出すファンクションを指定します。<br/>
   *
   * @param name ファンクション名
   */
  public void setPagingAjaxCompMethodName(String name) {
    this.pagingAjaxCompMethodName = name;
  }

  /**
   * ページングを「ls.exec」「ls.ajax」のどちらで呼び出すかを指定します。<br/>
   *
   * @param type 呼び出しファンクションのタイプ
   */
  public void setPagingCallMethodType(PagingCallMethodType type) {
    this.pagingCallMethodType = type;
  }

  /**
   * ページャーに表示するページ数の最大値を設定します。<br/>
   *
   * @param count ページ数
   */
  public void setPagingMaxPageCount(int count) {
    this.pagingMaxPageCount = count;
  }

  /**
   * ページング用のHTMLタグを取得します。<br/>
   *
   * @return  HTMLタグ
   */
  public String getPagerTag() {
    String link = "";
    //ページング未使用時は空文字を返却
    if (this.getShowRows() == -1) return "";
    //総ページ数算出
    int cnt = this.getPageCount();
    //総ページ数が「0」の場合は空文字を返却
    if (cnt == 0) return "";

    //リンク
    String strHref = "<a href=\"javascript:";
    switch(this.pagingCallMethodType) {
    case EXEC:
      strHref += "ls.exec(\'" + pagingJavaMethodName + "\',";
      break;
    case AJAX:
      strHref += "ls.ajax(\'" + pagingJavaMethodName + "\'," + pagingAjaxCompMethodName + ",";
      break;
    }

    //前のページが存在する
    if (this.hasPreviousPage()) {
      link += strHref+"\'prev\');\">" + strPrev + "</a>";
    }
    //前ページが存在しない
    else {
      link += strPrev;
    }

    //出力ページ番号数分、番号リンクを出力
    String num = "1";
    int startPage = 1;
    int endPage   = 1;
    int maxCount  = this.pagingMaxPageCount;
    if (maxCount == -1) maxCount = cnt;
    startPage = this.getCurrentPage() - maxCount / 2;
    if (startPage + maxCount - 1 > cnt) startPage = cnt - maxCount + 1;
    if (startPage < 1) startPage = 1;
    endPage = startPage + maxCount - 1;
    if (endPage > cnt) endPage = cnt;

    for (int i = startPage; i < endPage + 1; i++) {
      num = String.valueOf(i);
      //現在のページであれば非リンク
      if (this.getCurrentPage() == i) {
        link += strTagSpc + num;
      }
      //現在のページではない
      else {
        link += strTagSpc + strHref + "\'" + num + "\');\">" + num + "</a>";
      }
    }

    //次のページが存在する
    if (this.hasNextPage()) {
      link += strTagSpc + strHref + "\'next\');\">" + strNext + "</a>";
    }
    //次のページが存在しない
    else {
      link += strTagSpc + strNext;
    }
    return link;
  }

  /**
   * ページ数を取得します。<br/>
   *
   * @return ページ数
   */
  private int getPageCount() {
    int count = 0;
    if (isPageDivReserve()) {
      count = getAllRowCount() / showRows;
      if (getAllRowCount()%showRows > 0) count++;
    }
    else {
      count = this.getRowCount() / showRows;
      if (this.getRowCount()%showRows > 0) count++;
    }
    return count;
  }

  /**
   * WebSQLTableの内容を標準出力に表示します。
   *
   * @throws Exception 例外発生時
   */
  public void dump() throws Exception {
    String name = (getName() == null) ? "no name" : getName();

    System.out.println("");
    System.out.println("*** [DUMP START : " + name + "]");
    System.out.println("*** rowCount : " + allRows().size() + "");

    if (allRows().size() > 0) {
      System.out.print("[STATUS], ");
      for (int x = 0; x < getColumnCount(); x++) {
        if (x > 0) System.out.print(", ");
        System.out.print(getColumn(x).getName());
      }
      System.out.println("");

      for (Row<?> row : allRows()) {
        System.out.print("[" + row.getStatus() + "], ");
        for (int x = 0; x < getColumnCount(); x++) {
          if (x > 0) System.out.print(", ");
          Object tmp = row.getValue(getColumn(x).getName());
          tmp = (tmp == null) ? "(NULL)" : tmp;
          System.out.print(tmp);
        }
        System.out.println("");
      }
    }

    System.out.println("*** errorCount : " + getErrors().size() + "");
    if (getErrors().size() > 0) {
      System.out.println(getErrors().toString());
    }
    System.out.println("*** [DUMP END : " + name + "]");
    System.out.println("");
  }

  public boolean isAutoTrimRequest() {
    return autoTrimRequest;
  }

  public void setAutoTrimRequest(boolean autoTrimRequest) {
    this.autoTrimRequest = autoTrimRequest;
  }

  /**
   * 2009/08/10 WebDataTableのパラメータ名を固定にする対応に伴い作成 <br/>
   * WebDataTable オブジェクトのパラメータ名を設定します。<br/>
   */
  public void setId(String id){
      useStaticParameterName = true;
      this.id = id;
  }

  /**
   * 2009/08/10 WebDataTableのパラメータ名を固定にする対応に伴い作成 <br/>
   * WebDataTable オブジェクトのパラメータ名を取得します。<br/>
   */
  public String getId(){
      return this.id;
  }

  private String defaultInformation;

  /**
   * Please set up small pieces of information needed by default.
   */
  @Override
  public String getDefaultInformation() {
    return defaultInformation;
  }

  /**
   * Please set up small pieces of information needed by default.
   * @param defaultInformation
   */
  public void setDefaultInformation(String defaultInformation) {
    this.defaultInformation = defaultInformation;
  }

  @Override
  public String getLoadscript() {
    return "";
  }

}

