/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.Serializable;
import java.sql.Connection;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.ITableDef;
import jp.littlesoft.data.Params;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Column.NoAccessibleObjectException;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.ISQLApplication;
import jp.littlesoft.sql.SQL;
import jp.littlesoft.sql.ISQLDataProvider;
import jp.littlesoft.sql.ISQLDataResolver;
import jp.littlesoft.sql.SQL.BindParameter;
import jp.littlesoft.sql.SQL.Comparer;
import jp.littlesoft.sql.SQL.Condition;
import jp.littlesoft.sql.SQL.Delete;
import jp.littlesoft.sql.SQL.Expression;
import jp.littlesoft.sql.SQL.Insert;
import jp.littlesoft.sql.SQL.Select;
import jp.littlesoft.sql.SQL.Update;

/**
 * WebSQLテーブルクラス。<br/><br/>
 * 
 * SQLデータベースへの読み書きに対応したWebDataTable拡張クラスです。<br/>
 * このクラスはISQLDataProviderとISQLDataResolverを内蔵しています。<br/><br/>
 * 
 * [ISQLDataProvider]<br/>
 * SQLデータベースに対するSELECT文の結果をもとにRowオブジェクトを生成し、Tableオブジェクトの内容にセットするクラスです。<br/><br/>
 * 
 * [ISQLDataResolver]<br/>
 * Tableオブジェクトに対して行った編集（行の変更・削除・追加）をもとに、指定されたSQLデータベーステーブルへ内容を書き込むクラスです。<br/><br/>
 * 
 * このクラスの内部では、上記オブジェクトが自動的に生成され、DataProviderプロパティ、DataResolverプロパティに、それぞれセットされています。<br/><br/>
 * 
 * [基本的な使い方：SQLデータベースから値を取得する]
 * <pre>
 *   //このクラスのオブジェクトを生成します。
 *   WebSQLTable table = new WebSQLTable();
 *   //SELECT文を設定します。
 *   table.setSQL("select * from emp");
 *   //活性化すると自動的にSQLデータベースへのSELECT文を実行します。この際、JDBCコネクションを与えます。
 *   table.activate(conn);
 *   //SELECT文結果の値を表示します。
 *   table.first();  //先頭行へ移動
 *   while(table.inBounds()) {
 *     System.out.println("お名前：" + table.getValueAsString("NAME"));
 *     table.next(); //次の行へ移動
 *   }
 *   //非活性化すると行オブジェクトが解放されます。
 *   table.passivate();
 * </pre>
 * 
 * [基本的な使い方：SQLデータベーステーブルの値を変更する]
 * <pre>
 *   //n番目の行に移動します。
 *   table.moveTo(n);
 *   //NAMEカラムに値を設定します。
 *   table.getValueAsString("NAME","山田太郎");
 *   //行に対する変更内容を確認します。※必須項目に値を設定していないと例外となります。
 *   table.validateRow();
 *   //更新対象データベース・スキーマ名を指定します。※ORACLEで、接続ユーザー名とスキーマー名が同一の場合、省略可能です。
 *   table.setResolveTargetDBSchema("public");
 *   //更新対象データベース・テーブル名 を指定します。
 *   tableEmp.setResolveTargetDBTable("t_emp");
 *   //更新を実行します。この際、JDBCコネクションを与えます。
 *   tableEmp.resolveChanges(conn);  
 * </pre>
 * 
 * @see jp.littlesoft.sql.ISQLDataProvider
 * @see jp.littlesoft.sql.ISQLDataResolver
 * @author LittleSoft Corporation
 * @version 1.3.0
 *
 */
@SuppressWarnings("unchecked")
public class WebSQLTable<E extends Serializable> extends WebDataTable<E> implements ISQLApplication {

  private static final long serialVersionUID = 1L;
  private static String RESOLVER_IS_NULL_OR_INVALIDTYPE = "DataResolver is null or invalid type.";
  private SQL<E> sqlContext = null;
  /**
   * WebSQLTable オブジェクトを生成します。<br/>
   */
  public WebSQLTable(E...e) {
    super(e);
    sqlContext = new SQL<E>(this, e);
    sqlContext.setParams(getParams());
  }
  
  private SQL.Select select = null;

  /**
   * SELECTオブジェクトを作成します。.<br/>
   * valuesには選択するデータ項目[SELECT ...]を指定します。<br/>
   * 省略すると[SELECT * FROM ...] と同じ意味になります。
   * @param columns ダミーエンティティオブジェクトのフィールドまたはgetter、あるいは式（Expression）オブジェクト
   * @return Select
   */
  public SQL.Select select(Serializable...values) {
    if (select == null) {
      setSQL(sqlContext.select(values));
      return select;
    } else {
      return sqlContext.select(values);
    }
  }
  
  /**
   * SQL.Selectオブジェクトを取得します。
   * @return Select
   */
  public SQL.Select getSQLSelect() {
    return select;
  }
  
  /**
   * SQL.Selectオブジェクトを設定します。
   * @param Select
   */
  public void setSQL(SQL.Select select) {
    this.select = select;
  }
  
  private void setDefault(Connection conn) {
    try { 
      if (this.getDataProvider()==null) {
        this.setDataProvider(sqlContext.getProvider(conn)); 
      }
      if (this.getDataResolver()==null) {
        this.setDataResolver(sqlContext.getResolver(conn)); 
      }
    } catch (Exception x) {
      x.printStackTrace();
    };
  }
  
  /**
   * JDBCコネクションを設定します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void setConnection(Connection conn) throws DataException {
    setDefault(conn);
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      ((ISQLDataProvider)this.getDataProvider()).setConnection(conn);
    }
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setConnection(conn);
    }
  }

  /**
   * JDBCコネクションを取得します。<br/>
   * 
   * @return JDBCコネクション 
   */
  public Connection getConnection() {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      return ((ISQLDataProvider)this.getDataProvider()).getConnection();
    }
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getConnection();
    }
    return null;
  }
  
  /**
   * JDBCコネクションをクローズします。<br/>
   */
  public void closeConnection() {
    try { 
      if (this.getDataProvider() instanceof ISQLDataProvider) {
        ((ISQLDataProvider)this.getDataProvider()).close();
      }
      if (this.getDataResolver() instanceof ISQLDataResolver) {
        ((ISQLDataResolver)this.getDataResolver()).close();
      }
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }
  
  /**
   * JDBCコネクションを設定すると同時にテーブルを活性化します。<br/><br/>
   * 
   * この際、SQLプロパティに設定したSELECT文が実行され、テーブルオブジェクトにデータ行が読み込まれた状態になります。<br/><br/>
   * 
   * ※但し、ページングを使用する(setShowRows(int)で「-1」以外の値が設定されている)場合、<br/>
   *   SQLプロパティに設定したSELECT文ではなくページ取得用SELECT文が生成され、実行されます。<br/><br/>
   * 
   * ※OFFSET = (ページ数 - 1) * 表示行数 + 表示行数
   * ※LIMIT  = 表示行数
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public synchronized void activate(Connection conn) throws DataException {
    activate(conn, true);
  }
  
  private synchronized void activate(Connection conn, boolean resetPage) throws DataException {
    if (isActive()) passivate();
    setConnection(conn);
    if (select!=null) {
      setSQL("");
      setSQL(select.getSQL(false));
      ITableDef.MetaData meta = select.getPrimaryEntity(false);
      if (getResolveTargetDBSchema()==null || "".equals(getResolveTargetDBSchema())) {
        setResolveTargetDBSchema(meta.getDBSchema());
      }
      if (getResolveTargetDBTable()==null || "".equals(getResolveTargetDBTable())) {
        setResolveTargetDBTable(meta.getDBTableName(false));
      }
      select.createRemoteSnapshot(conn);
    }
    try {
      if (resetPage) this.setCurrentPage(1);
      if (isPageDivReserve()) setPageSQL();
      activate();
    } finally {
      if (select!=null) {
        select.dropRemoteSnapshot(conn);
        select.prepare(true);
        //setSQL(select.getSQL(true));
      }
    }
  }

  public synchronized void emptyActivate(Connection conn) throws DataException {
    if (select==null) throw new DataException("This operation requires SQL.Select object.");
    select.setUseAbsoluteFakeCondition(true);
    try {
      activate(conn, true);
    } finally {
      select.setUseAbsoluteFakeCondition(false);
    }
  }
   
  /**
   * ページ取得用のSQL文を生成し、WebSQLTable オブジェクトを活性化する際に実行される
   * SELECT文として設定します。<br/><br/>
   * 
   * @throws DataException 例外発生時
   */
  private void setPageSQL()  throws DataException {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      ISQLDataProvider prov = (ISQLDataProvider)this.getDataProvider();
      prov.setPageSQL(prov.getSQL(), (getCurrentPage()-1)*getShowRows() ,getShowRows());
    }
  }

  /**
   * ページング使用時の最終行の調整を行います。<br/>
   */
  private void adjustEndRows() {
    if (isPageDivReserve()) setPageEndRowIndex(super.getRowCount() - 1);
  }
  
  /**
   * 先頭ページの内容をデータベースより取得し、テーブルオブジェクトに格納します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void pageFirst(Connection conn) throws DataException {
//    if (!this.isActive() || getShowRows()==-1) return;
    if (isPageDivReserve()) {
      if (getCurrentPage()==1) return;
      pageMoveTo(1, conn);
    } else {
      if (!this.isActive() || getShowRows()==-1) return;
      pageFirst();
    }
  }
  
  /**
   * 次ページの内容をデータベースより取得し、テーブルオブジェクトに格納します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void pageNext(Connection conn) throws DataException {
//    if (!this.isActive() || getShowRows()==-1) return;
    if (isPageDivReserve()) {
      if (getPageEndRowIndex()==(this.getRowCount()-1))  return;
      pageMoveTo(getCurrentPage() + 1, conn);
    } else {
      if (!this.isActive() || getShowRows()==-1) return;
      pageNext();
    }
  }

  /**
   * 前ページの内容をデータベースより取得し、テーブルオブジェクトに格納します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void pagePrevious(Connection conn) throws DataException {
//    if (!this.isActive() || getShowRows()==-1) return;
    if (isPageDivReserve()) {
      if (getCurrentPage()==1) return;
      pageMoveTo(getCurrentPage() - 1, conn);
    } else {
      if (!this.isActive() || getShowRows()==-1) return;
      pagePrevious();
    }
  }
  
  /**
   * 最終ページの内容をデータベースより取得し、テーブルオブジェクトに格納します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void pageLast(Connection conn) throws DataException {
//    if (!this.isActive() || getShowRows()==-1) return;
    if (isPageDivReserve()) {
      pageMoveTo(getPageCount(), conn);
    } else {
      if (!this.isActive() || getShowRows()==-1) return;
      pageLast();
    }
  }

  /**
   * 引数で指定されたページの内容をデータベースより取得し、テーブルオブジェクトに格納します。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public void pageMoveTo(int page, Connection conn) throws DataException {
//    if (!this.isActive() || !isPageDivReserve()) return;
    if (!isPageDivReserve()) return;
      setCurrentPage(page);
      activate(conn, false);
      //最終行がずれるので再調整
      adjustEndRows();
      setCurrentPage(page);
  }
  
  /**
   * 全体のページ数を取得します。<br/>
   * 
   * @return 全体のページ数
   */
  public int getPageCount() {
    if (!this.isActive() || !isPageDivReserve()) return -1;
    return (int)Math.ceil(((double)this.getRowCount()/(double)this.getShowRows()));
  }

  /**
   * Tableオブジェクトに格納されているデータ件数を取得します。<br/><br/>
   * 
   * ページングを使用する(setShowRows(int)で「-1」以外の値が設定されている)場合、<br/>
   * Tableオブジェクトに格納されている件数ではなく、SQLプロパティに設定したSELECT文で<br/>
   * 取得される件数が返却されます。
   *   
   * @see jp.littlesoft.data.Table#getRowCount()
   */
  @Override
  public int getRowCount() {
    if (!isPageDivReserve() || getAllRowCount()==-1) return super.getRowCount();
    return getAllRowCount();
  }
  
  /**
   * 次のページが存在するかどうか調べます。<br/>
   * 
   * @see jp.littlesoft.web.WebDataTable#hasNextPage()
   */
  @Override
  public boolean hasNextPage() {
    if (!this.isActive() || getShowRows()==-1) return false;
    if (!isPageDivReserve()) return super.hasNextPage();
    return (getCurrentPage()!=getPageCount());
  }
  
  /**
   * 前のページが存在するかどうか調べます。<br/>
   * 
   * @see jp.littlesoft.web.WebDataTable#hasPreviousPage()
   */
  @Override
  public boolean hasPreviousPage() {
    if (!this.isActive() || getShowRows()==-1) return false;
    if (!isPageDivReserve()) return super.hasPreviousPage();
    return (getCurrentPage() > 1);
  }

  /**
   * 現在行を削除します。<br/><br/>
   * 
   * ページング使用時は削除後に最終行Indexの調整、全体行数の調整を行う。<br/>
   * 
   * @see jp.littlesoft.web.WebDataTable#deleteRow()
   */
  @Override
  public synchronized void deleteRow() throws DataException {
    super.deleteRow();
    adjustEndRows();
    if (this.isActive() && isPageDivReserve()) setAllRowCount(getAllRowCount()-1);
  }
  
  /**
   * SQLテーブル最後に新規行を追加します。<br/><br/>
   * 
   * ページング使用時は削除後に全体行数の調整を行う。<br/>
   * 
   * @see jp.littlesoft.web.WebDataTable#appendRow()
   */
  @Override
  public synchronized Row<E> appendRow() throws DataException {
    Row<E> row = super.appendRow();
    if (this.isActive() && isPageDivReserve()) setAllRowCount(getAllRowCount()+1);
    return row;
  }

  /**
   * SQLテーブルに新規行を追加します。<br/><br/>
   * 
   * ページング使用時は削除後に全体行数の調整を行う。<br/>
   * 
   * @see jp.littlesoft.web.WebDataTable#insertRow(int)
   */
  @Override
  public synchronized Row<E> insertRow(int index) throws DataException {
    Row<E> row = super.insertRow(index);
    if (this.isActive() && isPageDivReserve()) setAllRowCount(getAllRowCount()+1);
    return row;
  }

  /**
   * 指定されたカラムでSQLテーブル内の行をソートします。<br/>
   * 
   * @see jp.littlesoft.data.Table#sortRows(java.lang.String)
   */
  @Override
  public void sortRows(String columnName) {
    if (!isPageDivReserve()) {
      super.sortRows(columnName);
    } else {
      
    }
  }
  
  /**
   * SQLパラメータのセットを取得します。<br/><br/>
   * 
   * このメソッドで取得したParamsオブジェクトを利用して、SQLプロパティに設定したSQL文に含まれる
   * バインドパラメータへアクセスすることができます。<br/><br/>
   * 
   * 設定するSQLには
   * <pre>
   *   table.setSQL("select * from emp where empno = :EMPLOYEE_NO");
   * </pre>
   * の様に、バインドパラメータにする項目名の先頭にコロン(:)を指定します。<br/><br/>
   * 
   * EMPLOYEE_NOに値をセットする場合は
   * <pre>
   *   table.getParam.setValueAsString("1001");
   * </pre>
   * というコードになります。<br/>
   */
  @Override
  public Params<E> getParams() {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      return (Params<E>) ((ISQLDataProvider)this.getDataProvider()).getParams();
    } else {
      return super.getParams();
    }
  }
  
  /**
   * WebSQLTable オブジェクトを活性化する際に実行されるSELECT文を設定します。<br/><br/>
   * 
   * activate() メソッドを実行すると、SQLプロパティに設定されたSELECT文が実行され、
   * その結果セットがテーブルオブジェクトの内容として設定されます。<br/><br/>
   * 
   * これらは実際のデータベーステーブルとは切り離された状態のオブジェクトなので、値の変更や、
   * 行の追加削除などを自由に行うことが可能です。<br/><br/>
   * 
   * WebSQLTable オブジェクトに対して行われた編集内容は resolveChange() メソッドによって
   * データベーステーブルへ書き戻すことができます。<br/>
   * 
   * @param selectSQL SELECT文
   * @throws DataException 例外発生時
   */
  public void setSQL(String selectSQL)  throws DataException {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      ((ISQLDataProvider)this.getDataProvider()).setSQL(selectSQL);
    }
  }
  
  /**
   * WebSQLTable オブジェクトを活性化する際に実行されるSELECT文を取得します。<br/>
   * 
   * @return String SELECT文
   */
  public String getSQL() {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      return ((ISQLDataProvider)this.getDataProvider()).getSQL();
    } else {
      return null;
    }
  }
  
  /**
   * 更新対象データベーステーブルのカタログ名を設定します。<br/><br/>
   * 
   * このプロパティは resolveChange() メソッドの中で、更新対象テーブルのメタデータを検索する為に使用されます。<br/>
   * 
   * @param targetDBCatalog カタログ名
   * @throws DataException 例外発生時
   */
  public void setResolveTargetDBCatalog(String targetDBCatalog)  throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setTargetDBCatalog(targetDBCatalog);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }  

  /**
   * 更新対象データベーステーブルのスキーマ名を設定します。<br/><br/>
   * 
   * このプロパティは resolveChange() メソッドの中で、更新対象テーブルのメタデータを検索する為に使用されます。<br/>
   * 
   * @param targetDBSchema スキーマ名
   * @throws DataException 例外発生時
   */
  public void setResolveTargetDBSchema(String targetDBSchema)  throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setTargetDBSchema(targetDBSchema);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }  

  /**
   * 更新対象データベーステーブル名を設定します。<br/><br/>
   * 
   * このプロパティは resolveChange() メソッドの中で、更新対象テーブルのメタデータを検索する為、
   * また更新SQLを作成する為に使用されます。<br/>
   * 
   * @param targetDBTable テーブル名
   * @throws DataException 例外発生時
   */
  public void setResolveTargetDBTable(String targetDBTable)  throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setTargetDBTable(targetDBTable);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }
  
  /**
   * 更新対象データベーステーブルのカタログ名を取得します。<br/>
   * 
   * @return 更新対象データベーステーブルのカタログ名
   */
  public String getResolveTargetDBCatalog() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getTargetDBCatalog();
    } else {
      return null;
    }
  }
  
  /**
   * 更新対象データベーステーブルのスキーマ名を取得します。<br/>
   * 
   * @return 更新対象データベーステーブルのスキーマ名
   */
  public String getResolveTargetDBSchema() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getTargetDBSchema();
    } else {
      return null;
    }
  }

  /**
   * 更新対象データベーステーブル名を取得します。<br/>
   * 
   * @return 更新対象データベーステーブルのデータベーステーブル名
   */
  public String getResolveTargetDBTable() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getTargetDBTable();
    } else {
      return null;
    }
  }
  
  /**
   * データベーステーブルの更新対象行を検索するキーとなる列を設定します。<br/><br/>
   * 
   * このプロパティはresolveChangeメソッドの中で、UPDATE文、DELETE文のWHERE句を作成する為に使用されます。<br/><br/>
   * 
   * このプロパティを省略するとResolveTargetDBTableプロパティで指定されたデータベーステーブルに含まれれる
   * 全ての列がWHERE句に含まれるようになります。<br/>
   * <em>※注意：</em>
   * 但し、BLOB、CLOB等など特殊なデータ型のカラムは除かれます。<br/>
   * 
   * @param newResolveKeyColumns 検索するキーの配列
   * @throws DataException 例外発生時
   */
  public void setResolveKeyColumns(String[] newResolveKeyColumns) throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setKeyColumns(newResolveKeyColumns);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }    
  }
  
  /**
   * データベーステーブルの更新対象行を検索するキーとなる列名の配列を取得します。<br/>
   * 
   * @return 列名の配列
   */
  public String[] getResolveKeyColumns() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getKeyColumns();
    } else {
      return null;
    }    
  }

  /**
   * WebSQLTable オブジェクトに対して行われた編集結果をもとに、データベーステーブルの内容を書き換えます。<br/><br/>
   * 
   * WebSQLTable オブジェクト内の行の状態をもとにDELETE文、INSERT文、UPDATE文を自動的作成すると同時に、
   * データベースサーバーへ適用します。<br/>
   * <em>※注意：</em>
   * このメソッドの中ではコミットを行わないので、アプリケーションプログラム側で、適切なコミット処理を行ってください。<br/><br/>
   * 
   * 当メソッドを実行する為には以下プロパティが必須です。
   * <ul>
   *   <li>Connection</li>
   *   <li>ResolveTargetDBTable</li>
   *   <li>ResolveTargetDBSchema</li>
   * </ul>
   * <em>※注意：</em>
   * ORACLEで、接続ユーザー名とスキーマー名が同一の場合、省略可能です。<br/><br/>
   */
  @Override
  public synchronized void resolveChanges() throws DataException {
    internalResolveChanges(true);
  }

  /**
   * WebSQLTable オブジェクトに対して行われた編集結果をもとに、データベーステーブルの内容を書き換えます。<br/><br/>
   * 
   * 引数としてJDBCコネクションを与えることができる点を除き、引数無しの resolveChanges() メソッドと同じです。<br/>
   * 
   * @param conn JDBCコネクション
   * @throws DataException 例外発生時
   */
  public synchronized void resolveChanges(Connection conn) throws DataException {
    this.setConnection(conn);
    internalResolveChanges(true);
  }

  private synchronized void internalResolveChanges(boolean pendingCommit) throws DataException {
    if (getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)getDataResolver()).setPendingCommit(pendingCommit);
    }
    super.resolveChanges();
    // 複数行で削除された場合、pageStartRowが存在しない行の場合があるので、その場合最終ページにする。add hirayanagi
    if (getPageStartRowIndex() > getPageEndRowIndex()) {
      pagePrevious();
    }
    //ページング使用時、resolveChanges()により最終行がずれるので再調整
    adjustEndRows();
  }
  
  /**
   * resolveChange() メソッド使用するDELETE文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとresolveChange() メソッド内で、自動作成されるDELETE文が使用されます。<br/>
   * 
   * @param sqlDelete DELETE文
   * @throws DataException 例外発生時
   * @see jp.littlesoft.sql.ISQLDataResolver#setSQLDelete(String)
   */
  public void setSQLDelete(String sqlDelete) throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setSQLDelete(sqlDelete);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }
  
  /**
   * resolveChange() メソッド使用するDELETE文を取得します。<br/><br/>
   * 
   * setSQLDeleteで設定した値を取得することができますが、resolveChange() メソッド内で、
   * 自動作成される内容を取得する事はできません。<br/>
   * 
   * @return DELETE文
   */
  public String getSQLDelete() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getSQLDelete();
    } else {
      return null;
    }    
  }
  
  /**
   * resolveChange() メソッド使用するUPDATE文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとresolveChange() メソッド内で、自動作成されるUPDATE文が使用されます。<br/>
   * 
   * @param sqlUpdate UPDATE文
   * @throws DataException 例外発生時
   * @see jp.littlesoft.sql.ISQLDataResolver#setSQLUpdate(String)
   */
  public void setSQLUpdate(String sqlUpdate) throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setSQLUpdate(sqlUpdate);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }
  
  /**
   * resolveChange() メソッド使用するUPDATE文を取得します。<br/><br/>
   * 
   * setSQLUpdateで設定した値を取得することができますが、resolveChange() メソッド内で、
   * 自動作成される内容を取得する事はできません。<br/>
   * 
   * @return UPDATE文
   */
  public String getSQLUpdate() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getSQLUpdate();
    } else {
      return null;
    }    
  }

  /**
   * resolveChange() メソッド使用するINSERT文を設定します。<br/><br/>
   * 
   * このプロパティを省略するとresolveChange() メソッド内で、自動作成されるINSERT文が使用されます。<br/>
   * 
   * @param sqlInsert INSERT文
   * @throws DataException 例外発生時
   * @see jp.littlesoft.sql.ISQLDataResolver#setSQLInsert(String)
   */
  public void setSQLInsert(String sqlInsert) throws DataException {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      ((ISQLDataResolver)this.getDataResolver()).setSQLInsert(sqlInsert);
    } else {
      throw new DataException(RESOLVER_IS_NULL_OR_INVALIDTYPE);
    }
  }
  
  /**
   * resolveChange() メソッド使用するINSERT文を取得します。<br/><br/>
   * 
   * setSQLInsertで設定した値を取得することができますが、resolveChange() メソッド内で、
   * 自動作成される内容を取得する事はできません。<br/>
   * 
   * @return INSERT文
   */
  public String getSQLInsert() {
    if (this.getDataResolver() instanceof ISQLDataResolver) {
      return ((ISQLDataResolver)this.getDataResolver()).getSQLInsert();
    } else {
      return null;
    }    
  }

  /**
   * Tableに読み込む最大行数を設定します。<br/><br/>
   *
   * ０を設定すると無制限になります。（デフォルト）<br/>
   * 
   * @param  newMaxRows  最大行数(0以上の整数,0は無制限)
   */
  public void setMaxRows(int newMaxRows) {
    if (this.getDataProvider() instanceof ISQLDataProvider) {
      ((ISQLDataProvider)this.getDataProvider()).setMaxRows(newMaxRows);
    }
  }

  /* (非 Javadoc)
   * @see jp.littlesoft.data.Table#getRow(int)
   */
  @Override
  public Row<E> getRow(int rowIndex) throws DataException {
    // 行取得に失敗した場合
    // 失敗した行インデックスと取得元TableにセットされたSQLをException内容とするようにする
    try {
      return super.getRow(rowIndex);
    } catch (IndexOutOfBoundsException x) {
      String msg = "指定した行がありません[" + rowIndex + "]\n"+ 
        this.getClass().getName() + ".SQL：" + this.getSQL();
      for (int i=0; i < this.getParams().getColumnCount(); i++) {
        int row = i + 1;
        msg += "\n  [" + row + "] " + this.getParams().getColumn(i).getName() + ": " + this.getParams().getValueAsString(i);
      }
      throw new DataException(msg);
    }
  }

  @Override
  public Expression $(Serializable source) {
    return sqlContext.$(source);
  }

  @Override
  public BindParameter $bindParam(String name) {
    return sqlContext.$bindParam(name);
  }

  @Override
  public Column $column(Serializable source) {
    for (Column c : getColumnArray()) {
      try {
        if (c.getEntityValue($$)==source) {
          return c;
        }
      } catch (NoAccessibleObjectException e) {
        throw new RuntimeException(e);
      }
    }
    if (this.select!=null) {
      this.select.getColumn(source);
    }
    return null;
  }

  @Override
  public Condition<?> $if(Condition nestedCondition) {
    return sqlContext.$if(nestedCondition);
  }

  @Override
  public Comparer<?, ?> $if(Serializable value) {
    return sqlContext.$if(value);
  }

  @Override
  public <T extends Serializable> T $useDummy(Class<? extends T> type) {
    return sqlContext.$useDummy(type);
  }

  /**
   * 実行時パラメータがNOT NULLであれば評価される条件式を開始します.<br/>
   * @param nestedCondition 入れ子にする条件式（Conditionオブジェクト）
   */
  public Condition<?> $evaluateNonNulls(Condition<?> nestedCondition) {
    return sqlContext.$evaluateNonNulls(nestedCondition);
  }  
  
  /**
   * 実行時パラメータがNOT NULLであれば評価される条件式を開始します.<br/>
   * @param value  ダミーエンティティオブジェクトのフィールドまたはgetter
   */
  public Comparer $evaluateNonNulls(Serializable value) {
    return sqlContext.$evaluateNonNulls(value);
  }
  /**
   * DELETE文を構築します.<br/>
   * @param entity
   * @return Delete
   */
  public Delete delete(Serializable entity) {
    return sqlContext.delete(entity);
  }
  
  /**
   * DELETE文を構築します.<br/>
   * ResolveTargetDBTableが削除対象テーブルとなります。<br/>
   * @return Delete
   */
  public Delete delete() {
    return sqlContext.delete(getResolveTargetEntity());
  }

  /**
   * INSERT文を構築します.<br/>
   * @param entity
   * @return Insert
   */
  public Insert insert(Serializable entity) {
    return sqlContext.insert(entity);
  }
  
  /**
   * INSERT文を構築します.<br/>
   * ResolveTargetDBTableが挿入対象テーブルとなります。<br/>
   * @return Insert
   */
  public Insert insert() {
    return sqlContext.insert(getResolveTargetEntity());
  }
  
  /**
   * UPDATEオブジェクトを作成します。.<br/>
   * @param entity 更新対象テーブルのダミーエンティティオブジェクト
   * @return Update オブジェクト
   */
  public Update update(Serializable entity) {
    return sqlContext.update(entity);
  }
  
  /**
   * UPDATEオブジェクトを作成します。.<br/>
   * ResolveTargetDBTableが更新対象テーブルとなります。<br/>
   * @return Update オブジェクト
   */
  public Update update() {
    return sqlContext.update(getResolveTargetEntity());
  } 

  private Serializable getResolveTargetEntity() {
    String targetTable = getResolveTargetDBTable();
    if (targetTable==null || "".equals(targetTable)) {
      return select.getPrimaryEntity(false);
    } else {
      return select.getDummy(targetTable);
    }
  }
  

  @Override
  protected Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  /**
   * 設定済のダイナミック条件式を全てクリアします。
   */
  @Override
  public void resetDynamicCondtions() {
    sqlContext.resetDynamicCondtions();
  }

  /**
   * キーを指定してダイナミック条件式を設定します.<br/>
   * @param key 条件式を識別するキー
   */
  @Override
  public void setDynamicCondtion(String key, Condition<?> condition) {
    sqlContext.setDynamicCondtion(key, condition);
  }
  
  /**
   * リモートデータベースに接続するダミーエンティティオブジェクトを作成します。.<br/>
   * @param type エンティティオブジェクトのクラス
   * @param id   リモートデータベースを識別する任意のID
   */
  @Override
  public <T extends Serializable> T $useRemoteDummy(Class<? extends T> type, String id) {
    return (T) sqlContext.$useRemoteDummy(type, id);
  }
  
  /**
   * ダミーエンティティオブジェクトにリモートデータベースのIDを割り当てます.<br/>
   * @param dummy  ダミーエンティティオブジェクﾄ
   * @param id  リモートデータベースを識別する任意のID
   */
  @Override
  public void $useRemote(ITableDef dummy, String id) {
    sqlContext.$useRemote(dummy, id);
  }

  /**
   * リモートデータベースのコネクションを設定します.
   * @param id リモートデータベースを識別する任意のID
   * @param connection JDBCコネクション
   */
  @Override
  public void setRemoteConnection(String id, Connection connection) {
    sqlContext.setRemoteConnection(id, connection);
  }

}
